proc make_project { } {
  # Get the directory containing this script. Source file paths are relative to this.
  set origin_dir [file dirname [file normalize [info script]]]

  # Define versions of dependencies in one convenient place
  source "$origin_dir/version.tcl"

  set viv_ver [version -short]
  set viv_spl [split $viv_ver .]
  set viv_maj [lindex $viv_spl 0]
  set viv_min [lindex $viv_spl 1]

  if { $viv_maj < 2018 || ($viv_maj == 2018 && $viv_min < 3) } {
    send_msg_id {AD-MKXPR-001} {ERROR} "Vivado ${viv_ver} is not supported for this FPGA design. Minimum Vivado version is 2018.3."
  }

  # Create project
  set config {8g_2400_x32_axi4}
  set device {ku115_2i}
  set dir "${config}-${device}"
  if { [info exists ::env(ADMXRC3_PKG_TEST)] } {
    append dir {-} [version -short]
  }
  set part {xcku115-flva1517-2-i}
  set name {ddr4_test}
  set project [create_project -force -part $part $name "$origin_dir/$dir"]
  current_project $project

  # Get the directory path for the new project
  set proj_dir [get_property directory $project]

  # Set project properties
  set_property "default_lib" "xil_defaultlib" $project
  set_property "part" $part $project
  set_property "simulator_language" "Mixed" $project
  set_property "target_language" "Verilog" $project

  # Create 'sources_1' fileset (if not found)
  if {[string equal [get_filesets -quiet sources_1] ""]} {
    create_fileset -srcset sources_1
  }
  set source_set [get_filesets sources_1]

  # Set source set properties
  set_property "top" "ddr4_test" $source_set

  # Add HDL source files
  set hdl_files [list \
    [file normalize "$origin_dir/../../../../fpga/lib/memory_exerciser-$ver_memory_exerciser/src/rng_n1024_r32_t6_k32_s2999b26.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/memory_exerciser-$ver_memory_exerciser/src/memory_exerciser_rng.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/memory_exerciser-$ver_memory_exerciser/src/memory_exerciser_data_check.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/memory_exerciser-$ver_memory_exerciser/src/memory_exerciser_fifo.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/memory_exerciser-$ver_memory_exerciser/src/memory_exerciser_error_log.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/memory_exerciser-$ver_memory_exerciser/src/memory_exerciser_data_gen.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/memory_exerciser-$ver_memory_exerciser/src/memory_exerciser_rearranger.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/memory_exerciser-$ver_memory_exerciser/src/memory_exerciser_dearranger.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/memory_exerciser-$ver_memory_exerciser/src/memory_exerciser_axi4.vhd"] \
    [file normalize "$origin_dir/../src/ddr4_test_axi.vhd"] \
  ]
  add_files -norecurse -fileset $source_set $hdl_files

  # Import IP sources
  set ip_files [list \
    [file normalize "$origin_dir/../ip/2016.1/ddr4sdram_8g_2400_x32_axi4-${device}/ddr4sdram.xci"] \
  ]
  import_ip -srcset $source_set $ip_files
  upgrade_ip [get_ips]

  create_ip -vlnv "xilinx.com:ip:vio:*" -module_name vio_ip
  set vio_ip [get_ips vio_ip]

  # Set initial value of probe_out2 (memory exerciser number of 32-byte words) to 0x03FFFFFF, for 2x1 GiB parts per bank
  set_property -dict { \
    CONFIG.C_NUM_PROBE_IN {24} \
    CONFIG.C_NUM_PROBE_OUT {5} \
    CONFIG.C_PROBE_IN0_WIDTH {1} \
    CONFIG.C_PROBE_IN1_WIDTH {1} \
    CONFIG.C_PROBE_IN2_WIDTH {1} \
    CONFIG.C_PROBE_IN3_WIDTH {1} \
    CONFIG.C_PROBE_IN4_WIDTH {1} \
    CONFIG.C_PROBE_IN5_WIDTH {32} \
    CONFIG.C_PROBE_IN6_WIDTH {3} \
    CONFIG.C_PROBE_IN7_WIDTH {32} \
    CONFIG.C_PROBE_IN8_WIDTH {32} \
    CONFIG.C_PROBE_IN9_WIDTH {32} \
    CONFIG.C_PROBE_IN10_WIDTH {32} \
    CONFIG.C_PROBE_IN11_WIDTH {32} \
    CONFIG.C_PROBE_IN12_WIDTH {32} \
    CONFIG.C_PROBE_IN13_WIDTH {32} \
    CONFIG.C_PROBE_IN14_WIDTH {32} \
    CONFIG.C_PROBE_IN15_WIDTH {32} \
    CONFIG.C_PROBE_IN16_WIDTH {32} \
    CONFIG.C_PROBE_IN17_WIDTH {32} \
    CONFIG.C_PROBE_IN18_WIDTH {32} \
    CONFIG.C_PROBE_IN19_WIDTH {32} \
    CONFIG.C_PROBE_IN20_WIDTH {32} \
    CONFIG.C_PROBE_IN21_WIDTH {32} \
    CONFIG.C_PROBE_IN22_WIDTH {32} \
    CONFIG.C_PROBE_IN23_WIDTH {32} \
    CONFIG.C_PROBE_OUT0_WIDTH {1} \
    CONFIG.C_PROBE_OUT1_WIDTH {32} \
    CONFIG.C_PROBE_OUT2_WIDTH {32} \
    CONFIG.C_PROBE_OUT3_WIDTH {8} \
    CONFIG.C_PROBE_OUT4_WIDTH {8} \
    CONFIG.C_PROBE_OUT0_INIT_VAL {0x0} \
    CONFIG.C_PROBE_OUT1_INIT_VAL {0x00000000} \
    CONFIG.C_PROBE_OUT2_INIT_VAL {0x03FFFFFF} \
    CONFIG.C_PROBE_OUT3_INIT_VAL {0x00} \
    CONFIG.C_PROBE_OUT4_INIT_VAL {0xFF} \
  } $vio_ip


  # Add constraints files
  # Put target .xdc as LAST in list
  set constraint_set [get_filesets constrs_1]
  set constraint_files [list \
    [file normalize "$origin_dir/../src/refclk300.xdc"] \
    [file normalize "$origin_dir/../src/ddr4sdram.xdc"] \
    [file normalize "$origin_dir/../src/ddr4sdram_locs_b0_x32.xdc"] \
    [file normalize "$origin_dir/../src/ddr4sdram_locs_b1_x32.xdc"] \
    [file normalize "$origin_dir/../src/ddr4sdram_locs_b2_x32.xdc"] \
    [file normalize "$origin_dir/../src/ddr4sdram_locs_b3_x32.xdc"] \
    [file normalize "$origin_dir/../src/bitstream.xdc"] \
    [file normalize "$origin_dir/../src/ddr4_test_axi.xdc"] \
  ]
  add_files -norecurse -fileset $constraint_set $constraint_files
  set_property "target_constrs_file" [lindex $constraint_files end] $constraint_set

  # Add utilities files (Tcl hook scripts etc.)
  set utils_set [get_filesets -quiet utils_1]
  if {[string equal $utils_set ""]} {
    # Utilities fileset introduced in Vivado 2018.3; fall back to using sources set if not present.
    set utils_set $source_set
  }
  #set utils_files [list \
  #  [file normalize "$origin_dir/soak_test.tcl"] \
  #]
  #add_files -norecurse -fileset $utils_set $utils_files
  # Set used_in_XXX properties so that Vivado 2018.2 and earlier doesn't treat them as design sources.
  #set_property -dict {
  #  {used_in_synthesis} {false} \
  #  {used_in_implementation} {false} \
  #  {used_in_simulation} {false} \
  #} [get_files $utils_files]

  set sim_set [get_filesets sim_1]

  # Add simulation HDL source files
  set sim_files [list \
    [file normalize "$origin_dir/../../../../fpga/lib/ddr4_sdram_model-$ver_ddr4_sdram_model/src/ddr4_sdram_model_pkg.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/ddr4_sdram_model-$ver_ddr4_sdram_model/src/ddr4_sdram_model_int_pkg.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/ddr4_sdram_model-$ver_ddr4_sdram_model/src/ddr4_sdram_bank_state.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/ddr4_sdram_model-$ver_ddr4_sdram_model/src/ddr4_sdram_data.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/ddr4_sdram_model-$ver_ddr4_sdram_model/src/ddr4_sdram_data_sv.sv"] \
    [file normalize "$origin_dir/../../../../fpga/lib/ddr4_sdram_model-$ver_ddr4_sdram_model/src/ddr4_sdram_init_monitor.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/ddr4_sdram_model-$ver_ddr4_sdram_model/src/ddr4_sdram_power_state.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/ddr4_sdram_model-$ver_ddr4_sdram_model/src/ddr4_sdram_refresh_monitor.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/ddr4_sdram_model-$ver_ddr4_sdram_model/src/ddr4_sdram_su_h_check.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/ddr4_sdram_model-$ver_ddr4_sdram_model/src/ddr4_sdram_die.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/ddr4_sdram_model-$ver_ddr4_sdram_model/src/ddr4_sdram_chip.vhd"] \
    [file normalize "$origin_dir/../../../../fpga/lib/ddr4_sdram_model-$ver_ddr4_sdram_model/src/ddr4_sdram_mt40a_pkg.vhd"] \
    [file normalize "$origin_dir/../src/tb_ddr4_test.vhd"] \
  ]
  add_files -norecurse -fileset $sim_set $sim_files

  # Set simulation properties
  set_property "source_set" "sources_1" $sim_set
  set_property "top" "tb_ddr4_test" $sim_set
  # XSIM incremental compilation seems unreliable for large projects
  set_property "incremental" "false" $sim_set
  set_property "runtime" "25 us" $sim_set
  # Set runtime on per-simulator basis
  foreach simulator [get_simulators] {
    set_property "target_simulator" "$simulator" $project
    set_property "${simulator}.simulate.runtime" "25 us" $sim_set
  }
  set_property "xsim.simulate.custom_tcl" [file normalize "$origin_dir/xsim_wave_axi.tcl"] $sim_set
  set_property "modelsim.simulate.custom_udo" [file normalize "$origin_dir/vsim_wave_axi.udo"] $sim_set
  # Return to using XSim
  set_property target_simulator {XSim} $project

  set synth_run [get_runs synth_1]
  #set_property strategy {Flow_PerfOptimized_high} $synth_run
  set_property "needs_refresh" "1" $synth_run
  set_property "part" $part $synth_run
  current_run -synthesis $synth_run

  set impl_run [get_runs impl_1]
  set_property "needs_refresh" "1" $impl_run
  set_property "part" $part $impl_run
  set_property "steps.phys_opt_design.is_enabled" "true" $impl_run
  set_property "steps.post_route_phys_opt_design.is_enabled" "true" $impl_run
  set_property -name {steps.route_design.args.more options} -value -timing_summary -objects $impl_run
  current_run -implementation $impl_run

  update_compile_order -fileset $source_set
  if { [info exists sim_set] } {
    update_compile_order -fileset $sim_set
  }

  return $project
}

return [make_project]
