library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;

library work;
use work.axi4_128.all;
use work.axi4_128_tb_pkg.all;

entity tb_dma_demo_stim is
port(
  -- AXI
  axi_clk          : in    std_logic;
  direct_slave_m2s : out   axi4_128_m2s_t;
  direct_slave_s2m : in    axi4_128_s2m_t;
  dma_channels_m2s : out   axi4_128_m2s_t_array(0 to 1);
  dma_channels_s2m : in    axi4_128_s2m_t_array(0 to 1);
  -- Online
  axi_online       : in    std_logic);
end;

architecture behav of tb_dma_demo_stim is

  constant DMA_CHANNELS : natural := 2;

  constant DS_BEATS  : natural := 32;
  constant DMA_BEATS : natural := 32;

  type ds_dma_addr_array_t is array(0 to DMA_CHANNELS-1) of axi4_128_addr_s;
  constant DS_DMA_ADDR : ds_dma_addr_array_t := (X"0000000000000000",
                                                 X"0000000000080000");

  constant DMA_BASE_ADDR  : axi4_128_addr_s := (others => '0');

  signal ds_m2s : axi4_128_m2s_t := AXI4_128_DEFAULT_M2S;

  signal dma_m2s : axi4_128_m2s_t_array(0 to DMA_CHANNELS-1) := (others => AXI4_128_DEFAULT_M2S);

  signal ds_complete   : std_logic;
  signal dma_complete  : std_logic_vector(DMA_CHANNELS-1 downto 0);
  signal test_complete : boolean;

  signal ds_passed   : std_logic_vector(DMA_CHANNELS-1 downto 0);
  signal dma_passed  : std_logic_vector(DMA_CHANNELS-1 downto 0);
  signal test_passed : boolean;

begin

  direct_slave_m2s <= ds_m2s;
  dma_channels_m2s <= dma_m2s;


  ----------------------------
  -- Direct Slave interface --
  ----------------------------

  direct_slave_p : process
    variable ds_data_wr  : byte_vector_t(0 to AXI4_128_STRB_WIDTH-1);
    variable ds_data_rd  : byte_vector_t(0 to AXI4_128_STRB_WIDTH-1);
    variable ds_data_exp : byte_vector_t(0 to AXI4_128_STRB_WIDTH-1);
    variable ds_be       : byte_enable_t(0 to AXI4_128_STRB_WIDTH-1);
    variable ds_addr     : axi4_128_addr_s;
    variable ds_xfer     : natural;
  begin
    ds_complete <= '0';
    ds_passed <= (others => '1');

    wait until (axi_online = '1');  -- AXI commands ignored until MPTL online
    report "MPTL now online";

    ----------------------------
    -- DS access to BRAM Test --
    ----------------------------
    ds_be := conv_byte_enable(X"FFFF");

    for n in 0 to DMA_CHANNELS-1 loop
      for i in 0 to DS_BEATS-1 loop
        -- Write a pattern to BRAM
        -- Uses axi4_128_sim_write procedure in axi4_128_tb_pkg
        ds_addr := DS_DMA_ADDR(n) + i*AXI4_128_STRB_WIDTH;
        ds_data_wr(0 to 3)   := conv_byte_vector(ds_addr(33 downto 4) & "00");
        ds_data_wr(4 to 7)   := conv_byte_vector(ds_addr(33 downto 4) & "01");
        ds_data_wr(8 to 11)  := conv_byte_vector(ds_addr(33 downto 4) & "10");
        ds_data_wr(12 to 15) := conv_byte_vector(ds_addr(33 downto 4) & "11");

        axi4_128_sim_write(ds_addr,i,ds_be,ds_data_wr,ds_xfer,axi_clk,ds_m2s,direct_slave_s2m);
        report "Wrote DS DATA " & integer'image(ds_xfer) &
               " bytes 0x" & conv_string_hex(ds_data_wr) &
               " with enable 0b" & conv_string_bin(ds_be) &
               " to byte address 0x" & conv_string_hex(ds_addr);
      end loop;
    end loop;

    for n in 0 to DMA_CHANNELS-1 loop
      for i in 0 to DS_BEATS-1 loop
        -- Read back pattern from BRAM
        -- Uses axi4_128_sim_read procedure in axi4_128_tb_pkg
        ds_addr := DS_DMA_ADDR(n) + i*AXI4_128_STRB_WIDTH;
        axi4_128_sim_read(ds_addr,i,ds_data_rd,ds_xfer,axi_clk,ds_m2s,direct_slave_s2m);
        report "Read DS DATA " & integer'image(ds_xfer) &
               " bytes 0x" & conv_string_hex(ds_data_rd) &
               " from byte address 0x" & conv_string_hex(ds_addr);

        ds_data_exp(0 to 3)   := conv_byte_vector(ds_addr(33 downto 4) & "00");
        ds_data_exp(4 to 7)   := conv_byte_vector(ds_addr(33 downto 4) & "01");
        ds_data_exp(8 to 11)  := conv_byte_vector(ds_addr(33 downto 4) & "10");
        ds_data_exp(12 to 15) := conv_byte_vector(ds_addr(33 downto 4) & "11");

        if ds_data_rd /= ds_data_exp then
          assert false
            report "Address " & conv_string_hex(ds_addr) &
                   " Byte " & integer'image(i) &
                   " Expected data: 0x" & conv_string_hex(ds_data_exp) &
                   " Actual data: 0x" & conv_string_hex(ds_data_rd)
            severity error;
          ds_passed(n) <= '0';
        end if;
      end loop;
    end loop;

    wait until rising_edge(axi_clk);

    if (AND_reduce(ds_passed) = '1') then
      assert false report "Test DS (AXI) completed: PASSED." severity note;
    else
      assert false report "Test DS (AXI) completed: FAILED." severity error;
    end if;

    ds_complete <= '1';

    wait;
  end process;


  -- Generate process for each DMA channel
  dma_g : for n in 0 to DMA_CHANNELS-1 generate

    -------------------
    -- DMA interface --
    -------------------

    dma_p : process
      variable dma_st_byte  : natural;
      variable dma_data_wr  : byte_vector_t(0 to DMA_BEATS*AXI4_128_STRB_WIDTH-1);
      variable dma_data_rd  : byte_vector_t(0 to DMA_BEATS*AXI4_128_STRB_WIDTH-1);
      variable dma_data_exp : byte_vector_t(0 to DMA_BEATS*AXI4_128_STRB_WIDTH-1);
      variable dma_be       : byte_enable_t(0 to DMA_BEATS*AXI4_128_STRB_WIDTH-1);
      variable dma_addr     : axi4_128_addr_s;
      variable dma_xfer     : natural;
    begin
      dma_complete(n) <= '0';
      dma_passed(n) <= '1';

      wait until (ds_complete = '1');

      -----------------------------
      -- DMA access to BRAM Test --
      -----------------------------

      for i in 0 to DMA_BEATS-1 loop
        dma_st_byte := i*AXI4_128_STRB_WIDTH;
        dma_addr := DS_DMA_ADDR(n) + dma_st_byte;

        dma_be(dma_st_byte + 0 to dma_st_byte + 15) := conv_byte_enable(X"FFFF");

        dma_data_exp(dma_st_byte + 0  to dma_st_byte + 3)  := conv_byte_vector(dma_addr(33 downto 4) & "00");
        dma_data_exp(dma_st_byte + 4  to dma_st_byte + 7)  := conv_byte_vector(dma_addr(33 downto 4) & "01");
        dma_data_exp(dma_st_byte + 8  to dma_st_byte + 11) := conv_byte_vector(dma_addr(33 downto 4) & "10");
        dma_data_exp(dma_st_byte + 12 to dma_st_byte + 15) := conv_byte_vector(dma_addr(33 downto 4) & "11");
      end loop;

      dma_addr := DMA_BASE_ADDR;

      -- Read back pattern from BRAM
      -- Uses axi4_128_sim_read procedure in axi4_128_tb_pkg
      axi4_128_sim_read(dma_addr,n,dma_data_rd,dma_xfer,axi_clk,dma_m2s(n),dma_channels_s2m(n));
      report "Read DMA" & integer'image(n) & " DATA " & integer'image(dma_xfer) &
             " bytes from byte address 0x" & conv_string_hex(dma_addr);
 
      for i in 0 to DMA_BEATS*AXI4_128_STRB_WIDTH-1 loop
        if dma_data_rd(i) /= dma_data_exp(i) then
          assert false
            report "Address " & conv_string_hex(dma_addr) &
                   " Byte " & integer'image(i) &
                   " Expected data: 0x" & conv_string_hex(dma_data_exp(i)) &
                   " Actual data: 0x" & conv_string_hex(dma_data_rd(i))
            severity error;
          dma_passed(n) <= '0';
        end if;
      end loop;

      for i in 0 to DMA_BEATS-1 loop
        dma_st_byte := i*AXI4_128_STRB_WIDTH;
        dma_addr := DS_DMA_ADDR(n) + dma_st_byte;

        dma_be(dma_st_byte + 0 to dma_st_byte + 15) := conv_byte_enable(X"FFFF");

        dma_data_wr(dma_st_byte + 0  to dma_st_byte + 3)  := conv_byte_vector(dma_addr(33 downto 4) & "11");
        dma_data_wr(dma_st_byte + 4  to dma_st_byte + 7)  := conv_byte_vector(dma_addr(33 downto 4) & "10");
        dma_data_wr(dma_st_byte + 8  to dma_st_byte + 11) := conv_byte_vector(dma_addr(33 downto 4) & "01");
        dma_data_wr(dma_st_byte + 12 to dma_st_byte + 15) := conv_byte_vector(dma_addr(33 downto 4) & "00");
      end loop;

      dma_addr := DMA_BASE_ADDR;

      -- Write a pattern to BRAM
      -- Uses axi4_128_sim_write procedure in axi4_128_tb_pkg
      axi4_128_sim_write(dma_addr,n,dma_be,dma_data_wr,dma_xfer,axi_clk,dma_m2s(n),dma_channels_s2m(n));
      report "Wrote DMA" & integer'image(n) & " DATA " & integer'image(dma_xfer) &
             " bytes to byte address 0x" & conv_string_hex(dma_addr);

      -- Read back pattern from BRAM
      -- Uses axi4_128_sim_read procedure in axi4_128_tb_pkg
      axi4_128_sim_read(dma_addr,n,dma_data_rd,dma_xfer,axi_clk,dma_m2s(n),dma_channels_s2m(n));
      report "Read DMA" & integer'image(n) & " DATA " & integer'image(dma_xfer) &
             " bytes from byte address 0x" & conv_string_hex(dma_addr);

      for i in 0 to DMA_BEATS*AXI4_128_STRB_WIDTH-1 loop
        if dma_data_rd(i) /= dma_data_wr(i) then
          assert false
            report "Address " & conv_string_hex(dma_addr) &
                   " Byte " & integer'image(i) &
                   " Expected data: 0x" & conv_string_hex(dma_data_wr(i)) &
                   " Actual data: 0x" & conv_string_hex(dma_data_rd(i))
            severity error;
          dma_passed(n) <= '0';
        end if;
      end loop;

      wait until rising_edge(axi_clk);

      if (dma_passed(n) = '1') then
        assert false report "Test DMA" & integer'image(n) & " (AXI) completed: PASSED." severity note;
      else
        assert false report "Test DMA" & integer'image(n) & " (AXI) completed: FAILED." severity error;
      end if;

      dma_complete(n) <= '1';

      wait;
    end process;
  end generate;


  -------------------------
  -- Output test results --
  -------------------------

  gen_results_p : process(ds_passed, ds_complete, dma_passed, dma_complete)
  begin
    test_passed   <= (AND_reduce(ds_passed) = '1') and (AND_reduce(dma_passed) = '1');
    test_complete <= (ds_complete = '1') and (AND_reduce(dma_complete) = '1');
  end process;

  test_results_p : process(test_complete, test_passed)
  begin
    if (test_complete) then
      if (test_passed) then
        assert false report "Test of design DMA_DEMO (AXI) completed: PASSED." severity failure;
      else
        assert false report "Test of design DMA_DEMO (AXI) completed: FAILED." severity failure;
      end if;
    end if;
  end process;

end;