/*
** cmdline.cpp
**
** Test program for the DMA Demonstration FPGA Design (dma_demo)
**
** Notes
** =====
**
** This source file contains the main() entry point for Linux & Windows,
** and should not be compiled for VxWorks.
*/

#ifdef _WIN32

# include <windows.h>
# include <tchar.h>

typedef UINT32 uint32_t;

#else

# include <stdint.h>

# define _T(x) x
# define _tmain main
typedef char TCHAR;

#endif

#include <iostream>

#include <exappcmdline.h>

#include "dma_demo.h"

using namespace AlphaData::AppFramework;

#ifdef _UNICODE
typedef std::wstringstream SStreamType;
#define SSEMPTY(ss) ((ss).str(std::basic_string<wchar_t>()))
#else
typedef std::stringstream SStreamType;
#define SSEMPTY(ss) ((ss).str(std::string()))
#endif

#ifdef _WIN32
const TCHAR g_bitPathDefault[] = _T("..\\..\\..\\..\\fpga\\bit");
#else
const TCHAR g_bitPathDefault[] = _T("../../../fpga/bit");
#endif

class MyCmdLineArguments : public CExAppCmdLineArgs {
public:
  MyCmdLineArguments(int argc, TCHAR* argv[]) :
    CExAppCmdLineArgs(argc, argv),
    m_index(0U, _T("index"), _T("Zero-based index of reconfigurable computing card to use. Mutually exclusive with '-sn'.")),
    m_serialNumber(_T("sn"), _T("Serial number of reconfigurable computing card to use. Mutually exclusive with '-index'.")),
    m_duration(2000U, _T("duration"), _T("Duration of performance measurement, in milliseconds. A value of 0 means run for ever.")),
    m_verifyData(true, _T("verify"), _T("If true, verifies data transferred in the last DMA transfer of the performance test,")),
    m_maxErrorDisplayed(10U, _T("maxerr"), _T("Maximum number of errors to display in detail when verifying transferred data.")),
    m_bitPath(g_bitPathDefault, _T("bitpath"), _T("Default path under which to look for dma_demo.bit files.")),
    m_bitFile(_T("bitfile"), _T("The .bit file to use (overrides -bitpath).")),
    m_dmaEngineMask(0x1U, _T("channels"), _T("Bitmask of DMA engines to use, where a 1 at bit 'i' means use DMA engine 'i'.")),
    m_dmaDirectionMask(0x0U, _T("directions"), _T("Bitmask of DMA transfer directions, where a 1 at bit 'i' means read from FPGA to host and and 0 means write from host to FPGA.")),
    m_dmaTransferSize(0U, _T("size"), _T("Number of bytes to transfer for each DMA transfer, where 0 means use size of BlockRAMs in FPGA."))
  {
    AddOption(m_index);
    AddOption(m_serialNumber);
    AddOption(m_duration);
    AddOption(m_verifyData);
    AddOption(m_maxErrorDisplayed);
    AddOption(m_bitPath);
    AddOption(m_bitFile);
    AddPositional(m_dmaEngineMask);
    AddPositional(m_dmaDirectionMask);
    AddPositional(m_dmaTransferSize);
  }

  virtual void OnParseComplete() {
    if (m_index.IsSet() && m_serialNumber.IsSet()) {
      AddError(_T("'-index' and '-sn' options are mutually exclusive"));
      SetParseResult(ExAppParseResultMutuallyExclusive);
    }

    if (m_duration.GetValueUInt() > 100000) {
      FormatError(
        _T("Value %u for option '-%s' must be no greater than than 100000."),
        m_duration.GetValueUInt(),
        m_duration.GetName(0).c_str());
      SetParseResult(ExAppParseResultOutOfRange);
    }

    if (0 == m_dmaEngineMask.GetValueUInt32()) {
      FormatError(
        _T("At least one DMA engine must be selected via positional argument '%s'."),
        m_dmaEngineMask.GetName(0).c_str());
      SetParseResult(ExAppParseResultOutOfRange);
    }
  }
  
#ifdef _UNICODE
  virtual void DisplaySummary(OStreamType& ostream = std::wcout) {
#else
  virtual void DisplaySummary(OStreamType& ostream = std::cout) {
#endif
    m_bodyLayout.GetStream(0) << _T("This program tests the DMA Demonstration FPGA Design (dma_demo).") << std::endl << std::endl;
    m_bodyLayout.Flush(ostream);
    CExAppCmdLineArgs::DisplaySummary(ostream);
  }
  
public:
  CExAppOptionUInt m_index;
  CExAppOptionUInt32 m_serialNumber;
  CExAppOptionUInt m_duration;
  CExAppOptionBool m_verifyData;
  CExAppOptionUInt m_maxErrorDisplayed;
  CExAppOptionStringDef m_bitPath;
  CExAppOptionString m_bitFile;
  CExAppPositionalUInt32 m_dmaEngineMask;
  CExAppPositionalUInt32 m_dmaDirectionMask;
  CExAppPositionalUInt32 m_dmaTransferSize;
};

int
_tmain(
  int argc,
  TCHAR* argv[])
{
  int exitCode = ExitCodeSuccess;

  // Parse the command-line arguments; performs most of the required validation.
  MyCmdLineArguments args(argc, argv);
  args.SetMessageDisplay(&g_dmaDemoMessageDisplay);
  ExAppParseResult result = args.Parse();
  args.DisplayMessages();
  if (ExAppParseResultOk != result) {
    exitCode = args.GetExitCode();
    goto done;
  }
  
  if (args.m_serialNumber.IsSet()) {
    exitCode = admxrc3DMADemoSN(
      args.m_bitPath.GetValueString().c_str(),
      args.m_bitFile.IsSet() ? args.m_bitFile.GetValueString().c_str() : NULL,
      args.m_serialNumber.GetValueUInt32(),
      args.m_dmaEngineMask.GetValueUInt32(),
      args.m_dmaDirectionMask.GetValueUInt32(),
      args.m_dmaTransferSize.GetValueUInt32(),
      args.m_duration.GetValueUInt(),
      args.m_maxErrorDisplayed.GetValueUInt(),
      args.m_verifyData.GetValueBool() ? 1 : 0);
  } else {
    exitCode = admxrc3DMADemoIndex(
      args.m_bitPath.GetValueString().c_str(),
      args.m_bitFile.IsSet() ? args.m_bitFile.GetValueString().c_str() : NULL,
      args.m_index.GetValueUInt(),
      args.m_dmaEngineMask.GetValueUInt32(),
      args.m_dmaDirectionMask.GetValueUInt32(),
      args.m_dmaTransferSize.GetValueUInt32(),
      args.m_duration.GetValueUInt(),
      args.m_maxErrorDisplayed.GetValueUInt(),
      args.m_verifyData.GetValueBool() ? 1 : 0);
  }

done:
  args.DisplayMessages();
  return (int)exitCode;
}
