proc make_project { } {
  # Get the directory containing this script. Source file paths are relative to this.
  set origin_dir [file dirname [file normalize [info script]]]

  # Define versions of dependencies in one convenient place
  source "$origin_dir/version.tcl"

  set viv_ver [version -short]
  set viv_spl [split $viv_ver .]
  set viv_maj [lindex $viv_spl 0]
  set viv_min [lindex $viv_spl 1]

  if { $viv_maj < 2018 || ($viv_maj == 2018 && $viv_min < 3) } {
    send_msg_id {AD-MKXPR-001} {ERROR} "Vivado ${viv_ver} is not supported for this FPGA design. Minimum Vivado version is 2018.3."
  }

  # Create project
  set device {ku060_2i}
  set dir "dma_demo_pcie-${device}"
  if { [info exists ::env(ADMXRC3_PKG_TEST)] } {
    append dir {-} [version -short]
  }
  set part {xcku060-ffva1517-2-i}
  set name {dma_demo_pcie}
  set project [create_project -force -part $part $name "$origin_dir/$dir"]
  current_project $project

  # Set the directory path for the new project
  set proj_dir [get_property directory [current_project]]

  # Set project properties
  set_property "default_lib" "xil_defaultlib" $project
  set_property "part" $part $project
  set_property "simulator_language" "Mixed" $project
  set_property "target_language" "Verilog" $project

  set repo_paths [list \
    [file normalize "$origin_dir/../../../../fpga/repo/vivado-2015.1"] \
    [file normalize "$origin_dir/../../../../fpga/repo/vivado-2016.2"] \
  ]
  set_property "ip_repo_paths" $repo_paths $project
  update_ip_catalog

  # Create 'sources_1' fileset (if not found)
  if {[string equal [get_filesets -quiet sources_1] ""]} {
    create_fileset -srcset sources_1
  }
  set source_set [get_filesets sources_1]

  # Add HDL source files
  set hdl_files [list \
    [file normalize "$origin_dir/../src/adb3_admxrcku1_wrap.vhd"] \
    [file normalize "$origin_dir/../src/dma_demo_pcie.vhd"] \
  ]
  add_files -norecurse -fileset $source_set $hdl_files

  # Create IP
  create_ip -vlnv "alphadata:user:adb3_admxrcku1_x8_axi4:${p5hi_ver}" -module_name adb3_admxrcku1_x8_axi4_ip
  set adb3_admxrcku1_x8_axi4_ip [get_ips adb3_admxrcku1_x8_axi4_ip]
  set_property -dict { \
    {CONFIG.use_diff_ref_clock} {true} \
    {CONFIG.number_of_dma_engines} {4} \
  } $adb3_admxrcku1_x8_axi4_ip

  create_ip -vlnv "alpha-data.com:user:ram_interface_axi4:${ram_if_axi_ver}" -module_name ram_interface_axi4_ip
  set ram_interface_axi4_ip [get_ips ram_interface_axi4_ip]
  set_property -dict { \
    CONFIG.ram_order {14} \
    CONFIG.ram_read_latency {3} \
    CONFIG.s_axi_axaddr_width {19} \
    CONFIG.s_axi_xdata_width {256} \
  } $ram_interface_axi4_ip
  
  create_ip -vlnv "xilinx.com:ip:axi_crossbar:*" -module_name axi_crossbar_ip
  set axi_crossbar_ip [get_ips axi_crossbar_ip]
  set_property -dict { \
    CONFIG.CLKIF.FREQ_HZ {250000000} \
    CONFIG.NUM_SI {1} \
    CONFIG.NUM_MI {4} \
    CONFIG.STRATEGY {2} \
    CONFIG.PROTOCOL {AXI4} \
    CONFIG.ADDR_WIDTH {64} \
    CONFIG.DATA_WIDTH {256} \
    CONFIG.ID_WIDTH {0} \
    CONFIG.CONNECTIVITY_MODE {SAMD} \
    CONFIG.ADDR_RANGES {1} \
    CONFIG.M00_A00_BASE_ADDR {0x0000000000000000} \
    CONFIG.M01_A00_BASE_ADDR {0x0000000000080000} \
    CONFIG.M02_A00_BASE_ADDR {0x0000000000100000} \
    CONFIG.M03_A00_BASE_ADDR {0x0000000000180000} \
    CONFIG.M00_A00_ADDR_WIDTH {19} \
    CONFIG.M01_A00_ADDR_WIDTH {19} \
    CONFIG.M02_A00_ADDR_WIDTH {19} \
    CONFIG.M03_A00_ADDR_WIDTH {19} \
  } $axi_crossbar_ip

  create_ip -vlnv "xilinx.com:ip:blk_mem_gen:*" -module_name blk_mem_gen_ip
  set blk_mem_gen_ip [get_ips blk_mem_gen_ip]
  set_property -dict { \
    CONFIG.AXI_Slave_Type {Memory_Slave}  \
    CONFIG.AXI_Type {AXI4_Full} \
    CONFIG.Additional_Inputs_for_Power_Estimation {false} \
    CONFIG.Algorithm {Minimum_Area} \
    CONFIG.Byte_Size {8} \
    CONFIG.CTRL_ECC_ALGO {NONE} \
    CONFIG.EN_SLEEP_PIN {false} \
    CONFIG.Enable_32bit_Address {false} \
    CONFIG.Enable_A {Use_ENA_Pin} \
    CONFIG.Enable_B {Use_ENB_Pin} \
    CONFIG.Fill_Remaining_Memory_Locations {false} \
    CONFIG.Interface_Type {Native} \
    CONFIG.Load_Init_File {false} \
    CONFIG.Memory_Type {True_Dual_Port_RAM} \
    CONFIG.Operating_Mode_A {WRITE_FIRST} \
    CONFIG.Operating_Mode_B {WRITE_FIRST} \
    CONFIG.Pipeline_Stages {0} \
    CONFIG.Port_A_Clock {100} \
    CONFIG.Port_A_Enable_Rate {100} \
    CONFIG.Port_A_Write_Rate {50} \
    CONFIG.Port_B_Clock {100} \
    CONFIG.Port_B_Enable_Rate {100} \
    CONFIG.Port_B_Write_Rate {50} \
    CONFIG.Read_Width_A {256} \
    CONFIG.Read_Width_B {256} \
    CONFIG.Register_PortA_Output_of_Memory_Core {true} \
    CONFIG.Register_PortA_Output_of_Memory_Primitives {true} \
    CONFIG.Register_PortB_Output_of_Memory_Core {true} \
    CONFIG.Register_PortB_Output_of_Memory_Primitives {true} \
    CONFIG.Use_Byte_Write_Enable {true} \
    CONFIG.Use_REGCEA_Pin {false} \
    CONFIG.Use_RSTA_Pin {false} \
    CONFIG.Use_RSTB_Pin {false} \
    CONFIG.Write_Depth_A {16384} \
    CONFIG.Write_Width_A {256} \
    CONFIG.Write_Width_B {256} \
    CONFIG.use_bram_block {Stand_Alone} \
  } $blk_mem_gen_ip

  # Add constraints files
  # Put target .xdc as LAST in list
  set constraint_set [get_filesets constrs_1]
  set constraint_files [list \
    [file normalize "$origin_dir/../src/refclk300.xdc"] \
    [file normalize "$origin_dir/../src/adb3_admxrcku1-pcie_x0y0.xdc"] \
    [file normalize "$origin_dir/../src/bitstream.xdc"] \
    [file normalize "$origin_dir/../src/dma_demo_pcie.xdc"] \
  ]
  add_files -norecurse -fileset $constraint_set $constraint_files
  set_property "target_constrs_file" [lindex $constraint_files end] $constraint_set

  # Add utilities files (Tcl hook scripts etc.)
  set utils_set [get_filesets -quiet utils_1]
  if {[string equal $utils_set ""]} {
    # Utilities fileset introduced in Vivado 2018.3; fall back to using sources set if not present.
    set utils_set $source_set
  }
  #set utils_files [list \
  #  [file normalize "$origin_dir/write_bitstream_maxthread.tcl"] \
  #]
  #add_files -norecurse -fileset $utils_set $utils_files
  ## Set used_in_XXX properties so that Vivado 2018.2 and earlier doesn't treat them as design sources.
  #set_property -dict {
  #  {used_in_synthesis} {false} \
  #  {used_in_implementation} {false} \
  #  {used_in_simulation} {false} \
  #} [get_files $utils_files]

  set sim_set [get_filesets sim_1]

  # Add simulation HDL source files
  set sim_files [list \
    [file normalize "$origin_dir/../src/axi4_256_pkg.vhd"] \
    [file normalize "$origin_dir/../src/axi4_256_tb_pkg.vhd"] \
    [file normalize "$origin_dir/../src/tb_dma_demo_pcie_stim.vhd"] \
    [file normalize "$origin_dir/../src/adb3_admxrcku1_sim.vhd"] \
    [file normalize "$origin_dir/../src/tb_dma_demo_pcie.vhd"] \
  ]
  add_files -norecurse -fileset $sim_set $sim_files

  # Set simulation properties
  set_property "source_set" "sources_1" $sim_set
  set_property "top" "tb_dma_demo_pcie" $sim_set
  set_property "generic" "sim_axi=true" $sim_set
  # XSIM incremental compilation seems unreliable for large projects
  set_property "incremental" "false" $sim_set
  set_property "runtime" "50 us" $sim_set
  # Set runtime on per-simulator basis
  foreach simulator [get_simulators] {
    set_property "target_simulator" "$simulator" $project
    set_property "${simulator}.simulate.runtime" "50 us" $sim_set
  }
  set_property "xsim.simulate.custom_tcl" [file normalize "$origin_dir/xsim_wave.tcl"] $sim_set
  set_property "modelsim.simulate.custom_udo" [file normalize "$origin_dir/vsim_wave.udo"] $sim_set
  # Return to using XSim
  set_property target_simulator {XSim} $project

  set synth_run [get_runs synth_1]
  #set_property strategy {Flow_PerfOptimized_high} $synth_run
  set_property "needs_refresh" "1" $synth_run
  set_property "part" $part $synth_run
  current_run -synthesis $synth_run

  set impl_run [get_runs impl_1]
  set_property "needs_refresh" "1" $impl_run
  set_property "part" $part $impl_run
  set_property {steps.phys_opt_design.is_enabled} false $impl_run
  set_property {steps.post_route_phys_opt_design.is_enabled} false $impl_run
  set_property -name {steps.route_design.args.more options} -value -timing_summary -objects $impl_run
  current_run -implementation $impl_run

  update_compile_order -fileset $source_set
  if { [info exists sim_set] } {
    update_compile_order -fileset $sim_set
  }

  return $project
}

return [make_project]
