library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library std;
use std.textio.all;

library work;
use work.ddr4_sdram_model.all;
use work.ddr4_sdram_model_int.all;

entity ddr4_sdram_bank_state is
  generic(
    model_name          : in    string;
    bank                : in    natural;
    bank_group          : in    natural;
    part                : in    part_t
  );
  port(
    reset_n             : in    std_logic; -- Should be _X01 version of signal
    ck_t                : in    std_logic;
    cke                 : in    std_logic; -- Should be _X01 version of signal
    cs_n                : in    std_logic; -- Should be _X01 version of signal
    act_n               : in    std_logic; -- Should be _X01 version of signal
    a                   : in    std_logic_vector(part.pinout.a_width - 1 downto 0);  -- Should be _01 version of signal
    ba                  : in    std_logic_vector(part.pinout.ba_width - 1 downto 0); -- Should be _01 version of signal
    bg                  : in    std_logic_vector(part.pinout.bg_width - 1 downto 0); -- Should be _01 version of signal
    ras_n               : in    std_logic; -- Should be _X01 version of signal
    cas_n               : in    std_logic; -- Should be _X01 version of signal
    we_n                : in    std_logic; -- Should be _X01 version of signal
    bank_state          : out   bank_state_t;
    activated_row       : out   natural := 0
  );
end entity;

architecture behav of ddr4_sdram_bank_state is

  signal idle : boolean := true;
  signal activated : boolean := false;
  signal precharged : boolean := true;

  signal precharge_cmd : boolean;
  signal precharge_all : boolean;
  signal precharge_bank : boolean;
  signal activate_cmd : boolean;
  
  
begin

  -- Generate bank state output from internal signals
  bank_state <= bank_state_idle       when     idle and     precharged else
                bank_state_activating when not idle and not activated  else
                bank_state_activated  when not idle and     activated  else
                bank_state_precharging;
  
  
  --
  -- Keep track of bank state.
  -- A bank is considered to be in one of two major states: idle or not idle.
  -- In the idle state, the t_rp has either elapsed or not, indicated by 'precharged'.
  -- In the not-idle state, the t_act has either elapsed or not, indicated by 'activated'.
  -- Reset forces the bank to idle, and 'precharged' to true.
  --
  
  precharge_cmd <= cke = '1' and cs_n = '0' and act_n = '1' and ras_n = '0' and cas_n = '1' and we_n = '0';
  precharge_bank <= a(10) = '0' and ba = bank and bg = bank_group;
  precharge_all <= a(10) = '1';
  
  activate_cmd <= cke = '1' and cs_n = '0' and act_n = '0' and ba = bank and bg = bank_group;
  
  gen_bank_state : process(reset_n, ck_t)
  begin
    if reset_n = '0' then
      idle <= true;
      activated <= false;
      precharged <= true;
      activated_row <= 0;
    elsif rising_edge(ck_t) then
      case idle is
        when true =>
          if activate_cmd then
            if not precharged then
              report model_name & ": ACT command while bank G" & natural'image(bank_group) & "B" & natural'image(bank) & " precharging (not ignoring this command)" severity warning;
            end if;
            idle <= false;
            activated <= true after part.timing.operational.t_rcd;
            activated_row <= to_integer(unsigned(a(part.pinout.row_width - 1 downto 0)));
            precharged <= false;
          end if;
          if precharge_cmd and (precharge_bank or precharge_all) then
            -- Receiving another precharge command, while already in the process of precharging, resets the timer.
            precharged <= false;
            precharged <= true after part.timing.operational.t_rp;
          end if;
          
        when false =>
          if precharge_cmd and (precharge_bank or precharge_all) then
            if not activated then
              report model_name & ": PRE/PREA command while bank G" & natural'image(bank_group) & "B" & natural'image(bank) & " still activating (not ignoring this command)." severity warning;
            end if;
            idle <= true;
            activated <= false;
            activated_row <= 0;
            precharged <= true after part.timing.operational.t_rp;
          end if;
          if activate_cmd then
            if activated then
              report model_name & ": Ignoring ACT command while bank G" & natural'image(bank_group) & "B" & natural'image(bank) & " activated." severity warning;
            else
              report model_name & ": Ignoring ACT command while bank G" & natural'image(bank_group) & "B" & natural'image(bank) & " activating." severity warning;
            end if;
          end if;
      end case;
    end if;
  end process;
  
  
end architecture;
