`timescale 1ns/1ps

module ddr4_sdram_data_sv(
  reset_n,
  ck_t,
  cke,
  cs_n,
  act_n,
  a_col,
  a_ap,
  a_bc,
  ba,
  bg,
  ras_n,
  cas_n,
  we_n,
  dq,
  dm_dbi_n,
  dqs_t,
  dqs_c,
  bank_states_vec,
  activated_rows_vec,
  cl_vec,
  al_vec,
  cwl_vec,
  bl_vec,
  bt,
  read_preamble,
  rd_dbi_enabled,
  wr_dbi_enabled
);

const bit false = 0;
const bit true = 1;

parameter string model_name = "ddr4_sdram";
parameter int part_pinout_row_width = 0;
parameter int part_pinout_col_width = 0;
parameter int part_pinout_ba_width = 0;
parameter int part_pinout_bg_width = 0;
parameter int part_pinout_dq_nibbles = 0;
parameter int part_pinout_dm_dbi_width = 0;
parameter int part_pinout_dm_supported = 0;
parameter int part_pinout_dqs_width = 0;
parameter int num_bank_total = 0;
parameter read_undef_val = 1'bx;

localparam data_width = 4 * part_pinout_dq_nibbles;

input reset_n;
input ck_t;
input cke;
input cs_n;
input act_n;
input [part_pinout_col_width - 1 : 0] a_col;
input a_ap;
input a_bc;
input [part_pinout_ba_width - 1 : 0] ba;
input [part_pinout_bg_width - 1 : 0] bg;
input ras_n;
input cas_n;
input we_n;
inout [4 * part_pinout_dq_nibbles - 1 : 0] dq;
inout [part_pinout_dm_dbi_width - 1 : 0] dm_dbi_n;
inout [part_pinout_dqs_width - 1 : 0] dqs_t;
inout [part_pinout_dqs_width - 1 : 0] dqs_c;
input [2 * num_bank_total - 1 : 0] bank_states_vec;
input [part_pinout_row_width * num_bank_total - 1 : 0] activated_rows_vec;
input [5:0] cl_vec;
input [5:0] al_vec;
input [5:0] cwl_vec;
input [1:0] bl_vec;
input bit bt;
input bit read_preamble;
input bit rd_dbi_enabled;
input bit wr_dbi_enabled;

localparam actual_burst_length_4 = 0;
localparam actual_burst_length_8 = 1;

localparam burst_type_nibble_seq = 0;
localparam burst_type_interleave = 1;

localparam burst_length_bl8 = 0;
localparam burst_length_otf = 1;
localparam burst_length_bc4 = 2;

localparam read_preamble_1ck = 0;
localparam read_preamble_2ck = 1;

localparam min_cl = 9;
localparam max_cl = 32;
localparam min_cwl = 9;
localparam max_cwl = 20;
localparam min_al = 0;
localparam max_al = max_cl - 1;
localparam min_rl = min_cl + min_al;
localparam max_rl = max_cl + max_al;
localparam min_wl = min_cwl + min_al;
localparam max_wl = max_cwl + max_al;
// Min/max read or write latency
localparam min_rwl = min_rl < min_wl ? min_rl : min_wl;
localparam max_rwl = max_rl > max_wl ? max_rl : max_wl;

const int bl4_seq_offsets[0 : 3][0 : 3] = {
  { 0, 1, 2, 3 },
  { 1, 2, 3, 0 },
  { 2, 3, 0, 1 },
  { 3, 0, 1, 2 }
};
const int bl4_ilv_offsets[0 : 3][0 : 3] = {
  { 0, 1, 2, 3 },
  { 1, 0, 3, 2 },
  { 2, 3, 0, 1 },
  { 3, 2, 1, 0 }
};
const int bl8_seq_offsets[0 : 7][0 : 7] = {
  { 0, 1, 2, 3, 4, 5, 6, 7 },
  { 1, 2, 3, 0, 5, 6, 7, 4 },
  { 2, 3, 0, 1, 6, 7, 4, 5 },
  { 3, 0, 1, 2, 7, 4, 5, 6 },
  { 4, 5, 6, 7, 0, 1, 2, 3 },
  { 5, 6, 7, 4, 1, 2, 3, 0 },
  { 6, 7, 4, 5, 2, 3, 0, 1 },
  { 7, 4, 5, 6, 3, 0, 1, 2 }
};
const int bl8_ilv_offsets[0 : 7][0 : 7] = {
  { 0, 1, 2, 3, 4, 5, 6, 7 },
  { 1, 0, 3, 2, 5, 4, 7, 6 },
  { 2, 3, 0, 1, 6, 7, 4, 5 },
  { 3, 2, 1, 0, 7, 6, 5, 4 },
  { 4, 5, 6, 7, 0, 1, 2, 3 },
  { 5, 4, 7, 6, 1, 0, 3, 2 },
  { 6, 7, 4, 5, 2, 3, 0, 1 },
  { 7, 6, 5, 4, 3, 2, 1, 0 }
};

function logic compute_dbi_byte(
  input [7:0] data_byte);
  
  int num_zero;
  integer i;
  
  num_zero = 0;
  for (i = 0; i < 8; i++) begin
    if (data_byte[i] == 0) begin
      num_zero++;
    end
  end
  
  if (num_zero > 4) begin
    return 1'b0;
  end else begin
    return 1'b1;
  end
endfunction

function logic [part_pinout_dm_dbi_width - 1 : 0] compute_dbi(
  input [data_width - 1 : 0] data);
  
  logic [part_pinout_dm_dbi_width - 1 : 0] dbi;
  integer lane;
  
  //assert data'length mod 8 = 0 severity failure; // Sanity check: vector length must be integer number of bytes
  for (lane = 0; lane < part_pinout_dm_dbi_width; lane++) begin
    dbi[lane] = compute_dbi_byte(data[8 * (lane + 1) - 1 -: 8]);
  end
  return dbi;
endfunction

// Convert starting column address plus beat # (0..3 or 0..7) into a word address for addressing the column's RAM array
function int gen_rd_burst_addr(
  input int column, beat,
  input bit bl, bt);

  int base, offset;
  
  if (bl == actual_burst_length_8) begin
    offset = column % 8;
    base = column - offset;
    if (bt == burst_type_nibble_seq) begin
      return base + bl8_seq_offsets[offset][beat];
    end else begin
      return base + bl8_ilv_offsets[offset][beat];
    end
  end else begin
    // assert beat < 4 severity failure;
    offset = column % 4;
    base = column - offset;
    if (bt == burst_type_nibble_seq) begin
      return base + bl4_seq_offsets[offset][beat];
    end else begin
      return base + bl4_ilv_offsets[offset][beat];
    end
  end
endfunction;

// Convert starting column address plus beat # (0..3 or 0..7) into a word address for addressing the column's RAM array
function int gen_wr_burst_addr(
  input int column, beat,
  input bit bl);
  
  int base, offset;
  
  if (bl == actual_burst_length_8) begin
    // assert beat < 8 severity failure;
    offset = column % 8;
  end else begin
    // assert beat < 4 severity failure;
    offset = column % 4;
  end
  base = column - offset;
  return base + beat;
endfunction;

typedef struct {
  bit valid;
  bit is_write;
  int bank_index;
  int row;
  int column;
  bit ap;
  bit bl;
  bit bt;
} pipe_element_t;
const pipe_element_t pipe_element_invalid = '{ false, false, 0, 0, 0, false, actual_burst_length_8, burst_type_nibble_seq };

int rl;
int wl;
bit otf_bl;
bit actual_bl;
int bank_index;
int activated_row;
bit cmd;
bit write_cmd;
bit autoprecharge;
pipe_element_t pipe[0 : max_rwl];
pipe_element_t read_pipe_sel;
pipe_element_t write_pipe_sel;

typedef struct {
  bit valid;
  int bank_index;
  int row;
  int column;
  bit bl;
  bit bt;
  int pos;
} burst_info_t;
const burst_info_t burst_info_invalid = '{ false, 0, 0, 0, actual_burst_length_8, burst_type_nibble_seq, 0};

typedef logic [4 * part_pinout_dq_nibbles - 1 : 0] memory_location_t;
const memory_location_t memory_location_undef = '{default:read_undef_val};
typedef logic [part_pinout_dqs_width - 1 : 0] memory_dbi_t;

burst_info_t read_burst = burst_info_invalid;
burst_info_t read_burst_f = burst_info_invalid;
burst_info_t read_burst_r = burst_info_invalid;
bit drive_dq = 0;
memory_location_t dq_out;
bit drive_dm_dbi = 0;
memory_dbi_t dm_dbi_out;

bit [5:0] drive_dqs_pipe = 0;
bit drive_dqs;
bit drive_dqs_r = 0;
bit drive_dqs_f = 0;
logic dqs_out_t;
bit generate_dqs_r;

burst_info_t write_burst = burst_info_invalid;
burst_info_t write_burst_f = burst_info_invalid;
burst_info_t write_burst_r = burst_info_invalid;

localparam num_bank_per_group = 1 << part_pinout_ba_width;

localparam column_size = 1 << part_pinout_col_width;
class memory_column_t;
  memory_location_t locations[0 : column_size - 1];
  
  function new();
    integer i;
    for (i = 0; i < column_size; i++) begin
      locations[i] = memory_location_undef;
    end
  endfunction
endclass

localparam row_count = 1 << part_pinout_row_width;
class memory_rows_t;
  memory_column_t columns[0 : row_count - 1];

  function new();
    // Nothing to do yet
  endfunction
endclass
  
memory_rows_t memory_array[0 : num_bank_total - 1];

bit ignore_dm;


assign ignore_dm = !part_pinout_dm_supported || wr_dbi_enabled; // TO DO: also take into account TDQS enable in mode regs.
  
assign cmd = (cke == 1) && (cs_n == 0) && (act_n == 1) && (ras_n == 1) && (cas_n == 0);
assign write_cmd = (we_n == 0);
assign autoprecharge = (a_ap == 1);
  
assign otf_bl = (a_bc == 1) ? actual_burst_length_4 : actual_burst_length_8;

always @ (bl_vec, otf_bl) begin
  if (bl_vec == burst_length_bl8) begin
    actual_bl <= actual_burst_length_8;
  end else if (bl_vec == actual_burst_length_4) begin
    actual_bl <= actual_burst_length_4;
  end else begin
    actual_bl <= otf_bl;
  end
end

assign bank_index = bg * num_bank_per_group + ba;
assign activated_row = activated_rows_vec[part_pinout_row_width * (bank_index + 1) - 1 -: part_pinout_row_width];

assign pipe[0] = '{
  cmd,
  write_cmd,
  bank_index,
  activated_row,
  a_col,
  autoprecharge,
  actual_bl,
  bt
};
  
always @ (posedge ck_t or negedge reset_n) begin
  if (reset_n == 0) begin
    pipe[1 : max_rwl] <= '{default:pipe_element_invalid};
  end else begin
    pipe[1 : max_rwl] <= pipe[0 : max_rwl - 1];
  end
end

//
// Drive DQ on reads
//

// Read latency is CL + AL, but subtract 1 because read_burst_r is delayed by 1 ck.
assign rl = cl_vec + al_vec - 1;
assign read_pipe_sel = pipe[rl];

always @ (posedge ck_t or negedge reset_n) begin
  if (reset_n == 0) begin
    read_burst <= burst_info_invalid;
    read_burst_r <= burst_info_invalid;
  end else begin
    if (read_pipe_sel.valid && !read_pipe_sel.is_write) begin
      read_burst.valid <= true;
      read_burst.bank_index <= read_pipe_sel.bank_index;
      read_burst.row <= read_pipe_sel.row;
      read_burst.column <= read_pipe_sel.column;
      read_burst.bl <= read_pipe_sel.bl;
      read_burst.bt <= read_pipe_sel.bt;
      read_burst.pos <= 0;
    end else begin
      if (read_burst.valid) begin
        if (read_burst.bl == actual_burst_length_4) begin
          if (read_burst.pos == 2) begin
            read_burst.valid <= false;
          end
        end else begin
          if (read_burst.pos == 6) begin
            read_burst.valid <= false;
          end
        end
        read_burst.pos <= read_burst.pos + 2;
      end
    end
    
    read_burst_r <= read_burst_f;
  end
end

always @ (negedge ck_t or negedge reset_n) begin
  if (reset_n == 0) begin
    read_burst_f <= burst_info_invalid;
  end else begin
    read_burst_f <= read_burst;
  end
end

always @ (posedge ck_t or negedge reset_n) begin
  integer lane;
  int col_addr;
  memory_rows_t rows_ptr;
  memory_column_t column_ptr;
  memory_location_t dq_tmp;
  memory_dbi_t dm_dbi_tmp;
  
  if (reset_n == 0) begin
    dq_out <= '{default:read_undef_val};
    drive_dq <= false;
    dm_dbi_out <= '{default:read_undef_val};
    drive_dm_dbi <= false;
  end else begin
    if (read_burst_f.valid) begin
      col_addr = gen_rd_burst_addr(read_burst_f.column, read_burst_f.pos, read_burst_f.bl, read_burst_f.bt);
      rows_ptr = memory_array[read_burst_f.bank_index];
      if (null == rows_ptr) begin
        column_ptr = null;
      end else begin
        column_ptr = rows_ptr.columns[read_burst_f.row];
      end
      if (null == column_ptr) begin
        dq_tmp = '{default:read_undef_val};
        dm_dbi_tmp = '{default:read_undef_val};
      end else begin
        dq_tmp = column_ptr.locations[col_addr];
        if (rd_dbi_enabled) begin
          dm_dbi_tmp = compute_dbi(dq_tmp);
          for (lane = 0; lane < part_pinout_dqs_width; lane++) begin
            if (dm_dbi_tmp[lane] == 0) begin
              dq_tmp[8 * (lane + 1) - 1 -: 8] = ~dq_tmp[8 * (lane + 1) - 1 -: 8];
            end
          end
        end
      end
      dq_out <= dq_tmp;
      drive_dq <= true;
      dm_dbi_out <= dm_dbi_tmp;
      drive_dm_dbi <= rd_dbi_enabled;
    end else begin
      drive_dq <= false;
      drive_dm_dbi <= false;
    end
  end
end

always @ (negedge ck_t or negedge reset_n) begin
  integer lane;
  int col_addr;
  memory_rows_t rows_ptr;
  memory_column_t column_ptr;
  memory_location_t dq_tmp;
  memory_dbi_t dm_dbi_tmp;

  if (reset_n == 0) begin
    // Nothing to do here
  end else begin
    if (read_burst_r.valid) begin
      col_addr = gen_rd_burst_addr(read_burst_r.column, read_burst_r.pos + 1, read_burst_r.bl, read_burst_r.bt);
      rows_ptr = memory_array[read_burst_r.bank_index];
      if (null == rows_ptr) begin
        column_ptr = null;
      end else begin
        column_ptr = rows_ptr.columns[read_burst_r.row];
      end
      if (null == column_ptr) begin
        dq_tmp = '{default:read_undef_val};
        dm_dbi_tmp = '{default:read_undef_val};
      end else begin
        dq_tmp = column_ptr.locations[col_addr];
        if (rd_dbi_enabled) begin
          dm_dbi_tmp = compute_dbi(dq_tmp);
          for (lane = 0; lane < part_pinout_dqs_width; lane++) begin
            if (dm_dbi_tmp[lane] == 0) begin
              dq_tmp[8 * (lane + 1) - 1 -: 8] = ~dq_tmp[8 * (lane + 1) - 1 -: 8];
            end
          end
        end
      end
      dq_out <= dq_tmp;
      dm_dbi_out <= dm_dbi_tmp;
    end
  end
end

assign dq = drive_dq ? dq_out : {data_width{1'bZ}};
assign dm_dbi_n = drive_dm_dbi ? dm_dbi_out : {data_width{1'bZ}};


//
// Drive DQS on reads
//

// Generate DQS for read preamble and when driving DQ
always @ (posedge ck_t or negedge reset_n) begin
  if (reset_n == 0) begin
    drive_dqs_pipe <= 0;
    drive_dqs_r <= false;
    generate_dqs_r <= false;
  end else begin
    drive_dqs_pipe[0] <= pipe[rl - 2].valid && ~pipe[rl - 2].is_write;
    drive_dqs_pipe[5:1] <= drive_dqs_pipe[4:0];
    
    if (read_preamble == read_preamble_2ck) begin
      drive_dqs_r <= |drive_dqs_pipe[5:0];
    end else begin
      drive_dqs_r <= |drive_dqs_pipe[5:1];
    end
    
    generate_dqs_r <= |drive_dqs_pipe[5:1];
  end
end

always @ (negedge ck_t or negedge reset_n) begin
  if (reset_n == 0) begin
    drive_dqs_f <= false;
  end else begin
    drive_dqs_f <= drive_dqs_r;
  end
end

assign dqs_out_t = ~(generate_dqs_r && ~ck_t);
assign drive_dqs = drive_dqs_r || drive_dqs_f;
assign dqs_t = drive_dqs ? {part_pinout_dqs_width{dqs_out_t}} : {part_pinout_dqs_width{1'bZ}};
assign dqs_c = drive_dqs ? {part_pinout_dqs_width{~dqs_out_t}} : {part_pinout_dqs_width{1'bZ}};
  

//
// Writes
//

// Write latency is CWL + AL, but subtract 1 because write_burst_r is delayed by 1 ck.
assign wl = cwl_vec + al_vec - 1;
assign write_pipe_sel = pipe[wl];

always @ (posedge ck_t or negedge reset_n) begin
  if (reset_n == 0) begin
    write_burst <= burst_info_invalid;
    write_burst_r <= burst_info_invalid;
  end else begin
    if (write_pipe_sel.valid && write_pipe_sel.is_write) begin
      write_burst.valid <= true;
      write_burst.bank_index <= write_pipe_sel.bank_index;
      write_burst.row <= write_pipe_sel.row;
      write_burst.column <= write_pipe_sel.column;
      write_burst.bl <= write_pipe_sel.bl;
      write_burst.bt <= write_pipe_sel.bt;
      write_burst.pos <= 0;
    end else begin
      if (write_burst.valid) begin
        if (write_burst.bl == actual_burst_length_4) begin
          if (write_burst.pos == 2) begin
            write_burst.valid <= false;
          end
        end else begin
          if (write_burst.pos == 6) begin
            write_burst.valid <= false;
          end
        end
        write_burst.pos <= write_burst.pos + 2;
      end
    end
    
    write_burst_r <= write_burst_f;
  end
end
  
always @ (negedge ck_t or negedge reset_n) begin
  if (reset_n == 0) begin
    write_burst_f <= burst_info_invalid;
  end else begin
    write_burst_f <= write_burst;
  end
end

genvar lane;
generate
for (lane = 0; lane < part_pinout_dqs_width; lane++) begin
  
  always @ (posedge dqs_t[lane]) begin
    int col_addr;
    memory_rows_t rows_ptr;
    memory_column_t column_ptr;
    logic dm_dbi_10;
    
    // To be safe, explicitly check for dqs_t[lane] == 1 in case of posedge bugs.
    if (dqs_t[lane] == 1 && write_burst_f.valid) begin
      col_addr = gen_wr_burst_addr(write_burst_f.column, write_burst_f.pos, write_burst_f.bl);
      rows_ptr = memory_array[write_burst_f.bank_index];
      if (null == rows_ptr) begin
        //$info("write dqs_t rise: allocating rows for bank %d", write_burst_f.bank_index);
        rows_ptr = new();
        memory_array[write_burst_f.bank_index] = rows_ptr;
      end
      column_ptr = rows_ptr.columns[write_burst_f.row];
      if (null == column_ptr) begin
        column_ptr = new();
        rows_ptr.columns[write_burst_f.row] = column_ptr;
      end
      dm_dbi_10 = (dm_dbi_n[lane] == 0) ? 0 : 1;
      if (ignore_dm || dm_dbi_10 == 1) begin
        if (part_pinout_dqs_width == 1) begin
          if (wr_dbi_enabled && dm_dbi_10 == 0) begin
            column_ptr.locations[col_addr] = ~dq; // ~ will convert any Zs to Xs
          end else begin
            column_ptr.locations[col_addr] = ~(~dq); // ~ will convert any Zs to Xs
          end
        end else begin
          if (wr_dbi_enabled && dm_dbi_10 == 0) begin
            column_ptr.locations[col_addr][8 * (lane + 1) - 1 -: 8] = ~dq[8 * (lane + 1) - 1 -: 8]; // ~ will convert any Zs to Xs
          end else begin
            column_ptr.locations[col_addr][8 * (lane + 1) - 1 -: 8] = ~(~dq[8 * (lane + 1) - 1 -: 8]); // ~ will convert any Zs to Xs
          end
        end
      end
    end
  end
  
  always @ (negedge dqs_t[lane]) begin
    int col_addr;
    memory_rows_t rows_ptr;
    memory_column_t column_ptr;
    logic dm_dbi_10;
    
    // Bug in XSIM? 0 -> Z seems to be treated as negedge.
    // So explicitly check for dqs_t[lane] == 0.
    if (dqs_t[lane] == 0 && write_burst_r.valid) begin
      col_addr = gen_wr_burst_addr(write_burst_r.column, write_burst_r.pos + 1, write_burst_r.bl);
      rows_ptr = memory_array[write_burst_r.bank_index];
      if (null == rows_ptr) begin
        //$info("write dqs_t fall: allocating rows for bank %d", write_burst_r.bank_index);
        rows_ptr = new();
        memory_array[write_burst_r.bank_index] = rows_ptr;
      end
      column_ptr = rows_ptr.columns[write_burst_r.row];
      if (null == column_ptr) begin
        column_ptr = new();
        rows_ptr.columns[write_burst_r.row] = column_ptr;
      end
      dm_dbi_10 = (dm_dbi_n[lane] == 0) ? 0 : 1;
      if (ignore_dm || dm_dbi_10 == 1) begin
        if (part_pinout_dqs_width == 1) begin
          if (wr_dbi_enabled && dm_dbi_10 == 0) begin
            column_ptr.locations[col_addr] = ~dq; // ~ will convert any Zs to Xs
          end else begin
            column_ptr.locations[col_addr] = ~(~dq); // ~ will convert any Zs to Xs
          end
        end else begin
          if (wr_dbi_enabled && dm_dbi_10 == 0) begin
            column_ptr.locations[col_addr][8 * (lane + 1) - 1 -: 8] = ~dq[8 * (lane + 1) - 1 -: 8]; // ~ will convert any Zs to Xs
          end else begin
            column_ptr.locations[col_addr][8 * (lane + 1) - 1 -: 8] = ~(~dq[8 * (lane + 1) - 1 -: 8]); // ~ will convert any Zs to Xs
          end
        end
      end
    end
  end
    
end
endgenerate

endmodule
