library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library std;
use std.textio.all;

library work;
use work.ddr4_sdram_model.all;

package ddr4_sdram_model_int is

  --
  -- Internal-use datatypes
  --
  
  type boolean_array_t is array(natural range <>) of boolean;

  -- State for chip/die
  type power_state_t is (
    power_state_normal,
    power_state_powerdown,
    power_state_self_refresh);

  -- State for DLL
  type dll_state_t is (
    dll_state_unlocked,
    dll_state_resetting,
    dll_state_locked);

  -- State for ZQ calibration
  type zq_state_t is (
    zq_state_uncalibrated,
    zq_state_calibrating,
    zq_state_calibrated);

  -- State for a bank
  type bank_state_t is (
    bank_state_idle,
    bank_state_activating,
    bank_state_activated,
    bank_state_precharging);
  type bank_state_array_t is array(natural range <>) of bank_state_t;
  type bank_row_array_t is array(natural range <>) of natural;

  -- Burst lengths from MR0
  type burst_length_t is (
    burst_length_bl8,   -- Always BL8
    burst_length_otf,   -- On-the-fly BC4 or BL8
    burst_length_bc4);  -- Always BC4
    
  -- Actual burst length used for a given burst
  type actual_burst_length_t is (
    actual_burst_length_4,
    actual_burst_length_8);
    
  -- Burst types from MR0
  type burst_type_t is (
    burst_type_nibble_seq,
    burst_type_interleave);

  -- Read preamble types from MR4
  type read_preamble_t is (
    read_preamble_1ck,
    read_preamble_2ck);

  -- Latency values are bounded as follows:
  subtype cl_t  is natural range 9 to 32;
  subtype cwl_t is natural range 9 to 20;
  subtype al_t  is natural range 0 to 31; -- Can be 0, CL-2 or CL-1

  
  --
  -- Internal-use utility functions
  --
  
  -- Stand-in for ? operator that VHDL lacks
  function choice(
    constant cond : in boolean;
    constant t : in natural;
    constant f : in natural
  ) return natural;
  
  -- Compute DBI bit for a byte lane of data
  function compute_dbi_byte(
    constant data_byte : in std_logic_vector(7 downto 0)
  ) return std_logic;
  
  -- Compute DBI bits for one or more byte lanes of data
  function compute_dbi(
    constant data : in std_logic_vector
  ) return std_logic_vector;
  
  function min(
    constant a, b : in    integer
  ) return integer;
  
  function max(
    constant a, b : in    integer
  ) return integer;
  
  function and_reduce(
    constant arr : in    boolean_array_t
  ) return boolean;
  
  function or_reduce(
    constant arr : in    boolean_array_t
  ) return boolean;
  
  function make_X01(
    constant bit : in    std_logic
  ) return std_logic;
  
  function make_X01(
    constant vec : in    std_logic_vector
  ) return std_logic_vector;
  
  function make_01(
    constant bit : in    std_logic
  ) return std_logic;
  
  function make_01(
    constant vec : in    std_logic_vector
  ) return std_logic_vector;
  
  function make_10(
    constant bit : in    std_logic
  ) return std_logic;
  
  function make_10(
    constant vec : in    std_logic_vector
  ) return std_logic_vector;
  
  function classify_command(
    constant bl_otf   : in    boolean;
    constant cke_prev : in    std_logic;
    constant cke      : in    std_logic;
    constant cs_n     : in    std_logic;
    constant act_n    : in    std_logic;
    constant ras_n    : in    std_logic;
    constant cas_n    : in    std_logic;
    constant we_n     : in    std_logic;
    constant bg       : in    std_logic_vector;
    constant ba       : in    std_logic_vector;
    constant a        : in    std_logic_vector
  ) return string;
  
  -- Less detail version
  function classify_command(
    constant cke_prev : in    std_logic;
    constant cke      : in    std_logic;
    constant cs_n     : in    std_logic;
    constant act_n    : in    std_logic;
    constant ras_n    : in    std_logic;
    constant cas_n    : in    std_logic;
    constant we_n     : in    std_logic
  ) return string;
  
  function to_string_binary(
    constant vec      : in    std_logic_vector
  ) return string;

  
  --
  -- std_match stings for DDR4 commands
  -- (CKE prev., CKE, CS_n, ACT_n, RAS_n, CAS_n, WE_n)
  --
  -- For all commands except SRE, SRX, PDE, PDX & DES, we intentionally ignore the previous
  -- cycle state of CKE so that the model can recognise such commands even when the memory
  -- controller behaves badly and attempts to issue commands before properly exiting from
  -- power-down or self-refresh mode.
  --
  -- Note: Although we process such bad commands, we complain about them too.
  --
  
  subtype ddr4_cmd_t is std_logic_vector(6 downto 0);
  constant MATCH_MRS  : ddr4_cmd_t := "-101000";
  constant MATCH_REF  : ddr4_cmd_t := "-101001";
  constant MATCH_SRE  : ddr4_cmd_t := "1001001";
  constant MATCH_SRX  : ddr4_cmd_t := "0101001";
  constant MATCH_PRE  : ddr4_cmd_t := "-101010"; -- A10 = 1 => all banks (PREA), 0 => single bank (PRE)
  constant MATCH_RFU  : ddr4_cmd_t := "-101011";
  constant MATCH_ACT  : ddr4_cmd_t := "-100---";
  constant MATCH_WR   : ddr4_cmd_t := "-101100"; -- A10 = 1 => autoprecharge, A12 = 1 => burst chop
  constant MATCH_RD   : ddr4_cmd_t := "-101101"; -- A10 = 1 => autoprecharge, A12 = 1 => burst chop
  constant MATCH_NOP  : ddr4_cmd_t := "-101111";
  constant MATCH_DES  : ddr4_cmd_t := "111----";
  constant MATCH_PDE  : ddr4_cmd_t := "101----";
  constant MATCH_PDX  : ddr4_cmd_t := "011----";
  constant MATCH_ZQ   : ddr4_cmd_t := "-101110"; -- A10 = 1 => long cal (ZQCL), 0 => short cal (ZQCS)

  
  --
  -- Mode register definitions
  --
  
  subtype mode_reg_t is std_logic_vector(13 downto 0);
  type mode_reg_array_t is array(0 to 6) of mode_reg_t;
  
  subtype mode_reg_addr_t is std_logic_vector(2 downto 0);
  
  constant MR0_BL_OTF : std_logic_vector(1 downto 0) := "01";
  
  --
  -- Min. & max. read & write latencies
  --
  
  -- Min/max CAS latency
  constant min_cl : natural := 9;
  constant max_cl : natural := 32;
  -- Min/max CAS write latency
  constant min_cwl : natural := 9;
  constant max_cwl : natural := 20;
  -- Min/max additive latency
  constant min_al : natural := 0;
  constant max_al : natural := max_cl - 1;
  -- Min/max read latency
  constant min_rl : natural := min_cl + min_al;
  constant max_rl : natural := max_cl + max_al;
  -- Min/max write latency
  constant min_wl : natural := min_cwl + min_al;
  constant max_wl : natural := max_cwl + max_al;

  
  --
  -- Internal-use component definitions
  --
  
  component ddr4_sdram_die
    generic(
      -- String prepended to messages to give some indication where the message comes from (for multi-die chips)
      model_name          : in    string;
      -- Selects the DDR4 SDRAM part to be modeled.
      part                : in    part_t;
      -- The value returned (in all DQ bits) of a memory location that has not previously been written or initialized.
      read_undef_val      : in    std_logic := 'U';
      -- Assumes that simulation begins from power on, so that power on timing used; other uses reset timing.
      init_from_power_on  : in    boolean := false;
      -- Demotes warnings about deviations from recommended initialization sequence to informational messages.
      relaxed_init_timing : in    boolean := true;
      -- true  => Use SystemVerilog version of ddr4_sdram_data module to work around 2016.3 XSIM access type issue.
      -- false => Use VHDL version of ddr4_sdram_data module.
      xsim_workaround     : in    boolean := true
    );
    port(
      reset_n             : in    std_logic;
      ck_t                : in    std_logic;
      ck_c                : in    std_logic;
      ten                 : in    std_logic := '0';
      
      cke                 : in    std_logic;
      cs_n                : in    std_logic;
      odt                 : in    std_logic;
      act_n               : in    std_logic;
      -- A16 functions as RAS_n when ACT_n is high.
      -- A15 functions as CAS_n when ACT_n is high.
      -- A14 functions as WE_n when ACT_n is high.
      a                   : in    std_logic_vector(part.pinout.a_width - 1 downto 0);
      ba                  : in    std_logic_vector(part.pinout.ba_width - 1 downto 0);
      bg                  : in    std_logic_vector(part.pinout.bg_width - 1 downto 0);
      par                 : in    std_logic;
      
      -- x4: DM_DBI_n ignored.
      -- x8 & x16: DBI on same pins as DM and enabled via MR5.
      -- x8 only: DM_DBI_n[0] & TDQS_t are on same pin; TDQS function enabled via MR1 (takes precedence over DM & DBI functions).
      dm_dbi_n            : inout std_logic_vector(part.pinout.dm_dbi_width - 1 downto 0) := (others => 'U');
      -- x4 & x16: TDQS_c ignored.
      -- x8 only: TDQS_c function enabled via MR1; connect if used.
      tdqs_c              : in    std_logic := 'U';
      dq                  : inout std_logic_vector(4*part.pinout.dq_nibbles - 1 downto 0);
      dqs_t               : inout std_logic_vector(part.pinout.dqs_width - 1 downto 0);
      dqs_c               : inout std_logic_vector(part.pinout.dqs_width - 1 downto 0);
      
      alert_n             : out   std_logic
    );
  end component;

  component ddr4_sdram_bank_state
    generic(
      model_name          : in    string;
      bank_group          : in    natural;
      bank                : in    natural;
      part                : in    part_t
    );
    port(
      reset_n             : in    std_logic; -- Should be _X01 version of signal
      ck_t                : in    std_logic;
      cke                 : in    std_logic; -- Should be _X01 version of signal
      cs_n                : in    std_logic; -- Should be _X01 version of signal
      act_n               : in    std_logic; -- Should be _X01 version of signal
      a                   : in    std_logic_vector(part.pinout.a_width - 1 downto 0); -- Should be _01 version of signal
      ba                  : in    std_logic_vector(part.pinout.ba_width - 1 downto 0);  -- Should be _01 version of signal
      bg                  : in    std_logic_vector(part.pinout.bg_width - 1 downto 0);  -- Should be _01 version of signal
      ras_n               : in    std_logic; -- Should be _X01 version of signal
      cas_n               : in    std_logic; -- Should be _X01 version of signal
      we_n                : in    std_logic; -- Should be _X01 version of signal
      bank_state          : out   bank_state_t;
      activated_row       : out   natural := 0
    );
  end component;
  
  component ddr4_sdram_data
    generic(
      model_name          : in    string;
      part                : in    part_t;
      num_bank_total      : in    natural;
      read_undef_val      : in    std_logic := 'U'
    );
    port(
      reset_n             : in    std_logic; -- Should be _X01 version of signal
      ck_t                : in    std_logic;
      cke                 : in    std_logic; -- Should be _X01 version of signal
      cs_n                : in    std_logic; -- Should be _X01 version of signal
      act_n               : in    std_logic; -- Should be _X01 version of signal
      a_col               : in    std_logic_vector(part.pinout.col_width - 1 downto 0); -- Should be _01 version of signal
      a_ap                : in    std_logic; -- Should be _X01 version of signal
      a_bc                : in    std_logic; -- Should be _X01 version of signal
      ba                  : in    std_logic_vector(part.pinout.ba_width - 1 downto 0); -- Should be _01 version of signal
      bg                  : in    std_logic_vector(part.pinout.bg_width - 1 downto 0); -- Should be _01 version of signal
      ras_n               : in    std_logic; -- Should be _X01 version of signal
      cas_n               : in    std_logic; -- Should be _X01 version of signal
      we_n                : in    std_logic; -- Should be _X01 version of signal
      dq                  : inout std_logic_vector(4 * part.pinout.dq_nibbles - 1 downto 0);
      dm_dbi_n            : inout std_logic_vector(part.pinout.dm_dbi_width - 1 downto 0);
      dqs_t               : inout std_logic_vector(part.pinout.dqs_width - 1 downto 0);
      dqs_c               : inout std_logic_vector(part.pinout.dqs_width - 1 downto 0);
      bank_states         : in    bank_state_array_t;
      activated_rows      : in    bank_row_array_t;
      cl                  : in    cl_t;
      al                  : in    al_t;
      cwl                 : in    cwl_t;
      bl                  : in    burst_length_t;
      bt                  : in    burst_type_t;
      read_preamble       : in    read_preamble_t;
      rd_dbi_enabled      : in    boolean;
      wr_dbi_enabled      : in    boolean
    );
  end component;
  
    component ddr4_sdram_data_sv
      generic(
        model_name               : in    string  := "ddr4_sdram";
        part_pinout_row_width    : in    natural := 0;
        part_pinout_col_width    : in    natural := 0;
        part_pinout_ba_width     : in    natural := 0;
        part_pinout_bg_width     : in    natural := 0;
        part_pinout_dq_nibbles   : in    natural := 0;
        part_pinout_dm_dbi_width : in    natural := 0;
        part_pinout_dm_supported : in    natural := 0;
        part_pinout_dqs_width    : in    natural := 0;
        num_bank_total           : in    natural := 0;
        read_undef_val           : in    std_logic := 'X' -- SystemVerilog has only 4 logic values: 0, 1, X, Z
      );
      port(
        reset_n                  : in    std_logic; -- Should be _X01 version of signal
        ck_t                     : in    std_logic;
        cke                      : in    std_logic; -- Should be _X01 version of signal
        cs_n                     : in    std_logic; -- Should be _X01 version of signal
        act_n                    : in    std_logic; -- Should be _X01 version of signal
        a_col                    : in    std_logic_vector(part_pinout_col_width - 1 downto 0); -- Should be _01 version of signal
        a_ap                     : in    std_logic; -- Should be _X01 version of signal
        a_bc                     : in    std_logic; -- Should be _X01 version of signal
        ba                       : in    std_logic_vector(part_pinout_ba_width - 1 downto 0); -- Should be _01 version of signal
        bg                       : in    std_logic_vector(part_pinout_bg_width - 1 downto 0); -- Should be _01 version of signal
        ras_n                    : in    std_logic; -- Should be _X01 version of signal
        cas_n                    : in    std_logic; -- Should be _X01 version of signal
        we_n                     : in    std_logic; -- Should be _X01 version of signal
        dq                       : inout std_logic_vector(4 * part_pinout_dq_nibbles - 1 downto 0);
        dm_dbi_n                 : inout std_logic_vector(part_pinout_dm_dbi_width - 1 downto 0);
        dqs_t                    : inout std_logic_vector(part_pinout_dqs_width - 1 downto 0);
        dqs_c                    : inout std_logic_vector(part_pinout_dqs_width - 1 downto 0);
        bank_states_vec          : in    std_logic_vector(2 * num_bank_total - 1 downto 0);
        activated_rows_vec       : in    std_logic_vector(part_pinout_row_width * num_bank_total - 1 downto 0);
        cl_vec                   : in    std_logic_vector(5 downto 0);
        al_vec                   : in    std_logic_vector(5 downto 0);
        cwl_vec                  : in    std_logic_vector(5 downto 0);
        bl_vec                   : in    std_logic_vector(1 downto 0);
        bt                       : in    boolean;
        read_preamble            : in    boolean;
        rd_dbi_enabled           : in    boolean;
        wr_dbi_enabled           : in    boolean);
    end component;
    
  component ddr4_sdram_power_state
    generic(
      model_name          : in    string;
      part                : in    part_t
    );
    port(
      reset_n             : in    std_logic;
      ck_t                : in    std_logic;
      cke                 : in    std_logic;
      cke_prev            : in    std_logic;
      cs_n                : in    std_logic;
      act_n               : in    std_logic;
      ras_n               : in    std_logic;
      cas_n               : in    std_logic;
      we_n                : in    std_logic;
      power_state         : out   power_state_t
    );
  end component;
    
  component ddr4_sdram_init_monitor
    generic(
      model_name          : in    string;
      part                : in    part_t;
      init_from_power_on  : in    boolean;
      relaxed_init_timing : in    boolean
    );
    port(
      reset_n             : in    std_logic;                                           -- Should be _X01 version of signal
      ck_t                : in    std_logic;
      cke                 : in    std_logic;                                           -- Should be _X01 version of signal
      cs_n                : in    std_logic;                                           -- Should be _X01 version of signal
      act_n               : in    std_logic;                                           -- Should be _X01 version of signal
      ras_n               : in    std_logic;                                           -- Should be _X01 version of signal
      cas_n               : in    std_logic;                                           -- Should be _X01 version of signal
      we_n                : in    std_logic;                                           -- Should be _X01 version of signal
      a                   : in    std_logic_vector(part.pinout.a_width - 1 downto 0);  -- Should be _X01 version of signal
      ba                  : in    std_logic_vector(part.pinout.ba_width - 1 downto 0); -- Should be _X01 version of signal
      bg                  : in    std_logic_vector(part.pinout.bg_width - 1 downto 0); -- Should be _X01 version of signal
      init_done           : out   boolean;
      mr0                 : out   mode_reg_t;
      mr1                 : out   mode_reg_t;
      mr2                 : out   mode_reg_t;
      mr3                 : out   mode_reg_t;
      mr4                 : out   mode_reg_t;
      mr5                 : out   mode_reg_t;
      mr6                 : out   mode_reg_t;
      cl                  : out   cl_t;
      al                  : out   al_t;
      cwl                 : out   cwl_t;
      bl                  : out   burst_length_t;
      bt                  : out   burst_type_t;
      read_preamble       : out   read_preamble_t;
      rd_dbi_enabled      : out   boolean;
      wr_dbi_enabled      : out   boolean
    );
  end component;

  component ddr4_sdram_refresh_monitor
    generic(
      model_name          : in    string;
      part                : in    part_t;
      num_bank_total      : in    natural
    );
    port(
      reset_n             : in    std_logic; -- Should be _X01 version of signal
      ck_t                : in    std_logic;
      cke                 : in    std_logic; -- Should be _X01 version of signal
      cs_n                : in    std_logic; -- Should be _X01 version of signal
      act_n               : in    std_logic; -- Should be _X01 version of signal
      ras_n               : in    std_logic; -- Should be _X01 version of signal
      cas_n               : in    std_logic; -- Should be _X01 version of signal
      we_n                : in    std_logic; -- Should be _X01 version of signal
      bank_states         : in    bank_state_array_t(0 to num_bank_total - 1)
    );
  end component;

  component ddr4_sdram_su_h_check_bit
    generic(
      model_name          : in    string;
      clk_name            : in    string;
      sig_name            : in    string;
      posedge             : in    boolean;
      tsu                 : in    time;
      th                  : in    time
    );
    port(
      clk                 : in    std_logic;
      sig                 : in    std_logic; -- Should be _X01 version of signal
      qualified           : in    boolean
    );
  end component;

  component ddr4_sdram_su_h_check_vec
    generic(
      model_name          : in    string;
      clk_name            : in    string;
      sig_name            : in    string;
      posedge             : in    boolean;
      tsu                 : in    time;
      th                  : in    time
    );
    port(
      clk                 : in    std_logic;
      sig                 : in    std_logic_vector; -- Should be _X01 version of signal
      qualified           : in    boolean
    );
  end component;
  
  component ddr4_sdram_su_h_check_col_addr
    generic(
      model_name          : in    string;
      clk_name            : in    string;
      sig_name            : in    string;
      posedge             : in    boolean;
      tsu                 : in    time;
      th                  : in    time
    );
    port(
      clk                 : in    std_logic;
      sig                 : in    std_logic_vector; -- Should be _X01 version of signal
      qualified           : in    boolean
    );
  end component;
  
end package;

package body ddr4_sdram_model_int is

  function choice(
    constant cond : in boolean;
    constant t : in natural;
    constant f : in natural
  ) return natural is
  begin
    if cond then
      return t;
    else
      return f;
    end if;
  end;
    
  function compute_dbi_byte(
    constant data_byte : in std_logic_vector(7 downto 0)
  ) return std_logic is
    variable num_zero : natural;
  begin
    num_zero := 0;
    for i in data_byte'range loop
      if data_byte(i) = '0' then
        num_zero := num_zero + 1;
      end if;
    end loop;
    if num_zero > 4 then
      return '0';
    else
      return '1';
    end if;
  end;
  
  function compute_dbi(
    constant data : in std_logic_vector
  ) return std_logic_vector is
    constant num_lane : natural := data'length / 8;
    variable num_zero : natural;
    variable dbi : std_logic_vector(num_lane - 1 downto 0);
  begin
    assert data'length mod 8 = 0 severity failure; -- Sanity check: vector length must be integer number of bytes
    if data'ascending then
      for lane in 0 to num_lane - 1 loop
        dbi(lane) := compute_dbi_byte(data(8 * lane to 8 * (lane + 1) - 1));
      end loop;
    else
      for lane in 0 to num_lane - 1 loop
        dbi(lane) := compute_dbi_byte(data(8 * (lane + 1) - 1 downto 8 * lane));
      end loop;
    end if;
    return dbi;
  end;
  
  function min(
    constant a, b : in integer
  ) return integer is
  begin
    if a < b then
      return a;
    else
      return b;
    end if;
  end;
  
  function max(
    constant a, b : in integer
  ) return integer is
  begin
    if a > b then
      return a;
    else
      return b;
    end if;
  end;
  
  function and_reduce(
    constant arr : in boolean_array_t
  ) return boolean is
  begin
    for i in arr'range loop
      if not arr(i) then
       return false;
      end if;
    end loop;
    return true;
  end;
  
  function or_reduce(
    constant arr : in boolean_array_t
  ) return boolean is
  begin
    for i in arr'range loop
      if arr(i) then
       return true;
      end if;
    end loop;
    return false;
  end;
  
  function make_X01(
    constant bit : in    std_logic
  ) return std_logic is
  begin
    case bit is
      when '1' | 'H' =>
        return '1';
      when '0' | 'L' =>
        return '0';
      when others =>
        return 'X';
    end case;
  end;
  
  function make_X01(
    constant vec : in    std_logic_vector
  ) return std_logic_vector is
    variable result : std_logic_vector(vec'range);
  begin
    for i in vec'range loop
      result(i) := make_X01(vec(i));
    end loop;
    return result;
  end;
  
  function make_01(
    constant bit : in    std_logic
  ) return std_logic is
  begin
    case bit is
      when '1' | 'H' =>
        return '1';
      when others =>
        return '0';
    end case;
  end;
  
  function make_01(
    constant vec : in    std_logic_vector
  ) return std_logic_vector is
    variable result : std_logic_vector(vec'range);
  begin
    for i in vec'range loop
      result(i) := make_01(vec(i));
    end loop;
    return result;
  end;
  
  function make_10(
    constant vec : in    std_logic_vector
  ) return std_logic_vector is
    variable result : std_logic_vector(vec'range);
  begin
    for i in vec'range loop
      result(i) := make_10(vec(i));
    end loop;
    return result;
  end;
  
  function make_10(
    constant bit : in    std_logic
  ) return std_logic is
  begin
    case bit is
      when '0' | 'L' =>
        return '0';
      when others =>
        return '1';
    end case;
  end;
  
  function classify_command(
    constant bl_otf   : in    boolean;
    constant cke_prev : in    std_logic;
    constant cke      : in    std_logic;
    constant cs_n     : in    std_logic;
    constant act_n    : in    std_logic;
    constant ras_n    : in    std_logic;
    constant cas_n    : in    std_logic;
    constant we_n     : in    std_logic;
    constant bg       : in    std_logic_vector;
    constant ba       : in    std_logic_vector;
    constant a        : in    std_logic_vector
  ) return string is
    variable l : line;
    variable cmd : std_logic_vector(6 downto 0);
    variable mrs_addr : std_logic_vector(2 downto 0);
    variable n : natural;
  begin
    cmd := cke_prev & cke & cs_n & act_n & ras_n & cas_n & we_n;
    if std_match(cmd, MATCH_MRS) then
      write(l, string'("MRS"));
      if is_X(bg) then
        write(l, string'(", invalid BG"));
      end if;
      if is_X(ba) then
        write(l, string'(", invalid BA"));
      end if;
      mrs_addr := bg(0) & ba(1 downto 0);
      if not is_X(mrs_addr) then
        n := to_integer(unsigned(mrs_addr));
        write(l, string'(" for MR" & natural'image(n)));
      else
        write(l, string'(" for MR?"));
      end if;
      if is_X(a(13 downto 0)) then
        write(l, string'(", invalid A[13:0]"));
      end if;
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    elsif std_match(cmd, MATCH_REF) then
      write(l, string'("REF"));
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    elsif std_match(cmd, MATCH_SRE) then
      write(l, string'("SRE"));
    elsif std_match(cmd, MATCH_SRX) then
      write(l, string'("SRX"));
    elsif std_match(cmd, MATCH_PRE) then
      if a(10) = '0' then
        write(l, string'("PRE"));
      elsif a(10) = '1' then
        write(l, string'("PREA"));
        if is_X(bg) then
          write(l, string'(", invalid BG"));
        end if;
        if is_X(ba) then
          write(l, string'(", invalid BA"));
        end if;
      else
        write(l, string'("PRE?, invalid A[10]"));
      end if;
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    elsif std_match(cmd, MATCH_RFU) then
      write(l, string'("RFU"));
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    elsif std_match(cmd, MATCH_ACT) then
      write(l, string'("ACT"));
      if is_X(bg) then
        write(l, string'(", invalid BG"));
      end if;
      if is_X(ba) then
        write(l, string'(", invalid BA"));
      end if;
      if is_X(a) then
        write(l, string'(", invalid A"));
      end if;
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    elsif std_match(cmd, MATCH_WR) then
      if a(10) = '0' then
        if bl_otf then
          if a(12) = '0' then
            write(l, string'("WR OTF BL4"));
          elsif a(12) = '1' then
            write(l, string'("WR OTF BL8"));
          else
            write(l, string'("WR OTF BL?, invalid A[12]"));
          end if;
        else
          write(l, string'("WR FIXED"));
        end if;
      elsif a(10) = '1' then
        if bl_otf then
          if a(12) = '0' then
            write(l, string'("WRA OTF BL4"));
          elsif a(12) = '1' then
            write(l, string'("WRA OTF BL8"));
          else
            write(l, string'("WRA OTF BL?, invalid A[12]"));
          end if;
        else
          write(l, string'("WRA FIXED"));
        end if;
      else
        if bl_otf then
          if a(12) = '0' then
            write(l, string'("WR? OTF BL4, invalid A[10]"));
          elsif a(12) = '1' then
            write(l, string'("WR? OTF BL8, invalid A[10]"));
          else
            write(l, string'("WR? OTF BL?, invalid A[12,10]"));
          end if;
        else
          write(l, string'("WR? FIXED, invalid A[10]"));
        end if;
      end if;
      if is_X(bg) then
        write(l, string'(", invalid BG"));
      end if;
      if is_X(ba) then
        write(l, string'(", invalid BA"));
      end if;
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    elsif std_match(cmd, MATCH_RD) then
      if a(10) = '0' then
        if bl_otf then
          if a(12) = '0' then
            write(l, string'("RD OTF BL4"));
          elsif a(12) = '1' then
            write(l, string'("RD OTF BL8"));
          else
            write(l, string'("RD OTF BL?, invalid A[12]"));
          end if;
        else
          write(l, string'("RD FIXED"));
        end if;
      elsif a(10) = '1' then
        if bl_otf then
          if a(12) = '0' then
            write(l, string'("RDA OTF BL4"));
          elsif a(12) = '1' then
            write(l, string'("RDA OTF BL8"));
          else
            write(l, string'("RDA OTF, invalid A[12]"));
          end if;
        else
          write(l, string'("RDA FIXED"));
        end if;
      else
        if bl_otf then
          if a(12) = '0' then
            write(l, string'("RD? OTF BL4, invalid A[10]"));
          elsif a(12) = '1' then
            write(l, string'("RD? OTF BL8, invalid A[10]"));
          else
            write(l, string'("RD? OTF BL?, invalid A[12,10]"));
          end if;
        else
          write(l, string'("RD? FIXED, invalid A[10]"));
        end if;
      end if;
      if is_X(bg) then
        write(l, string'(", invalid BG"));
      end if;
      if is_X(ba) then
        write(l, string'(", invalid BA"));
      end if;
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    elsif std_match(cmd, MATCH_NOP) then
      write(l, string'("NOP"));
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    elsif std_match(cmd, MATCH_DES) then
      write(l, string'("DES"));
    elsif std_match(cmd, MATCH_PDE) then
      write(l, string'("PDE"));
    elsif std_match(cmd, MATCH_PDX) then
      write(l, string'("PDX"));
    elsif std_match(cmd, MATCH_ZQ) then
      if a(10) = '0' then
        write(l, string'("ZQCS"));
      elsif a(10) = '0' then
        write(l, string'("ZQCL"));
      else
        write(l, string'("ZQ??, invalid A[10]"));
      end if;
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    else
      write (l, string'("Invalid command (CKE'=" & std_logic'image(cke_prev) & " CKE=" & std_logic'image(cke) &
                        " CS_n=" & std_logic'image(cs_n) & " ACT_n=" & std_logic'image(act_n) &
                        " RAS_n=" & std_logic'image(ras_n) & " CAS_n=" & std_logic'image(cas_n) &
                        " WE_n=" & std_logic'image(we_n) & " BG=" & to_string_binary(bg) &
                        " BA=" & to_string_binary(ba) & " A=" & to_string_binary(a) & ")"));
    end if;
    
    return l.all;
  end;
  
  function classify_command(
    constant cke_prev : in    std_logic;
    constant cke      : in    std_logic;
    constant cs_n     : in    std_logic;
    constant act_n    : in    std_logic;
    constant ras_n    : in    std_logic;
    constant cas_n    : in    std_logic;
    constant we_n     : in    std_logic
  ) return string is
    variable l : line;
    variable cmd : std_logic_vector(6 downto 0);
    variable n : natural;
  begin
    cmd := cke_prev & cke & cs_n & act_n & ras_n & cas_n & we_n;
    if std_match(cmd, MATCH_MRS) then
      write(l, string'("MRS"));
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    elsif std_match(cmd, MATCH_REF) then
      write(l, string'("REF"));
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    elsif std_match(cmd, MATCH_SRE) then
      write(l, string'("SRE"));
    elsif std_match(cmd, MATCH_SRX) then
      write(l, string'("SRX"));
    elsif std_match(cmd, MATCH_PRE) then
      write(l, string'("PRE/PREA"));
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    elsif std_match(cmd, MATCH_RFU) then
      write(l, string'("RFU"));
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    elsif std_match(cmd, MATCH_ACT) then
      write(l, string'("ACT"));
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    elsif std_match(cmd, MATCH_WR) then
      write(l, string'("WR"));
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    elsif std_match(cmd, MATCH_RD) then
      write(l, string'("RD"));
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    elsif std_match(cmd, MATCH_NOP) then
      write(l, string'("NOP"));
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    elsif std_match(cmd, MATCH_DES) then
      write(l, string'("DES"));
    elsif std_match(cmd, MATCH_PDE) then
      write(l, string'("PDE"));
    elsif std_match(cmd, MATCH_PDX) then
      write(l, string'("PDX"));
    elsif std_match(cmd, MATCH_ZQ) then
      write(l, string'("ZQCS/ZQCL"));
      if cke_prev /= '1' then
        write(l, string'(", CKE not asserted in prev. cycle"));
      end if;
    else
      write (l, string'("Invalid command (CKE'=" & std_logic'image(cke_prev) & " CKE=" & std_logic'image(cke) &
                        " CS_n=" & std_logic'image(cs_n) & " ACT_n=" & std_logic'image(act_n) &
                        " RAS_n=" & std_logic'image(ras_n) & " CAS_n=" & std_logic'image(cas_n) &
                        " WE_n=" & std_logic'image(we_n)));
    end if;
    
    return l.all;
  end;
  
  function to_string_binary(
    constant vec      : in    std_logic_vector
  ) return string is
    variable s : string(1 to vec'length);
    variable j : positive;
  begin
    j := 1;
    for i in vec'range loop
      case vec(i) is
        when '0' =>
          s(j) := '0';
        when '1' =>
          s(j) := '1';
        when 'L' =>
          s(j) := 'L';
        when 'H' =>
          s(j) := 'H';
        when 'X' =>
          s(j) := 'X';
        when 'U' =>
          s(j) := 'U';
        when 'W' =>
          s(j) := 'W';
        when '-' =>
          s(j) := '-';
        when 'Z' =>
          s(j) := 'Z';
      end case;
      j := j + 1;
    end loop;
    return s;
  end;
  

end package body;

