--
-- This package defines datatypes used in the DDR4 SDRAM model, and the
-- component definition for the DDR4 SDRAM model itself.
--
-- Part definitions are provided in packages for a specific device or family
-- of devices.
--

library ieee;
use ieee.std_logic_1164.all;

package ddr4_sdram_model is

  type stack_t is (stack_none, stack_twin_die, stack_stacked);
  
  type pinout_t is record
    stack_type      : stack_t;
    stack_size      : natural range 1 to 8; -- Can be >= 1 iff stack_type /= stack_none
    a_width         : natural range 16 to 20;
    ba_width        : natural range 1 to 2;
    bg_width        : natural range 1 to 2;
    dq_nibbles      : natural range 1 to 4; -- Number of nibbles (4 bits) making up DQ
    dqs_width       : natural range 1 to 2;
    dm_dbi_width    : natural range 1 to 2;
    dm_supported    : boolean;
    tdqs_supported  : boolean;
    row_width       : natural range 1 to 20;
    col_width       : natural range 1 to 20;
  end record;

  -- Timing for recommended power-on sequence.
  type power_on_timing_t is record
    t_reset_hold    : time;   -- Min. reset assertion time.
    t_cke_reset_off : time;   -- Min. time from CKE low to reset deasserted.
  end record;
  
  -- Timing for recommended reset sequence with power uninterrupted.
  type reset_timing_t is record
    t_reset_hold    : time;   -- Min. reset assertion time.
    t_cke_reset_off : time;   -- Min. time from CKE low to reset deasserted.
  end record;
  
  -- Timing for recommended initialization sequence.
  type init_timing_t is record
    -- Specify delay from reset deasserted to CKE registered high.
    t_reset_cke     : time;    -- Min. delay from reset deasserted to CKE registered high.
    
    -- Specify delay from CK running to CKE registered high.
    n_ck_cke        : natural; -- Min. no. of cycles of CK before CKE registered high.
    t_ck_cke        : time;    -- Min. delay from CK running to CKE registered high.
    
    -- From CKE registered high to 1st MRS command must be >= MAX(nXPR * tCK, tXS).
    n_xpr           : natural; -- nXPR
    t_xs_extra      : time;    -- Added to tRFC(min) to get tXS
    
    -- DLL locking time, tDLLK.
    n_dllk          : natural; -- Max. no. of CK cycles needed for DLL to achieve lock.
    
    -- ZQ long calibration time from reset / power on.
    n_zqinit        : natural; -- Max. no. of CK cycles needed for ZQ calibration.
  end record;
  
  type operational_timing_t is record
    -- Mode register command to mode register command delay
    -- tMRD = MAX(tCK * n_mrd, t_mrd), to allow delay to be specified in cycles or as a delay in ns
    n_mrd           : natural;
    t_mrd           : time;
    
    -- Input setup time for address / control
    t_is            : time; -- tIS
    
    -- Input hold time for address / control
    t_ih            : time; -- tIH
    
    -- RAS to CAS delay
    t_rcd           : time; -- tRCD
    
    -- Row precharge time
    t_rp            : time; -- tRP
  end record;
  
  type refresh_timing_t is record
    -- Refresh cycle time; typically depends upon device size (8 Gb, 16 Gb etc.)
    t_rfc           : time; -- tRFC
    t_refi          : time; -- tREFI for normal temp. range (0C to 85C)
    t_refi_ext      : time; -- tREFI for extended temp. range (0C to 95C)
  end record;
  
  type timing_t is record
    power_on    : power_on_timing_t;
    reset       : reset_timing_t;
    init        : init_timing_t;
    operational : operational_timing_t;
    refresh     : refresh_timing_t;
  end record;
  
  type part_t is record
    timing   : timing_t;
    pinout   : pinout_t;
  end record;
  
  component ddr4_sdram_chip
    generic(
      -- String prepended to messages to give some indication where the message comes from (for multi-chip modules)
      model_name          : in    string := "ddr4_sdram_chip";
      -- Selects the DDR4 SDRAM part to be modeled.
      part                : in    part_t;
      -- The value returned (in all DQ bits) of a memory location that has not previously been written or initialized.
      read_undef_val      : in    std_logic := 'U';
      -- Assumes that simulation begins from power on, so that power on timing used; other uses reset timing.
      init_from_power_on  : in    boolean := false;
      -- Demotes warnings about deviations from recommended initialization sequence to informational messages.
      relaxed_init_timing : in    boolean := true;
      -- true  => Use SystemVerilog version of ddr4_sdram_data module to work around 2016.3 XSIM access type issue.
      -- false => Use VHDL version of ddr4_sdram_data module.
      xsim_workaround     : in    boolean := true
    );
    port(
      reset_n             : in    std_logic;
      ck_t                : in    std_logic;
      ck_c                : in    std_logic;
      ten                 : in    std_logic := '0';
      
      cke                 : in    std_logic;
      cs_n                : in    std_logic;
      odt                 : in    std_logic;
      -- C0 functions as CKE1 if traditional twin-die part.
      -- C0 used to select die if stacked-die part.
      -- C0 ignored if neither twin-die nor stacked-die part (leave unconnected).
      c0_cke1             : in    std_logic := 'U';
      -- C1 functions as CS1_n if traditional twin-die part.
      -- C1 used to select die if stacked-die part.
      -- C1 ignored if neither twin-die nor stacked-die part (leave unconnected).
      c1_cs1_n            : in    std_logic := 'U';
      -- C2 functions as ODT1 if traditional twin-die part.
      -- C2 used to select die if stacked-die part.
      -- C2 ignored if neither twin-die nor stacked-die part (leave unconnected).
      c2_odt1             : in    std_logic := 'U';
      act_n               : in    std_logic;
      -- A16 functions as RAS_n when ACT_n is high.
      -- A15 functions as CAS_n when ACT_n is high.
      -- A14 functions as WE_n when ACT_n is high.
      a                   : in    std_logic_vector(part.pinout.a_width - 1 downto 0);
      ba                  : in    std_logic_vector(part.pinout.ba_width - 1 downto 0);
      bg                  : in    std_logic_vector(part.pinout.bg_width - 1 downto 0);
      par                 : in    std_logic;
      
      -- x4: DM_DBI_n ignored.
      -- x8 & x16: DBI on same pins as DM and enabled via MR5.
      -- x8 only: DM_DBI_n[0] & TDQS_t are on same pin; TDQS function enabled via MR1 (takes precedence over DM & DBI functions).
      dm_dbi_n            : inout std_logic_vector(part.pinout.dm_dbi_width - 1 downto 0) := (others => 'U');
      -- x4 & x16: TDQS_c ignored.
      -- x8 only: TDQS_c function enabled via MR1; connect if supported by part and used in design.
      tdqs_c              : in    std_logic := 'U';
      dq                  : inout std_logic_vector(4*part.pinout.dq_nibbles - 1 downto 0);
      dqs_t               : inout std_logic_vector(part.pinout.dqs_width - 1 downto 0);
      dqs_c               : inout std_logic_vector(part.pinout.dqs_width - 1 downto 0);
      
      alert_n             : out   std_logic
    );
  end component;

  
end package;

package body ddr4_sdram_model is


end package body;