library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

entity memory_exerciser_data_check is
  generic(
    memory_data_width  : in    natural;
    memory_addr_width  : in    natural
  );
  port(                        
    rst                : in    std_logic;
    clk                : in    std_logic;
    -- Initial address
    init               : in    std_logic;
    init_offset        : in    std_logic_vector(memory_addr_width - 1 downto 0);
    -- Expected and actual data
    ce                 : in    std_logic;
    actual             : in    std_logic_vector(memory_data_width - 1 downto 0);
    expected           : in    std_logic_vector(memory_data_width - 1 downto 0);
    -- Error reporting
    error_valid        : out   std_logic;
    error              : out   std_logic;
    error_addr         : out   std_logic_vector(memory_addr_width - 1 downto 0);
    error_bits         : out   std_logic_vector(memory_data_width - 1 downto 0);
    error_data         : out   std_logic_vector(memory_data_width - 1 downto 0);
    error_lanes        : out   std_logic_vector(memory_data_width / 8 - 1 downto 0)
  );
end entity;

architecture rtl of memory_exerciser_data_check is

  signal addr : std_logic_vector(memory_addr_width - 1 downto 0);

  signal bad_bits : std_logic_vector(memory_data_width - 1 downto 0);
  signal bad_data : std_logic_vector(memory_data_width - 1 downto 0);
  signal bad_bits_valid : std_logic := '0';
  
  constant num_lanes : natural := memory_data_width / 8;
  signal bad_bits_q1 : std_logic_vector(memory_data_width - 1 downto 0);
  signal bad_data_q1 : std_logic_vector(memory_data_width - 1 downto 0);
  signal bad_lanes : std_logic_vector(num_lanes - 1 downto 0);
  signal bad_lanes_valid : std_logic := '0';
  
  constant num_lanes32 : natural := (num_lanes + 31) / 32;
  signal bad_bits_q2 : std_logic_vector(memory_data_width - 1 downto 0);
  signal bad_data_q2 : std_logic_vector(memory_data_width - 1 downto 0);
  signal bad_lanes_q : std_logic_vector(num_lanes - 1 downto 0);
  signal bad_lanes32 : std_logic_vector(num_lanes32 - 1 downto 0);
  signal bad_lanes32_valid : std_logic := '0';
  
  signal i_error : std_logic := '0';
  signal i_error_valid : std_logic := '0';
  
  
begin

  track_addr : process(clk)
  begin
    if rising_edge(clk) then
      if init = '1' then
        addr <= init_offset;
      else
        if i_error_valid = '1' then
          addr <= addr + 1;
        end if;
      end if;
    end if;
  end process;
  
  error_addr <= addr;
  
  check_lanes : process(rst, clk)
    variable tmp : std_logic_vector(num_lanes32 * 32 - 1 downto 0);
    -- Map bits of vector as follows: 0 -> 0, L -> 0, anything else -> 1
    function map_bad_bits_to_1(
      constant v : in std_logic_vector)
    return std_logic_vector is
      variable ret : std_logic_vector(v'range);
    begin
      for i in v'range loop
        if v(i) = '0' or v(i) = 'L' then
          ret(i) := '0';
        else
          ret(i) := '1';
        end if;
      end loop;
      return ret;
    end;
  begin
    if rising_edge(clk) then
      if ce = '1' then
        bad_bits <= map_bad_bits_to_1(actual xor expected);
        bad_data <= actual;
      end if;
      bad_bits_valid <= ce;
      
      if bad_bits_valid = '1' then
        for i in 0 to num_lanes - 1 loop
          bad_lanes(i) <= OR_reduce(bad_bits(8 * (i + 1) - 1 downto 8 * i));
        end loop;
        bad_bits_q1 <= bad_bits;
        bad_data_q1 <= bad_data;
      end if;
      bad_lanes_valid <= bad_bits_valid;
      
      if bad_lanes_valid = '1' then
        tmp := std_logic_vector(resize(unsigned(bad_lanes), num_lanes32 * 32));
        for i in 0 to num_lanes32 - 1 loop
          bad_lanes32(i) <= OR_reduce(tmp(32 * (i + 1) - 1 downto 32 * i));
        end loop;
        bad_lanes_q <= bad_lanes;
        bad_bits_q2 <= bad_bits_q1;
        bad_data_q2 <= bad_data_q1;
      end if;
      bad_lanes32_valid <= bad_lanes_valid;
      
      if bad_lanes32_valid = '1' then
        i_error <= OR_reduce(bad_lanes32);
        error_lanes <= bad_lanes_q;
        error_bits <= bad_bits_q2;
        error_data <= bad_data_q2;
      end if;
      i_error_valid <= bad_lanes32_valid;
    end if;
  end process;
  
  error <= i_error;
  error_valid <= i_error_valid;
  
  
end architecture;
