--
-- FIFO for storing generated data and expected data
--
-- Holds up to 2**ORDER items.
--
-- Provides 'accounting' facility, where accounting pointer can be
-- advanced without actually reading an item from the FIFO. This permits

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

entity memory_exerciser_fifo is
  generic(
    ORDER       : in    natural range 2 to 6;
    WIDTH       : in    natural);
  port(
    rst         : in    std_logic;                             -- Asynchronous reset
    clk         : in    std_logic;                             -- Clock
    sr          : in    std_logic;                             -- Synchronous reset
    -- Writer interface
    wr          : in    std_logic;                             -- '1' => write value on 'd' into FIFO
    wd          : in    std_logic_vector(WIDTH - 1 downto 0);  -- Item to write into FIFO, qualified by 'wr'
    wfull       : out   std_logic;                             -- '1' => full (from point of view of write side)
    -- Accounting interface
    aadv        : in    std_logic;                             -- Advance accounting pointer
    aempty      : out   std_logic;                             -- '1' => empty
    alevel      : out   std_logic_vector(ORDER downto 0);      -- Number of items in FIFO
    -- Reader interface
    radv        : in    std_logic;                             -- '1' => present next item on 'q' to next item
    rempty      : out   std_logic;                             -- '1' => empty
    rq          : out   std_logic_vector(WIDTH - 1 downto 0)); -- Current item, qualified by 'valid'
end;

architecture rtl of memory_exerciser_fifo is

  constant DEPTH : natural := 2**ORDER;
  constant MAX_LEVEL : natural := DEPTH - 1;

  -- Writing
  signal wptr : std_logic_vector(ORDER downto 0) := (others => '0');
  signal i_full : std_logic := '0';
  
  -- Reading
  signal rptr : std_logic_vector(ORDER downto 0) := (others => '0');
  signal rn : std_logic_vector(ORDER downto 0) := (others => '0');
  signal qce : std_logic;
  signal rd : std_logic;
  signal i_q : std_logic_vector(WIDTH - 1 downto 0);
  signal i_rempty : std_logic := '1';

  -- Accounting
  signal aptr : std_logic_vector(ORDER downto 0) := (others => '0');
  signal an : std_logic_vector(ORDER downto 0) := (others => '0');
  signal i_aempty : std_logic := '1';
  
  -- Storage for items
  subtype ram_data_t is std_logic_vector(WIDTH - 1 downto 0);
  type ram_array_t is array(0 to DEPTH - 1) of ram_data_t;
  signal ram : ram_array_t;
  attribute RAM_STYLE : string;
  attribute RAM_STYLE of ram : signal is "distributed";
  signal sel_ram : ram_data_t;

begin

  wfull <= i_full;
  rempty <= i_rempty;
  rq <= i_q;
  alevel <= an;
  aempty <= i_aempty;

  write_logic : process(rst, clk)
    variable tmp : std_logic_vector(ORDER downto 0);
  begin
    if rst = '1' then
      wptr <= (others => '0');
      i_full <= '0';
    elsif rising_edge(clk) then
      if sr = '1' then
        wptr <= (others => '0');
        i_full <= '0';
      else
        if wr = '1' then
          wptr <= wptr + 1;
        end if;
  
        if wr = '1' then
          assert conv_integer(wptr - rptr) < MAX_LEVEL severity error;
          tmp := wptr - rptr + 1;
        else
          tmp := wptr - rptr;
        end if;
        i_full <= AND_reduce(tmp(ORDER - 1 downto 2));
      end if;
    end if;

    if rising_edge(clk) then
      if wr = '1' then
        ram(conv_integer(wptr(ORDER - 1 downto 0))) <= wd;
      end if;
    end if;
  end process;

  accounting_logic : process(rst, clk)
    variable tmp : std_logic_vector(ORDER downto 0);
  begin
    if rst = '1' then
      aptr <= (others => '0');
    elsif rising_edge(clk) then
      if sr = '1' then
        aptr <= (others => '0');
      else
        if aadv = '1' then
          aptr <= aptr + 1;
        end if;
      end if;
    end if;

    if rst = '1' then
      null;
    elsif rising_edge(clk) then
      if sr = '1' then
        null;
      else
        if aadv = '1' then
          assert i_aempty = '0' severity error;
        end if;
      end if;
    end if;

    if rst = '1' then
      an <= (others => '0');
      i_aempty <= '1';
    elsif rising_edge(clk) then
      if sr = '1' then
        an <= (others => '0');
        i_aempty <= '1';
      else
        if aadv = '1' then
          tmp := wptr - aptr - 1;
        else
          tmp := wptr - aptr;
        end if;
        an <= tmp;
        i_aempty <= not OR_reduce(tmp);
      end if;
    end if;
  end process;

  qce <= radv or i_rempty;
  rd <= (radv or i_rempty) and OR_reduce(rn);

  read_logic : process(rst, clk)
  begin
    if rst = '1' then
      rptr <= (others => '0');
      rn <= (others => '0');
    elsif rising_edge(clk) then
      if sr = '1' then
        rptr <= (others => '0');
        rn <= (others => '0');
      else
        if rd = '1' then
          rptr <= rptr + 1;
        end if;
        if rd = '1' then
          rn <= wptr - rptr - 1;
        else
          rn <= wptr - rptr;
        end if;
      end if;
    end if;

    if rst = '1' then
      null;
    elsif rising_edge(clk) then
      if sr = '1' then
        null;
      else
        if rd = '1' then
          assert conv_integer(wptr - rptr) > 0 severity error;
        end if;
      end if;
    end if;
  end process;

  output_valid : process(rst, clk)
  begin
    if rst = '1' then
      i_rempty <= '1';
    elsif rising_edge(clk) then
      if sr = '1' then
        i_rempty <= '1';
      else
        if qce = '1' then
          i_rempty <= not OR_reduce(rn);
        end if;
      end if;
    end if;
  end process;

  sel_ram <= ram(conv_integer(rptr(ORDER - 1 downto 0)));

  output_data : process(clk)
  begin
    if rising_edge(clk) then
      if qce = '1' then
        -- Assign i_q to undefined on FIFO empty to assist in simulation debug only
        if OR_reduce(rn) = '1' then
          i_q <= sel_ram;
        else
          i_q <= (others => '-');
        end if;
      end if;
    end if;
  end process;

end;