--
-- adb3_ocp_cross_clock_dom.vhd
-- 
-- (C) 2009 Alpha Data Ltd.
--
-- Asychronous connection between 2 Clock Domains for ADB3 OCP Transactions


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.fifo.all;
use work.adb3_ocp.all;

entity adb3_ocp_cross_clock_dom is

  port (
    master_clk : in  std_logic;
    slave_clk  : in  std_logic;
    master_m2s : in  adb3_ocp_m2sT;
    master_s2m : out adb3_ocp_s2mT;
    slave_m2s  : out adb3_ocp_m2sT;
    slave_s2m  : in  adb3_ocp_s2mT);

end adb3_ocp_cross_clock_dom;

architecture mixed of adb3_ocp_cross_clock_dom is

 
  
  signal cmd_fifo_data_in, cmd_fifo_data_out                                                   : std_logic_vector(ADB3_OCP_TAG_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+2 downto 0);
  signal cmd_fifo_wadv, cmd_fifo_radv, cmd_fifo_empty, cmd_fifo_nfull, cmd_fifo_full           : std_logic;
  signal mdata_fifo_data_in, mdata_fifo_data_out                                               : std_logic_vector(ADB3_OCP_DATA_WIDTH+ADB3_OCP_DATA_WIDTH/8-1 downto 0);
  signal mdata_fifo_wadv, mdata_fifo_radv, mdata_fifo_empty, mdata_fifo_nfull, mdata_fifo_full : std_logic;
  signal sdata_fifo_data_in, sdata_fifo_data_out                                               : std_logic_vector(ADB3_OCP_DATA_WIDTH+1+ADB3_OCP_TAG_WIDTH downto 0);
  signal sdata_fifo_wadv, sdata_fifo_radv, sdata_fifo_empty, sdata_fifo_nfull, sdata_fifo_full : std_logic;

  signal local_rst : std_logic := '1';
  signal rst       : std_logic := '0';
  
begin  -- mixed

  -- Generate a Power on Reset for FIFOs
  lrst0 : process (master_clk)
  begin  -- process lrst0
    if master_clk'event and master_clk = '1' then     -- rising clock edge
      local_rst <= '0';
      rst       <= local_rst;
    end if;
  end process lrst0;

  
  -- Generate Asychronous FIFOs
  cmd_fifo : afifo
    generic map(
      width  => ADB3_OCP_TAG_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3,
      order  => 4,
      wpfl   => 4,
      wpfl2  => 0,
      rpel   => 0,
      family => family_virtex5,
      safety => false,
      fflush => false)
    port map(
      rst    => rst,
      wk     => master_clk,
      wd     => cmd_fifo_data_in,
      wadv   => cmd_fifo_wadv,
      we     => open,
      wpf    => cmd_fifo_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => open,
      rk     => slave_clk,
      rq     => cmd_fifo_data_out,
      radv   => cmd_fifo_radv,
      rflush => '0',
      re     => cmd_fifo_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => open);

  mdata_fifo : afifo
    generic map(
      width  => ADB3_OCP_DATA_WIDTH+ADB3_OCP_DATA_WIDTH/8,
      order  => 4,
      wpfl   => 4,
      wpfl2  => 0,
      rpel   => 2,
      family => family_virtex5,
      safety => false,
      fflush => false)
    port map(
      rst    => rst,
      wk     => master_clk,
      wd     => mdata_fifo_data_in,
      wadv   => mdata_fifo_wadv,
      we     => open,
      wpf    => mdata_fifo_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => open,
      rk     => slave_clk,
      rq     => mdata_fifo_data_out,
      radv   => mdata_fifo_radv,
      rflush => '0',
      re     => mdata_fifo_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => open);

  sdata_fifo : afifo
    generic map(
      width  => ADB3_OCP_DATA_WIDTH+2+ADB3_OCP_TAG_WIDTH,
      order  => 4,
      wpfl   => 4,
      wpfl2  => 0,
      rpel   => 2,
      family => family_virtex5,
      safety => false,
      fflush => false)
    port map(
      rst    => rst,
      wk     => slave_clk,
      wd     => sdata_fifo_data_in,
      wadv   => sdata_fifo_wadv,
      we     => open,
      wpf    => sdata_fifo_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => open,
      rk     => master_clk,
      rq     => sdata_fifo_data_out,
      radv   => sdata_fifo_radv,
      rflush => '0',
      re     => sdata_fifo_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => open);

  -- Register Nearly Full Signal
  process (master_clk)
  begin  -- process
    if master_clk'event and master_clk = '1' then  -- rising clock edge
      cmd_fifo_full   <= cmd_fifo_nfull;
      mdata_fifo_full <= mdata_fifo_nfull;

      cmd_fifo_data_in         <= master_m2s.Tag & master_m2s.BurstLength & master_m2s.Addr & master_m2s.Cmd;
      if master_m2s.Cmd /= ocp_Cmd_IDLE and cmd_fifo_full = '0' then
        cmd_fifo_wadv  <= '1';
      else
        cmd_fifo_wadv <= '0';
      end if;     


      mdata_fifo_data_in    <= master_m2s.DataByteEn & master_m2s.Data;
      mdata_fifo_wadv       <= master_m2s.DataValid and not mdata_fifo_full;
      
    end if;
  end process;
  process (slave_clk)
  begin  -- process
    if slave_clk'event and slave_clk = '1' then    -- rising clock edge
      sdata_fifo_full <= sdata_fifo_nfull;

      sdata_fifo_data_in   <= slave_s2m.Tag & slave_s2m.Resp & slave_s2m.Data;
      if slave_s2m.Resp /= ocp_Resp_NONE and sdata_fifo_full = '0' then
        sdata_fifo_wadv <= '1';
      else
        sdata_fifo_wadv <= '0';  
      end if;
    end if;
  end process;


  -- Map Control Signals to FIFOs

  --cmd_fifo_data_in         <= master_m2s.Tag & master_m2s.BurstLength & master_m2s.Addr & master_m2s.Cmd;
  --cmd_fifo_wadv            <= '1'                           when master_m2s.Cmd /= ocp_Cmd_IDLE and cmd_fifo_full = '0' else '0';
  master_s2m.CmdAccept     <= not cmd_fifo_full;
  slave_m2s.Cmd            <= cmd_fifo_data_out(2 downto 0) when cmd_fifo_empty = '0'                                   else ocp_Cmd_IDLE;
  slave_m2s.Addr           <= cmd_fifo_data_out(ADB3_OCP_ADDR_WIDTH+2 downto 3);
  slave_m2s.BurstLength    <= cmd_fifo_data_out(ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+2 downto ADB3_OCP_ADDR_WIDTH+3);
  slave_m2s.Tag <= cmd_fifo_data_out(ADB3_OCP_TAG_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+2 downto ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
  cmd_fifo_radv            <= slave_s2m.CmdAccept and not cmd_fifo_empty;

  --mdata_fifo_data_in    <= master_m2s.DataByteEn & master_m2s.Data;
  --mdata_fifo_wadv       <= master_m2s.DataValid and not mdata_fifo_full;
  master_s2m.DataAccept <= not mdata_fifo_full;
  slave_m2s.Data        <= mdata_fifo_data_out(ADB3_OCP_DATA_WIDTH-1 downto 0);
  slave_m2s.DataByteEn  <= mdata_fifo_data_out(ADB3_OCP_DATA_WIDTH+ADB3_OCP_DATA_WIDTH/8-1 downto ADB3_OCP_DATA_WIDTH);
  slave_m2s.DataValid   <= not mdata_fifo_empty;
  mdata_fifo_radv       <= slave_s2m.DataAccept and not mdata_fifo_empty;

  --sdata_fifo_data_in   <= slave_s2m.Tag & slave_s2m.Resp & slave_s2m.Data;
  --sdata_fifo_wadv <= '1'                                 when slave_s2m.Resp /= ocp_Resp_NONE and sdata_fifo_full = '0' else '0';
  slave_m2s.RespAccept <= not sdata_fifo_full;
  master_s2m.Data      <= sdata_fifo_data_out(ADB3_OCP_DATA_WIDTH-1 downto 0);
  master_s2m.Resp      <= sdata_fifo_data_out(ADB3_OCP_DATA_WIDTH+1 downto ADB3_OCP_DATA_WIDTH) when sdata_fifo_empty = '0'                                    else ocp_Resp_NONE;
  master_s2m.Tag       <= sdata_fifo_data_out(ADB3_OCP_DATA_WIDTH+1+ADB3_OCP_TAG_WIDTH downto ADB3_OCP_DATA_WIDTH+2);
  sdata_fifo_radv      <= master_m2s.RespAccept and not sdata_fifo_empty;
  
  
end mixed;
