--==============================================================================
--
--     File Name: adb3_ocp_dm_adapter_chop.vhd
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: adb3_ocp_dm_adapter
--
--   Description: Submodule of adb3_ocp_dm_adapter responsible for implementing
--                constraints R1 - R3 &  W1 - W4. In other words, this module
--                chops general OCP commands into one or more pieces that
--                satisify maximum read request and maximum payload
--                constraints as well as PCI-E payload first & last byte enable
--                rules.
--
--   Limitations:
--
--         Notes:
--
--         TO DO: 1. Pipeline the 60-bit address counters / accumulators.
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2013 Alpha Data, All rights reserved
--
--     Revisions: 0.1, TW, 14 Nov 2013: Initial design
--                0.2, TW, 18 Mar 2014:
--                  - Added PCIE_HINTS generic and modified Addr / Tag
--                    generation accordingly.
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_dm_int.all;

entity adb3_ocp_dm_adapter_chop is
generic(
  -- This parameter overrides 'ctl_max_payload' if 'ctl_max_payload' represents
  -- a number of OCP words larger than 2**MAX_WR_BURST_ORDER. This ensures
  -- that the DM data & byte enable FIFO is large enough to hold an entire
  -- burst whilst its byte enables are being analyzed.
  MAX_WR_BURST_ORDER : in    natural range 3 to 9;
  -- Number of bits used for generating tags for chopped commands
  DM_TAG_BITS        : in    natural range 1 to ADB3_OCP_TAG_WIDTH := 6;
  -- true  => Generated write bursts are also compatible with MPTL payload.
  -- false => Generated write bursts in general are not compatible with MPTL
  --          payload.
  MPTL_COMPATIBLE    : in    boolean;
  -- true  => For write bursts, low bits of master_m2s_addr indicate the
  --          first valid byte lane of the first word of the burst, and
  --          master_m2s_tag (normally ignored by slave) indicates the last
  --          valid byte of the the last word of the burst.
  -- false => Low bits of master_m2s_addr are zero and master_m2s_tag must be
  --          considered invalid unless the burst is a read.
  PCIE_HINTS         : in    boolean;
  -- OCP data width
  DATA_WIDTH         : in    natural);
port(
  ocp_rst         : in    std_logic;
  ocp_clk         : in    std_logic;
  -- Commands in from OCP-side command FIFO
  ocf_cmd         : in    ocp_CmdT;
  ocf_addr        : in    adb3_ocp_addr_s;
  ocf_blen        : in    adb3_ocp_burst_s;
  ocf_tag         : in    adb3_ocp_tag_s;
  ocf_cmd_valid   : in    std_logic;
  ocf_cmd_adv     : out   std_logic;
  -- Data in from OCP write data / byte enable FIFO
  owf_data        : in    std_logic_vector(DATA_WIDTH - 1 downto 0);
  owf_be          : in    std_logic_vector(DATA_WIDTH / 8 - 1 downto 0);
  owf_data_valid  : in    std_logic;
  owf_data_adv    : out   std_logic;
    -- Chopped commands to DM command FIFO
  chop_cmd        : out   ocp_CmdT;
  chop_addr       : out   adb3_ocp_addr_s;
  chop_blen       : out   adb3_ocp_burst_s;
  chop_ocp_tag    : out   adb3_ocp_tag_s;
  chop_dm_tag     : out   adb3_ocp_tag_s;
  chop_cmd_valid  : out   std_logic;        -- '1' => above chop_* values are valid
    -- Chopped data & byte enables to DM data FIFO
  chop_data       : out   std_logic_vector(DATA_WIDTH - 1 downto 0);
  chop_be         : out   std_logic_vector(DATA_WIDTH / 8 - 1 downto 0);
  chop_data_valid : out   std_logic;        -- '1' => above chop_* values are valid
  -- Control
  ctl_cmd_enable  : in    std_logic;                    -- '1' => allow processing of commands
  ctl_data_enable : in    std_logic;                    -- '1' => allow processing of write data & byte enables
  ctl_max_rd_req  : in    std_logic_vector(2 downto 0); -- PCIE maximum read request code, as per PCI Express specification
  ctl_max_payload : in    std_logic_vector(2 downto 0); -- PCIE maximum payload code, as per PCI Express specification
  debug_info      : out   std_logic_vector(63 downto 0));
end;


architecture syn of adb3_ocp_dm_adapter_chop is

  function match(
    constant a      : in std_logic_vector;
    constant b      : in std_logic_vector;
    constant ignore : in std_logic_vector)
  return boolean is
    constant tmp_b : std_logic_vector(a'range) := b;
    constant tmp_ignore : std_logic_vector(a'range) := ignore;
  begin
    if AND_reduce(tmp_ignore OR (NOT tmp_ignore AND NOT (a XOR tmp_b))) = '1' then
      return true;
    else
      return false;
    end if;
  end;

  function is_last_word_in_burst(
    constant remaining : std_logic_vector)
  return std_logic is
  begin
    if remaining'ascending then
      return AND_reduce(NOT remaining(1 to remaining'high));
    else
      return AND_reduce(NOT remaining(remaining'high downto 1));
    end if;
  end;

  function is_at_addr_boundary(
    constant addr      : std_logic_vector;
    constant addr_mask : std_logic_vector)
  return std_logic is
  begin
    return AND_reduce(addr(addr_mask'range) OR addr_mask);
  end;

  function is_next_last_word_in_burst(
    constant remaining : std_logic_vector)
  return std_logic is
  begin
    if remaining'ascending then
      return (AND_reduce(NOT remaining(2 to remaining'high)) AND NOT remaining(0));
    else
      return (AND_reduce(NOT remaining(remaining'high downto 2)) AND NOT remaining(0));
    end if;
  end;

   -- Generate a mask for an 8-bit slice of the byte enables (corresponding to
  -- a 64-bit slice of the data), for generating 'cp_be_mask_out' when in the
  -- 'cp_state_special' state.
  function gen_be_mask_out(
    constant lane_count : in natural;
    constant pos_1h     : in std_logic_vector)
  return std_logic_vector is
    variable mask : std_logic_vector(pos_1h'length * 8 - 1 downto 0);
  begin
    for i in pos_1h'range loop
      mask(8 * (i + 1) - 1 downto 8 * i) := (others => NOT pos_1h(i));
    end loop;
    return mask;
  end;

  -- true => instantiate ChipScope
  constant CHIPSCOPE_ON : boolean := false;

 -- Limits length of generated write bursts to DM OCP port
  constant MAX_WR_BURST_LENGTH : natural := 2**MAX_WR_BURST_ORDER;

  -- Log base 2 of number of byte lanes
  constant BE_WIDTH : natural := DATA_WIDTH / 8;
  constant LOG2_BE_WIDTH : natural := natural(floor(log2(real(BE_WIDTH) + 0.5)));

  -- Number of bits needed for generated burst length = log2(4096) - log2(BE_WIDTH) + 1
  constant GEN_BLEN_WIDTH : natural := 12 - LOG2_BE_WIDTH + 1;

  -- Number of bits needed for burst length mask = log2(4096) - log2(BE_WIDTH)
  constant BLEN_MASK_WIDTH : natural := 12 - LOG2_BE_WIDTH;

  -- Registered versions of control signals
  signal cmd_enable_q : std_logic;
  signal data_enable_q : std_logic;
  signal max_rd_req_q : std_logic_vector(2 downto 0);
  signal max_payload_q : std_logic_vector(2 downto 0);
  -- Decoded versions of control signals
  signal max_rd_req_value : std_logic_vector(GEN_BLEN_WIDTH - 1 downto 0);
  signal max_payload_mask : std_logic_vector(11 downto LOG2_BE_WIDTH);

  -- Burst length arithmetic
  signal offset_to_rd_req_boundary : std_logic_vector(GEN_BLEN_WIDTH - 1 downto 0);
  signal cross_rd_req_boundary : std_logic;
  signal max_payload_exceeded : std_logic;
  signal constrained_blen_rd : std_logic_vector(GEN_BLEN_WIDTH - 1 downto 0);

  -- Pre-analysis of byte enables, into 4-bit groups
  constant NUM_BE_GROUP : natural := BE_WIDTH / 4;
  constant LOG2_NUM_BE_GROUP : natural := LOG2_BE_WIDTH - 2;
  signal be_grp_not_empty : std_logic_vector(NUM_BE_GROUP - 1 downto 0);
  signal be_grp_full : std_logic_vector(NUM_BE_GROUP - 1 downto 0);
  signal be_grp_left : std_logic_vector(NUM_BE_GROUP - 1 downto 0);
  signal be_grp_left_or_full : std_logic_vector(NUM_BE_GROUP - 1 downto 0);
  signal be_grp_right : std_logic_vector(NUM_BE_GROUP - 1 downto 0);
  signal be_grp_right_or_full : std_logic_vector(NUM_BE_GROUP - 1 downto 0);
  signal be_grp_0to1_one : std_logic_vector(NUM_BE_GROUP - 1 downto 0);
  signal be_grp_0to1_two : std_logic_vector(NUM_BE_GROUP - 1 downto 0);

  -- Overall classification of byte enables
  signal be_class_en : std_logic; -- Clock enable for be_* signals below
  signal be_class_not_empty : std_logic;
  signal be_grp2_not_empty_q : std_logic_vector(NUM_BE_GROUP / 2 - 1 downto 0);
  signal be_class_full : std_logic;
  signal be_class_left : std_logic;
  signal be_class_right : std_logic;
  signal be_class_center : std_logic;
  signal be_class_be : std_logic_vector(BE_WIDTH - 1 downto 0); -- The actual byte enables
  signal be_class_data : std_logic_vector(DATA_WIDTH - 1 downto 0); -- Data accompanying byte enables
  signal be_class_valid : std_logic; -- Qualifies above be_* signals
  signal be_class_first_idx : std_logic_vector(LOG2_BE_WIDTH - 1 downto 0); -- Index of first (rightmost) occupied byte lane
  signal be_class_last_idx : std_logic_vector(LOG2_BE_WIDTH - 1 downto 0); -- Index of last (leftmost) occupied byte lane

  -- Command processing state machine
  type burst_action_t is (action_normal, action_end_burst, action_new_burst, action_suppress, action_special);
  type cp_state_t is (cp_state_idle, cp_state_read1, cp_state_read2, cp_state_write_normal, cp_state_write_general);
  signal cp_state : cp_state_t := cp_state_idle;
  signal n_cp_state : cp_state_t;
  signal cp_rd_addr, n_cp_rd_addr : std_logic_vector(ADB3_OCP_ADDR_WIDTH - 1 downto LOG2_BE_WIDTH);
  signal cp_wr_addr, n_cp_wr_addr : std_logic_vector(ADB3_OCP_ADDR_WIDTH - 1 downto LOG2_BE_WIDTH);
  signal cp_wr_start_addr, n_cp_wr_start_addr : std_logic_vector(ADB3_OCP_ADDR_WIDTH - 1 downto LOG2_BE_WIDTH);
  signal cp_remaining, n_cp_remaining : adb3_ocp_burst_s;
  signal cp_gen_cmd, n_cp_gen_cmd : ocp_CmdT;
  signal cp_gen_addr, n_cp_gen_addr : std_logic_vector(ADB3_OCP_ADDR_WIDTH - 1 downto LOG2_BE_WIDTH);
  signal cp_gen_blen, n_cp_gen_blen : std_logic_vector(GEN_BLEN_WIDTH - 1 downto 0);
  signal cp_gen_tag : adb3_ocp_tag_s := (others => '0');
  signal cp_gen_cmd_valid : std_logic := '0';
  signal n_cp_gen_cmd_valid : std_logic;
  signal cp_ocp_tag, n_cp_ocp_tag : adb3_ocp_tag_s;
  signal cp_bcount, n_cp_bcount : std_logic_vector(GEN_BLEN_WIDTH - 1 downto 0);
  signal cp_bfirst, n_cp_bfirst : std_logic;
  signal cp_blast, n_cp_blast : std_logic;
  signal cp_cmd_acc : std_logic := '0';
  signal n_cp_cmd_acc : std_logic;
  signal cp_be_mask_out, n_cp_be_mask_out : std_logic_vector(BE_WIDTH - 1 downto 0);
  signal cp_data_valid : std_logic := '0';
  signal n_cp_data_valid : std_logic;
  signal cp_data_acc : std_logic := '0';
  signal n_cp_data_acc : std_logic;
  signal cp_special_pos, n_cp_special_pos : std_logic_vector(LOG2_NUM_BE_GROUP - 2 downto 0);
  signal cp_special_pos_1h, n_cp_special_pos_1h : std_logic_vector(NUM_BE_GROUP / 2 - 1 downto 0);
  signal cp_be_first_idx, n_cp_be_first_idx : std_logic_vector(LOG2_BE_WIDTH - 1 downto 0);
  signal cp_be_last_idx, n_cp_be_last_idx : std_logic_vector(LOG2_BE_WIDTH - 1 downto 0);
  signal cp_gen_1_qword, n_cp_gen_1_qword : std_logic;
  -- Signals associated with command processing state machine
  signal cp_debug_action, n_cp_debug_action : burst_action_t;
  signal cp_discriminant : std_logic_vector(6 downto 0);
  signal cp_addr_rd_masked : std_logic_vector(11 downto LOG2_BE_WIDTH);
  signal cp_bbound : std_logic;
  signal cp_data_out : std_logic_vector(DATA_WIDTH - 1 downto 0);
  signal cp_be_out : std_logic_vector(BE_WIDTH - 1 downto 0);

  -- Debug vector
  signal i_debug_info : std_logic_vector(debug_info'range) := (others => '0');

begin

  debug_info <= i_debug_info;

  debug_state : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      i_debug_info(0) <= data_enable_q;
      i_debug_info(1) <= cmd_enable_q;
      i_debug_info(4 downto 2) <= max_rd_req_q;
      i_debug_info(7 downto 5) <= max_payload_q;
      case cp_state is
        when cp_state_idle =>
          i_debug_info(11 downto 8) <= "0000";

        when cp_state_read1 =>
          i_debug_info(11 downto 8) <= "0001";

        when cp_state_read2 =>
          i_debug_info(11 downto 8) <= "0010";

        when cp_state_write_normal =>
          i_debug_info(11 downto 8) <= "0011";

        when cp_state_write_general =>
          i_debug_info(11 downto 8) <= "0100";
      end case;
      i_debug_info(23 downto 12) <= cp_remaining;
      i_debug_info(31 downto 24) <= cp_wr_addr(LOG2_BE_WIDTH + 7 downto LOG2_BE_WIDTH);
      i_debug_info(43 downto 32) <= cp_wr_start_addr(LOG2_BE_WIDTH + 11 downto LOG2_BE_WIDTH);
      i_debug_info(52 downto 44) <= std_logic_vector(resize(unsigned(cp_bcount), 9));
      i_debug_info(53) <= cp_bfirst;
      i_debug_info(54) <= cp_blast;
      i_debug_info(55) <= cp_bbound;
      i_debug_info(63 downto 56) <= cp_gen_tag;
    end if;
  end process;

  ocf_cmd_adv <= n_cp_cmd_acc;
  owf_data_adv <= (n_cp_data_acc OR NOT be_class_valid) AND owf_data_valid;

  chop_cmd_valid <= cp_gen_cmd_valid;
  chop_cmd <= cp_gen_cmd;
  chop_addr(ADB3_OCP_ADDR_WIDTH - 1 downto LOG2_BE_WIDTH) <= cp_gen_addr;
  chop_addr(LOG2_BE_WIDTH - 1 downto 0) <= cp_be_first_idx when PCIE_HINTS else (others => '0');
  chop_blen <= std_logic_vector(resize(unsigned(cp_gen_blen), ADB3_OCP_BURST_WIDTH));
  chop_ocp_tag <= cp_ocp_tag;
  
  gen_chop_dm_tag : process(
    cp_gen_cmd,
    cp_gen_tag,
    cp_be_last_idx,
    cp_gen_1_qword)
  begin
    if PCIE_HINTS then
      case cp_gen_cmd is
        when OCP_CMD_READ =>
          chop_dm_tag <= cp_gen_tag;
       
        when OCP_CMD_WRITE | OCP_CMD_WRITENONPOST =>
          chop_dm_tag(LOG2_BE_WIDTH - 1 downto 0) <= cp_be_last_idx;
          chop_dm_tag(5 downto LOG2_BE_WIDTH) <= (others => '0');
          chop_dm_tag(6) <= cp_gen_1_qword;
          chop_dm_tag(7) <= '1';
          
        when others =>
          chop_dm_tag <= (others => '-');
      end case;
    else
      chop_dm_tag <= cp_gen_tag;
    end if;
  end process;
    

  -- Register control signals for better timing closure.
  register_control : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      cmd_enable_q <= ctl_cmd_enable;
      data_enable_q <= ctl_data_enable AND ctl_cmd_enable;
      max_rd_req_q <= ctl_max_rd_req;
      max_payload_q <= ctl_max_payload;
    end if;
  end process;

  -- Decode 'max_rd_req' value encoded as per PCI-E spec into a number of OCP
  -- words.
  decode_max_rd_req : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      case max_rd_req_q is
        when "000" =>
          -- 128 bytes
          max_rd_req_value <= std_logic_vector(to_unsigned(128 / BE_WIDTH, GEN_BLEN_WIDTH));

        when "001" =>
          -- 256 bytes
          max_rd_req_value <= std_logic_vector(to_unsigned(256 / BE_WIDTH, GEN_BLEN_WIDTH));

        when "010" =>
          -- 512 bytes
          max_rd_req_value <= std_logic_vector(to_unsigned(512 / BE_WIDTH, GEN_BLEN_WIDTH));

        when "011" =>
          -- 1 kiB
          max_rd_req_value <= std_logic_vector(to_unsigned(1024 / BE_WIDTH, GEN_BLEN_WIDTH));

        when "100" =>
          -- 2 kiB
          max_rd_req_value <= std_logic_vector(to_unsigned(2048 / BE_WIDTH, GEN_BLEN_WIDTH));

        when "101" =>
          -- 4 kiB
          max_rd_req_value <= std_logic_vector(to_unsigned(4096 / BE_WIDTH, GEN_BLEN_WIDTH));

        when "110" | "111" =>
          -- These values are reserved in current PCI-E spec. but we will map
          -- them to 128 bytes.
          max_rd_req_value <= std_logic_vector(to_unsigned(128 / BE_WIDTH, GEN_BLEN_WIDTH));

        when others =>
          max_rd_req_value <= (others => '-');
      end case;
    end if;
  end process;

  -- Decode 'max_payload' value encoded as per PCI-E spec into a number of OCP
  -- words.
  decode_max_payload : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      case max_payload_q is
        when "000" =>
          -- 128 B / 8 OCP words.
          max_payload_mask <= NOT std_logic_vector(to_unsigned(128 / BE_WIDTH - 1, BLEN_MASK_WIDTH));

        when "001" =>
          if MAX_WR_BURST_ORDER < 4 then
            -- Override with MAX_WR_BURST_LENGTH
            max_payload_mask <= NOT std_logic_vector(to_unsigned(MAX_WR_BURST_LENGTH - 1, BLEN_MASK_WIDTH));
          else
            -- 256 B / 16 OCP words.
            max_payload_mask <= NOT std_logic_vector(to_unsigned(256 / BE_WIDTH - 1, BLEN_MASK_WIDTH));
          end if;

        when "010" =>
          if MAX_WR_BURST_ORDER < 5 then
            -- Override with MAX_WR_BURST_LENGTH
            max_payload_mask <= NOT std_logic_vector(to_unsigned(MAX_WR_BURST_LENGTH - 1, BLEN_MASK_WIDTH));
          else
            -- 512 B / 32 OCP words.
            max_payload_mask <= NOT std_logic_vector(to_unsigned(512 / BE_WIDTH - 1, BLEN_MASK_WIDTH));
          end if;

        when "011" =>
          if MAX_WR_BURST_ORDER < 6 then
            -- Override with MAX_WR_BURST_LENGTH
            max_payload_mask <= NOT std_logic_vector(to_unsigned(MAX_WR_BURST_LENGTH - 1, BLEN_MASK_WIDTH));
          else
            -- 1 kiB / 64 OCP words.
            max_payload_mask <= NOT std_logic_vector(to_unsigned(1024 / BE_WIDTH - 1, BLEN_MASK_WIDTH));
          end if;

        when "100" =>
          if MAX_WR_BURST_ORDER < 7 then
            -- Override with MAX_WR_BURST_LENGTH
            max_payload_mask <= NOT std_logic_vector(to_unsigned(MAX_WR_BURST_LENGTH - 1, BLEN_MASK_WIDTH));
          else
            -- 2 kiB / 128 OCP words.
            max_payload_mask <= NOT std_logic_vector(to_unsigned(2048 / BE_WIDTH - 1, BLEN_MASK_WIDTH));
          end if;

        when "101" =>
          if MAX_WR_BURST_ORDER < 8 then
            -- Override with MAX_WR_BURST_LENGTH
            max_payload_mask <= NOT std_logic_vector(to_unsigned(MAX_WR_BURST_LENGTH - 1, BLEN_MASK_WIDTH));
          else
            -- 4 kiB / 256 OCP words.
            max_payload_mask <= NOT std_logic_vector(to_unsigned(4096 / BE_WIDTH - 1, BLEN_MASK_WIDTH));
          end if;

        when "110" | "111" =>
          -- These values are reserved in current PCI-E spec. but we will map
          -- them to 128 B / 8 OCP words.
          max_payload_mask <= NOT std_logic_vector(to_unsigned(128 / BE_WIDTH - 1, BLEN_MASK_WIDTH));

        when others =>
          max_payload_mask <= (others => '-');
      end case;
    end if;
  end process;

  -- Mask off address bits according to 'max_rd_req', in order to be able to
  -- perform boundary-crossing detection.
  mask_addr_for_max_rd_req : process(cp_rd_addr, max_rd_req_q)
  begin
    case max_rd_req_q is
      when "000" =>
        -- 128 B / 8 OCP words.
        cp_addr_rd_masked <= "00000" & cp_rd_addr(6 downto LOG2_BE_WIDTH);

      when "001" =>
        -- 256 B / 16 OCP words.
        cp_addr_rd_masked <= "0000" & cp_rd_addr(7 downto LOG2_BE_WIDTH);

      when "010" =>
        -- 512 B / 32 OCP words.
        cp_addr_rd_masked <= "000" & cp_rd_addr(8 downto LOG2_BE_WIDTH);

      when "011" =>
        -- 1 kiB / 64 OCP words.
        cp_addr_rd_masked <= "00" & cp_rd_addr(9 downto LOG2_BE_WIDTH);

      when "100" =>
        -- 2 kiB / 128 OCP words.
        cp_addr_rd_masked <= "0" & cp_rd_addr(10 downto LOG2_BE_WIDTH);

      when "101" =>
        -- 4 kiB / 256 OCP words.
        cp_addr_rd_masked <= cp_rd_addr(11 downto LOG2_BE_WIDTH);

      when "110" | "111" =>
        -- These values are reserved in current PCI-E spec. but we will map
        -- them to 128 B / 8 OCP words.
        cp_addr_rd_masked <= "00000" & cp_rd_addr(6 downto LOG2_BE_WIDTH);

      when others =>
        cp_addr_rd_masked <= (others => '-');
    end case;
  end process;

  do_comparisons : process(ocp_clk)
    variable tmp : std_logic_vector(LOG2_BE_WIDTH + 12 downto LOG2_BE_WIDTH);
  begin
    if rising_edge(ocp_clk) then
      if cp_state /= cp_state_idle then
        -- Calculate number of OCP words until 'max_rd_req' address boundary
        -- Need to_01 to prevent a lot of warnings about undefined values
        offset_to_rd_req_boundary <= max_rd_req_value - ("0" & std_logic_vector(to_01(unsigned(cp_addr_rd_masked))));

        -- Determine if would cross a 'max_rd_req' boundary
        tmp := std_logic_vector(to_01(unsigned(cp_addr_rd_masked))) + ("0" & cp_remaining);
        case max_rd_req_q is
          when "000" =>
            -- 128 B / 8 OCP words.
            cross_rd_req_boundary <= OR_reduce(tmp(LOG2_BE_WIDTH + 12 downto 7));

          when "001" =>
            -- 256 B / 16 OCP words.
            cross_rd_req_boundary <= OR_reduce(tmp(LOG2_BE_WIDTH + 12 downto 8));

          when "010" =>
            -- 512 B / 32 OCP words.
            cross_rd_req_boundary <= OR_reduce(tmp(LOG2_BE_WIDTH + 12 downto 9));

          when "011" =>
            -- 1 kiB / 64 OCP words.
            cross_rd_req_boundary <= OR_reduce(tmp(LOG2_BE_WIDTH + 12 downto 10));

          when "100" =>
            -- 2 kiB / 128 OCP words.
            cross_rd_req_boundary <= OR_reduce(tmp(LOG2_BE_WIDTH + 12 downto 11));

          when "101" =>
            -- 4 kiB / 256 OCP words.
            cross_rd_req_boundary <= OR_reduce(tmp(LOG2_BE_WIDTH + 12 downto 12));

          when "110" | "111" =>
            -- These values are reserved in current PCI-E spec. but we will map
            -- them to 128 B / 8 OCP words.
            cross_rd_req_boundary <= OR_reduce(tmp(LOG2_BE_WIDTH + 12 downto 7));

          when others => -- Should never happen
            cross_rd_req_boundary <= '-';
        end case;

      else
        offset_to_rd_req_boundary <= (others => '-');
        cross_rd_req_boundary <= '-';
      end if;
    end if;
  end process;

  -- Figure out OCP read command length to Direct Master, constrained by 'max_rd_req'.
  constrained_blen_rd <= offset_to_rd_req_boundary when cross_rd_req_boundary = '1' else
                         cp_remaining(GEN_BLEN_WIDTH - 1 downto 0);

  --
  -- Analysis of 'owf_be'.
  --

  -- Pre-analysis of 4-bit strobe groups
  gen_be_groups : process(
    owf_be)
    variable tmp : std_logic_vector(3 downto 0);
    variable transition_0_to_1 : std_logic_vector(BE_WIDTH - 1 downto 0);
  begin
    for i in 0 to NUM_BE_GROUP - 1 loop
      tmp := owf_be(4 * (i + 1) - 1 downto 4 * i);

      be_grp_full(i) <= AND_reduce(tmp);

      be_grp_not_empty(i) <= OR_reduce(tmp);

      case tmp is
        when "1000" | "1100" | "1110" =>
          be_grp_left(i) <= '1';
          be_grp_left_or_full(i) <= '1';

        when "1111" =>
          be_grp_left(i) <= '0';
          be_grp_left_or_full(i) <= '1';

        when others =>
          be_grp_left(i) <= '0';
          be_grp_left_or_full(i) <= '0';
      end case;

      case tmp is
        when "0001" | "0011" | "0111" =>
          be_grp_right(i) <= '1';
          be_grp_right_or_full(i) <= '1';

        when "1111" =>
          be_grp_right(i) <= '0';
          be_grp_right_or_full(i) <= '1';

        when others =>
          be_grp_right(i) <= '0';
          be_grp_right_or_full(i) <= '0';
      end case;
    end loop;

    -- Find 0 to 1 transitions in write strobes vector, going from bit 'n-1'
    -- to bit 'n'.
    transition_0_to_1(0) := '0';
    transition_0_to_1(BE_WIDTH - 1) := '0';
    for i in 1 to BE_WIDTH - 2 loop
      transition_0_to_1(i) := NOT owf_be(i - 1) AND owf_be(i);
    end loop;

    -- Find groups of 4 bits in 'transition_0_to_1' that indicate exactly one
    -- 0 -> 1 transition. If a 4-bit group has two 0 -> 1 transitions, then
    -- a special 'bust' flag is set.
    for i in 0 to NUM_BE_GROUP - 1 loop
      tmp := transition_0_to_1(4 * (i + 1) - 1 downto 4 * i);
      case tmp is
        when "0000" =>
          -- No 0->1 transitions.
          be_grp_0to1_one(i) <= '0';
          be_grp_0to1_two(i) <= '0';

        when "0001" | "0010" | "0100" | "1000" =>
          -- Exactly one 0->1 transition.
          be_grp_0to1_one(i) <= '1';
          be_grp_0to1_two(i) <= '0';

        when "0101" | "1010" | "1001" =>
          -- Two one 0->1 transitions.
          be_grp_0to1_one(i) <= '0';
          be_grp_0to1_two(i) <= '1';

        when others =>
          -- Other cases should never occur!
          be_grp_0to1_one(i) <= '-';
          be_grp_0to1_two(i) <= '-';
      end case;
    end loop;
  end process;

  --
  -- Classify the entire strobe word in a high-level way (examples for 64 bit
  -- data width => 8 bits of strobe):
  --
  -- (a) Entirely full:        1111.1111
  -- (b) Left-justified:       1110.0000, 1111.1110 etc.
  -- (c) Right-justified:      0000.0111, 0011.1111 etc.
  -- (d) Center:               0111.1100, 0000.1000, 0000.0110 etc.
  -- (e) Non-empty:            0001.0100, 0110.1010 etc.
  --
  -- Classes (a) to (d) are mutually exclusive with one another. Class (e) is
  -- not mutually exclusive with classes (a) to (d).
  --

  be_class_en <= n_cp_data_acc OR NOT be_class_valid;

  be_analysis : process(ocp_rst, ocp_clk)
    constant m : natural := NUM_BE_GROUP - 1;
    variable tmp, check : std_logic;
    variable tmp2 : std_logic_vector(be_grp_0to1_one'range);
    variable be_grp_idx : std_logic_vector(LOG2_NUM_BE_GROUP - 1 downto 0);
    variable be_grp : std_logic_vector(3 downto 0);
  begin
    if ocp_rst = '1' then
      be_class_valid <= '0';
    elsif rising_edge(ocp_clk) then
      if be_class_en = '1' then
        be_class_valid <= owf_data_valid;
      end if;
    end if;

    if rising_edge(ocp_clk) then
      if be_class_en = '1' then
        -- Non-empty classification
        be_class_not_empty <= OR_reduce(be_grp_not_empty);
        for i in 0 to NUM_BE_GROUP / 2 - 1 loop
          be_grp2_not_empty_q(i) <= OR_reduce(be_grp_not_empty(2 * (i + 1) - 1 downto 2 * i));
        end loop;

        -- Entirely full classification
        be_class_full <= AND_reduce(be_grp_full);

        -- Left-justified (but not entirely full) classification
        tmp := '0';
        for i in m downto 0 loop
          if AND_reduce("1" & be_grp_full(m downto i + 1)) = '1' AND
             ((i = 0 AND be_grp_left(i) = '1') OR (i /= 0 AND be_grp_left_or_full(i) = '1')) AND
             OR_reduce("0" & be_grp_not_empty(i - 1 downto 0)) = '0'
          then
            tmp := '1';
          end if;
        end loop;
        be_class_left <= tmp;

        -- Right-justified (but not entirely full) classification
        tmp := '0';
        for i in 0 to m loop
          if AND_reduce("1" & be_grp_full(i - 1 downto 0)) = '1' AND
             ((i = m AND be_grp_right(i) = '1') OR (i /= m AND be_grp_right_or_full(i) = '1')) AND
             OR_reduce("0" & be_grp_not_empty(m downto i + 1)) = '0'
          then
            tmp := '1';
          end if;
        end loop;
        be_class_right <= tmp;

        -- Center classification
        --
        -- We are looking for exactly one 0 -> 1 transition in 'owf_be' AND
        -- lowermost and uppermost byte enables are both 0.
        --
        -- 'owf_be' has already been partially analyzed for 0 -> 1 transitions
        -- into 'be_grp_0to1_one' & 'be_grp_0to1_two'.
        --
        -- If at least one bit of 'be_grp_0to1_two' is 1, it means that two
        -- or more 0 -> 1 transitions have been seen, so 'owf_be' cannot be
        -- center-classified.
        if owf_data_valid = '1' then
          -- Look for exactly one bit set in 'be_grp_0to1_one'.
          tmp := '0';
          for i in 0 to NUM_BE_GROUP - 1 loop
            tmp2 := (others => '1');
            tmp2(i) := '0';
            tmp := tmp OR AND_reduce(be_grp_0to1_one XOR tmp2);
          end loop;
          -- Check that lowermost and uppermost byte enables are both 0.
          tmp := tmp AND NOT owf_be(0) AND NOT owf_be(BE_WIDTH - 1);
          -- Check that all bits of 'be_grp_0to1_two' are zero.
          tmp := tmp AND NOT OR_reduce(be_grp_0to1_two);
          be_class_center <= tmp;
          -- As a sanity check on the above computation, we do the 'center'
          -- classification another way and compare the result.
          -- synthesis translate_off
          check := is_center_classification(owf_be);
          assert tmp = check
            report "adb3_ocp_dm_adapter_base: center classification sanity check failed, owf_be=0x" & conv_string_hex(owf_be) &
                   " be_class_center=" & std_logic'image(tmp) & " check=" & std_logic'image(check)
            severity error;
          -- synthesis translate_on
        else
          be_class_center <= '-';
        end if;

        if owf_data_valid = '1' then
          -- Figure out which is the first occupied DWORD lane, starting from DWORD lane 0.
          be_grp_idx := (others => '0');
          be_grp := (others => '0');
          for i in 0 to NUM_BE_GROUP - 1 loop
            if be_grp_not_empty(i) = '1' then
              if OR_reduce("0" & be_grp_not_empty(i - 1 downto 0)) = '0' then
                be_grp_idx := be_grp_idx OR std_logic_vector(to_unsigned(i, LOG2_NUM_BE_GROUP));
                be_grp := be_grp OR owf_be(4 * (i + 1) - 1 downto 4 * i);
              end if;
            end if;
          end loop;
          -- Figure out which is the first occupied byte lane, starting from byte lane 0.
          be_class_first_idx(LOG2_BE_WIDTH - 1 downto 2) <= be_grp_idx;
          case be_grp is
            when "0001" | "0011" | "0101" | "0111" | "1001" | "1011" | "1101" | "1111" =>
              be_class_first_idx(1 downto 0) <= "00";
            
            when "0010" | "0110" | "1010" | "1110" =>
              be_class_first_idx(1 downto 0) <= "01";
            
            when "0100" | "1100" =>
              be_class_first_idx(1 downto 0) <= "10";
            
            when "1000" =>
              be_class_first_idx(1 downto 0) <= "11";
            
            when "0000" =>
              -- Can happen only with all zero byte enables, in which case we don't care.
              be_class_first_idx(1 downto 0) <= "--";
            
            when others =>
              -- Should never happen!
              be_class_first_idx(1 downto 0) <= "--";
              assert false severity failure;
          end case;
        else
          be_class_first_idx(1 downto 0) <= "--";
        end if;
        
        if owf_data_valid = '1' then
          -- Figure out which is the last occupied DWORD lane, starting from DWORD lane 0.
          be_grp_idx := (others => '0');
          be_grp := (others => '0');
          for i in 0 to NUM_BE_GROUP - 1 loop
            if be_grp_not_empty(i) = '1' then
              if OR_reduce("0" & be_grp_not_empty(NUM_BE_GROUP - 1 downto i + 1)) = '0' then
                be_grp_idx := be_grp_idx OR std_logic_vector(to_unsigned(i, LOG2_NUM_BE_GROUP));
                be_grp := be_grp OR owf_be(4 * (i + 1) - 1 downto 4 * i);
              end if;
            end if;
          end loop;
          -- Figure out which is the last occupied byte lane, starting from byte lane 0.
          be_class_last_idx(LOG2_BE_WIDTH - 1 downto 2) <= be_grp_idx;
          case be_grp is
            when "0001" =>
              be_class_last_idx(1 downto 0) <= "00";
            
            when "0010" | "0011" =>
              be_class_last_idx(1 downto 0) <= "01";
            
            when "0100" | "0110" | "0101" | "0111" =>
              be_class_last_idx(1 downto 0) <= "10";
            
            when "1000" | "1100" | "1010" | "1110" | "1001" | "1101" | "1011" | "1111" =>
              be_class_last_idx(1 downto 0) <= "11";
            
            when "0000" =>
              -- Can happen only with all zero byte enables, in which case we don't care.
              be_class_last_idx(1 downto 0) <= "--";
            
            when others =>
              -- Should never happen!
              be_class_last_idx(1 downto 0) <= "--";
              assert false severity failure;
          end case;
        else
          be_class_last_idx(1 downto 0) <= "--";
        end if;
        
        -- Need registered versions of 'owf_data' and 'owf_be' to match timing of 'be_class_*' signals.
        be_class_data <= owf_data;
        be_class_be <= owf_be;
      end if;
    end if;
  end process;

  --
  -- State machine that generates constrained OCP bursts for master OCP port.
  --

  cp_bbound <= is_at_addr_boundary(cp_wr_addr, max_payload_mask);

  cp_discriminant <= cp_bfirst & (cp_blast OR cp_bbound) & be_class_full & be_class_left & be_class_right & be_class_center & be_class_not_empty;

  gen_cp_state : process(ocp_rst, ocp_clk)
  begin
    if ocp_rst = '1' then
      cp_state <= cp_state_idle;
      cp_gen_cmd <= (others => '-');
      cp_gen_addr <= (others => '-');
      cp_gen_blen <= (others => '-');
      cp_gen_cmd_valid <= '0';
      cp_rd_addr <= (others => '-');
      cp_wr_addr <= (others => '-');
      cp_wr_start_addr <= (others => '-');
      cp_ocp_tag <= (others => '-');
      cp_remaining <= (others => '-');
      cp_cmd_acc <= '0';
      cp_data_valid <= '0';
      cp_data_acc <= '0';
      cp_be_mask_out <= (others => '-');
      cp_bfirst <= '-';
      cp_blast <= '-';
      cp_bcount <= (others => '-');
      cp_special_pos <= (others => '-');
      cp_special_pos_1h <= (others => '-');
      cp_be_first_idx <= (others => '-');
      cp_be_last_idx <= (others => '-');
      cp_gen_1_qword <= '-';
      cp_gen_tag <= (others => '0');
      cp_data_out <= (others => '-');
      cp_be_out <= (others => '-');
    elsif rising_edge(ocp_clk) then
      cp_state <= n_cp_state;
      cp_gen_cmd <= n_cp_gen_cmd;
      cp_gen_addr <= n_cp_gen_addr;
      cp_gen_blen <= n_cp_gen_blen;
      cp_gen_cmd_valid <= n_cp_gen_cmd_valid;
      cp_rd_addr <= n_cp_rd_addr;
      cp_wr_addr <= n_cp_wr_addr;
      cp_wr_start_addr <= n_cp_wr_start_addr;
      cp_ocp_tag <= n_cp_ocp_tag;
      cp_remaining <= n_cp_remaining;
      cp_cmd_acc <= n_cp_cmd_acc;
      cp_data_valid <= n_cp_data_valid;
      cp_data_acc <= n_cp_data_acc;
      cp_be_mask_out <= n_cp_be_mask_out;
      cp_bfirst <= n_cp_bfirst;
      cp_blast <= n_cp_blast;
      cp_bcount <= n_cp_bcount;
      cp_special_pos <= n_cp_special_pos;
      cp_special_pos_1h <= n_cp_special_pos_1h;
      cp_be_first_idx <= n_cp_be_first_idx;
      cp_be_last_idx <= n_cp_be_last_idx;
      cp_gen_1_qword <= n_cp_gen_1_qword;
      
      if cp_gen_cmd_valid = '1' AND cp_gen_cmd = OCP_CMD_READ then
        -- Generate tag for chopped read commands
        cp_gen_tag(DM_TAG_BITS - 1 downto 0) <= cp_gen_tag(DM_TAG_BITS - 1 downto 0) + 1;
      end if;
      
      cp_data_out <= be_class_data;
      cp_be_out <= be_class_be;
    end if;
    
    if rising_edge(ocp_clk) then
      cp_debug_action <= n_cp_debug_action;
    end if;
  end process;

  gen_n_cp_state : process(
    cp_state,
    cp_wr_addr,
    cp_rd_addr,
    cp_wr_start_addr,
    cp_ocp_tag,
    cp_remaining,
    cp_bcount,
    cp_bfirst,
    cp_blast,
    cp_special_pos,
    cp_special_pos_1h,
    cp_bbound,
    cp_discriminant,
    cp_be_first_idx,
    cp_be_last_idx,
    ctl_cmd_enable,
    data_enable_q,
    constrained_blen_rd,
    max_payload_mask,
    ocf_cmd,
    ocf_cmd_valid,
    ocf_addr,
    ocf_blen,
    ocf_tag,
    be_grp2_not_empty_q,
    be_class_valid,
    be_class_first_idx,
    be_class_last_idx)
    variable action : burst_action_t;
  begin
    -- Defaults, overridden as necessary below
    n_cp_cmd_acc <= '0';
    n_cp_wr_addr <= cp_wr_addr;
    n_cp_rd_addr <= cp_rd_addr;
    n_cp_wr_start_addr <= cp_wr_start_addr;
    n_cp_gen_cmd <= (others => '-');
    n_cp_gen_addr <= (others => '-');
    n_cp_gen_blen <= (others => '-');
    n_cp_gen_cmd_valid <= '0';
    n_cp_ocp_tag <= cp_ocp_tag;
    n_cp_remaining <= cp_remaining;
    n_cp_bcount <= cp_bcount;
    n_cp_bfirst <= cp_bfirst;
    n_cp_blast <= cp_blast;
    n_cp_be_mask_out <= (others => '-');
    n_cp_data_valid <= '0';
    n_cp_data_acc <= '0';
    n_cp_special_pos <= cp_special_pos;
    n_cp_special_pos_1h <= cp_special_pos_1h;
    n_cp_debug_action <= action_special;
    n_cp_be_first_idx <= cp_be_first_idx;
    n_cp_be_last_idx <= cp_be_last_idx;
    n_cp_gen_1_qword <= '-';
    
    case cp_state is
      when cp_state_idle =>
        if ctl_cmd_enable = '1' AND ocf_cmd_valid = '1' then
          case ocf_cmd is
            when OCP_CMD_READ =>
              n_cp_state <= cp_state_read1;
              n_cp_cmd_acc <= '1';
              n_cp_wr_addr <= (others => '-');
              n_cp_rd_addr <= ocf_addr(ADB3_OCP_ADDR_WIDTH - 1 downto LOG2_BE_WIDTH);
              n_cp_wr_start_addr <= (others => '-');
              n_cp_ocp_tag <= ocf_tag;
              n_cp_remaining <= ocf_blen;
              n_cp_bcount <= (others => '-');
              n_cp_bfirst <= '-';
              n_cp_blast <= '-';
              n_cp_special_pos <= (others => '-');
              n_cp_special_pos_1h <= (others => '-');
              n_cp_be_first_idx <= (others => '-');
              n_cp_be_last_idx <= (others => '-');

            when OCP_CMD_WRITE | OCP_CMD_WRITENONPOST =>
              n_cp_state <= cp_state_write_normal;
              n_cp_cmd_acc <= '1';
              n_cp_wr_addr <= ocf_addr(ADB3_OCP_ADDR_WIDTH - 1 downto LOG2_BE_WIDTH);
              n_cp_rd_addr <= (others => '-');
              n_cp_wr_start_addr <= ocf_addr(ADB3_OCP_ADDR_WIDTH - 1 downto LOG2_BE_WIDTH);
              n_cp_ocp_tag <= ocf_tag;
              n_cp_remaining <= ocf_blen;
              n_cp_bcount <= (others => '0');
              n_cp_bfirst <= '1';
              n_cp_blast <= is_last_word_in_burst(ocf_blen);
              n_cp_special_pos <= (others => '-');
              n_cp_special_pos_1h <= (others => '-');
              n_cp_be_first_idx <= (others => '-');
              n_cp_be_last_idx <= (others => '-');

            when others => -- Illegal commands; discard them
              n_cp_state <= cp_state_idle;
              n_cp_cmd_acc <= '1';
              n_cp_wr_addr <= (others => '-');
              n_cp_rd_addr <= (others => '-');
              n_cp_wr_start_addr <= (others => '-');
              n_cp_remaining <= (others => '-');
              n_cp_ocp_tag <= (others => '-');
              n_cp_bcount <= (others => '-');
              n_cp_bfirst <= '-';
              n_cp_blast <= '-';
              n_cp_special_pos <= (others => '-');
              n_cp_special_pos_1h <= (others => '-');
              n_cp_be_first_idx <= (others => '-');
              n_cp_be_last_idx <= (others => '-');
          end case;
        else
          n_cp_state <= cp_state_idle;
          n_cp_wr_addr <= (others => '-');
          n_cp_rd_addr <= (others => '-');
          n_cp_wr_start_addr <= (others => '-');
          n_cp_remaining <= (others => '-');
          n_cp_ocp_tag <= (others => '-');
          n_cp_bcount <= (others => '-');
          n_cp_bfirst <= '-';
          n_cp_blast <= '-';
          n_cp_special_pos <= (others => '-');
          n_cp_special_pos_1h <= (others => '-');
          n_cp_be_first_idx <= (others => '-');
          n_cp_be_last_idx <= (others => '-');
        end if;

      when cp_state_read1 =>
        if cp_remaining = 0 then
          n_cp_state <= cp_state_idle;
          n_cp_rd_addr <= (others => '-');
          n_cp_remaining <= (others => '-');
          n_cp_ocp_tag <= (others => '-');
        else
          n_cp_state <= cp_state_read2;
        end if;
        n_cp_wr_addr <= (others => '-');
        n_cp_wr_start_addr <= (others => '-');
        n_cp_bcount <= (others => '-');
        n_cp_bfirst <= '-';
        n_cp_blast <= '-';
        n_cp_special_pos <= (others => '-');
        n_cp_special_pos_1h <= (others => '-');
        n_cp_be_first_idx <= (others => '-');
        n_cp_be_last_idx <= (others => '-');

      when cp_state_read2 =>
        if ctl_cmd_enable = '1' then
          n_cp_state <= cp_state_read1;
          n_cp_rd_addr <= cp_rd_addr + constrained_blen_rd;
          n_cp_remaining <= cp_remaining - constrained_blen_rd;
          n_cp_gen_cmd <= OCP_CMD_READ;
          n_cp_gen_blen <= constrained_blen_rd;
          n_cp_gen_addr <= cp_rd_addr;
          n_cp_gen_cmd_valid <= '1';
          n_cp_be_first_idx <= (others => '0');
        else
          n_cp_state <= cp_state_read2;
          n_cp_be_first_idx <= (others => '-');
        end if;
        n_cp_wr_addr <= (others => '-');
        n_cp_wr_start_addr <= (others => '-');
        n_cp_bcount <= (others => '-');
        n_cp_bfirst <= '-';
        n_cp_blast <= '-';
        n_cp_special_pos <= (others => '-');
        n_cp_special_pos_1h <= (others => '-');
        n_cp_be_last_idx <= (others => '-');

      -- +----+----------------------------------------------------------+------------+--------------------------------------------------+
      -- | mc | bfirst  last*   full    l       r       c       ne       | mnemonic   | consume op_data op_cmd  new_cmd special n_state  |
      -- +----+----------------------------------------------------------+------------+--------------------------------------------------+
      -- | -  | -       0       1       -       -       -       -        | normal     | y       y       n       n       n       burst*   |
      -- | F  | 1       0       -       1       -       -       -        |            | y       y       n       n       n       burst*   |
      -- +----+----------------------------------------------------------+------------+--------------------------------------------------+
      -- | -  | -       1       1       -       -       -       -        | end_burst  | y       y       y1      y1      n       burst*   |
      -- | T  | 1       0       -       1       -       -       -        |            | y       y       n       n       n       burst*   |
      -- | -  | 1       1       -       1       -       -       -        |            | y       y       y1      y1      n       burst*   |
      -- | F  | -       -       -       -       1       -       -        |            | y       y       y1      y1      n       burst*   |
      -- | T  | 1       -       -       -       1       -       -        |            | y       y       y1      y1      n       burst*   |
      -- | -  | 1       -       -       -       -       1       -        |            | y       y       y1      y1      n       burst*   |
      -- +----+----------------------------------------------------------+------------+--------------------------------------------------+
      -- | -  | 0       -       -       1       -       -       -        | new_burst  | n       n       y0      y0      n       burst*   |
      -- | T  | 0       -       -       -       1       -       -        |            | y       y       y1      y1      n       burst*   |
      -- | -  | 0       -       -       -       -       1       -        |            | n       n       y0      y0      n       burst*   |
      -- +----+----------------------------------------------------------+------------+--------------------------------------------------+
      -- | -  | 1       -       0       0       0       0       0        | suppress   | y       n       n       y1      n       burst*   |
      -- | -  | 0       -       0       0       0       0       0        |            | y       n       y0      y1      n       burst*   |
      -- +----+----------------------------------------------------------+------------+--------------------------------------------------+
      -- | -  | 1       -       0       0       0       0       1        | go_special | n       n       n       y0      y       special  |
      -- | -  | 0       -       0       0       0       0       1        |            | n       n       y0      y0      y       special  |
      -- +----+----------------------------------------------------------+------------+--------------------------------------------------+
      when cp_state_write_normal =>
        n_cp_rd_addr <= (others => '-');
        if be_class_valid = '1' AND data_enable_q = '1' then
          action := action_special; -- Overriden if any match below occurs.
          if match(cp_discriminant, "-01----", "1001111") then
            action := action_normal;
          end if;
          if match(cp_discriminant, "10-1---", "0010111") AND NOT MPTL_COMPATIBLE then
            action := action_normal;
          end if;
          if match(cp_discriminant, "-11----", "1001111") then
            action := action_end_burst;
          end if;
          if match(cp_discriminant, "10-1---", "0010111") AND MPTL_COMPATIBLE then
            -- MPTL write payload doesn't support burst of length > 1 that doesn't have full byte enables for every word
            action := action_end_burst;
          end if;
          if match(cp_discriminant, "11-1---", "0010111") then
            action := action_end_burst;
          end if;
          if match(cp_discriminant, "----1--", "1111011") AND NOT MPTL_COMPATIBLE then
            action := action_end_burst;
          end if;
          if match(cp_discriminant, "1---1--", "0111011") AND MPTL_COMPATIBLE then
            action := action_end_burst;
          end if;
          if match(cp_discriminant, "1----1-", "0111101") then
            action := action_end_burst;
          end if;
          if match(cp_discriminant, "0--1---", "0110111") then
            action := action_new_burst;
          end if;
          if match(cp_discriminant, "0---1--", "0111011") AND MPTL_COMPATIBLE then
            action := action_new_burst;
          end if;
          if match(cp_discriminant, "0----1-", "0111101") then
            action := action_new_burst;
          end if;
          if match(cp_discriminant, "1-00000", "0100000") then
            action := action_suppress;
          end if;
          if match(cp_discriminant, "0-00000", "0100000") then
            action := action_suppress;
          end if;
          n_cp_debug_action <= action;
          -- In the synthesized design, the above cases should catch every
          -- possibility, so that if none of them matched, the only remaining
          -- possibilities are "1-00001" and "0-00001":
          -- if match(cp_discriminant, "1-00001", "0100000") then
          --   action := action_special;
          -- end if;
          -- if match(cp_discriminant, "0-00001", "0100000") then
          --   action := action_special;
          -- end if;
          case action is
            when action_normal =>
              n_cp_state <= cp_state_write_normal;
              n_cp_wr_addr <= cp_wr_addr + 1;
              n_cp_remaining <= cp_remaining - 1;
              n_cp_data_valid <= '1';
              if cp_bfirst = '1' then
                n_cp_be_first_idx <= be_class_first_idx;
              end if;
              n_cp_be_last_idx <= be_class_last_idx;
              n_cp_data_acc <= '1';
              n_cp_be_mask_out <= (others => '0');
              n_cp_bfirst <= '0';
              n_cp_blast <= is_next_last_word_in_burst(cp_remaining);
              n_cp_bcount <= cp_bcount + 1;
              n_cp_special_pos <= (others => '-');
              n_cp_special_pos_1h <= (others => '-');

            when action_end_burst =>
              if cp_blast = '1' then
                n_cp_state <= cp_state_idle;
                n_cp_wr_addr <= (others => '-');
                n_cp_wr_start_addr <= (others => '-');
                n_cp_remaining <= (others => '-');
                n_cp_bfirst <= '-';
                n_cp_blast <= '-';
                n_cp_bcount <= (others => '-');
              else
                n_cp_state <= cp_state_write_normal;
                n_cp_wr_addr <= cp_wr_addr + 1;
                n_cp_wr_start_addr <= cp_wr_addr + 1;
                n_cp_remaining <= cp_remaining - 1;
                n_cp_bfirst <= '1';
                n_cp_blast <= is_next_last_word_in_burst(cp_remaining);
                n_cp_bcount <= (others => '0');
              end if;
              n_cp_gen_cmd <= OCP_CMD_WRITE;
              n_cp_gen_addr <= cp_wr_start_addr;
              n_cp_gen_blen <= cp_bcount + 1;
              n_cp_gen_cmd_valid <= '1';
              n_cp_data_valid <= '1';
              if cp_bfirst = '1' then
                n_cp_be_first_idx <= be_class_first_idx;
              end if;
              n_cp_be_last_idx <= be_class_last_idx;
              n_cp_gen_1_qword <= '0';
              n_cp_data_acc <= '1';
              n_cp_be_mask_out <= (others => '0');
              n_cp_special_pos <= (others => '-');
              n_cp_special_pos_1h <= (others => '-');

            when action_new_burst =>
              n_cp_state <= cp_state_write_normal;
              n_cp_wr_start_addr <= cp_wr_addr;
              n_cp_bfirst <= '1';
              n_cp_bcount <= (others => '0');
              n_cp_gen_cmd <= OCP_CMD_WRITE;
              n_cp_gen_addr <= cp_wr_start_addr;
              n_cp_gen_blen <= cp_bcount;
              n_cp_gen_cmd_valid <= '1';
              n_cp_special_pos <= (others => '-');
              n_cp_special_pos_1h <= (others => '-');
              n_cp_gen_1_qword <= '0';

            when action_suppress =>
              if cp_blast = '1' then
                n_cp_state <= cp_state_idle;
                n_cp_wr_addr <= (others => '-');
                n_cp_wr_start_addr <= (others => '-');
                n_cp_remaining <= (others => '-');
                n_cp_bfirst <= '-';
                n_cp_blast <= '-';
                n_cp_bcount <= (others => '-');
                n_cp_data_acc <= '1';
              else
                n_cp_state <= cp_state_write_normal;
                n_cp_wr_addr <= cp_wr_addr + 1;
                n_cp_wr_start_addr <= cp_wr_addr + 1;
                n_cp_remaining <= cp_remaining - 1;
                n_cp_bfirst <= '1';
                n_cp_blast <= is_next_last_word_in_burst(cp_remaining);
                n_cp_bcount <= (others => '0');
                n_cp_data_acc <= '1';
              end if;
              if cp_bfirst = '0' then
                n_cp_gen_cmd <= OCP_CMD_WRITE;
                n_cp_gen_addr <= cp_wr_start_addr;
                n_cp_gen_blen <= cp_bcount;
                n_cp_gen_cmd_valid <= '1';
                n_cp_gen_1_qword <= '0';
              else
                n_cp_be_first_idx <= (others => '-');
                n_cp_be_last_idx <= (others => '-');
              end if;
              n_cp_special_pos <= (others => '-');
              n_cp_special_pos_1h <= (others => '-');

            when action_special =>
              n_cp_state <= cp_state_write_general;
              n_cp_wr_start_addr <= cp_wr_addr;
              n_cp_bfirst <= '1';
              n_cp_blast <= is_last_word_in_burst(cp_remaining);
              n_cp_bcount <= (others => '0');
              n_cp_special_pos <= (others => '0');
              n_cp_special_pos_1h(0) <= '1';
              n_cp_special_pos_1h(n_cp_special_pos_1h'high downto 1) <= (others => '0');
              if cp_bfirst = '0' then
                n_cp_gen_cmd <= OCP_CMD_WRITE;
                n_cp_gen_addr <= cp_wr_start_addr;
                n_cp_gen_blen <= cp_bcount;
                n_cp_gen_cmd_valid <= '1';
                n_cp_gen_1_qword <= '0';
              else
                n_cp_be_first_idx <= (others => '-');
                n_cp_be_last_idx <= (others => '-');
              end if;
          end case;
        else
          n_cp_state <= cp_state_write_normal;
          n_cp_special_pos <= (others => '-');
          n_cp_special_pos_1h <= (others => '-');
        end if;

      when cp_state_write_general =>
        n_cp_rd_addr <= (others => '-');
        if be_class_valid = '1' AND data_enable_q = '1' then
          if AND_reduce(cp_special_pos) = '1' then
            if cp_blast = '1' then
              n_cp_state <= cp_state_idle;
              n_cp_wr_addr <= (others => '-');
              n_cp_wr_start_addr <= (others => '-');
              n_cp_remaining <= (others => '-');
              n_cp_bfirst <= '-';
              n_cp_blast <= '-';
              n_cp_bcount <= (others => '-');
            else
              n_cp_state <= cp_state_write_normal;
              n_cp_wr_addr <= cp_wr_addr + 1;
              n_cp_wr_start_addr <= cp_wr_addr + 1;
              n_cp_remaining <= cp_remaining - 1;
              n_cp_bfirst <= '1';
              n_cp_blast <= is_next_last_word_in_burst(cp_remaining);
              n_cp_bcount <= (others => '0');
            end if;
            n_cp_data_acc <= '1';
          else
            n_cp_state <= cp_state_write_general;
            n_cp_bfirst <= '1';
            n_cp_bcount <= (others => '0');
          end if;
          if OR_reduce(be_grp2_not_empty_q AND cp_special_pos_1h) = '1' then
            n_cp_gen_cmd <= OCP_CMD_WRITE;
            n_cp_gen_addr <= cp_wr_start_addr;
            n_cp_gen_blen <= cp_bcount + 1;
            n_cp_gen_cmd_valid <= '1';
            n_cp_data_valid <= '1';
            n_cp_be_first_idx <= cp_special_pos & "000";
            n_cp_be_last_idx <= cp_special_pos & "111";
            n_cp_gen_1_qword <= '1';
            n_cp_be_mask_out <= gen_be_mask_out(BE_WIDTH, cp_special_pos_1h);
          else
            n_cp_be_first_idx <= (others => '-');
            n_cp_be_last_idx <= (others => '-');
          end if;
          n_cp_special_pos <= cp_special_pos + 1;
          n_cp_special_pos_1h <= cp_special_pos_1h(cp_special_pos_1h'high - 1 downto 0) & "0";
        else
          n_cp_state <= cp_state_write_general;
          n_cp_be_first_idx <= (others => '-');
          n_cp_be_last_idx <= (others => '-');
        end if;
    end case;
  end process;

  chop_data <= cp_data_out;
  chop_be <= cp_be_out AND NOT cp_be_mask_out;
  chop_data_valid <= cp_data_valid;


  --
  -- Chipscope debug
  --

  gen_chipscope : if chipscope_on generate

    component adb3_ocp_dm_adapter_chop_icon
    port(
      control0 : inout std_logic_vector(35 downto 0));
    end component;

    component adb3_ocp_dm_adapter_chop_ila
    port(
      control  : inout std_logic_vector(35 downto 0);
      clk      : in    std_logic;
      trig0    : in    std_logic_vector(158 downto 0);
      trig_out : out   std_logic);
    end component;

    signal control : std_logic_vector(35 downto 0);
    signal trig0 : std_logic_vector(158 downto 0);
    signal trig_out : std_logic;

  begin

    trig0(0) <= be_class_en;
    trig0(1) <= be_class_not_empty;
    trig0(9 downto 2) <= std_logic_vector(resize(unsigned(be_grp2_not_empty_q), 8));
    trig0(10) <= be_class_full;
    trig0(11) <= be_class_left;
    trig0(12) <= be_class_right;
    trig0(13) <= be_class_center;
    trig0(14) <= be_class_valid;
    trig0(15) <= owf_data_valid;
    trig0(79 downto 16) <= std_logic_vector(resize(unsigned(be_class_be), 64));
    trig0(82 downto 80) <= cp_gen_cmd;
    trig0(83) <= cp_gen_cmd_valid;
    trig0(87 downto 84) <= i_debug_info(11 downto 8);
    trig0(88) <= cp_data_valid;
    encode_debug_action : process(ocp_clk)
    begin
      if rising_edge(ocp_clk) then
        case cp_debug_action is
          when action_normal =>
            trig0(91 downto 89) <= "000";
          
          when action_end_burst =>
            trig0(91 downto 89) <= "001";
          
          when action_new_burst =>
            trig0(91 downto 89) <= "010";
          
          when action_suppress =>
            trig0(91 downto 89) <= "011";
          
          when action_special =>
            trig0(91 downto 89) <= "100";
          
          when others =>
            trig0(91 downto 89) <= "111";
        end case;
      end if;
    end process;
    trig0(94 downto 92) <= (others => '0');
    trig0(158 downto 95) <= std_logic_vector(resize(unsigned(cp_be_out AND NOT cp_be_mask_out), 64));
    
    icon_0 : adb3_ocp_dm_adapter_chop_icon
    port map(
      control0 => control);

    ila_0 : adb3_ocp_dm_adapter_chop_ila
    port map(
      control => control,
      clk => ocp_clk,
      trig0 => trig0,
      trig_out => trig_out);

  end generate;


end;
