--------------------------------------------------------------------------------
--
-- Module Name      : fpga_int_adb3
--
--                    Copyright 2009 by
--                    Alpha Data Parallel Systems Ltd.
--                    All Rights Reserved.
--
-- Original Author  : David Miller
-- Created          : 
--
-- Modified By      : DRM
-- Date             : 01-Oct-2010
-- Change Notes     : Added temp_alert input from sensor averaging in IIC I/F
--
-- Modified By      : AM
-- Date             : 28-Jun-2011
-- Change Notes     : Modified to be a general IRQ handling module
--
-- Modified By      : DRM
-- Date             : 4-Mar-2012
-- Change Notes     : Added "buffer_type" attribute to stop BUFGs being placed
--
-- Description      : 
--
--
-- Dependencies     : adb3_board_ctrl
--
-- Disclaimer       : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                    WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                    WARRANTIES IMPLIED OF MERCHANTABILITY, FITNESS FOR A
--                    PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library unisim;
use unisim.vcomponents.all;

library work;
-- use work.adb3.all;
use work.adb3_board_ctrl.all;

entity fpga_int_adb3 is
  generic(
    NUMBER_OF_INTERRUPTS     : natural                                           := 16;
    LEVEL_SENSITIVE_DEFAULTS : std_logic_vector(15 downto 0) := X"0000";
    ACTIVE_LOW               : std_logic_vector(15 downto 0) := X"0003");
  port(
    rst        : in  std_logic;
    lclk       : in  std_logic;
    write      : in  std_logic;
    be         : in  std_logic_vector(7 downto 0);
    data       : in  std_logic_vector(63 downto 0);
    status     : out std_logic_vector(63 downto 0);
    int_local  : out std_logic;
    interrupts : in  std_logic_vector(NUMBER_OF_INTERRUPTS-1 downto 0));
end fpga_int_adb3;

------------------------------------------------------------------
architecture rtl of fpga_int_adb3 is

  constant FICTL_DEFAULTS : std_logic_vector(31 downto 0) := LEVEL_SENSITIVE_DEFAULTS(15 downto 8) &
                                                             X"00" &
                                                             LEVEL_SENSITIVE_DEFAULTS(7 downto 0) &
                                                             X"00";                                                          
  
  signal fictl           : std_logic_vector(31 downto 0)                     := FICTL_DEFAULTS;
  signal fistat          : std_logic_vector(31 downto 0);
  signal level_sensitive : std_logic_vector(15 downto 0);
  signal clear_int       : std_logic_vector(NUMBER_OF_INTERRUPTS-1 downto 0) := (others => '1');
  signal edge_detected   : std_logic_vector(NUMBER_OF_INTERRUPTS-1 downto 0) := (others => '0');
  signal interrupt       : std_logic_vector(15 downto 0) := (others => '0');
  signal interrupts_int  : std_logic_vector(NUMBER_OF_INTERRUPTS-1 downto 0);

  --
  -- Stop BUFG being instantiated on the interrupts
  --
  attribute buffer_type : string;
  attribute buffer_type of interrupts : signal is "none";
  

  
begin

  assert NUMBER_OF_INTERRUPTS<16 report "Maximum number of interrupts available is 16" severity error;

  status <= fistat & fictl;

  fix_polarity: for i in 0 to NUMBER_OF_INTERRUPTS-1 generate
    interrupts_int(i) <= interrupts(i) when ACTIVE_LOW(i) = '0' else not interrupts(i);
  end generate fix_polarity;
  
  ---------------------------------------
  generate_clear_int : process(rst, lclk)
    ---------------------------------------
  begin
    if rst = '1' then
      clear_int <= (others => '1');
    elsif lclk'event and lclk = '1' then
      for i in 0 to NUMBER_OF_INTERRUPTS-1 loop
        if i < 8 then
          clear_int(i) <= level_sensitive(i) or (write and be(4) and data(32+i));
        else
          clear_int(i) <= level_sensitive(i) or (write and be(6) and data(32+8+i));
        end if;
      end loop;  -- i
    end if;
  end process;

  gen_capture_irq : for i in 0 to NUMBER_OF_INTERRUPTS-1 generate
    ---------------------------------------
    capture_edge_irq_i : process(clear_int, interrupts_int)
    ---------------------------------------
    begin
      if clear_int(i) = '1' then
        edge_detected(i) <= '0';
      elsif rising_edge(interrupts_int(i)) then
        edge_detected(i) <= '1';
      end if;
    end process;
  end generate gen_capture_irq;

  --
  -- Generate the FPGA and system monitor interrupts
  --
  ---------------------------------------
  capture_level : process(rst, lclk)
    ---------------------------------------
  begin
    if rst = '1' then
      interrupt <= (others => '0');
    elsif lclk'event and lclk = '1' then
      for i in 0 to NUMBER_OF_INTERRUPTS-1 loop
        interrupt(i) <= (interrupts_int(i) and level_sensitive(i)) or edge_detected(i);
      end loop;  -- i
    end if;
  end process;

  fistat(7 downto 0)  <= interrupt(7 downto 0);
  fistat(15 downto 8) <= X"00";
  fistat(23 downto 16)  <= interrupt(15 downto 8);
  fistat(31 downto 24) <= X"00";

  ---------------------------------------
  implement_registers : process(rst, lclk)
    ---------------------------------------
  begin
    if rst = '1' then
      fictl <= FICTL_DEFAULTS;
    elsif rising_edge(lclk) then
      -- default
      fictl <= fictl;

      if write = '1' and be(0) = '1' then
        fictl(7 downto 0) <= data(7 downto 0);
      end if;

      if write = '1' and be(2) = '1' then
        fictl(23 downto 16) <= data(23 downto 16);
      end if;
      
      if write = '1' and be(1) = '1' then
        fictl(15 downto 8)   <= data(15 downto 8);
      end if;

      if write = '1' and be(3) = '1' then
        fictl(31 downto 24)   <= data(31 downto 24);
      end if;
      
    end if;
  end process;

  level_sensitive <= fictl(31 downto 24) & fictl(15 downto 8);

  ---------------------------------------
  gen_int_local : process(lclk)
    ---------------------------------------
  begin
    if rising_edge(lclk) then
      int_local <= OR_reduce(fistat(7 downto 0) and fictl(7 downto 0)) or OR_reduce(fistat(23 downto 16) and fictl(23 downto 16));
    end if;
  end process;

end;
