--
-- model_clocks_xrc6t1.vhd
--
-- ADM-XRC6T1 Model Specific Clocks
-- Alpha Data, 2010
--
---------------------------------------------------------------
-- clocks
-- rclk (300MHz) is used as ref
-- ocp_clk = 800/OCPCLKDIV
-- lclk = 40MHz,
-- smap interface uses lclk_x2 (40MHz)
-- mclk is variable, using the progclk_mux
---------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_ARITH.all;
use IEEE.STD_LOGIC_UNSIGNED.all;
use IEEE.STD_LOGIC_MISC.all;

library work;
use work.adb3_model_defs.all;
use work.adb3_board_ctrl.all;

library unisim;
use unisim.vcomponents.all;


---------------------------------------------------------------
entity model_clocks_ku1tgt is
   generic (
      OCPCLKDIV      : integer := 5
   );                 
   port (
     rclk_bufg          : in  std_logic;
      --
      lclk_locked     : out std_logic;
      --
      rclk            : out std_logic;
      ocp_ref_clk     : out std_logic;
      lclk            : out std_logic;
      lclk270         : out std_logic;
      lclk2x          : out std_logic;
      lclkEn400k      : out std_logic;
      lclkEn1pps      : out std_logic
   );
end model_clocks_ku1tgt;

---------------------------------------------------------------
architecture rtl of model_clocks_ku1tgt is


signal dcm_rst_count   : std_logic_vector(7 downto 0) := x"00";
signal dcm_rst         : std_logic := '1';
signal CLKFBOUT        : std_logic;
signal rclk0           : std_logic;
signal lclk0           : std_logic := '0';
signal lclk_int        : std_logic := '0';
signal lclk1           : std_logic := '0';
signal lclk2           : std_logic := '0';
signal ocpclk0         : std_logic := '0';
signal clk400k         : std_logic := '0';
signal clk400k_pipe    : std_logic_vector(2 downto 0) := (others => '0');
signal lclkEn400k_int  : std_logic := '0';
signal lclk1pps_cntval : std_logic_vector(18 downto 0) := (others => '0');

signal clk400k_counter : std_logic_vector(6 downto 0) := (others => '0');

--------------------------------------
begin

    -------------------------------------------
  
   -------------------------------------------
   dcm_rst0 : process (rclk_bufg)
   -------------------------------------------
   begin  -- process dcm_rst0
      if rising_edge(rclk_bufg) then
         if dcm_rst_count /= x"FF" then
            dcm_rst_count <= dcm_rst_count + 1;
         end if;
         dcm_rst <= not and_reduce(dcm_rst_count);
      end if;  -- rclk_bufg
   end process dcm_rst0;
   
   -------------------------------------------
   pll_lclk : MMCM_ADV
   -------------------------------------------
      generic map(
         BANDWIDTH             => "OPTIMIZED",
         CLKFBOUT_MULT_F       => 8.0,   -- VCO = 800MHz
         CLKFBOUT_PHASE        => 0.0,
         CLKIN1_PERIOD         => 3.333,
         CLKIN2_PERIOD         => 0.000,
         --
         CLKOUT0_DIVIDE_F      => 4.0,  -- 200MHz
         CLKOUT1_DIVIDE        => 20,  -- 40MHz
         CLKOUT2_DIVIDE        => 20,  -- 40MHz + 270deg
         CLKOUT3_DIVIDE        => 10,  -- 80MHz
         CLKOUT4_DIVIDE        => adb3_400kHz_div,  -- 400kHz
         CLKOUT5_DIVIDE        => OCPCLKDIV,
         CLKOUT6_DIVIDE        => 80,  -- 10MHz for clkout4
         --
         CLKOUT0_PHASE         => 0.000,
         CLKOUT1_PHASE         => 0.000,
         CLKOUT2_PHASE         => 270.000,
         CLKOUT3_PHASE         => 0.000,
         CLKOUT4_PHASE         => 0.000,
         CLKOUT5_PHASE         => 0.000,
         CLKOUT6_PHASE         => 0.000,
         --
         CLKOUT0_DUTY_CYCLE    => 0.500,
         CLKOUT1_DUTY_CYCLE    => 0.500,
         CLKOUT2_DUTY_CYCLE    => 0.500,
         CLKOUT3_DUTY_CYCLE    => 0.500,
         CLKOUT4_DUTY_CYCLE    => 0.500,
         CLKOUT5_DUTY_CYCLE    => 0.500,
         CLKOUT6_DUTY_CYCLE    => 0.500,
         --
         CLKOUT4_CASCADE       => TRUE,
         DIVCLK_DIVIDE         => 3,
         REF_JITTER1           => 0.1
      )
      port map (
         -- Clock Outputs: 1-bit (each) User configurable clock outputs
         CLKOUT0 => rclk0,
         CLKOUT0B => open,
         CLKOUT1 => lclk0,
         CLKOUT1B => open,
         CLKOUT2 => lclk1,
         CLKOUT2B => open,
         CLKOUT3 => lclk2,
         CLKOUT3B => open,
         CLKOUT4 => clk400k,
         CLKOUT5 => ocpclk0,
         CLKOUT6 => open,
         -- DRP Ports: 16-bit (each) Dynamic reconfigration ports
         DO => open,
         DRDY => open,
         -- Dynamic Phase Shift Ports: 1-bit (each) Ports used for dynamic phase shifting of the outputs
         PSDONE => open,
         -- Feedback Clocks: 1-bit (each) Clock feedback ports
         CLKFBOUT => CLKFBOUT,
         CLKFBOUTB => open,
         -- Status Ports: 1-bit (each) MMCM status ports
         CLKFBSTOPPED => open,
         CLKINSTOPPED => open,
         LOCKED => lclk_locked,
         -- Clock Inputs: 1-bit (each) Clock inputs
         CLKIN1 => rclk_bufg,
         CLKIN2 => '0',
         -- Control Ports: 1-bit (each) MMCM control ports
         CLKINSEL => '1',
         PWRDWN => '0',
         RST => dcm_rst,
         -- DRP Ports: 7-bit (each) Dynamic reconfigration ports
         DADDR => "0000000",
         DCLK => '0',
         DEN => '0',
         DI => x"0000",
         DWE => '0',
         -- Dynamic Phase Shift Ports: 1-bit (each) Ports used for dynamic phase shifting of the outputs
         PSCLK => '0',
         PSEN => '0',
         PSINCDEC => '0',
         -- Feedback Clocks: 1-bit (each) Clock feedback ports
         CLKFBIN => CLKFBOUT
      );    --  pll_lclk

  
   -------------------------------------------
   rclk_bufg_1 : BUFG
   -------------------------------------------
      port map (
         I => rclk0,
         O => rclk
      );
      
      

   -------------------------------------------
   lclk_bufg_0 : BUFG
   -------------------------------------------
      port map (
         I => lclk0,
         O => lclk_int
      );
      
   lclk <= lclk_int;

   -------------------------------------------
   lclk_bufg_1 : BUFG
   -------------------------------------------
      port map (
         I => lclk1,
         O => lclk270
      );

   
   -------------------------------------------
   lclk_bufg_2 : BUFR
   -------------------------------------------
      port map (
         CLR => '0',
         CE => '1',
         I => lclk2,
         O => lclk2x
      );
 

   -------------------------------------------
   ocpclk_bufg_3 : BUFG
   -------------------------------------------
      port map (
         I => ocpclk0,
         O => ocp_ref_clk
      );

   -------------------------------------------
   -- 400kHz clock enable, synchronous to 40MHz lclk 
   -------------------------------------------
   clken_proc : process(lclk_int)
   begin
      if rising_edge(lclk_int) then
        --clk400k_pipe <= clk400k_pipe(1 downto 0) & clk400k;
         
        --if clk400k_pipe(1) = '1' and clk400k_pipe(2) = '0' then
        --   lclkEn400k_int <= '1';
        --else
        --   lclkEn400k_int <= '0';
        --end if;
        -- Divide 40MHz by 100
        if clk400k_counter = "1100011" then
          clk400k_counter <= (others => '0');
          lclkEn400k_int <= '1';
        else
          clk400k_counter <=  clk400k_counter +1;
          lclkEn400k_int <= '0';
        end if;
      end if;
   end process clken_proc;
   
   lclkEn400k <= lclkEn400k_int;

   
   -------------------------------------------
   -- 1pps clock enable, synchronous to 40MHz lclk 
   -------------------------------------------
   onepps_proc : process(lclk_int)
   begin
      if rising_edge(lclk_int) then
         if lclkEn400k_int = '1' then
            if lclk1pps_cntval = (adb3_1pps_div -1) then
               lclk1pps_cntval <= (others => '0');
            else
               lclk1pps_cntval <= lclk1pps_cntval + 1;
            end if;
         end if;  -- lclken_400k_int (clock enable)

         if (lclk1pps_cntval = (adb3_1pps_div -1)) and
               (clk400k_pipe(1) = '1') and 
               (clk400k_pipe(2) = '0') then
            lclkEn1pps      <= '1';
         else
            lclkEn1pps      <= '0';
         end if;
   
      end if;  -- lclk_int
   end process onepps_proc;
   

end rtl; -- model_clocks
