--------------------------------------------------------------------------------
--
-- Module Name      : descriptor_fetch
--
--                    Copyright 2009 by
--                    Alpha Data Parallel Systems Ltd.
--                    All Rights Reserved.
--
-- Original Author  : Andrew McCormick
-- Created          : 
--
-- Modified By      : 
-- Date             :
-- Change Notes     :     
--
-- Description      : 
--
--
-- Dependencies     : adb3_ocp,adb3_dma
--
-- Disclaimer       : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                    WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                    WARRANTIES IMPLIED OF MERCHANTABILITY, FITNESS FOR A
--                    PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.adb3_dma.all;


entity descriptor_fetch256_pre is
  generic (
    registering_for_speed : boolean := true);
  port (
    clk              : in  std_logic;
    -- ABORT/Cleanup Control
    abort            : in  std_logic;
    cleanup          : in  std_logic;
    -- FSM Status
    fsm_idle         : out std_logic;
    fsm_aborting     : out std_logic;
    fsm_fail         : out std_logic;
    -- Register Control
    new_desc_addr    : in  std_logic_vector(63 downto 0);
    new_desc_addr_we : in  std_logic;
    -- OCP Master Port (to MUX to PCIe/or FPGA)
    ocp_m2s          : out adb3w_ocp_m2sT;
    ocp_s2m          : in  adb3w_ocp_s2mT;
    -- PCIE Worker DMA Engine Control
    pcie_worker_cmd       : out adb3_dma_worker_command_type := ADB3_DMA_WORKER_COMMAND_TYPE_DEFAULT;
    pcie_worker_cmd_valid : out std_logic := '0';
    pcie_worker_ready     : in  std_logic;
    -- MPTL Worker DMA Engine Control
    mptl_worker_cmd       : out adb3_dma_worker_command_type := ADB3_DMA_WORKER_COMMAND_TYPE_DEFAULT;
    mptl_worker_cmd_valid : out std_logic := '0';
    mptl_worker_ready     : in  std_logic;
    -- IRQ Control
    irq              : out std_logic;
    irq_ack          : in  std_logic;
    -- Wait Control
    waiting          : out std_logic;
    waiting_ack      : in  std_logic;
    write_finished   : in  std_logic;
    -- Debug Counter
    debug_count      : out std_Logic_vector(31 downto 0);
    -- Alignment Control
    alignment        : out std_logic_vector(4 downto 0) := "00000";
    alignment_ifw    : out std_logic := '0';
    alignment_set    : out std_logic := '0';
    enable_prefetch  : in  std_logic);

end descriptor_fetch256_pre;

architecture rtl of descriptor_fetch256_pre is

  
  
  
  type state_type is (IDLE, FETCH_RDCMD, FETCH_RDATA0,  WAIT_1_CYCLE, SET_IRQ_NOW, CALC_ALIGNMENT, PARSE_DESC, WAIT_FLUSH, ABORT0, ABORT1, FAIL, FLUSH_PREFETCH_BUFFER);
  constant DESC_BURST_LENGTH : integer := 1;
  constant DESC_BURST_LENGTH_PREFETCH : integer := 8;
  
  signal state : state_type := ABORT0;

  signal current_desc_addr  : std_logic_vector(63 downto 0) := (others => '0');
  signal current_desc_valid : std_logic := '0';
  

  signal mcmd         : ocp_CmdT                                          := OCP_CMD_IDLE;
  signal maddr        : std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0)  := (others => '0');
  signal mburstlength : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0) := (others => '0');
  signal mrespaccept  : std_logic                                         := '0';

  signal current_desc_slv : std_logic_vector(255 downto 0) := (others => '0');
  signal current_desc_slv_r1 : std_logic_vector(255 downto 0) := (others => '0');
  signal current_desc_slv_valid : std_logic := '0';
  signal current_desc_slv_valid_r1 : std_logic := '0';
  
  signal irq_state : std_logic := '0';
  signal waiting_state : std_logic := '0';
  signal current_desc : adb3_dma_descriptor_type;

  signal workers_ready : std_logic := '0';
  signal workers_ready1 : std_logic := '0';

  signal mptl_worker_cmd1       : adb3_dma_worker_command_type := ADB3_DMA_WORKER_COMMAND_TYPE_DEFAULT;
  signal mptl_worker_cmd_valid1 : std_logic := '0';

  signal local_length_adjustment : std_logic_vector(5 downto 0) := (others => '0');
  
  constant ones : std_logic_vector(63 downto 0) := X"FFFFFFFFFFFFFFFF";


  signal calculate_alignment : std_logic := '0';
  signal wait_for_flush      : std_logic := '0';

  signal try_prefetch_read : std_logic := '0';
  signal flush_prefetch    : std_logic := '0';
  signal ocp_s2m_resp_valid : std_logic;
  signal use_pre_fetch : std_logic := '0';
  signal resp_data : std_logic_vector(255 downto 0);
  signal resp_fifo_empty : std_logic;
  signal resp_fifo_radv : std_logic;

  signal resp_in_flight_cnt : std_logic_vector(3 downto 0) := (others => '0');
  signal resp_in_flight : std_logic := '0';

  signal next_desc_matches : std_logic := '0';
  signal rst_fifo : std_logic := '1';

  signal debug_counter : std_logic_vector(31 downto 0) := (others => '0');
  
begin  -- rtl

  ocp_m2s.Cmd         <= mcmd;
  ocp_m2s.Addr        <= maddr;
  ocp_m2s.BurstLength <= CONV_STD_LOGIC_VECTOR(DESC_BURST_LENGTH, ADB3_OCP_BURST_WIDTH) when
                         try_prefetch_read = '0' else
                         CONV_STD_LOGIC_VECTOR(DESC_BURST_LENGTH_PREFETCH, ADB3_OCP_BURST_WIDTH);
  ocp_m2s.Data        <= (others => '0');
  ocp_m2s.DataByteEn  <= (others => '0');
  ocp_m2s.DataValid   <= '0';
  ocp_m2s.RespAccept  <= '1';
  ocp_m2s.Tag         <= (others => '0');

  irq <= irq_state;
  waiting <= waiting_state;

  process (clk) is
  begin  -- process
    if rising_edge(clk) then            -- rising clock edge
      if state = ABORT0 or state = ABORT1 or abort = '1' then
        rst_fifo <= '1';
      else
        rst_fifo <= '0';
      end if;
    end if;
  end process;
  

  small_sfifo_1: entity work.small_sfifo
    generic map (
      width => 256)
    port map (
      clk    => clk,
      rst    => rst_fifo,
      wadv   => ocp_s2m_resp_valid,
      wdata  => ocp_s2m.Data,
      nfull  => open,
      radv   => resp_fifo_radv,
      rempty => resp_fifo_empty,
      rdata  => resp_data);
  resp_fifo_radv <= mrespaccept or flush_prefetch;
  ocp_s2m_resp_valid <= '1' when ocp_s2m.Resp = OCP_RESP_VALID else '0';
  
  
  gen_rfs0: if not registering_for_speed generate
    current_desc <= slv_to_adb3_dma_descriptor(current_desc_slv);
    workers_ready <= pcie_worker_ready and mptl_worker_ready;
    mptl_worker_cmd_valid <= mptl_worker_cmd_valid1;
    mptl_worker_cmd.start_addr <= mptl_worker_cmd1.start_addr;   
    mptl_worker_cmd.xfer_length       <=
      mptl_worker_cmd1.xfer_length + (ones(adb3_dma_xfer_bits_used-1 downto 5)  & local_length_adjustment(4 downto 0))
      when local_length_adjustment(5) = '1'
      else mptl_worker_cmd1.xfer_length + EXT(local_length_adjustment(4 downto 0),adb3_dma_xfer_bits_used);
    mptl_worker_cmd.read_data <= mptl_worker_cmd1.read_data; 
    mptl_worker_cmd.irq_on_completion <= mptl_worker_cmd1.irq_on_completion;
    mptl_worker_cmd.last_descriptor <= mptl_worker_cmd1.last_descriptor;
    mptl_worker_cmd.fixed_addr <= mptl_worker_cmd1.fixed_addr;
  end generate gen_rfs0;
  
  fsm0 : process (clk)
    variable mptl_addr,pcie_addr,alignment_diff : std_logic_vector(5 downto 0);
    variable end_of_desc_addr_sum : std_logic_vector(4 downto 0);
  begin  -- process fsm0
    if clk'event and clk = '1' then     -- rising clock edge
      if registering_for_speed then
        current_desc_slv_valid_r1 <= current_desc_slv_valid;
        if current_desc_slv_valid_r1 = '1' then
          current_desc <= slv_to_adb3_dma_descriptor(current_desc_slv_r1);
        end if;       
        current_desc_slv_r1<= current_desc_slv;      
        workers_ready1 <= pcie_worker_ready and mptl_worker_ready;
        workers_ready <=  workers_ready1;
        mptl_worker_cmd_valid <= mptl_worker_cmd_valid1;
        
        mptl_worker_cmd.start_addr <= mptl_worker_cmd1.start_addr;
        
        if local_length_adjustment(5) = '1' then
          mptl_worker_cmd.xfer_length <= mptl_worker_cmd1.xfer_length + (ones(adb3_dma_xfer_bits_used-1 downto 5)  & local_length_adjustment(4 downto 0));
        else
          mptl_worker_cmd.xfer_length  <= mptl_worker_cmd1.xfer_length + EXT(local_length_adjustment(4 downto 0),adb3_dma_xfer_bits_used);
        end if;
        mptl_worker_cmd.read_data <= mptl_worker_cmd1.read_data; 
        mptl_worker_cmd.irq_on_completion <= mptl_worker_cmd1.irq_on_completion;
        mptl_worker_cmd.last_descriptor <= mptl_worker_cmd1.last_descriptor;
        mptl_worker_cmd.fixed_addr <= mptl_worker_cmd1.fixed_addr;
      end if;
      pcie_worker_cmd_valid <= '0';
      mptl_worker_cmd_valid1 <= '0';
      current_desc_slv_valid <= '0';
      alignment_set <='0';
      try_prefetch_read <= '0';
      case state is
        when IDLE =>
          wait_for_flush <= '0';
          use_pre_fetch <= '0';
          flush_prefetch <= '0';
          if abort = '1' or cleanup = '1' then
            current_desc_valid <= '0';
          elsif current_desc_valid = '1' then
            state              <= FETCH_RDCMD;
            mcmd               <= OCP_CMD_READ;
            maddr              <= current_desc_addr;
            current_desc_valid <= '0';
            if current_desc_addr(7 downto 4) = "0000" then
              try_prefetch_read <= enable_prefetch;
            else
              try_prefetch_read <= '0';
            end if;
            
          else
            mcmd <= OCP_CMD_IDLE;
            if new_desc_addr_we = '1' then
              current_desc_valid <= '1';
              current_desc_addr  <= new_desc_addr;
              calculate_alignment <= '1';
              if new_desc_addr(7 downto 4) = "0000" then
                try_prefetch_read <= enable_prefetch;
              else
                try_prefetch_read <= '0';
              end if;
            end if;
          end if;        
          mrespaccept <= '0';
        when FETCH_RDCMD =>
          flush_prefetch <= '0';
          mrespaccept <= '0';
          if ocp_s2m.CmdAccept = '1' then
            mcmd <= OCP_CMD_IDLE;
            if abort = '1' then
              state <= ABORT0;
            else
              state <= FETCH_RDATA0;            
            end if;         
          else
            try_prefetch_read <= try_prefetch_read;
            if abort = '1' then
              mcmd  <= OCP_CMD_IDLE;
              state <= ABORT0;
            end if;
          end if;
        when FETCH_RDATA0 =>
          mrespaccept <= '1';
          flush_prefetch <= '0';
          --if ocp_s2m.Resp = OCP_RESP_VALID then
          if resp_fifo_empty = '0' and mrespaccept = '1' then
            if abort = '1' then
              state <= ABORT0;
            else
              state                          <= WAIT_1_CYCLE;            
            end if;
            --current_desc_slv <= ocp_s2m.Data;
            current_desc_slv <= resp_data;
            mrespaccept <= '0';
            current_desc_slv_valid <= '1';
          --elsif ocp_s2m.Resp = OCP_RESP_FAILED or ocp_s2m.Resp = OCP_RESP_ERROR then
          --  state <= FAIL;
          else
            if abort = '1' then
              state <= ABORT0;
            end if;
          end if;
          
        --when FETCH_RDATA1 =>
        --  if ocp_s2m.Resp = OCP_RESP_VALID then
        --    if abort = '1' then
        --      state <= IDLE;             
        --    else
        --      state                            <= WAIT_1_CYCLE;           
        --    end if;
        --    mrespaccept <= '0';
        --    current_desc_slv_valid <= '1';
        --    current_desc_slv(255 downto 128) <= ocp_s2m.Data;
        --    current_desc_slv(127 downto 0) <= current_desc_slv(255 downto 128);
        --  elsif ocp_s2m.Resp = OCP_RESP_FAILED or ocp_s2m.Resp = OCP_RESP_ERROR then
        --    state <= FAIL;
        --  else
        --    if abort = '1' then
        --      state <= ABORT1;
        --    end if;
        --  end if;
          
        when WAIT_1_CYCLE =>
          mrespaccept <= '0';
          flush_prefetch <= '0';
          if abort = '1' then
            state <= ABORT0;
          else
            state <= SET_IRQ_NOW;
          end if;
        when SET_IRQ_NOW =>
          mrespaccept <= '0';
          flush_prefetch <= '0';
          if abort = '1' then
            state <= ABORT0;
          else
            state <= CALC_ALIGNMENT;
          end if;
        when CALC_ALIGNMENT =>
          mrespaccept <= '0';
          flush_prefetch <= '0';
          if abort = '1' then
            state <= ABORT0;      
          else
            state <= PARSE_DESC;
            alignment_set <= calculate_alignment;                      
          end if;
          calculate_alignment <= '0';
          mptl_addr := '0' & current_desc.local_addr(4 downto 0);
          pcie_addr := '0' & current_desc.start_addr(4 downto 0);
          alignment_diff := mptl_addr - pcie_addr;
          alignment <= alignment_diff(4 downto 0);
          
          if current_desc.read_data = '1' then
            alignment_ifw <= alignment_diff(5); 
          else
            alignment_ifw <= (not alignment_diff(5)) and OR_reduce(alignment_diff(4 downto 0)); 
          end if;       
          
          end_of_desc_addr_sum := current_desc.start_addr(4 downto 0) + current_desc.xfer_length(4 downto 0);
          if (OR_reduce(end_of_desc_addr_sum) = '1' or current_desc.force_flush = '1') and current_desc.last_descriptor = '0' then
             wait_for_flush <= '1';
          end if;
        when PARSE_DESC =>
          flush_prefetch <= '0';
          mrespaccept <= '0';
          current_desc_valid <= '0';
          if abort = '1' then
            state <= ABORT0;
          elsif (workers_ready = '1' and (current_desc.wait_irq_ack = '0' or waiting_state = '0')) then
            state <= WAIT_FLUSH;
            pcie_worker_cmd_valid             <= '1';
            mptl_worker_cmd_valid1             <= '1';
            if current_desc.last_descriptor = '0' then
              current_desc_valid <= '1';
              current_desc_addr  <= current_desc.next_descriptor;
              if (current_desc_addr(11 downto 0)+32 = current_desc.next_descriptor(11 downto 0)) and (next_desc_matches = '1') then
                use_pre_fetch <= (not resp_fifo_empty) or resp_in_flight; 
              else
                use_pre_fetch <= '0';
              end if;
            else
              use_pre_fetch <= '0';
            end if;     
          end if;
--          if (workers_ready = '1' and (current_desc.wait_irq_ack = '0' or waiting_state = '0')) then        
            if registering_for_speed then
              pcie_worker_cmd.start_addr        <= current_desc.start_addr;
              pcie_worker_cmd.xfer_length       <= current_desc.xfer_length;
              pcie_worker_cmd.read_data         <= current_desc.read_data;
              pcie_worker_cmd.irq_on_completion <= current_desc.irq_on_pcie_completion;
              pcie_worker_cmd.last_descriptor   <= current_desc.last_descriptor or wait_for_flush;
              pcie_worker_cmd.fixed_addr        <= '0';
              mptl_worker_cmd1.start_addr        <= EXT(current_desc.local_addr,64);
              mptl_worker_cmd1.xfer_length       <= current_desc.xfer_length;
              local_length_adjustment <= current_desc.local_length_adjustment(5 downto 0); 
--               if current_desc.local_length_adjustment(4) = '1' then
--                 mptl_worker_cmd1.xfer_length       <= current_desc.xfer_length + (ones(adb3_dma_xfer_bits_used-1 downto 4)  & current_desc.local_length_adjustment(3 downto 0));
--               else
--                 mptl_worker_cmd1.xfer_length       <= current_desc.xfer_length + EXT(current_desc.local_length_adjustment(3 downto 0),adb3_dma_xfer_bits_used);
--               end if;              
              mptl_worker_cmd1.read_data         <= not current_desc.read_data;
              mptl_worker_cmd1.irq_on_completion <= current_desc.irq_on_mptl_completion;
              mptl_worker_cmd1.last_descriptor   <= current_desc.last_descriptor or wait_for_flush;
              mptl_worker_cmd1.fixed_addr        <= current_desc.fixed_addr;
            end if;
                 
 --         end if;
        when WAIT_FLUSH =>
          flush_prefetch <= '0';
          if abort = '1' then
            state <= ABORT0;
            wait_for_flush <= '0';
          elsif wait_for_flush = '0' or write_finished = '1' then
            if use_pre_fetch = '1' then
              state <= FETCH_RDATA0;
            elsif resp_fifo_empty = '0' or resp_in_flight = '1' then
              state <= FLUSH_PREFETCH_BUFFER;
            else
              state <= IDLE;
            end if;
            
            wait_for_flush <= '0';
            if wait_for_flush = '1' then
              calculate_alignment <= '1';
            end if;         
          end if;
        when ABORT0 =>
          flush_prefetch <= '0';
          current_desc_valid <= '0';
          if cleanup = '1' then
            state <= IDLE;
          elsif ocp_s2m.Resp = OCP_RESP_VALID then
            state <= ABORT1;
          elsif ocp_s2m.Resp = OCP_RESP_FAILED or ocp_s2m.Resp = OCP_RESP_ERROR then
            state <= FAIL;
          end if;
          mrespaccept <= '1';
        when ABORT1 =>
          flush_prefetch <= '0';
          current_desc_valid <= '0';
          if cleanup = '1' then
            state <= IDLE;
--          elsif ocp_s2m.Resp = OCP_RESP_VALID then
--            state <= IDLE;
          elsif ocp_s2m.Resp = OCP_RESP_FAILED or ocp_s2m.Resp = OCP_RESP_ERROR then
            state <= FAIL;
          end if;
          mrespaccept <= '1';
        when FAIL =>
          flush_prefetch <= '0';
          current_desc_valid <= '0';
          if cleanup = '1' then
            state <= IDLE;
          end if;
          mrespaccept <= '1';
        when FLUSH_PREFETCH_BUFFER =>
          flush_prefetch <=  (not resp_in_flight) and (not flush_prefetch) and (not resp_fifo_empty);
          --current_desc_valid <= '0';
          if resp_fifo_empty = '1' and resp_in_flight = '0' then
            state <= IDLE;
            flush_prefetch <= '0';
          end if;
        when others =>
          state <= IDLE;
      end case;


      if abort = '1' or cleanup = '1' then
        irq_state <= '0';
      elsif state = SET_IRQ_NOW and current_desc.irq_now = '1' then
        irq_state <= '1';
      elsif irq_ack = '1' then
        irq_state <= '0';
      end if;

      if abort = '1' or cleanup = '1' then
        waiting_state <= '0';
      elsif state = SET_IRQ_NOW and current_desc.irq_now = '1' then
        waiting_state <= '1';
      elsif waiting_ack = '1' then
        waiting_state <= '0';
      end if;

      -- Output Status Info
      -- Only for host read status regs, not for FPGA control
      if state = IDLE and current_desc_valid = '0' then
        fsm_idle <= '1';
      else
        fsm_idle <= '0';
      end if;
      if state = ABORT0 or state = ABORT1 then
        fsm_aborting <= '1';
      else
        fsm_aborting <= '0';
      end if;
      
      if state = FAIL then
        fsm_fail <= '1';
      else
        fsm_fail <= '0';
      end if;
      --fsm_fail <= OR_reduce(current_desc.start_addr(63 downto 32));

      if abort = '1' then
        resp_in_flight_cnt <= (others => '0');
      elsif mcmd = OCP_CMD_READ and ocp_s2m.CmdAccept = '1' then
        if try_prefetch_read = '1' then
          resp_in_flight_cnt <= CONV_STD_LOGIC_VECTOR(DESC_BURST_LENGTH_PREFETCH, 4);
        else
          resp_in_flight_cnt <= CONV_STD_LOGIC_VECTOR(DESC_BURST_LENGTH, 4);
        end if;
      elsif ocp_s2m_resp_valid = '1' then
         resp_in_flight_cnt <=  resp_in_flight_cnt -1;        
      end if;
      resp_in_flight <= OR_reduce(resp_in_flight_cnt);

      if current_desc_addr(63 downto 12) = current_desc.next_descriptor(63 downto 12) then
        next_desc_matches <= '1';
      else
        next_desc_matches <= '0';
      end if;

      if state = FETCH_RDATA0 then
        debug_counter <= debug_counter+1;
      else
        debug_counter <= (others => '0');
      end if;
      debug_count <= debug_counter;
      
    end if;
  end process fsm0;
  
  gen_worker_cmd_small: if not registering_for_speed generate
     pcie_worker_cmd.start_addr        <= current_desc.start_addr;
     pcie_worker_cmd.xfer_length       <= current_desc.xfer_length;
     pcie_worker_cmd.read_data         <= current_desc.read_data;
     pcie_worker_cmd.irq_on_completion <= current_desc.irq_on_pcie_completion;
     pcie_worker_cmd.last_descriptor   <= current_desc.last_descriptor or wait_for_flush;
     pcie_worker_cmd.fixed_addr        <= '0';

     mptl_worker_cmd1.start_addr        <= EXT(current_desc.local_addr,64);
--      mptl_worker_cmd1.xfer_length       <=
--        current_desc.xfer_length + (ones(adb3_dma_xfer_bits_used-1 downto 4)  & current_desc.local_length_adjustment(3 downto 0))
--      when current_desc.local_length_adjustment(4) = '1'
--        else current_desc.xfer_length + EXT(current_desc.local_length_adjustment(3 downto 0),adb3_dma_xfer_bits_used);

     mptl_worker_cmd1.xfer_length       <= current_desc.xfer_length;
     local_length_adjustment <= current_desc.local_length_adjustment(5 downto 0); 
     
     mptl_worker_cmd1.read_data         <= not current_desc.read_data;
     mptl_worker_cmd1.irq_on_completion <= current_desc.irq_on_mptl_completion;
     mptl_worker_cmd1.last_descriptor   <= current_desc.last_descriptor or wait_for_flush;
     mptl_worker_cmd1.fixed_addr        <= current_desc.fixed_addr;
  end generate gen_worker_cmd_small;
  --mrespaccept <= '1' when state = ABORT0 or state = ABORT1 or state = FETCH_RDATA0 or state = FETCH_RDATA1 else '0';
  
  
end rtl;

