--------------------------------------------------------------------------------
-- Company:         Alpha-Data Parallel Systems
-- Engineer:        Alan Crawford
--
-- Create Date:
-- Design Name:     ADB3
-- Module Name:     
-- Project Name:    Gen3 adb3
-- Target Device:   V7
-- Tool versions:
-- Description:     
-- Dependencies:    PCIe/OCP Packages for signal records
--
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
--
--------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.pcie.all;
use work.adb3_ocp.all;
use work.pciegen3_pkg.all;
use work.pcie_config.all;


entity bar_tag_manager is
  generic(
    max_timeout             : std_logic_vector(15 downto 0) := x"FFFF";
    single_req              : boolean := true
  );
  port(
    pcie_rst                : in std_logic;
    pcie_clk                : in std_logic;
    us_counter              : in std_logic;
    
    -- Tag requester I/F
    tx_tag_request          : in  std_logic;
    tx_tag_data_in          : in  std_logic_vector(71 downto 0);
    tx_tag_number           : out std_logic_vector(2 downto 0);
    tx_tag_valid            : out std_logic;

    -- Tag checker I/F
    rx_tag_value            : in  std_logic_vector(2 downto 0);
    rx_tag_clear            : in  std_logic;
    rx_tag_valid            : out std_logic;
    rx_tag_info             : out std_logic_vector(71 downto 0);
    
    -- status and error I/O
    tag_timeout_enable      : in  std_logic;                      -- enables timeout logic
    clr_timeout             : in  std_logic;                      -- Clear timeout overflow flag
    tag_timeout             : out std_logic;                      -- no response within desired timeout period
    flush_tags              : out std_logic;                      -- flush any tags active during the timeout
    clear_tags              : out std_logic;                      -- hold of requests when timeout has been cleared
    tag_error               : out std_logic;                      -- when a tag is returned but is not active
    status                  : out std_logic_vector(15 downto 0)   -- contains activity flags and timeout settings
  );  
end bar_tag_manager;

architecture rtl of bar_tag_manager is


  constant no_tags_available            : natural := 8;

  subtype tag_data_t                    is std_logic_vector(71 downto 0);
  type    tag_array_t                   is array(0 to no_tags_available-1) of tag_data_t;

  signal tag_array                      : tag_array_t;
  signal tag_active                     : std_logic_vector(no_tags_available-1 downto 0);
  signal tag_counter                    : std_logic_vector(2 downto 0);
  signal tags_in_flight                 : std_logic_vector(2 downto 0);


  signal timeout_counter                : std_logic_vector(15 downto 0);  -- full timeout counter

  signal timeout_overflow               : std_logic;  -- timeout has occurred
  signal no_tags_active                 : std_logic;  -- only activate timeout when we have a read request in flight

  signal tag_return                     : std_logic := '0';
  signal tag_return_enable              : std_logic := '0';
  signal tag_return_enable_r1           : std_logic := '0';
  signal tag_delay                      : std_logic_vector(4 downto 0);

  signal clear_tags_int                 : std_logic := '0';
  
  signal local_rst,local_sync_rst : std_logic := '1';
  attribute keep : string;
  attribute max_fanout : string;
  attribute equivalent_register_removal : string;
  attribute keep of local_rst : signal is "TRUE";
  attribute equivalent_register_removal of local_rst : signal is "no";
  attribute max_fanout of local_rst : signal is "1";
  attribute keep of local_sync_rst : signal is "TRUE";
  attribute equivalent_register_removal of local_sync_rst : signal is "no";

  signal local_rst_r1: std_logic := '1';
  attribute shreg_extract : string;
  attribute keep of local_rst_r1 : signal is "TRUE";  
  attribute equivalent_register_removal of local_rst_r1 : signal is "no";
  attribute max_fanout of local_sync_rst : signal is "16";
  attribute shreg_extract of local_rst_r1 : signal is "no";
  attribute shreg_extract of local_sync_rst : signal is "no";
  
begin

  process (pcie_clk,pcie_rst) is
  begin  -- process
    if pcie_rst = '1' then
      local_rst <= '1';
    elsif rising_edge(pcie_clk) then            -- rising clock edge
      local_rst <= '0';
    end if;
  end process;
  
  process (pcie_clk) is
  begin  -- process
    if rising_edge(pcie_clk) then       -- rising clock edge
      local_rst_r1 <= local_rst;
      local_sync_rst <= local_rst_r1;
    end if;
  end process;
  
  tag_timeout <= timeout_overflow;
  tag_error <= '0';
  status <= (others => '0');
-- ----------------------------------------------------------------------------
-- timeout
  timeout_count : process(pcie_clk)
  begin
    if pcie_clk'event and pcie_clk = '1' then
      if(local_sync_rst = '1')then
        timeout_counter  <= max_timeout;
        timeout_overflow <= '0';
      else
        if(tags_in_flight = "000")then
          no_tags_active <= '1';
        else
          no_tags_active <= '0';
        end if;
        
        -- only operate when enabled
        
        if(tag_timeout_enable = '1')then

          -- only change on 1us Strobe = '1'
          
          if(us_counter = '1')then

            -- only count when there are reads in flight (this should pulse when a new packet is requested to ensure
            -- the timeout is from the last read request sent (stop the timeout occuring when there is always a PIF
            -- even though it they are different requests
            
            if((no_tags_active = '1')or(tx_tag_request = '1'))then
              timeout_counter <= max_timeout;
            else
              if(timeout_counter /= X"0000")then
                timeout_counter <= timeout_counter - '1';
              else
                timeout_overflow <= '1';
              end if;
            end if;
          else
            if((no_tags_active = '1')or(tx_tag_request = '1'))then
              timeout_counter <= max_timeout;
            end if;
          end if;
        else
          timeout_overflow <= '0';
          timeout_counter  <= max_timeout;
        end if;

        -- allow software to clear the overflow

        if(clr_timeout = '1')then
          timeout_overflow <= '0';
        end if;
      end if;
    end if;
    
  end process timeout_count;

  gen_clear_tags : process(pcie_clk)
  begin
    if pcie_clk'event and pcie_clk = '1' then
      if(local_sync_rst = '1')then
        clear_tags_int   <= '0';
        flush_tags <=  '0';
      else
        if(timeout_overflow = '1')then
          flush_tags <= '1';
        end if;
        
        if(clr_timeout = '1')then
          clear_tags_int <= '1';
        end if;
        if((clear_tags_int = '1')and(tag_active = x"00"))then
          clear_tags_int <= '0';
          flush_tags <= '0';
        end if;
      end if;
    end if;
  end process;
  
  clear_tags <= clear_tags_int;
  
  gen_tag_val : process(pcie_clk)
  begin
    if pcie_clk'event and pcie_clk = '1' then
      if(local_sync_rst = '1')then
        tx_tag_valid   <= '1';
      else
        if(tags_in_flight = "111")then
          tx_tag_valid <= '0';
        else
          if((tag_active(conv_integer(tag_counter)) = '0'))then
            if(tags_in_flight = "110")then
              tx_tag_valid <= '0';
            else
              tx_tag_valid <= '1';
            end if;
          else
            tx_tag_valid <= '0';
          end if;
        end if;
      end if;
    end if;
  end process;
  
  gen_tag_count : process(pcie_clk)
  begin
    if pcie_clk'event and pcie_clk = '1' then
      if(local_sync_rst = '1')then
        tag_counter    <= "000";
      else
        if(tags_in_flight /= "111")then
          if((tx_tag_request = '1'))then
            tag_counter  <= tag_counter+'1';
          end if;
        end if;
      end if;
    end if;
  end process;
  
  gen_tag_active : process(pcie_clk)
  begin
    if pcie_clk'event and pcie_clk = '1' then
      if(local_sync_rst = '1')then
        tag_active     <= (others => '0');
      else
        if(tags_in_flight /= "111")then
          if((tx_tag_request = '1'))then
            tag_active(conv_integer(tag_counter)) <= '1';
          end if;
        end if;
        if(rx_tag_clear = '1')then
          tag_active(conv_integer(rx_tag_value)) <= '0';
        end if;
      end if;
    end if;
  end process;
  
  gen_tag_inflight : process(pcie_clk)
  begin
    if pcie_clk'event and pcie_clk = '1' then
      if(local_sync_rst = '1')then
        tags_in_flight <= "000";
      else
        if(tags_in_flight /= "111")then
          if((tx_tag_request = '1'))then
            if(rx_tag_clear = '0')then
              tags_in_flight <= tags_in_flight + '1';
            end if;
          end if;
        end if;
        if(rx_tag_clear = '1')then
          if(tags_in_flight > "000")then
            if(tx_tag_request = '0')then
              tags_in_flight <= tags_in_flight - '1';
            end if;
          end if;
        end if;
        if(timeout_overflow = '1')then
          tags_in_flight <= (others => '0');
        end if;
      end if;
    end if;
  end process;

  tx_tag_number <= tag_counter;

  get_pcie_tag_val : process(pcie_clk)
  begin
    if pcie_clk'event and pcie_clk = '1' then
      if(local_sync_rst = '1')then
        rx_tag_valid  <= '0';
      else
        rx_tag_valid <= tag_active(conv_integer(rx_tag_value));
      end if;
    end if;
  end process get_pcie_tag_val;
  
  access_ram : process(pcie_clk)
  begin
    if pcie_clk'event and pcie_clk = '1' then
      rx_tag_info <= tag_array(conv_integer(rx_tag_value));
      if((tx_tag_request = '1')and(tags_in_flight /= "111"))then
        tag_array(conv_integer(tag_counter))  <= tx_tag_data_in;
      end if;
    end if;
  end process;
           
  
end rtl;

