--------------------------------------------------------------------------------
-- Company:         Alpha-Data Parallel Systems
-- Engineer:        Alan Crawford
--
-- Create Date:
-- Design Name:
-- Module Name:     pcieg3_bar_arbiter
-- Project Name:    ADB3G3
-- Target Device:   V7
-- Tool versions:
-- Description:     Interfaces to the Gen3 Core BAR I/F and provides the
--                  required number of OCP ports
--
-- Dependencies:    PCIe/OCP Packages for signal records
--
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
--
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_ARITH.all;
use IEEE.STD_LOGIC_UNSIGNED.all;

library work;
use work.pcie.all;
use work.adb3_ocp.all;
use work.pciegen3_pkg.all;
use work.pcie_config.all;

entity pcieg3_bar_arbiter is
  generic(
    datawidthTX : integer := 128
    );
  port(
    pcie_clk : in std_logic;
    pcie_rst : in std_logic;

    rx_bar0_data_out     : in  std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
    rx_bar0_data_pop     : out std_logic := '0';
    rx_bar0_data_pkt_rdy : in  std_logic;

    rx_bar1_data_out     : in  std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
    rx_bar1_data_pop     : out std_logic := '0';
    rx_bar1_data_pkt_rdy : in  std_logic;

    rx_bar2_data_out     : in  std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
    rx_bar2_data_pop     : out std_logic := '0';
    rx_bar2_data_pkt_rdy : in  std_logic;

    rx_bar_data_out   : out std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
    rx_bar_data_push  : out std_logic := '0';
    rx_bar_data_space : in  std_logic

    );
end pcieg3_bar_arbiter;

architecture synth of pcieg3_bar_arbiter is

  signal sequencer  : std_logic_vector(1 downto 0) := "01";
  signal select_src : std_logic_vector(2 downto 0) := "001";


  signal rx_bar0_data_pop_int : std_logic;
  signal rx_bar1_data_pop_int : std_logic;
  signal rx_bar2_data_pop_int : std_logic;

  signal local_rst,local_sync_rst : std_logic := '1';
  attribute keep : string;
  attribute max_fanout : string;
  attribute equivalent_register_removal : string;
  attribute keep of local_rst : signal is "TRUE";
  attribute equivalent_register_removal of local_rst : signal is "no";
  attribute max_fanout of local_rst : signal is "1";
  attribute keep of local_sync_rst : signal is "TRUE";
  attribute equivalent_register_removal of local_sync_rst : signal is "no";
  signal local_rst_r1: std_logic := '1';
  attribute shreg_extract : string;
  attribute keep of local_rst_r1 : signal is "TRUE";  
  attribute equivalent_register_removal of local_rst_r1 : signal is "no";
  attribute max_fanout of local_sync_rst : signal is "16";
  attribute shreg_extract of local_rst_r1 : signal is "no";
  attribute shreg_extract of local_sync_rst : signal is "no";
  
begin

  process (pcie_clk,pcie_rst) is
  begin  -- process
    if pcie_rst = '1' then
      local_rst <= '1';
    elsif rising_edge(pcie_clk) then            -- rising clock edge
      local_rst <= '0';
    end if;
  end process;
  
  process (pcie_clk) is
  begin  -- process
    if rising_edge(pcie_clk) then       -- rising clock edge
      local_rst_r1 <= local_rst;
      local_sync_rst <= local_rst_r1;
    end if;
  end process;
  
  seq_run : process(pcie_clk)
  begin
    if pcie_clk'event and pcie_clk = '1' then
      if(local_sync_rst = '1')then
        sequencer            <= "01";
        select_src           <= "001";
        rx_bar0_data_pop_int <= '0';
        rx_bar1_data_pop_int <= '0';
        rx_bar2_data_pop_int <= '0';
        rx_bar_data_push     <= '0';
      else
        if(sequencer(0) = '1')and(rx_bar_data_space = '1')then
          if(rx_bar0_data_pkt_rdy = '1')then
            select_src           <= "001";
            sequencer            <= "10";
            rx_bar0_data_pop_int <= '1';
            rx_bar1_data_pop_int <= '0';
            rx_bar2_data_pop_int <= '0';
            rx_bar_data_push     <= '1';
          elsif(rx_bar1_data_pkt_rdy = '1')then
            select_src           <= "010";
            sequencer            <= "10";
            rx_bar0_data_pop_int <= '0';
            rx_bar1_data_pop_int <= '1';
            rx_bar2_data_pop_int <= '0';
            rx_bar_data_push     <= '1';
          elsif(rx_bar2_data_pkt_rdy = '1')then
            select_src           <= "100";
            sequencer            <= "10";
            rx_bar0_data_pop_int <= '0';
            rx_bar1_data_pop_int <= '0';
            rx_bar2_data_pop_int <= '1';
            rx_bar_data_push     <= '1';
          else
            select_src           <= "001";
            sequencer            <= "01";
            rx_bar0_data_pop_int <= '0';
            rx_bar1_data_pop_int <= '0';
            rx_bar2_data_pop_int <= '0';
            rx_bar_data_push     <= '0';
          end if;
        end if;
        if(sequencer(1) = '1')then
          if(select_src(0) = '1')then
            if(rx_bar0_data_out(datawidthTX+datawidthTX/32) = '1')then
              if(rx_bar1_data_pkt_rdy = '1')then
                rx_bar0_data_pop_int <= '0';
                rx_bar1_data_pop_int <= not rx_bar1_data_out(datawidthTX+datawidthTX/32);
                rx_bar2_data_pop_int <= '0';
                rx_bar_data_push     <= not rx_bar1_data_out(datawidthTX+datawidthTX/32);
                select_src           <= "010";
                sequencer            <= "10";
              elsif(rx_bar2_data_pkt_rdy = '1')then
                rx_bar0_data_pop_int <= '0';
                rx_bar1_data_pop_int <= '0';
                rx_bar2_data_pop_int <= not rx_bar2_data_out(datawidthTX+datawidthTX/32);
                rx_bar_data_push     <= not rx_bar2_data_out(datawidthTX+datawidthTX/32);
                select_src           <= "100";
                sequencer            <= "10";
              else
                select_src           <= "010";
                sequencer            <= "01";
                rx_bar0_data_pop_int <= '0';
                rx_bar1_data_pop_int <= '0';
                rx_bar2_data_pop_int <= '0';
                rx_bar_data_push     <= '0';
              end if;
            end if;
          elsif(select_src(1) = '1')then
            if(rx_bar1_data_out(datawidthTX+datawidthTX/32) = '1')then
              if(rx_bar2_data_pkt_rdy = '1')then
                rx_bar0_data_pop_int <= '0';
                rx_bar1_data_pop_int <= '0';
                rx_bar2_data_pop_int <= not rx_bar2_data_out(datawidthTX+datawidthTX/32);
                rx_bar_data_push     <= not rx_bar2_data_out(datawidthTX+datawidthTX/32);
                select_src           <= "100";
                sequencer            <= "10";
              elsif(rx_bar0_data_pkt_rdy = '1')then
                rx_bar0_data_pop_int <= not rx_bar0_data_out(datawidthTX+datawidthTX/32);
                rx_bar1_data_pop_int <= '0';
                rx_bar2_data_pop_int <= '0';
                rx_bar_data_push     <= not rx_bar0_data_out(datawidthTX+datawidthTX/32);
                select_src           <= "001";
                sequencer            <= "10";
              else
                select_src           <= "100";
                sequencer            <= "01";
                rx_bar0_data_pop_int <= '0';
                rx_bar1_data_pop_int <= '0';
                rx_bar2_data_pop_int <= '0';
                rx_bar_data_push     <= '0';
              end if;
            end if;
          elsif(select_src(2) = '1')then
            if(rx_bar2_data_out(datawidthTX+datawidthTX/32) = '1')then
              if(rx_bar0_data_pkt_rdy = '1')then
                rx_bar0_data_pop_int <= not rx_bar0_data_out(datawidthTX+datawidthTX/32);
                rx_bar1_data_pop_int <= '0';
                rx_bar2_data_pop_int <= '0';
                rx_bar_data_push     <= not rx_bar0_data_out(datawidthTX+datawidthTX/32);
                select_src           <= "001";
                sequencer            <= "10";
              elsif(rx_bar1_data_pkt_rdy = '1')then
                rx_bar0_data_pop_int <= '0';
                rx_bar1_data_pop_int <= not rx_bar1_data_out(datawidthTX+datawidthTX/32);
                rx_bar2_data_pop_int <= '0';
                rx_bar_data_push     <= not rx_bar1_data_out(datawidthTX+datawidthTX/32);
                select_src           <= "010";
                sequencer            <= "10";
              else
                select_src           <= "001";
                sequencer            <= "01";
                rx_bar0_data_pop_int <= '0';
                rx_bar1_data_pop_int <= '0';
                rx_bar2_data_pop_int <= '0';
                rx_bar_data_push     <= '0';
              end if;
            end if;
          end if;
        end if;
      end if;
    end if;
  end process seq_run;


  rx_bar0_data_pop <= rx_bar0_data_pop_int;
  rx_bar1_data_pop <= rx_bar1_data_pop_int;
  rx_bar2_data_pop <= rx_bar2_data_pop_int;

  sel_src : process(select_src, rx_bar0_data_out, rx_bar1_data_out, rx_bar2_data_out)
  begin
    case (select_src) is
      when "001"  => rx_bar_data_out <= rx_bar0_data_out;
      when "010"  => rx_bar_data_out <= rx_bar1_data_out;
      when "100"  => rx_bar_data_out <= rx_bar2_data_out;
      when others => rx_bar_data_out <= (others => '0');
    end case;
  end process sel_src;

end architecture;
