--------------------------------------------------------------------------------
-- Company:         Alpha-Data Parallel Systems
-- Engineer:        Alan Crawford
--
-- Create Date:
-- Design Name:     ADB3
-- Module Name:
-- Project Name:    Gen3 adb3
-- Target Device:   V7
-- Tool versions:
-- Description:
-- Dependencies:    PCIe/OCP Packages for signal records
--
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
--
--------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.pcie.all;
use work.adb3_ocp.all;
use work.pciegen3_pkg.all;
use work.pcie_config.all;



entity pcie_g3_bridge is
  generic(
    max_timeout  : std_logic_vector(15 downto 0) := x"0FFF";
    no_slaves    : integer := 3;
    no_masters   : integer := 5;
    dmChannel    : integer := 4
    );

  port(
    pcie_rst                            : in  std_logic;
    pcie_clk                            : in  std_logic;
    
    -- Bar MAsks from config block
    bar_masks                           : in  pcie_bar_mask_t;

    -- bar Interface Requester

    cq_in                               : in  cq_c2b_t;
    cq_out                              : out cq_b2c_t;

    -- bar Interface Completer

    cc_in                               : in  cc_c2b_t;
    cc_out                              : out cc_b2c_t;

    -- DMA Interface Requester

    rq_in                               : in  rq_c2b_t;
    rq_out                              : out rq_b2c_t;

    -- DMA Interface Completer

    rc_in                               : in  rc_c2b_t;
    rc_out                              : out rc_b2c_t;

    -- Message interface
    messages                            : in  message_c2b_t;

    -- Interrupts to from core
    int_in                              : in  interrupts_c2b_t;
    int_out                             : out interrupts_b2c_t;

    -- OCP Masters (DMA and Direct Master)
    m_s2m_o                             : out adb3w_ocp_s2mT_array(0 to no_masters-1);
    m_m2s_i                             : in  adb3w_ocp_m2sT_array(0 to no_masters-1);

    -- OCP Slaves (BARS)
    s_s2m_i                             : in  adb3_ocp_s2mT_array(0 to 1);
    s_m2s_o                             : out adb3_ocp_m2sT_array(0 to 1);
    s_s2m_iw                            : in  adb3w_ocp_s2mT;
    s_m2s_ow                            : out adb3w_ocp_m2sT;

    -- Interrupt signal
    interrupt                           : in  std_logic;

    -- status and control to/from register bank
    pcie_reg_ctrl                       : in  pcie_reg_ctrl_t;

    pcie_reg_status                     : in  pcie_reg_status_t;
    maxTotalPIF                         : in  std_logic_vector(7 downto 0) := "00010000";
    maxPIF                              : in  std_logic_vector(7 downto 0) := "00000100";
    dmaif_status                        : out std_logic_vector(31 downto 0);
    dmaClr                              : in std_Logic_vector(7 downto 0);
    tag_timeout_error                   : out std_logic_vector(no_masters-1 downto 0);
    clear_to_error                      : in  std_logic_vector(no_masters-1 downto 0);
    timeout_bar23                       : out std_logic;
    clr_timeout_bar23                   : in  std_logic
  );

end pcie_g3_bridge;

architecture rtl of pcie_g3_bridge is

  signal    interrupt_msi               : std_logic_vector(1 downto 0);
  constant  timeout_enable              : std_logic := '1';

  signal    core_interrupt              : std_logic := '0';
  signal    core_msi_interrupt          : std_logic := '0';


  signal    us_strobe                   : std_logic;
  signal    us_reset                    : std_logic_vector(1 downto 0);
  signal    us_counter                  : std_logic_vector(7 downto 0);
  constant  us_limit                    : std_logic_vector(7 downto 0) := X"FA";

  signal local_rst,local_sync_rst : std_logic := '1';
  attribute keep : string;
  attribute max_fanout : string;
  attribute equivalent_register_removal : string;
  attribute keep of local_rst : signal is "TRUE";
  attribute equivalent_register_removal of local_rst : signal is "no";
  attribute max_fanout of local_rst : signal is "1";
  attribute keep of local_sync_rst : signal is "TRUE";
  attribute equivalent_register_removal of local_sync_rst : signal is "no";
  
  signal local_rst_r1: std_logic := '1';
  attribute shreg_extract : string;
  attribute keep of local_rst_r1 : signal is "TRUE";  
  attribute equivalent_register_removal of local_rst_r1 : signal is "no";
  attribute max_fanout of local_sync_rst : signal is "16";
  attribute shreg_extract of local_rst_r1 : signal is "no";
  attribute shreg_extract of local_sync_rst : signal is "no";
  
begin

  process (pcie_clk,pcie_rst) is
  begin  -- process
    if pcie_rst = '1' then
      local_rst <= '1';
    elsif rising_edge(pcie_clk) then            -- rising clock edge
      local_rst <= '0';
    end if;
  end process;
  
  process (pcie_clk) is
  begin  -- process
    if rising_edge(pcie_clk) then       -- rising clock edge
      local_rst_r1 <= local_rst;
      local_sync_rst <= local_rst_r1;
    end if;
  end process;
  
-- -----------------------------------------------------------------------------
-- Create 1us strobe for timeout logic
-- -----------------------------------------------------------------------------
  us_strobe_gen : process(pcie_clk)
  begin
    if((pcie_clk'event)and(pcie_clk = '1'))then
      if(local_sync_rst = '1')then
        us_counter <= us_limit;
        us_strobe  <= '0';
        us_reset  <= "00";
        
      else
        if(timeout_enable = '1')then
          us_reset  <= us_reset(0) & us_counter(7);
          if(us_counter /= X"FF")then
            us_counter <= us_counter +'1';
          else
            us_counter <= X"00";
          end if;
          if(us_reset = "01")then
            us_strobe <= '1';
          else
            us_strobe <= '0';
          end if;
        else
          us_counter <= us_limit;
          us_strobe  <= '0';
        end if;
      end if;
    end if;
  end process us_strobe_gen;

-- -----------------------------------------------------------------------------
-- Basic message decoding
-- -----------------------------------------------------------------------------
-- there will be no message handling initially so the message signals shall
-- remain unconnected.
-- -----------------------------------------------------------------------------
-- messages => unused
-- -----------------------------------------------------------------------------


-- -----------------------------------------------------------------------------
-- interrupt_msi[1] == interrupt_msi source for the CORE
-- -----------------------------------------------------------------------------
  process (pcie_clk)
  begin
    if(rising_edge(pcie_clk))then
      int_out.cfg_interrupt_int                  <= "000" & core_interrupt;
      int_out.cfg_interrupt_msi_int(31 downto 1) <= (others => '0');
      int_out.cfg_interrupt_msi_int(0)           <= core_msi_interrupt;
  
      core_interrupt <= interrupt;
      interrupt_msi  <= interrupt_msi(0) & interrupt;
  
      if(interrupt_msi = "01")then
        core_msi_interrupt <= '1';
      end if;
      if(int_in.cfg_interrupt_msi_sent = '1')then
        core_msi_interrupt <= '0';
      end if;
      if(int_in.cfg_interrupt_msi_fail = '1')then
        interrupt_msi <= "00";
      end if;
    end if;
  end process;

  int_out.cfg_interrupt_pending <= (others => '0');
  int_out.cfg_interrupt_msi_pending_status <= (others => '0');
  int_out.cfg_interrupt_msi_select <= (others => '0');

  tag_timeout_error <= (others => '0');
  
-- -----------------------------------------------------------------------------
-- BAR Interface
-- -----------------------------------------------------------------------------

  bars : pcieg3_bar_if
    generic map(
      max_timeout   => max_timeout,
      datawidthRX   => axi4_data_widthRX,
      datawidthTX   => axi4_data_widthTX,
      datawidthBAR2 => 256
      )
    port map(
      pcie_clk => pcie_clk,
      pcie_rst => pcie_rst,
      us_strobe => us_strobe,
      bar_masks => bar_masks,
      cq_c2b   => cq_in,
      cq_b2c   => cq_out,
      cc_b2c   => cc_out,
      cc_c2b   => cc_in,
      m_m2s_o  => s_m2s_o,
      m_s2m_i  => s_s2m_i,
      m_m2s_ow  => s_m2s_ow,
      m_s2m_iw  => s_s2m_iw,
      timeout_bar23 => timeout_bar23,
      clr_timeout_bar23 => clr_timeout_bar23
      );



  dma : dma_if
    generic map(
      dmChannel => dmChannel,
      channelCount => no_masters,
      dataWidth    => 256
      )
    port map(
      clk         => pcie_clk,
      sr          => pcie_rst,
      m2s_i       => m_m2s_i,
      s2m_o       => m_s2m_o,
      rq_b2c      => rq_out,
      rq_c2b      => rq_in,
      maxTotalPIF => maxTotalPIF,
      maxPIF      => maxPIF,
      dmaClr      => dmaClr,
      status      => dmaif_status,
      rc_b2c      => rc_out,
      rc_c2b      => rc_in
      );

end;

