--
-- pcieg3_pkg.vhd - definitions file for pcie Gen3 bridge signals in adb3
--
--
-- (c) copyright alpha data ltd. 2009
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.ALL;
use work.pcie_config.all;
use work.pcie.all;

package pciegen3_pkg is

  constant axi4_data_widthRX    : natural := 256;
  constant axi4_data_widthTX    : natural := axi4_data_widthRX;
  
  constant MAX_TIMEOUT          : std_logic_vector(15 downto 0) := X"01F0";
                                
  constant no_masters           : natural := cfg_no_masters;
  constant no_slaves            : natural := cfg_no_slaves;
  constant source_size          : natural := cfg_source_size;
  constant no_32bit_bars        : natural := cfg_no_32bit_bars;
  constant max_data_width       : natural := cfg_max_data_width;
  constant no_control_bits      : natural := cfg_no_control_bits;
  constant pcie_data_size       : natural := (max_data_width + no_control_bits);
  
  type cq_c2b_t is record -- AXI4 interface completer Requester (C2B)
    m_axis_cq_tdata             : std_logic_vector(axi4_data_widthRX-1 downto 0);
    m_axis_cq_tparity           : std_logic_vector(axi4_data_widthRX/8-1 downto 0);
    m_axis_cq_tvalid            : std_logic;
    m_axis_cq_tlast             : std_logic;
    m_axis_cq_tkeep             : std_logic_vector(axi4_data_widthRX/32-1 downto 0);
    m_axis_cq_tuser             : std_logic_vector(84 downto 0);
    pcie_cq_np_req_count        : std_logic_vector(5 downto 0);
  end record cq_c2b_t;
  
  type cq_b2c_t is record -- AXI4 interface completer Requester (B2C)
    m_axis_cq_tready            : std_logic;
  end record cq_b2c_t;
  
  type cc_c2b_t is record -- AXI4 interface completer completer (C2B)
    s_axis_cc_tready            :std_logic;
  end record cc_c2b_t;
  
  type cc_b2c_t is record -- AXI4 interface completer completer (B2C)
    s_axis_cc_tdata             : std_logic_vector(axi4_data_widthTX-1 downto 0);
    s_axis_cc_tparity           : std_logic_vector(axi4_data_widthTX/8-1 downto 0);
    s_axis_cc_tvalid            : std_logic;
    s_axis_cc_tuser             : std_logic_vector(32 downto 0);
    s_axis_cc_tkeep             : std_logic_vector(axi4_data_widthTX/32-1 downto 0);
    s_axis_cc_tlast             : std_logic;
  end record cc_b2c_t;
  
  type rq_b2c_t is record -- AXI4 interface requester requester (B2C)
    s_axis_rq_tdata             : std_logic_vector(axi4_data_widthRX-1 downto 0);
    s_axis_rq_tparity           : std_logic_vector(axi4_data_widthRX/8-1 downto 0);
    s_axis_rq_tvalid            : std_logic;
    s_axis_rq_tlast             : std_logic;
    s_axis_rq_tkeep             : std_logic_vector(axi4_data_widthRX/32-1 downto 0);
    s_axis_rq_tuser             : std_logic_vector(axi4_data_widthRX/8+27 downto 0);
  end record rq_b2c_t;
  
  type rq_c2b_t is record -- AXI4 interface requester requester (C2B)
    s_axis_rq_tready            : std_logic;
    pcie_rq_tag                 : std_logic_vector(5 downto 0);
    pcie_rq_tag_vld             : std_logic;
    pcie_tfc_nph                : std_logic_vector(1 downto 0);
    pcie_tfc_npd                : std_logic_vector(1 downto 0);
    pcie_rq_tab_av              : std_logic_vector(1 downto 0);
    pcie_rq_seq_num             : std_logic_vector(3 downto 0);
    pcie_rq_seq_num_vld         : std_logic;
  end record rq_c2b_t;
  
  type rc_c2b_t is record -- AXI4 interface requester completer (C2B)
    m_axis_rc_tdata             : std_logic_vector(axi4_data_widthRX-1 downto 0);
    m_axis_rc_tparity           : std_logic_vector(axi4_data_widthRX/8-1 downto 0);
    m_axis_rc_tvalid            : std_logic;
    m_axis_rc_tlast             : std_logic;
    m_axis_rc_tkeep             : std_logic_vector(axi4_data_widthRX/32-1 downto 0);
    m_axis_rc_tuser             : std_logic_vector(74 downto 0);
  end record rc_c2b_t;
    
  type rc_b2c_t is record -- AXI4 interface requester completer (B2C)
    m_axis_rc_tready            : std_logic;
  end record rc_b2c_t;
  
  type message_c2b_t is record -- Message Interface C2B only
    cfg_msg_received            : std_logic;
    cfg_msg_received_type       : std_logic_vector(4 downto 0);
    cfg_msg_received_data       : std_logic_vector(7 downto 0);
  end record message_c2b_t;
  
  type config_c2b_t is record -- Configuration/operation Information Interface C2B only
    cfg_phy_link_down           : std_logic;
    cfg_phy_link_status         : std_logic_vector(1 downto 0);
    cfg_negotiated_width        : std_logic_vector(3 downto 0);
    cfg_current_speed           : std_logic_vector(2 downto 0);
    cfg_max_payload             : std_logic_vector(2 downto 0);
    cfg_max_read_req            : std_logic_vector(2 downto 0);
    cfg_function_status         : std_logic_vector(15 downto 0);
    cfg_vf_status               : std_logic_vector(15 downto 0);
    cfg_function_power_state    : std_logic_vector(11 downto 0);
    cfg_vf_power_state          : std_logic_vector(23 downto 0);
    cfg_link_power_state        : std_logic_vector(1 downto 0);
    cfg_ltssm_state             : std_logic_vector(5 downto 0);
  end record config_c2b_t;
  
  type interrupts_c2b_t is record  -- interrupt flags from core (i.e. interrupt mode ctrl)
    cfg_interrupt_sent            : std_logic; 
    cfg_interrupt_msi_enable      : std_logic_vector(3 downto 0); -- enable MSI per physical function
    cfg_interrupt_msi_sent        : std_logic;
    cfg_interrupt_msi_fail        : std_logic;
    cfg_interrupt_msi_mmenable    : std_logic_vector(11 downto 0); -- multiple message enables
    cfg_interrupt_msi_mask_update : std_logic; -- mask register has been updated
    cfg_interrupt_msi_data        : std_logic_vector(31 downto 0); --  NO IDEA
  end record interrupts_c2b_t;
  
  type interrupts_b2c_t is record  -- interrupt flags to core (i.e. interrupts)
    cfg_interrupt_int                 : std_logic_vector(3 downto 0); -- legacy INTx pins
    cfg_interrupt_pending             : std_logic_vector(3 downto 0); -- one bit per physical function (only use [0])
    cfg_interrupt_msi_int             : std_logic_vector(31 downto 0); -- 32 MSI vectors only use [0]
    cfg_interrupt_msi_pending_status  : std_logic_vector(31 downto 0); -- pending stae for all MSI ints on both PFs
    cfg_interrupt_msi_select          : std_logic_vector(3 downto 0); -- NO IDEA
  end record interrupts_b2c_t;
  
  constant cfg_phy_link_status_not_connected  : std_logic_vector(1 downto 0) := "00";
  constant cfg_phy_link_status_in_training    : std_logic_vector(1 downto 0) := "01";
  constant cfg_phy_link_status_link_up        : std_logic_vector(1 downto 0) := "10";
  
  constant cfg_current_speed_2_5g             : std_logic_vector(2 downto 0) := "001";
  constant cfg_current_speed_5g               : std_logic_vector(2 downto 0) := "010";
  constant cfg_current_speed_8g               : std_logic_vector(2 downto 0) := "100";

  constant max_payload_128                    : std_logic_vector(2 downto 0) := "000";
  constant max_payload_256                    : std_logic_vector(2 downto 0) := "001";
  constant max_payload_512                    : std_logic_vector(2 downto 0) := "010";
  constant max_payload_1024                   : std_logic_vector(2 downto 0) := "011";
  constant max_payload_2048                   : std_logic_vector(2 downto 0) := "100";
  constant max_payload_4096                   : std_logic_vector(2 downto 0) := "101";

-- Values for the Device Power state (d0 - operating state)

  constant power_state_d0_uninit              : std_logic_vector(2 downto 0) := "000";
  constant power_state_d0_active              : std_logic_vector(2 downto 0) := "001";
  constant power_state_d1                     : std_logic_vector(2 downto 0) := "010";
  constant power_state_d3_hot                 : std_logic_vector(2 downto 0) := "100";

-- Values for the Physical Link Status (L0 - operating state)

  constant link_power_state_l0                : std_logic_vector(1 downto 0) := "00";
  constant link_power_state_l0s               : std_logic_vector(1 downto 0) := "01";
  constant link_power_state_l1                : std_logic_vector(1 downto 0) := "10";
  constant link_power_state_l2_reserved       : std_logic_vector(1 downto 0) := "11";

-- Values for the completion status

  constant completion_ok                      : std_logic_vector(3 downto 0) := "0000";
  constant completion_poisoned                : std_logic_vector(3 downto 0) := "0001";
  constant completion_terminated_ur           : std_logic_vector(3 downto 0) := "0010";
  constant completion_no_data                 : std_logic_vector(3 downto 0) := "0011";
  constant completion_mismatch                : std_logic_vector(3 downto 0) := "0100";
  constant completion_address_err             : std_logic_vector(3 downto 0) := "0101";
  constant completion_tag_err                 : std_logic_vector(3 downto 0) := "0110";
  constant completion_timeout                 : std_logic_vector(3 downto 0) := "0111";
  constant completion_req_terminated          : std_logic_vector(3 downto 0) := "1000";
  
  constant dataWidth                          : integer  := axi4_data_widthRX;
  constant channelCount                       : integer  := 5;

  subtype   dataChannel      is std_logic_vector(dataWidth+dataWidth/32+3 downto 0);
  type      channelArray     is array (0 to channelCount-1) of dataChannel;

  component pcie_g3_bridge is
    generic(
      max_timeout       : std_logic_vector(15 downto 0) := x"0FFF";
      no_slaves         : integer := 3;
      no_masters        : integer := 5;
      dmChannel           : integer := 4
      );
    port(
      pcie_rst          : in  std_logic;
      pcie_clk          : in  std_logic;
      bar_masks         : in  pcie_bar_mask_t;
      cq_in             : in  cq_c2b_t;
      cq_out            : out cq_b2c_t;
      cc_in             : in  cc_c2b_t;
      cc_out            : out cc_b2c_t;
      rq_in             : in  rq_c2b_t;
      rq_out            : out rq_b2c_t;
      rc_in             : in  rc_c2b_t;
      rc_out            : out rc_b2c_t;
      messages          : in  message_c2b_t;
      int_in            : in  interrupts_c2b_t;
      int_out           : out interrupts_b2c_t;
      m_s2m_o           : out adb3w_ocp_s2mT_array(0 to NO_MASTERS-1);
      m_m2s_i           : in  adb3w_ocp_m2sT_array(0 to NO_MASTERS-1);
      s_s2m_i           : in  adb3_ocp_s2mT_array(0 to 1);
      s_m2s_o           : out adb3_ocp_m2sT_array(0 to 1);
      s_s2m_iw          : in  adb3w_ocp_s2mT;
      s_m2s_ow          : out adb3w_ocp_m2sT;
      interrupt         : in  std_logic;
      pcie_reg_ctrl     : in  pcie_reg_ctrl_t;
      pcie_reg_status   : in  pcie_reg_status_t;
      maxTotalPIF       : in  std_logic_vector(7 downto 0) := "00010000";
      maxPIF            : in  std_logic_vector(7 downto 0) := "00000100";
      dmaif_status      : out std_logic_vector(31 downto 0);
      dmaClr            : in  std_Logic_vector(7 downto 0);
      tag_timeout_error : out std_logic_vector(NO_MASTERS-1 downto 0);
      clear_to_error    : in  std_logic_vector(NO_MASTERS-1 downto 0);
      timeout_bar23     : out std_logic;
      clr_timeout_bar23 : in  std_logic
    );
  end component;
  
  component pcieg3_bar_if is
    generic(
      max_timeout           : std_logic_vector(15 downto 0) := x"0FFF";
      datawidthRX           : integer := axi4_data_widthRX;
      datawidthTX           : integer := axi4_data_widthTX;
      datawidthBAR2         : integer := 256
    );
    port(
      pcie_clk              : in  std_logic;
      pcie_rst              : in  std_logic;
      us_strobe             : in  std_logic;
      bar_masks             : in  pcie_bar_mask_t;
      cq_c2b                : in  cq_c2b_t;
      cq_b2c                : out cq_b2c_t;
      cc_b2c                : out cc_b2c_t;
      cc_c2b                : in  cc_c2b_t;
      m_m2s_o               : out adb3_ocp_m2sT_array(0 to 1);
      m_s2m_i               : in  adb3_ocp_s2mT_array(0 to 1);
      m_m2s_ow              : out adb3w_ocp_m2sT;
      m_s2m_iw              : in  adb3w_ocp_s2mT;
      timeout_bar23         : out std_logic;
      clr_timeout_bar23     : in  std_logic
    );
  end component;
  
  component dma_if is
    generic(
      channelCount          : integer := 5;
      dataWidth             : integer := 256;
      dmChannel           : integer := -1
    );                    
    port(                   
      clk                   : in  std_logic;
      sr                    : in  std_logic;
      m2s_i                 : in  adb3w_ocp_m2sT_array;
      s2m_o                 : out adb3w_ocp_s2mT_array;
      rq_b2c                : out rq_b2c_t;
      rq_c2b                : in  rq_c2b_t;
      maxTotalPIF           : in  std_logic_vector(7 downto 0);
      maxPIF                : in  std_logic_vector(7 downto 0);
      status                : out std_logic_vector(31 downto 0);
      dmaClr              : in std_Logic_vector(7 downto 0);
      rc_b2c                : out rc_b2c_t;
      rc_c2b                : in  rc_c2b_t
    );
  end component;
  
  component pcieg3_bar_decoder is
    generic(
      datawidthRX         : integer := axi4_data_widthRX;
      datawidthTX         : integer := axi4_data_widthTX;
      single_tag_request    : boolean := false
    );
    port(
      pcie_clk              : in  std_logic;
      pcie_rst              : in  std_logic;
      bar_mask              : in  std_logic_vector(63 downto 0);
      bar_page              : in  std_logic_vector(63 downto 0);
      data                  : in  std_logic_vector(datawidthRX-1 downto 0);
      data_valid            : in  std_logic;
      data_ben              : in  std_logic_vector(datawidthRX/8-1 downto 0);
      data_user             : in  std_logic_vector(84 downto 0);
      cmd_push              : in  std_logic;
      data_push             : in  std_logic;
      bar_full              : out std_logic;
      rx_data_out           : out std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
      rx_data_pop           : in  std_logic;
      rx_data_pkt_rdy       : out std_logic;
      m_m2s_o               : out adb3_ocp_m2sT;
      m_s2m_i               : in  adb3_ocp_s2mT
    );
  end component;
  
  component pcieg3_bar2_decoder is
    generic(
      max_timeout           : std_logic_vector(15 downto 0) := x"0FFF";
      datawidthRX           : integer := 256;
      datawidthTX           : integer := 256;
      single_tag_request    : boolean := false
    );
    port(
      pcie_clk              : in  std_logic;
      pcie_rst              : in  std_logic;
      us_strobe             : in  std_logic;
      bar_mask              : in  std_logic_vector(63 downto 0);
      bar_page              : in  std_logic_vector(63 downto 0);
      data                  : in  std_logic_vector(datawidthRX-1 downto 0);
      data_valid            : in  std_logic;
      data_ben              : in  std_logic_vector(datawidthRX/8-1 downto 0);
      data_user             : in  std_logic_vector(84 downto 0);
      cmd_push              : in  std_logic;
      data_push             : in  std_logic;
      bar_full              : out std_logic;
      rx_data_out           : out std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
      rx_data_pop           : in  std_logic;
      rx_data_pkt_rdy       : out std_logic;
      m_m2s_o               : out adb3w_ocp_m2sT;
      m_s2m_i               : in  adb3w_ocp_s2mT;
      timeout               : out std_logic;
      clr_timeout           : in  std_logic
    );
  end component;
  
  component pcieg3_bar_arbiter is
    generic(
      datawidthTX : integer := 128
    );
    port(
      pcie_clk              : in  std_logic;
      pcie_rst              : in  std_logic;
      rx_bar0_data_out      : in  std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
      rx_bar0_data_pop      : out std_logic;
      rx_bar0_data_pkt_rdy  : in  std_logic;
      rx_bar1_data_out      : in  std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
      rx_bar1_data_pop      : out std_logic;
      rx_bar1_data_pkt_rdy  : in  std_logic;
      rx_bar2_data_out      : in  std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
      rx_bar2_data_pop      : out std_logic;
      rx_bar2_data_pkt_rdy  : in  std_logic;
      rx_bar_data_out       : out std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
      rx_bar_data_push      : out std_logic;
      rx_bar_data_space     : in  std_logic
    );
  end component;
  
  component pcieg3_cc_interface is
    generic(
      datawidthTX : integer := 128
    );
    port(
      pcie_clk              : in  std_logic;
      pcie_rst              : in  std_logic;
      rx_bar_data_out       : in  std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
      rx_bar_data_push      : in  std_logic := '0';
      rx_bar_data_space     : out std_logic;
      cc_b2c                : out cc_b2c_t;
      cc_c2b                : in  cc_c2b_t
    );
  end component;
  
  component dma_merger is
    generic(
      channelCount          : integer := 16;
      dataWidth             : integer := 256
    );
    port (
      Ck                    : in  std_logic;
      srcReq                : in  std_logic_vector(channelCount-1 downto 0);
      srcData               : in  channelArray;
      srcPop                : out std_logic_vector(channelCount-1 downto 0);
      opData                : out std_logic_vector(dataWidth+dataWidth/32+3 downto 0);
      opPush                : out std_logic;
      opFull                : in  std_logic
    );
  end component;
  
  component pargen is
    port(
      data_in               : in  std_logic_vector(7 downto 0);
      parity_odd            : out std_logic
    );
  end component;
  
  component dma_ocp_if is
    generic(
      chIndexVector           : std_logic_vector(4 downto 0) := "00000";
      dataWidth               : integer := 256
    );
    port(
      clk                     : in  std_logic;
      sr                      : in  std_logic;
      us_counter              : in  std_logic_vector(9 downto 0);
      m2s_i                   : in  adb3w_ocp_m2sT;
      s2m_o                   : out adb3w_ocp_s2mT;
      slave_o2p_push          : out std_logic := '0';
      slave_o2p_data          : out std_logic_vector(dataWidth+dataWidth/32+3 downto 0);
      slave_o2p_notfull       : in  std_logic;
      slave_p2o_push          : in  std_logic;
      slave_p2o_data          : in  std_logic_vector(dataWidth+3 downto 0);
--       slave_p2o_tag           : in  std_logic_vector(3 downto 0);
       slave_p2o_notfull       : out std_logic;
      disableReads            : in std_logic;
      cpldComplete            : in std_logic;
      actualPIF               : out std_logic_vector(7 downto 0);        
      maxPIF                  : in std_logic_vector(7 downto 0)
    );
  end component;
  
--   component pcie_pkt_fifo3 is
--     generic(
--       fifo_depth              : integer := 4;
--       fifo_width              : integer := 128;
--       ram_style_def           : string  := "distributed"
--       );
--     port(
--       clk                     : in  std_logic;
--       sr                      : in  std_logic;
--       max_pkt_size            : in  std_logic_vector(fifo_depth-1 downto 0);
--       din                     : in  std_logic_vector(fifo_width-1 downto 0);
--       dout                    : out std_logic_vector(fifo_width-1 downto 0) := (others => '0');
--       push                    : in  std_logic;
--       pop                     : in  std_logic;
--       full                    : out std_logic;
--       empty                   : out std_logic;
--       last                    : out std_logic;
--       free_pkt                : out std_logic;
--       pkt_available           : out std_logic
--     );
--   end component;
  
--   component pcie_pkt_fifo4 is
--     generic(
--       fifo_depth              : integer := 4;
--       fifo_width              : integer := 128;
--       ram_style_def           : string  := "distributed"
--       );
--     port(
--       clk                     : in  std_logic;
--       sr                      : in  std_logic;
--       max_pkt_size            : in  std_logic_vector(fifo_depth-1 downto 0);
--       din                     : in  std_logic_vector(fifo_width-1 downto 0);
--       dout                    : out std_logic_vector(fifo_width-1 downto 0) := (others => '0');
--       push                    : in std_logic;
--       pop                     : in std_logic;
--       full                    : out std_logic;
--       empty                   : out std_logic;
--       last                    : out std_logic;
--       free_pkt                : out std_logic;
--       data_available          : out std_logic_vector(fifo_depth-1 downto 0)
--     );
--   end component;
--   
  component pcie_pkt_fifo5 is
    generic(
      fifo_depth              : integer := 4;
      fifo_width              : integer := 128;
      ram_style_def           : string  := "distributed"
      );
    port(
      clk                     : in  std_logic;
      sr                      : in  std_logic;
      max_pkt_size            : in  std_logic_vector(fifo_depth-1 downto 0);
      din                     : in  std_logic_vector(fifo_width-1 downto 0);
      dout                    : out std_logic_vector(fifo_width-1 downto 0) := (others => '0');
      push                    : in  std_logic;
      pop                     : in  std_logic;
      full                    : out std_logic;
      empty                   : out std_logic;
      last                    : out std_logic;
      free_pkt                : out std_logic;
      pkt_available           : out std_logic
    );
  end component;

--   component pcie_pkt_fifo6 is
--     generic(
--       fifo_depth              : integer := 4;
--       fifo_width              : integer := 128;
--       ram_style_def           : string  := "distributed"
--       );
--     port(
--       clk                     : in  std_logic;
--       sr                      : in  std_logic;
--       max_pkt_size            : in  std_logic_vector(fifo_depth-1 downto 0);
--       din                     : in  std_logic_vector(fifo_width-1 downto 0);
--       dout                    : out std_logic_vector(fifo_width-1 downto 0) := (others => '0');
--       push                    : in  std_logic;
--       pop                     : in  std_logic;
--       full                    : out std_logic;
--       empty                   : out std_logic;
-- --       last                    : out std_logic;
--       free_pkt                : out std_logic
--     );
--   end component;
  
 component packet_sfifo is
generic(
  RAM_TYPE    : in    string := "AUTO";
  ORDER       : in    natural range 1 to 15;
  WIDTH       : in    natural;
  AFULL_LEVEL : in    natural);
port(
  rst         : in    std_logic;                             -- Asynchronous reset
  sr          : in    std_logic;                             -- Synchronous reset
  clk         : in    std_logic;                             -- Clock
  wen         : in    std_logic;                             -- '1' => write value on 'wd' into FIFO
  wd          : in    std_logic_vector(WIDTH - 1 downto 0);  -- Item to write into FIFO, qualified by 'wen'
  wfull       : out   std_logic;                             -- '1' => full
  wafull      : out   std_logic;                             -- '1' => almost full
  wlevel             : out   std_logic_vector(ORDER downto 0);     -- No. of items in FIFO from writer P.O.V.
  radv        : in    std_logic;                             -- '1' => present next item on 'rq'
--   rvalid      : out   std_logic;                             -- '1' => not empty
  rq          : out   std_logic_vector(WIDTH - 1 downto 0);  -- Current item, qualified by 'rvalid'
  rlevel      : out   std_logic_vector(ORDER downto 0);      -- Number of items currently in FIFO, but ignore if 'rempty' asserted
  rempty      : out   std_logic;
  debug_info  : out   std_logic_vector(15 downto 0));
end component;

-- 
--   component pcie_pkt_fifo7 is
--     generic(
--       fifo_depth              : integer := 4;
--       fifo_width              : integer := 128;
--       ram_style_def           : string  := "distributed"
--       );
--     port(
--       clk                     : in  std_logic;
--       sr                      : in  std_logic;
--       max_pkt_size            : in  std_logic_vector(fifo_depth-1 downto 0);
--       din                     : in  std_logic_vector(fifo_width-1 downto 0);
--       dout                    : out std_logic_vector(fifo_width-1 downto 0) := (others => '0');
--       push                    : in  std_logic;
--       pop                     : in  std_logic;
--       full                    : out std_logic;
--       empty                   : out std_logic;
-- --       last                    : out std_logic;
--       data_available          : out std_logic_vector(fifo_depth-1 downto 0)
--     );
--   end component;

  component bar_tag_manager is
    generic(
      max_timeout             : std_logic_vector(15 downto 0) := x"0FFF";
      single_req              : boolean := true
    );
    port(
      pcie_rst                : in  std_logic;
      pcie_clk                : in  std_logic;
      us_counter              : in  std_logic;
      tx_tag_request          : in  std_logic;
      tx_tag_data_in          : in  std_logic_vector(71 downto 0);
      tx_tag_number           : out std_logic_vector(2 downto 0);
      tx_tag_valid            : out std_logic;
      rx_tag_value            : in  std_logic_vector(2 downto 0);
      rx_tag_clear            : in  std_logic;
      rx_tag_valid            : out std_logic;
      rx_tag_info             : out std_logic_vector(71 downto 0);
      tag_timeout_enable      : in  std_logic;  -- enables timeout logic
      clr_timeout             : in  std_logic;  -- Clear timeout overflow flag
      tag_timeout             : out std_logic;  -- no response within desired timeout period
      flush_tags              : out std_logic;  -- Clear any active TAGs during the timeout
      clear_tags              : out std_logic;  -- when timeout cleared ensure tags are cleared out before going back to normal operation
      tag_error               : out std_logic;  -- when a tag is returned but is not active
      status                  : out std_logic_vector(15 downto 0)  -- contains activity flags and timeout settings
    );
  end component;


  component pcieg3_bar2_rx_if is
    generic(
      datawidthTX             : integer := 256
    );
    port(
      pcie_clk                : in  std_logic;
      pcie_rst                : in  std_logic;
      m_m2s_o                 : out adb3w_ocp_m2sT;
      m_s2m_i                 : in  adb3w_ocp_s2mT;
      rx_data_out             : out std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
      rx_data_pop             : in  std_logic;
      rx_data_pkt_rdy         : out std_logic;
      bin_responses           : in  std_logic;
      pause_rx                : out std_logic;
      rx_tag_value            : out std_logic_vector(7 downto 0);
      rx_tag_clear            : out std_logic;
      rx_tag_info             : in  std_logic_vector(71 downto 0);
      rx_tag_valid            : in  std_logic
    );
  end component;
  
  component pcieg3_bar_rx_if is
    generic(
      datawidthTX             : integer := 128
    );
    port(
      pcie_clk                : in  std_logic;
      pcie_rst                : in  std_logic;
      m_m2s_o                 : out adb3_ocp_m2sT;
      m_s2m_i                 : in  adb3_ocp_s2mT;
      rx_data_out             : out std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
      rx_data_pop             : in  std_logic;
      rx_data_pkt_rdy         : out std_logic;
      rx_tag_value            : out std_logic_vector(7 downto 0);
      rx_tag_clear            : out std_logic;
      rx_tag_info             : in  std_logic_vector(71 downto 0);
      rx_tag_valid            : in  std_logic
    );
  end component;
  
  component dma_arbiter is
    generic(
      maxChannels             : integer := 16
    );
    port (
      Ck                    : in  std_logic;
      PtrUpdate             : in  std_logic;
      Req                   : in  std_logic_vector(maxChannels-1 downto 0);
      AnySet                : out std_logic;
      LPVal                 : out std_logic_vector(maxChannels-1 downto 0)
    );
  end component;


component pcie_pkt_counter is
  port(
    clk                 : in  std_logic;
    sr                  : in  std_logic;
    eop_in              : in  std_logic;
    eop_out             : in  std_logic;

    push                : in  std_logic;
    pop                 : in  std_logic;

    pkt_available       : out std_logic := '0');


end component;

  
end package pciegen3_pkg;


package body pciegen3_pkg is
end package body pciegen3_pkg;

