--
-- ADB3 Core Bridge
--
-- Core bridge module, containing generic registers, DMA engines and PCIe OCP interface
-- Contains modules common to all bridges
--

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_ARITH.all;
use IEEE.STD_LOGIC_UNSIGNED.all;
use ieee.std_logic_misc.all;

library work;
use work.pcie.all;
use work.adb3_ocp.all;
use work.adb3_dma.all;
use work.pciegen3_pkg.all;
use work.pcie_config.all;

library unisim;
use unisim.vcomponents.all;

entity adb3_core_bridge_gen3w_ku is

  generic (
    no_dma_engines     : integer := 4;
    dma_config         : dma_engine_configuration_array(0 to 7) := (others =>  DMA_ENGINE_CONFIGURATION_DEFAULT);
    register_dma_regs  : boolean := false;
    bar_mask_init_val  : std_logic_vector(31 downto 0) := X"003FFFFF";
    bar2_size          : natural range 15 to 21 := 15;
    pif_init_val       : std_logic_vector(7 downto 0)  := X"02";
    direct_master_channels  : integer := 1;
    reg_channels       : integer := 0;
    use_icap           : boolean := true);
  port (
    -- PCIE
    sys_reset_n            : in  std_logic;
    sys_clk                : in  std_logic;
    sys_clk_gt                : in  std_logic;
    ocpclkout              : out std_logic;
    pci_exp_txn            : out std_logic_vector(pcie_no_lanes-1 downto 0);
    pci_exp_txp            : out std_logic_vector(pcie_no_lanes-1 downto 0);
    pci_exp_rxn            : in  std_logic_vector(pcie_no_lanes-1 downto 0);
    pci_exp_rxp            : in  std_logic_vector(pcie_no_lanes-1 downto 0);
    -- OCP interface to Model Specific Registers
    model_regs_m2s         : out adb3_ocp_m2sT;
    model_regs_s2m         : in  adb3_ocp_s2mT;
    model_regs_irq         : in  std_logic;
    model_regs_reset       : out std_logic := '1';
    model_clk              : in   std_logic;
    model_clk_locked       : in   std_logic;
    vpd_valid              : in  std_Logic;
    -- OCP interface to MPTL (or logic if target only)
    direct_slave_m2s       : out adb3w_ocp_m2sT;
    direct_slave_s2m       : in  adb3w_ocp_s2mT;
    dma_channels_m2s       : out adb3w_ocp_m2sT_array(0 to no_dma_engines-1);
    dma_channels_s2m       : in  adb3w_ocp_s2mT_array(0 to no_dma_engines-1);
    direct_master_pcie_m2s : in  adb3w_ocp_m2sT;
    direct_master_pcie_s2m : out adb3w_ocp_s2mT;
    direct_master_regs_m2s : in  adb3l_ocp_m2sT;
    direct_master_regs_s2m : out adb3l_ocp_s2mT;
    target_irq             : in  std_logic;
    target_dma_abort       : in  std_logic_vector(no_dma_engines-1 downto 0);
    core_status            : out std_logic_vector(63 downto 0);
    core_rst_out           : out std_logic);

end adb3_core_bridge_gen3w_ku;


architecture synth of adb3_core_bridge_gen3w_ku is




  constant NO_GENERAL_REGS : integer := 16;
  constant no_slaves : natural := 3;
  constant no_masters : natural := no_dma_engines+direct_master_channels;
    
  signal general_regs : register32_array(0 to NO_GENERAL_REGS-1) := (X"00000000",
                                                                     X"00000000",
                                                                     X"00000000",
                                                                     X"00000000",
                                                                     X"00000000",
                                                                     pif_init_val & X"100100",
                                                                     X"00000000",
                                                                     bar_mask_init_val,
                                                                     X"00000000",
                                                                     X"00000000",
                                                                     bar_mask_init_val,
                                                                     X"00000000",
                                                                     X"00000000",
                                                                     X"00000000",
                                                                     X"00000000",
                                                                     X"00000000");

  constant NO_DMA_REGS : integer := 10;

  constant bar2_aperture : std_logic_vector(4 downto 0) := CONV_STD_LOGIC_VECTOR(bar2_size,5);
  
  type dma_reg_2d_array is array (0 to no_dma_engines-1) of register32_array(0 to NO_DMA_REGS-1);
  signal dma_regs                     : dma_reg_2d_array := (others => (others => X"00000000"));
  signal dma_regs_out0, dma_regs_out1 : dma_reg_2d_array := (others => (others => X"00000000"));

  type dma_reg_written_2d_array is array (0 to no_dma_engines-1) of std_logic_vector(NO_DMA_REGS-1 downto 0);
  signal dma_regs_written0, dma_regs_written1 : dma_reg_written_2d_array := (others => (others => '0'));
  signal dma_regs_written                     : dma_reg_written_2d_array := (others => (others => '0'));

  signal general_regs_out0, general_regs_out1                               : register32_array(0 to NO_GENERAL_REGS-1);
  signal general_regs_written0, general_regs_written1, general_regs_written : std_logic_vector(NO_GENERAL_REGS-1 downto 0);

  signal dummy_regs0, dummy_regs1                               : register32_array(0 to 0) := (others => X"00000000");

  signal dma_channels_i_m2s : adb3w_ocp_m2sT_array(0 to 3) := (ADB3W_OCP_DEFAULT_M2S, ADB3W_OCP_DEFAULT_M2S, ADB3W_OCP_DEFAULT_M2S, ADB3W_OCP_DEFAULT_M2S);
  signal dma_channels_i_s2m : adb3w_ocp_s2mT_array(0 to 3);

  signal dma_channels_f_m2s : adb3w_ocp_m2sT_array(0 to 3) := (ADB3W_OCP_DEFAULT_M2S, ADB3W_OCP_DEFAULT_M2S, ADB3W_OCP_DEFAULT_M2S, ADB3W_OCP_DEFAULT_M2S);
  signal dma_channels_f_s2m : adb3w_ocp_s2mT_array(0 to 3);


  signal m45_m2s : adb3_ocp_m2sT;
  signal m45_s2m : adb3_ocp_s2mT := ADB3_OCP_DEFAULT_S2M;

  signal   m_m2s_o               : adb3_ocp_m2sT_array(0 to 1);
  signal   m_s2m_i               : adb3_ocp_s2mT_array(0 to 1);
  signal   m_m2s_ow               : adb3w_ocp_m2sT;
  signal   m_s2m_iw               : adb3w_ocp_s2mT;


  signal generic_regs_m2s : adb3_ocp_m2sT;
  signal generic_regs_s2m : adb3_ocp_s2mT;

  signal generic_regs_m2s_i : adb3_ocp_m2sT;
  signal generic_regs_s2m_i : adb3_ocp_s2mT;

  signal generic_regs_m2s_r1 : adb3_ocp_m2sT;
  signal generic_regs_s2m_r1 : adb3_ocp_s2mT;

  signal generic_regs_m2s_r11 : adb3_ocp_m2sT;
  signal generic_regs_s2m_r11 : adb3_ocp_s2mT;

  signal generic_regs_m2s_fpga : adb3_ocp_m2sT;
  signal generic_regs_s2m_fpga : adb3_ocp_s2mT;

  type adb3_dma_control_register_array is array (0 to 3) of adb3_dma_control_register_type;
  type adb3_dma_status_register_array is array (0 to 3) of adb3_dma_status_register_type;
  signal dma_control : adb3_dma_control_register_array:= (others => ADB3_DMA_CONTROL_REGISTER_DEFAULT);
  signal dma_control1 : adb3_dma_control_register_array:= (others => ADB3_DMA_CONTROL_REGISTER_DEFAULT);
  signal dma_status  : adb3_dma_status_register_array;
  signal dma_status1  : adb3_dma_status_register_array;
  signal dma_status2  : adb3_dma_status_register_array;

  signal host_generic_regs_m2s : adb3_ocp_m2sT_array(0 to no_dma_engines+1);
  signal host_generic_regs_s2m : adb3_ocp_s2mT_array(0 to no_dma_engines+1);


  signal host_generic_regs_m2s_r1 : adb3_ocp_m2sT_array(0 to no_dma_engines+1);
  signal host_generic_regs_s2m_r1 : adb3_ocp_s2mT_array(0 to no_dma_engines+1);

  signal generic_regs_m2s_r2 : adb3_ocp_m2sT_array(0 to no_dma_engines);
  signal generic_regs_s2m_r2 : adb3_ocp_s2mT_array(0 to no_dma_engines);

  signal pcie_irq   : std_logic := '0';
  signal dma_irq    : std_logic_vector(no_dma_engines-1 downto 0) := (others => '0');
  signal irq_enable : std_logic_vector(no_dma_engines+1 downto 0) := (others => '0');

  signal direct_slave_m2s_i       : adb3w_ocp_m2sT;
  signal direct_slave_s2m_i       : adb3w_ocp_m2sT;
  signal direct_master_pcie_m2s_i : adb3w_ocp_m2sT;
  signal direct_master_pcie_m2s_reg : adb3w_ocp_m2sT;

  signal direct_master_enable : std_logic                    := '0';
  signal dma_enable           : std_logic_vector(no_dma_engines-1 downto 0) := (others => '0');
  signal direct_master_enable_r1 : std_logic                    := '0';
  signal dma_enable_r1           : std_logic_vector(no_dma_engines-1 downto 0) := (others => '0');
  signal model_regs_m2s_i     : adb3_ocp_m2sT;
  signal model_regs_s2m_i     : adb3_ocp_s2mT;


  signal s_m2s_i : adb3w_ocp_m2sT_array(0 to no_dma_engines-1+direct_master_channels);
  signal s_s2m_o : adb3w_ocp_s2mT_array(0 to no_dma_engines-1+direct_master_channels);


  signal clk                  : std_logic;
  signal reset3               : std_logic := '1';
  signal reset2               : std_logic := '1';
  signal reset1               : std_logic := '1';
  signal reset                : std_logic := '1';
  signal reset3u               : std_logic := '1';
  signal reset2u               : std_logic := '1';
  signal reset1u               : std_logic := '1';
  signal resetu                : std_logic := '1';
  signal trn_reset_n          : std_logic;
    signal pcie_reg_status      : pcie_reg_status_t := (
    maxpayloadsize      => (others => '0'),
    act_link_width       => (others => '0'),
    link_speed          => (others => '0'),
    link_up               => '0',
    extended_tag          => '0',
    interrupt_en          => '0',
    bus_master_en         => '0',
    maxreadreqsize        => (others => '0'),
    cfg_bus_number        => (others => '0'),
    cfg_device_number     => (others => '0'),
    trn_fc_info          => (others => '0'));
  signal pcie_reg_ctrl        : pcie_reg_ctrl_t := (tx_pkt_size => (others => '0'),
                                                    rx_pkt_size => (others => '0'),
                                                    rx_pkts_in_flight      => (others => '0'),
                                                    total_tags_inflight  => (others => '0'),
                                                    interrupt_en          => '0',
                                                    dma_to_clears          => (others => '0'),
                                                    pl_directed_link_auton  => '0',
                                                    pl_directed_link_change => (others => '0'),
                                                    pl_directed_link_speed  => '0',
                                                    pl_directed_link_width => (others => '0'),
                                                    pl_upstream_prefer_deemph => '0') ;
 

  signal cq_in              : cq_c2b_t;
  signal cq_out             : cq_b2c_t;
  signal cc_in              : cc_c2b_t;
  signal cc_out             : cc_b2c_t;
  signal rq_in              : rq_c2b_t;
  signal rq_out             : rq_b2c_t;
  signal rc_in              : rc_c2b_t;
  signal rc_out             : rc_b2c_t;
  signal messages           : message_c2b_t;
  signal int_in             : interrupts_c2b_t;
  signal int_out            : interrupts_b2c_t := (
    cfg_interrupt_int        => (others => '0'),
    cfg_interrupt_pending   => (others => '0'),
    cfg_interrupt_msi_int   => (others => '0'),
    cfg_interrupt_msi_pending_status  => (others => '0'),
    cfg_interrupt_msi_select   => (others => '0'));

  signal maxTotalPIF        : std_logic_vector(7 downto 0) := "00010000";
  signal maxPIF             : std_logic_vector(7 downto 0) := "00000100";

  signal pcie_config_status : config_c2b_t;
  signal pcie_config_status_i : config_c2b_t;
  

  
  signal bar_masks : pcie_bar_mask_t;


  signal df_not_idle_count   : std_logic_vector(8 downto 0) := (others => '0');
  signal df_not_idle_timeout : std_logic                    := '0';

  signal pcie_cscope_bus : std_logic_vector(135 downto 0);

  signal gr6, gr13 : std_logic_vector(31 downto 0) := (others => '0');


  signal local_rst : std_logic := '1';
  signal rst       : std_logic := '0';

  signal abort_count : std_logic_vector(9 downto 0) := (others => '0');
  signal auto_cleanup : std_logic := '0';
  signal auto_abort : std_logic := '0';

  signal tag_timeout_error     : std_logic_vector(no_masters-1 downto 0) := (others => '0');
  signal clear_to_error        : std_logic_vector(no_masters-1 downto 0) := (others => '0');
  signal bar_timeout : std_logic;

  signal dma_to_errors : std_logic_vector(4 downto 0) := (others => '0');

  signal dmaClr1 : std_logic_vector(7 downto 0);
  signal dmaClr2 : std_logic_vector(7 downto 0);
  signal dmaClr3 : std_logic_vector(7 downto 0);
  signal dmaClr : std_logic_vector(7 downto 0);
  
  -----------------------------------------------------------------------------
  --
  -----------------------------------------------------------------------------
  component adb3_ocp_multiport2_sc
    generic (
      round_robin_arb : boolean;
      use_timeout     : boolean;
      timeout_bits    : integer);
    port (
      ocp_m2s_memif : out adb3_ocp_m2sT;
      ocp_s2m_memif : in  adb3_ocp_s2mT;
      ocp_m2s0      : in  adb3_ocp_m2sT;
      ocp_s2m0      : out adb3_ocp_s2mT;
      ocp_m2s1      : in  adb3_ocp_m2sT;
      ocp_s2m1      : out adb3_ocp_s2mT;
      status        : out std_logic_vector(3 downto 0);
      Clk           : in  std_logic);
  end component;

  signal target_dma_abort_r1,target_dma_abort_r2 : std_logic_vector(no_dma_engines-1 downto 0);
  signal pcie_bridge_statusx_dma_to_errors : std_logic_vector(no_dma_engines downto 0) := (others => '0');
  
  
  signal pcie_td_log : std_logic_vector(127 downto 0);
  signal pcie_rd_log : std_logic_vector(255 downto 0);

  -- Debug code to check for missing DMA data
  signal dma_data0 : std_logic_vector(3 downto 0) := (others => '1');
  signal dma_data1 : std_logic_vector(3 downto 0) := "1110";

  signal pcie_write_pkt : std_logic := '0';
  signal pcie_write_toggle : std_logic := '0';
  signal dma_data2 : std_logic_vector(3 downto 0) := (others => '1');
  signal dma_data3 : std_logic_vector(3 downto 0) := "1110";
  signal pcie_write_oosync : std_logic := '0';

  signal pcie_irq1 : std_logic := '0';
  signal pcie_irq_count : std_logic_vector(7 downto 0) := (others => '0');

  signal irq_readreq : std_logic := '0';
  signal pcie_resp_pkt : std_logic := '0';

  signal tag_timeout_status : std_logic_vector(4 downto 0);

  attribute shreg_extract : string;
  attribute shreg_extract of dma_control : signal is "no";
  attribute shreg_extract of dma_control1 : signal is "no";
  attribute shreg_extract of dma_status2 : signal is "no";
  attribute shreg_extract of dma_status1 : signal is "no";
  attribute shreg_extract of dma_regs : signal is "no";
  attribute shreg_extract of general_regs : signal is "no";
  attribute shreg_extract of gr13 : signal is "no";
  attribute shreg_extract of gr6 : signal is "no";
  attribute shreg_extract of reset : signal is "no";
  attribute shreg_extract of reset1 : signal is "no";
  attribute shreg_extract of reset2 : signal is "no";
  attribute shreg_extract of reset3 : signal is "no";
  attribute shreg_extract of resetu : signal is "no";
  attribute shreg_extract of reset1u : signal is "no";
  attribute shreg_extract of reset2u : signal is "no";
  attribute shreg_extract of reset3u : signal is "no";
  attribute shreg_extract of model_regs_reset : signal is "no";
 

  signal trn_clk : std_logic;


  -----------------------------------------------------------------------------
  -----------------------------------------------------------------------------
  -- PCIe transaction capture
  type pcie_memory_type is array (0 to 1023) of std_logic_vector(63 downto 0);
  signal td_ram : pcie_memory_type;
  signal rd_ram : pcie_memory_type;
  signal td_reg,td_ram_reg,td_ram_out : std_logic_vector(63 downto 0);
  signal rd_reg,rd_ram_reg,rd_ram_out : std_logic_vector(63 downto 0);
  signal td_raddr,td_waddr : std_logic_vector(9 downto 0) := (others => '0');
  signal rd_raddr,rd_waddr : std_logic_vector(9 downto 0) := (others => '0');
  signal td_wadv : std_logic := '0';
  signal td_sync : std_logic := '0';
  signal rd_wadv : std_logic := '0';
  signal rd_sync : std_logic := '0';

  signal d1_ram : pcie_memory_type;
  signal d1_reg,d1_ram_reg,d1_ram_out : std_logic_vector(63 downto 0);
  signal d1_raddr,d1_waddr : std_logic_vector(9 downto 0) := (others => '0');
  signal d1_wadv : std_logic := '0';
  signal d1_sync : std_logic := '0';

 
  signal dm_fail,dm_fail1,dm_fail2,dm_fail3 : std_logic := '0';
  signal dm_last_burst : std_logic_vector(11 downto 0);
  signal dm_wburst_limit : std_logic_vector(11 downto 0);
  signal dm_rburst_limit : std_logic_vector(11 downto 0);
  signal maxpayloadsize_reg : std_logic_vector(2 downto 0) := "000";
  signal dm_cmd_accepted : std_logic := '0';
  signal dm_data_accepted : std_logic := '0';
  
  signal maxpayloadsize : std_logic_vector(2 downto 0);
  signal maxreadreqsize : std_logic_vector(2 downto 0);
  signal dmaif_status : std_logic_vector(31 downto 0);
  
  signal dbg_count : std_logic_vector(7 downto 0) := (others => '0');
  signal dbg_trig : std_logic := '0';
  signal dbg_rd_req : std_logic := '0';
  signal dbg_rd_resp : std_logic := '0';
  signal dbg_rd_resp_cplt : std_logic := '0';

  signal ltssm_is_L0 : std_logic := '0';

  signal msi_cap_ver : std_logic_vector(7 downto 0) := X"01";
  signal msi_mode : std_logic := '0';
 
  signal irq_reg0,irq_reg1,irq_reg2,irq_reg3,irq_reg4 : std_logic_vector(no_dma_engines+1 downto 0) := (others => '0');

 attribute shreg_extract of irq_reg0 : signal is "no";
  attribute shreg_extract of irq_reg1 : signal is "no";
  attribute shreg_extract of irq_reg2 : signal is "no";
  attribute shreg_extract of irq_reg3 : signal is "no";
  attribute shreg_extract of irq_reg4 : signal is "no";
  
  signal ltssm_state : std_logic_vector(5 downto 0);
  signal ltssm_state_is_l0 : std_logic;
  signal ltssm_state_is_l0_r1 : std_logic;
  signal ltssm_state_l0_lost : std_logic;
  signal ltssm_state_l0_lost_count : std_logic_vector(15 downto 0);

  signal reset_bufg : std_logic;

  signal pcie_config_cfg_fc_sel : std_logic_vector(2 downto 0) := "000";


  signal dma_fetch_not_idle_count : std_logic_vector(15 downto 0) := (others => '0');
  signal dma_fetch_not_idle : std_logic := '0';                                                                  
  signal dm_order_status : std_logic_vector(63 downto 0);
  signal dm_clr_error : std_logic_vector(15 downto 0);

  signal pcie_rst_delay_cnt : std_logic_vector(13 downto 0) := (others => '1');
  signal pcie_rst_held : std_logic :=  '0';
  signal pcie_rst_held_r1 : std_logic := '0';
  signal pcie_rst_heldu : std_logic :=  '0';
  signal pcie_rst_heldu_r1 : std_logic := '0';

  signal local_reset,local_reset_r1 : std_logic := '0';
  
  attribute keep : string;
  attribute max_fanout : string;
  attribute equivalent_register_removal : string;
  attribute keep of local_reset : signal is "TRUE";
  attribute equivalent_register_removal of local_reset : signal is "no";
  attribute max_fanout of local_reset : signal is "1";
  attribute keep of local_reset_r1 : signal is "TRUE";
  attribute equivalent_register_removal of local_reset_r1 : signal is "no";

  signal local_reset_r2: std_logic := '1';
  attribute keep of local_reset_r2 : signal is "TRUE";  
  attribute equivalent_register_removal of local_reset_r2 : signal is "no";
  attribute max_fanout of local_reset_r2 : signal is "16";
  attribute shreg_extract of local_reset_r2 : signal is "no";
  attribute shreg_extract of local_reset_r1 : signal is "no";
  
begin  -- synth

  ocpclkout <= clk;
  core_rst_out <= pcie_rst_heldu_r1;
  
  clk_core_status: process (model_clk) is
  begin  -- process clk_core_status
    if rising_edge(model_clk) then        -- rising clock edge
      pcie_config_status <= pcie_config_status_i;
    end if;
  end process clk_core_status;
  core_status(23 downto 0) <= EXT(        
        pcie_config_status.cfg_ltssm_state      &  -- 21:16
        pcie_config_status.cfg_max_read_req         &  -- 15:13
        pcie_config_status.cfg_max_payload          &  -- 12:10
        pcie_config_status.cfg_current_speed    &  -- 9:7 
        pcie_config_status.cfg_negotiated_width &  -- 6:3
        pcie_config_status.cfg_phy_link_status &   -- 2:1
        pcie_config_status.cfg_phy_link_down       -- 0
        ,24);
  core_status(63 downto 32) <= (others => '0');
  gen_abrt_cleanup: for i in 0 to 3 generate
    gen_ac0: if i< no_dma_engines generate
      core_status(24+i) <= dma_control(i).abort;
      core_status(28+i) <= dmaClr(i);
    end generate gen_ac0;
    gen_ac1: if i>= no_dma_engines generate
      core_status(24+i) <= '0';
      core_status(28+i) <= '0';
    end generate gen_ac1;  
  end generate gen_abrt_cleanup;
  
  

 pcie3_7x_0_wrapper_1: entity work.pcie3_ku_0_wrapper
   generic map (
     pf0_bar2_aperture_size => bar2_aperture,
     pf0_bar4_aperture_size => bar2_aperture)
  port map (
    sys_reset_n        => sys_reset_n,
    sys_clk            => sys_clk,
    sys_clk_gt            => sys_clk_gt,
    pci_exp_txn        => pci_exp_txn,
    pci_exp_txp        => pci_exp_txp,
    pci_exp_rxn        => pci_exp_rxn,
    pci_exp_rxp        => pci_exp_rxp,
    trn_clk            => trn_clk,
    trn_reset_n        => trn_reset_n,
    cq_in              => cq_in,
    cq_out             => cq_out,
    cc_in              => cc_in,
    cc_out             => cc_out,
    rq_in              => rq_in,
    rq_out             => rq_out,
    rc_in              => rc_in,
    rc_out             => rc_out,
    messages           => messages,
    int_in             => int_in,
    int_out            => int_out,
    pcie_config_status => pcie_config_status_i,
    pcie_config_cfg_fc_sel => pcie_config_cfg_fc_sel);
 

 
  clk <= trn_clk;
 

  -- Delay Reset of USER Logic for 16k x 10ns clock cycles
  -- after PCIE_RST siganl applied
  -- Allows excessive amount of time for reset to propagate
  -- through PCIE Core and logic, and also for trn_clk
  -- to be stable
  -- PCIE should still be in Gen 1 training pattern
  process (model_clk,rst) is
  begin  -- process
    if rst = '1' then
      pcie_rst_delay_cnt <= (others => '1');
    elsif rising_edge(model_clk) then
      if OR_reduce(pcie_rst_delay_cnt) = '1' then
        pcie_rst_delay_cnt <= pcie_rst_delay_cnt-1;
      end if;
    end if;
  end process;
  
  process (model_clk) is
  begin  -- process
    if rising_edge(model_clk) then        -- rising clock edge
      if OR_reduce(pcie_rst_delay_cnt(5 downto 0)) = '1' and OR_reduce(pcie_rst_delay_cnt(13 downto 8)) = '0' and pcie_rst_delay_cnt(7 downto 6) = "10" then
        pcie_rst_held <= '1';
      else
        pcie_rst_held <= '0';
      end if;
      pcie_rst_held_r1 <= pcie_rst_held;
       if OR_reduce(pcie_rst_delay_cnt(6 downto 0)) = '1' and OR_reduce(pcie_rst_delay_cnt(13 downto 8)) = '0' then
        pcie_rst_heldu <= '1';
      else
        pcie_rst_heldu <= '0';
      end if;
      pcie_rst_heldu_r1 <= pcie_rst_heldu;
    end if;
  end process;
  -- Cross clock delayed reset from sys_clk to trn_clk
  process (clk) is
  begin  -- process
    if rising_edge(clk) then            -- rising clock edge
      reset1 <= reset2;
      reset  <= reset1;
      reset2 <= reset3;
      reset3 <= pcie_rst_held_r1;
--      reset2u <= reset3u;
--      reset3u <= pcie_rst_heldu_r1;
--      reset1u <= reset2u;
--      resetu  <= reset1u;
    end if;
  end process;
  
  process (clk,reset_bufg)
  begin  -- process
    if reset_bufg = '1' then
      abort_count <= (others => '0');
      abort_count(0) <= '1';
      auto_abort <= '1';
    elsif clk'event and clk = '1' then     -- rising clock edge
      if OR_reduce(abort_count) = '1' then
        abort_count <= abort_count+1;
        auto_abort <= '0';
      end if;
      auto_cleanup <= AND_reduce(abort_count);
    end if;
  end process;
  
  model_regs_reset <= reset_bufg;

  -- Register Control of Direct Master Access to PCIe
  direct_master_pcie_m2s_i.Cmd         <= direct_master_pcie_m2s.Cmd when direct_master_enable = '1' else OCP_CMD_IDLE;
  direct_master_pcie_m2s_i.Addr        <= direct_master_pcie_m2s.Addr;
  direct_master_pcie_m2s_i.BurstLength <= direct_master_pcie_m2s.BurstLength;
  direct_master_pcie_m2s_i.Data        <= direct_master_pcie_m2s.Data;
  direct_master_pcie_m2s_i.DataByteEn  <= direct_master_pcie_m2s.DataByteEn;
  direct_master_pcie_m2s_i.DataValid   <= direct_master_pcie_m2s.DataValid;
  direct_master_pcie_m2s_i.RespAccept  <= direct_master_pcie_m2s.RespAccept;
  direct_master_pcie_m2s_i.Tag         <= direct_master_pcie_m2s.Tag;


  -- Generate a Power on Reset for PCIE bridge
  rst <= (not trn_reset_n) or (not model_clk_locked);
  --rst <= not trn_reset_n;

  reset_bufg <= pcie_rst_held_r1;
  

 pcie_g3_bridge_1: entity work.pcie_g3_bridge
    generic map (
      no_slaves  => no_slaves,
      no_masters => no_masters,
      dmChannel  => no_dma_engines)
    port map (
      pcie_rst           => reset_bufg,
      pcie_clk           => trn_clk,
      bar_masks          => bar_masks,
      cq_in              => cq_in,
      cq_out             => cq_out,
      cc_in              => cc_in,
      cc_out             => cc_out,
      rq_in              => rq_in,
      rq_out             => rq_out,
      rc_in              => rc_in,
      rc_out             => rc_out,
      messages           => messages,
      int_in             => int_in,
      int_out            => int_out,
      m_s2m_o            => s_s2m_o,
      m_m2s_i            => s_m2s_i,
      s_s2m_i            => m_s2m_i,
      s_m2s_o            => m_m2s_o,
      s_s2m_iw            => m_s2m_iw,
      s_m2s_ow            => m_m2s_ow,
      interrupt          => pcie_irq,
      pcie_reg_ctrl      => pcie_reg_ctrl,
      pcie_reg_status    => pcie_reg_status,
      maxTotalPIF(5 downto 0)        => pcie_reg_ctrl.total_tags_inflight,
      maxTotalPIF(7 downto 6)       => "00",
      maxPIF(3 downto 0)             => pcie_reg_ctrl.rx_pkts_in_flight,
      maxPIF(7 downto 4)             => X"0",
       dmaif_status => dmaif_status,
      dmaClr => dmaClr,
      tag_timeout_error  => tag_timeout_error,
      clear_to_error     => clear_to_error,
      timeout_bar23      => bar_timeout,
      clr_timeout_bar23  => clear_to_error(0));

 
  generic_regs_m2s <= m_m2s_o(0);
  model_regs_m2s_i <= m_m2s_o(1);
  direct_slave_m2s_i <= m_m2s_ow;
  m_s2m_i(0)           <= generic_regs_s2m;
  m_s2m_i(1)          <= model_regs_s2m_i;
  m_s2m_iw          <= direct_slave_s2m;


  direct_slave_m2s <= direct_slave_m2s_i;

  gen_dm: if direct_master_channels = 1 generate
    adb3w_ocp_reorder_1: entity work.adb3w_ocp_reorder
      generic map (
        buffer_order       => 5,
        max_read_in_flight => 8,
        enforce_max_read   => true,
        unified_ram        => true,
        output_pipe        => true)
      port map (
        rst       => reset_bufg,
        sr        => '0',
        clk       => clk,
        io_m2s    => direct_master_pcie_m2s_i,
        io_s2m    => direct_master_pcie_s2m,
        uo_m2s    => s_m2s_i(no_dma_engines),
        uo_s2m    => s_s2m_o(no_dma_engines),
        status    => dm_order_status,
        clr_error => dm_clr_error);
   
  end generate gen_dm;
  gen_nodm: if direct_master_channels = 0 generate
    direct_master_pcie_s2m <= ADB3W_OCP_DEFAULT_S2M;
  end generate gen_nodm;
  s_m2s_i(0 to no_dma_engines-1) <= dma_channels_i_m2s(0 to no_dma_engines-1);
  dma_channels_i_s2m(0 to no_dma_engines-1) <= s_s2m_o(0 to no_dma_engines-1);

  process (model_regs_m2s_i, model_regs_s2m, pcie_config_status)
  begin  -- process
    model_regs_m2s           <= model_regs_m2s_i;
    model_regs_s2m_i         <= model_regs_s2m;
    model_regs_m2s.Addr(63 downto 0) <= X"00000000_00000000";
    model_regs_m2s.Addr(11 downto 4) <= model_regs_m2s_i.Addr(11 downto 4);
    --if pcie_config_status.cfg_phy_link_down = '1' then
    --  model_regs_m2s.Cmd <= OCP_CMD_IDLE;
    --  model_regs_m2s.DataValid <= '0';
    --  model_regs_s2m_i.Resp    <=  OCP_RESP_NONE;
    --end if;
  end process;

  process (generic_regs_m2s, generic_regs_s2m_i, pcie_config_status)
  begin  -- process
    generic_regs_m2s_i           <= generic_regs_m2s;
    generic_regs_s2m             <= generic_regs_s2m_i;
    generic_regs_m2s_i.Addr(63 downto 0) <= X"00000000_00000000";
    generic_regs_m2s_i.Addr(11 downto 4) <= generic_regs_m2s.Addr(11 downto 4);
    --if pcie_config_status.cfg_phy_link_down = '1' then
    --  generic_regs_m2s_i.Cmd <= OCP_CMD_IDLE;
    --  generic_regs_m2s_i.DataValid <= '0';
    --  generic_regs_s2m.Resp    <=  OCP_RESP_NONE;
    --end if;
  end process;

  gen_dma_engines0 : for i in 0 to no_dma_engines-1 generate
    adb3_dma_engine_1 : entity work.adb3_dma_engine256
      generic map (
        registering_for_speed => dma_config(i).registering_for_speed,
        fpga_descriptors => dma_config(i).fpga_descriptors,
        fpga_to_pci_only => dma_config(i).fpga_to_pci_only,
        pci_to_fpga_only => dma_config(i).pci_to_fpga_only,
        read_worker_chipscope => false)
      port map (
        clk          => clk,
        irq          => dma_irq(i),
        pcie_ocp_m2s => dma_channels_i_m2s(i),
        pcie_ocp_s2m => dma_channels_i_s2m(i),
        mptl_ocp_m2s => dma_channels_f_m2s(i),
        mptl_ocp_s2m => dma_channels_f_s2m(i),
        control      => dma_control1(i),
        status       => dma_status1(i));

    dma_channels_m2s(i)   <= dma_channels_f_m2s(i);
    dma_channels_f_s2m(i) <= dma_channels_s2m(i);

    gen_dma_control_no_regs: if not register_dma_regs generate
      dma_control1(i) <= dma_control(i);
    end generate gen_dma_control_no_regs;

    gen_dma_control_regs: if register_dma_regs generate
      process (clk) is
      begin  -- process
        if rising_edge(clk) then        -- rising clock edge
          dma_control1(i) <= dma_control(i);
        end if;
      end process;
    end generate gen_dma_control_regs;

  end generate gen_dma_engines0;


  gen_fpga_reg_access: if reg_channels = 1 generate

    adb3l_to_adb3_1: entity work.adb3l_to_adb3
    port map (
      pri_m2s => direct_master_regs_m2s,
      pri_s2m => direct_master_regs_s2m,
      sec_m2s => generic_regs_m2s_fpga,
      sec_s2m => generic_regs_s2m_fpga);

  adb3_ocp_multiport2_sc_1: entity work.adb3_ocp_multiport2_sc
    generic map (
      round_robin_arb => false,
      use_timeout     => false,
      timeout_bits    => 16)
    port map (
      ocp_m2s_memif => generic_regs_m2s_r11,
      ocp_s2m_memif => generic_regs_s2m_r11,
      ocp_m2s0      => generic_regs_m2s_i,
      ocp_s2m0      => generic_regs_s2m_i,
      ocp_m2s1      => generic_regs_m2s_fpga,
      ocp_s2m1      => generic_regs_s2m_fpga,
      status        => open,
      Clk           => clk);

     adb3_ocp_register_1 : adb3_ocp_register
    port map (
      clk        => clk,
      master_m2s => generic_regs_m2s_r11,
      master_s2m => generic_regs_s2m_r11,
      slave_m2s  => generic_regs_m2s_r1,
      slave_s2m  => generic_regs_s2m_r1);
  end generate gen_fpga_reg_access;
  
  
  gen_no_fpga_regs: if reg_channels = 0 generate
    adb3_ocp_register_1 : adb3_ocp_register
    port map (
      clk        => clk,
      master_m2s => generic_regs_m2s_i,
      master_s2m => generic_regs_s2m_i,
      slave_m2s  => generic_regs_m2s_r1,
      slave_s2m  => generic_regs_s2m_r1);

    direct_master_regs_s2m <= ADB3L_OCP_DEFAULT_S2M;
  end generate gen_no_fpga_regs;
 


  adb3_ocp_demux_1 : adb3_ocp_demux
    generic map (
      number_of_slaves  => no_dma_engines+2,
      demux_address_lsb => 6)
    port map (
      clk        => clk,
      master_m2s => generic_regs_m2s_r1,
      master_s2m => generic_regs_s2m_r1,
      slaves_m2s => host_generic_regs_m2s,
      slaves_s2m => host_generic_regs_s2m);


  gen_register_ocp_regs : for i in 0 to no_dma_engines+1 generate
    adb3_ocp_register_3 : adb3_ocp_register
      port map (
        clk        => clk,
        master_m2s => host_generic_regs_m2s(i),
        master_s2m => host_generic_regs_s2m(i),
        slave_m2s  => host_generic_regs_m2s_r1(i),
        slave_s2m  => host_generic_regs_s2m_r1(i));

  end generate gen_register_ocp_regs;

  dummy_regs_0 : adb3_ocp_register_interface32
    generic map (
      number_of_registers => 1)
    port map (
      ocp_clk          => clk,
      ocp_m2s          => host_generic_regs_m2s_r1(no_dma_engines+1),
      ocp_s2m          => host_generic_regs_s2m_r1(no_dma_engines+1),
      registers_out    => open,
      registers_back   => dummy_regs0,
      register_written => open);

  
  generic_regs_m2s_r2(0) <= host_generic_regs_m2s_r1(0);
  host_generic_regs_s2m_r1(0) <= generic_regs_s2m_r2(0);

  general_regs0 : adb3_ocp_register_interface32
    generic map (
      number_of_registers => no_general_regs)
    port map (
      ocp_clk          => clk,
      ocp_m2s          => generic_regs_m2s_r2(0),
      ocp_s2m          => generic_regs_s2m_r2(0),
      registers_out    => general_regs_out0,
      registers_back   => general_regs,
      register_written => general_regs_written0);

  process (clk,reset_bufg) is
  begin  -- process
    if reset_bufg = '1' then
      local_reset <= '1';
    elsif rising_edge(clk) then            -- rising clock edge
      local_reset <= '0';
    end if;
  end process;
  
  general_regs_p0 : process (clk)
  begin  -- process general_regs_p0
    if clk'event and clk = '1' then     -- rising clock edge
      local_reset_r1 <= local_reset;
      local_reset_r2 <= local_reset_r1;
      if local_reset_r2 = '1' then
        general_regs(1) <= (others => '0');
        general_regs(2) <= (others => '0');
        general_regs(5) <= pif_init_val & X"100100";
        general_regs(7) <= bar_mask_init_val;
        general_regs(8) <= (others => '0');
        general_regs(9) <= (others => '0');
      end if;
      for i in 0 to no_general_regs-1 loop
        if general_regs_written0(i) = '1' then
          general_regs(i)         <= general_regs_out0(i);
          general_regs_written(i) <= '1';
        else
          general_regs_written(i) <= '0';
        end if;
      end loop;  -- i

      -- IRQ Status in Reg 0
      general_regs(0)            <= EXT(dma_irq & model_regs_irq & (OR_reduce(tag_timeout_error) or dma_to_errors(0)), 32);
      irq_reg0 <= dma_irq & model_regs_irq & (OR_reduce(tag_timeout_error) or dma_to_errors(0));
      irq_reg1 <= irq_reg0;
      irq_reg2 <= irq_reg1;
      irq_reg3 <= irq_reg2;
      irq_reg4 <= irq_reg3;
      
      -- IRQ Enable in Reg 1
      irq_enable                 <= general_regs(1)(no_dma_engines+1 downto 0);

      if msi_mode = '0' then
        pcie_irq                   <= OR_reduce(irq_enable and (dma_irq & model_regs_irq &
                                    (OR_reduce(tag_timeout_error) or dma_to_errors(0))));
      else
        pcie_irq <= OR_reduce((irq_enable and irq_reg0) and not irq_reg4);
      end if;
      msi_mode <= int_in.cfg_interrupt_msi_enable(0);
                                                                                           
      if irq_reg0(2) = '0' then
        pcie_irq_count <= (others => '0');
      else
        pcie_irq_count <=  pcie_irq_count +1;
      end if;

      
      pcie_reg_ctrl.interrupt_en <= OR_reduce(general_regs(1)(no_dma_engines+1 downto 0));
      -- DMA and Direct Master Enable in Reg 2
      direct_master_enable_r1       <= general_regs(2)(0);
      dma_enable_r1                 <= general_regs(2)(no_dma_engines downto 1);
      direct_master_enable       <= direct_master_enable_r1  and not dma_to_errors(0);

      if vpd_valid = '1' then
        dma_enable                 <= dma_enable_r1 and not dma_to_errors(no_dma_engines downto 1);
      else
        dma_enable <= (others => '0');
        general_regs(2)(no_dma_engines downto 1) <= (others => '0');
      end if;
      
      
      

      -- Reg 3 : Features Register
      -- Bits 0:3 - MSI Version Number
      -- Bit 4 : DMA Descriptor fetch latches NDL writes
      -- Bit 5 : Includes ICAP
      if use_icap then
        general_regs(3) <= X"0000003" & EXT(msi_cap_ver,4);
      else
        general_regs(3) <= X"0000001" & EXT(msi_cap_ver,4);
      end if;
     

        
      -- Reg 4 LTSSM Error count
      ltssm_state <= pcie_config_status.cfg_ltssm_state;
      if ltssm_state = "010000" then
        ltssm_state_is_l0 <= '1';
      else
        ltssm_state_is_l0 <= '0';
      end if;
      ltssm_state_is_l0_r1 <= ltssm_state_is_l0;
      ltssm_state_l0_lost <= ltssm_state_is_l0_r1 and not ltssm_state_is_l0;
      if reset_bufg = '1' then
        ltssm_state_l0_lost_count <= (others => '0');
      else
        if ltssm_state_l0_lost = '1' then
          ltssm_state_l0_lost_count <= ltssm_state_l0_lost_count +1; 
        end if;
      end if;
      general_regs(4) <= X"0000" & ltssm_state_l0_lost_count;
      
      -- Reg 5,6 PCIe Control and Status
      pcie_reg_ctrl.tx_pkt_size       <= general_regs(5)(9 downto 0);
      pcie_reg_ctrl.rx_pkt_size       <= general_regs(5)(21 downto 12);
      pcie_reg_ctrl.rx_pkts_in_flight <= general_regs(5)(27 downto 24);
      pcie_config_cfg_fc_sel <= general_regs(5)(2 downto 0);
        
      pcie_reg_ctrl.total_tags_inflight <= ("00" & general_regs(5)(27 downto 24)) + ("00" & general_regs(5)(31 downto 28));

      general_regs(6) <= gr6;
      gr6 <= EXT(pcie_config_status.cfg_phy_link_down &                -- bit 30 
                 pcie_config_status.cfg_current_speed &             -- bits 29:27  
                 '0' & pcie_config_status.cfg_negotiated_width &         -- bits 25:22 
                 pcie_config_status.cfg_max_payload &         -- bits 21:19
                 '1' & --pcie_reg_status.extended_tag &           -- bit 18 
                 '0' & --pcie_reg_status.interrupt_en &           -- bit 17 (L)
                 '1' & --pcie_reg_status.bus_master_en &          -- bit 16 
                 pcie_config_status.cfg_max_read_req &         -- bits 15:13
                 --pcie_config_status.cfg_ds_bus_number &         -- bits 12:5 
                 --pcie_config_status.cfg_ds_device_number, 32);  -- bits 4:0
                 "0000000000000",32);

      bar_masks.bar0_mask <= X"00000000_00000FFF";
      bar_masks.bar0_page <= X"00000000_00000000";
      bar_masks.bar1_mask <= X"00000000_00000FFF";
      bar_masks.bar1_page <= X"00000000_00000000";

      bar_masks.bar23_mask(31 downto bar2_size+7) <= (others => '0');
      bar_masks.bar23_mask(bar2_size+6 downto 0)  <= (others => '1');
      bar_masks.bar23_mask(63 downto 32) <= (others => '0');
      bar_masks.bar45_mask(31 downto bar2_size+7) <= (others => '0');
      bar_masks.bar45_mask(bar2_size+6 downto 0)  <= (others => '1');
      bar_masks.bar45_mask(63 downto 32) <= (others => '0');
      general_regs(7)(bar2_size+6 downto 0)       <= (others => '1');
      general_regs(7)(31 downto bar2_size+7)       <= (others => '0'); 
      bar_masks.bar23_page               <= general_regs(9) & general_regs(8);
      bar_masks.bar45_page               <= general_regs(9) & general_regs(8);

      general_regs(12) <= dmaif_status;
      general_regs(13) <= gr13;

      gr13 <=
              --pcie_core_rd_i.dbg_rdst_rdy_n_sticky & pcie_core_rd_i.dbg_rdst_fifo_afull_sticky &
             
              "00" & dm_fail & --tag_timeout_status &            -- 31:27
              --EXT(tag_timeout_error,5) &             -- 26:24
              "00" & bar_timeout & "00" &
              dma_to_errors &                 -- 23:19
              bar_timeout & 
              "00" &
              X"0000";
      
              --pcie_bridge_status.interrupt &  -- 18
              --pcie_bridge_status.link_up_n &  -- 17
              --pcie_bridge_status.rx_bar_not_full & -- 16:13
              --pcie_bridge_status.rx_cpld_not_full &  --12:8
              --pcie_bridge_status.tx_fsm &  -- 7:5
              --pcie_bridge_status.tx_tags_active;  -- 4:0


      
      if direct_master_channels = 1 then
        general_regs(14) <= dm_order_status(31 downto 0);
        general_regs(15) <= dm_order_status(63 downto 32);
      end if;
      if general_regs_written0(14) = '1' then
        dm_clr_error <= general_regs_out0(14)(15 downto 0);
      else
        dm_clr_error <= (others => '0');
      end if;
      
      if auto_cleanup = '1' or general_regs_written(13) = '1' then
        clear_to_error <= (others => '1');
        pcie_reg_ctrl.dma_to_clears <= (others => '1');
      else
        clear_to_error <= (others => '0');
        pcie_reg_ctrl.dma_to_clears <= (others => '0');
      end if;
      dma_to_errors(no_masters-1 downto 1) <= (others => '0'); --pcie_bridge_status.dma_to_errors(no_slaves-1 downto 1);
      dma_to_errors(0) <= dm_fail; -- pcie_bridge_status.dma_to_errors(0) or dm_fail;
      

      
      -- Check Direct Master Writes
      --if auto_cleanup = '1' or general_regs_written(13) = '1' then
      --  dm_fail <= '0';
      --else
      --  if dm_fail1 = '1' or dm_fail2 = '1' or dm_fail3 = '1' then
      --    dm_fail <= '1';
      --  end if;       
      --end if;
      --if dm_data_accepted = '1' and dm_last_burst /= X"001" and direct_master_pcie_m2s_reg.DataValid = '1' and direct_master_pcie_m2s_reg.DataByteEn /= X"FFFF" then
      --  dm_fail1 <= '1';
      --else
      --  dm_fail1 <= '0';
      --end if;
        
      --if dm_cmd_accepted = '1' and direct_master_pcie_m2s_reg.Cmd = OCP_CMD_WRITE and direct_master_pcie_m2s_reg.BurstLength > dm_wburst_limit then
      --  dm_fail2 <= '1';
      --else
      --  dm_fail2 <= '0';
      --end if;

      --if dm_cmd_accepted = '1' and direct_master_pcie_m2s_reg.Cmd = OCP_CMD_READ and direct_master_pcie_m2s_reg.BurstLength > dm_rburst_limit then
      --  dm_fail3 <= '1';
      --else
      --  dm_fail3 <= '0';
      --end if;
      
      --dm_cmd_accepted <=  s_s2m_o(0).CmdAccept;
      --dm_data_accepted <=  s_s2m_o(0).DataAccept;

      --direct_master_pcie_m2s_reg <= direct_master_pcie_m2s;
      
      --if direct_master_pcie_m2s_reg.Cmd = OCP_CMD_WRITE then
      --  dm_last_burst <= direct_master_pcie_m2s.BurstLength;
      --end if;

      maxpayloadsize <= pcie_config_status.cfg_max_payload;--pcie_reg_status.maxpayloadsize;
        case maxpayloadsize  is
          when "000" => dm_wburst_limit <= "000000000100";
          when "001" =>  dm_wburst_limit <= "000000001000";
          when "010" =>  dm_wburst_limit <= "000000010000";
          when "011" =>  dm_wburst_limit <= "000000010000";   
          when others => dm_wburst_limit <= "000000010000";
        end case;
      --  maxpayloadsize_reg <= pcie_reg_status.maxpayloadsize;
      --end if;
      
      maxreadreqsize <= pcie_config_status.cfg_max_read_req;--pcie_reg_status.maxreadreqsize;
      case  maxreadreqsize  is
        when "000" => dm_rburst_limit <=  "000000000100";
        when "001" =>  dm_rburst_limit <= "000000001000";
        when "010" =>  dm_rburst_limit <= "000000010000";
        when "011" =>  dm_rburst_limit <= "000000100000";
        when "100" =>  dm_rburst_limit <= "000001000000";
        when others => dm_rburst_limit <= "000010000000";
      end case;

      if dma_status(0).fsm_idle(0) = '0' then
        dma_fetch_not_idle_count <=  dma_fetch_not_idle_count+1;
      else
        dma_fetch_not_idle_count <= (others => '0');
      end if;

      if dma_fetch_not_idle_count = X"FFFF" then
        dma_fetch_not_idle <= '1';
      end if;
      
      
    end if;
  end process general_regs_p0;

  



  -- DMA Registers
  gen_dma_registers : for i in 0 to no_dma_engines-1 generate

 -- Do not connect FPGA Direct Master Access to Regs
  generic_regs_m2s_r2(i+1) <= host_generic_regs_m2s_r1(i+1);
  host_generic_regs_s2m_r1(i+1) <= generic_regs_s2m_r2(i+1);
    
    dma_regs0 : adb3_ocp_register_interface32
      generic map (
        number_of_registers => no_dma_regs)
      port map (
        ocp_clk          => clk,
        ocp_m2s          => generic_regs_m2s_r2(i+1),
        ocp_s2m          => generic_regs_s2m_r2(i+1),
        registers_out    => dma_regs_out0(i),
        registers_back   => dma_regs(i),
        register_written => dma_regs_written0(i));

    gen_direct_dma_regs: if not register_dma_regs generate
      overwrite_dma_regs0 : process (dma_regs_out0, dma_status,dma_status1)
      begin  -- process overwrite_dma_regs0
        for j in 0 to no_dma_regs-1 loop
          dma_regs(i)(j)         <= dma_regs_out0(i)(j);
          dma_regs_written(i)(j) <= dma_regs_written0(i)(j);
        end loop;  -- j

        -- Overwrite Status bits
        dma_regs(i)(0) <= dma_status(i).fsm_fail & dma_status(i).fsm_aborting & dma_status(i).fsm_idle;
        dma_regs(i)(1) <= dma_status(i).fifo_info;

        --if dma_regs_written0(i)(2) = '0' then
        dma_regs(i)(2)(7 downto 0) <= dma_status(i).irq_status;

        if pcie_bridge_statusx_dma_to_errors(i+1) = '1' then
          dma_regs(i)(2)(8) <= '1';
        elsif dma_regs_written0(i)(2) = '1' then
          dma_regs(i)(2)(8) <= '0';
        end if;
        if target_dma_abort(i) = '1' then
          dma_regs(i)(2)(9) <= '1';
        elsif dma_regs_written0(i)(2) = '1' then
          dma_regs(i)(2)(9) <= '0';
        end if;
        
        --end if;
        dma_regs(i)(9) <= dma_status(i).illegal_we & EXT(dma_status(i).df_waiting, 31);
        dma_status(i) <= dma_status1(i);
      end process overwrite_dma_regs0;

    end generate gen_direct_dma_regs;

    gen_reg_dma_regs: if register_dma_regs generate
      overwrite_dma_regs1 : process (clk)
      begin  -- process overwrite_dma_regs0
        if rising_edge(clk) then


          for j in 0 to no_dma_regs-1 loop
            dma_regs(i)(j)         <= dma_regs_out0(i)(j);
            dma_regs_written(i)(j) <= dma_regs_written0(i)(j);
          end loop;  -- j

          -- Overwrite Status bits
          dma_regs(i)(0) <= dma_status(i).fsm_fail & dma_status(i).fsm_aborting & dma_status(i).fsm_idle;
          dma_regs(i)(1) <= dma_status(i).fifo_info;

          --if dma_regs_written0(i)(2) = '0' then
          dma_regs(i)(2)(7 downto 0) <= dma_status(i).irq_status;
          
          if pcie_bridge_statusx_dma_to_errors(i+1) = '1' then
            dma_regs(i)(2)(8) <= '1';
          elsif dma_regs_written0(i)(2) = '1' then
            dma_regs(i)(2)(8) <= '0';
          end if;
          if target_dma_abort(i) = '1' then
            dma_regs(i)(2)(9) <= '1';
          elsif dma_regs_written0(i)(2) = '1' then
            dma_regs(i)(2)(9) <= '0';
          end if;
          
          --end if;
          dma_regs(i)(9) <= dma_status(i).illegal_we & EXT(dma_status(i).df_waiting, 31);
          dma_status2(i) <= dma_status1(i);
          dma_status(i) <= dma_status2(i);
          --dma_regs(i)(10) <= dma_status(i).debug(31 downto 0);
        end if;
      end process overwrite_dma_regs1;

    end generate gen_reg_dma_regs;


    dma_regs_p0 : process (clk)
    begin  -- process general_regs_p0
      if clk'event and clk = '1' then   -- rising clock edge
        
        -- DMA Register Mapping

        -- Trigger DMA Abort on rising edge of target abort signal
        pcie_bridge_statusx_dma_to_errors(i+1) <= '0'; --pcie_bridge_status.dma_to_errors(i+1);
        target_dma_abort_r1(i) <= target_dma_abort(i) or pcie_bridge_statusx_dma_to_errors(i+1);
        target_dma_abort_r2(i) <= target_dma_abort_r1(i);
        -- In clocked process to help with timing
        dma_control(i).abort   <= dma_regs_written(i)(0) or auto_abort or (target_dma_abort_r1(i) and not target_dma_abort_r2(i));
        dma_control(i).cleanup <= dma_regs_written(i)(1) or auto_cleanup;
        dmaClr1(i) <= dma_control(i).cleanup;
        dmaClr2(i) <= dmaClr1(i);
        dmaClr3(i) <= dmaClr2(i);
        dmaClr(i) <= dmaClr1(i) or dmaClr2(i) or dmaClr3(i);
        dma_control(i).irq_ack_register <= (dma_regs(i)(2)(7 downto 0)) or (auto_cleanup & "0000000");
        dma_control(i).irq_ack_we       <= dma_regs_written(i)(2) or auto_cleanup;

        dma_control(i).irq_enable    <= dma_regs(i)(3)(7 downto 0);
        dma_control(i).irq_enable_we <= dma_regs_written(i)(3);

        dma_control(i).descriptor_host_addr        <= dma_regs(i)(5) & dma_regs(i)(4);
        dma_control(i).descriptor_host_we          <= dma_regs_written(i)(4) and not dma_regs(i)(6)(18);
        dma_control(i).descriptor_host_wait_ack    <= dma_regs_written(i)(9);
        dma_control(i).descriptor_fpga_addr        <= dma_regs(i)(5) & dma_regs(i)(4);
        dma_control(i).descriptor_fpga_we          <= dma_regs_written(i)(4) and dma_regs(i)(6)(18);
        dma_control(i).descriptor_fpga_wait_ack    <= dma_regs_written(i)(9);
   
        if dma_regs(i)(6)(16) = '1' then
          dma_control(i).host_write_worker_max_burst <= "0000000010";
        else
          case  maxpayloadsize  is
            when "000" => dma_control(i).host_write_worker_max_burst <= "0000000100";
            when "001" => dma_control(i).host_write_worker_max_burst <= "0000001000";
            when "010" => dma_control(i).host_write_worker_max_burst <= "0000010000";
            when "011" => dma_control(i).host_write_worker_max_burst <= "0000010000";  
            when others =>  dma_control(i).host_write_worker_max_burst <= "0000010000";
          end case;
        end if;
 
        case maxreadreqsize  is
           when "000" =>
             if  dma_regs(i)(6)(9 downto 0) > "0000000100" then
               dma_control(i).host_read_worker_max_burst  <= "0000000100";
             else
               dma_control(i).host_read_worker_max_burst  <= dma_regs(i)(6)(9 downto 0);   
             end if;
          when "001" =>
             if  dma_regs(i)(6)(9 downto 0) > "0000001000" then
               dma_control(i).host_read_worker_max_burst  <= "0000001000";
             else
               dma_control(i).host_read_worker_max_burst  <= dma_regs(i)(6)(9 downto 0);   
             end if;
           when "010" =>
             if  dma_regs(i)(6)(9 downto 0) > "0000010000" then
               dma_control(i).host_read_worker_max_burst  <= "0000010000";
             else
               dma_control(i).host_read_worker_max_burst  <= dma_regs(i)(6)(9 downto 0);   
             end if;
          when "011" =>
             if  dma_regs(i)(6)(9 downto 0) > "0000100000" then
               dma_control(i).host_read_worker_max_burst  <= "0000100000";
             else
               dma_control(i).host_read_worker_max_burst  <= dma_regs(i)(6)(9 downto 0);   
             end if;
          when "100" =>
             if  dma_regs(i)(6)(9 downto 0) > "0001000000" then
               dma_control(i).host_read_worker_max_burst  <= "0001000000";
             else
               dma_control(i).host_read_worker_max_burst  <= dma_regs(i)(6)(9 downto 0);   
             end if;
          when others =>
             if  dma_regs(i)(6)(9 downto 0) > "0010000000" then
               dma_control(i).host_read_worker_max_burst  <= "0010000000";
             else
               dma_control(i).host_read_worker_max_burst  <= dma_regs(i)(6)(9 downto 0);   
             end if;
        end case;
             
          
        dma_control(i).fpga_write_worker_max_burst <= dma_regs(i)(7)(9 downto 0);
        dma_control(i).fpga_read_worker_max_burst  <= dma_regs(i)(7)(25 downto 16);
        

        --dma_control(i).alignment_rotate    <= dma_regs(i)(8)(3 downto 0);
        --dma_control(i).ignore_first_word   <= dma_regs(i)(8)(4);
        --dma_control(i).alignment_rotate_we <= dma_regs_written(i)(8);

        dma_control(i).enable <= dma_enable(i);

      end if;
    end process dma_regs_p0;



  end generate gen_dma_registers;


  
end synth;
