library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

library unisim;
use unisim.vcomponents.all;

entity sysmon2_layer1 is
  port
  (
    -- External Interface
    ser_rx                    : in  std_logic;
    ser_tx                    : out std_logic;
    ser_ck                    : in  std_logic;
    -- Internal Interface
    clk                       : in  std_logic;
    rst                       : in  std_logic;
    tx_rst                    : in  std_logic;
    tx_data                   : in  std_logic_vector( 7 downto 0 );
    tx_soh                    : in  std_logic;
    tx_eot                    : in  std_logic;
    tx_err                    : out std_logic;
    tx_wr                     : in  std_logic;
    tx_full                   : out std_logic;
    tx_empty                  : out std_logic;
    rx_rst                    : in  std_logic;
    rx_data                   : out std_logic_vector( 7 downto 0 );
    rx_soh                    : out std_logic;
    rx_eot                    : out std_logic;
    rx_rd                     : in  std_logic;
    rx_err                    : out std_logic;
    rx_full                   : out std_logic;
    rx_empty                  : out std_logic
  );
end entity;

architecture rtl of sysmon2_layer1 is
 
  type tx_encode_state is (
    fifo_read,
    fifo_send_dly,
    fifo_send_dly_encode,
    fifo_send_encode,
    fifo_send_eot
  );
  
  type rx_decode_state is (
    fifo_read_soh,
    fifo_read_eot,
    fifo_read_decode
  );
  
  function needs_encode( raw : std_logic_vector( 7 downto 0 ) ) return boolean is
  begin
    if ( raw = X"01" ) or ( raw = X"04" ) or ( raw = X"10" ) then
      return true;
    else
      return false;
    end if;
  end function;
  
  signal tx_state       : tx_encode_state := fifo_read;
  signal tx_err_stat    : std_logic := '0';
  signal tx_inframe     : std_logic := '0';
  signal tx_data_ext    : std_logic_vector( 31 downto 0 );
  signal tx_por_vec     : std_logic_vector( 2 downto 0 ) := ( others => '1' );
  signal rx_por_vec     : std_logic_vector( 2 downto 0 ) := ( others => '1' );
  signal txu_data_ext   : std_logic_vector( 31 downto 0 );
  
  signal n_tx_state     : tx_encode_state;
  signal n_uart_tx_w    : std_logic;
  signal n_uart_tx_d    : std_logic_vector( 7 downto 0 );
  signal n_txu_adv      : std_logic;
  signal n_tx_inframe   : std_logic;
  signal n_tx_err_stat  : std_logic;

  signal txu_data       : std_logic_vector( 7 downto 0 ) := ( others => '0' );
  signal txu_empty      : std_logic;
  signal txu_soh        : std_logic;
  signal txu_eot        : std_logic;
  signal txu_adv        : std_logic := '0';
  signal tx_por         : std_logic := '1';
  signal rx_por         : std_logic := '1';
  
  signal o_tx_full      : std_logic;
  
  signal uart_tx_d      : std_logic_vector( 7 downto 0 ) := ( others => '-' );
  signal uart_tx_w      : std_logic := '0';
  signal uart_tx_f      : std_logic;
  
  signal tx_rst_or_rst  : std_logic;
  signal rx_rst_or_rst  : std_logic;
  
  signal rx_state       : rx_decode_state := fifo_read_soh;
  
  signal reg_rx_data    : std_logic_vector( 7 downto 0 ) := ( others => '-' );
  signal reg_rx_data_z1 : std_logic_vector( 7 downto 0 ) := ( others => '-' );
  signal reg_rx_soh     : std_logic := '-';
  signal reg_rx_eot     : std_logic := '-';
  signal reg_rx_wr      : std_logic := '0';
  signal reg_rx_wr_dly  : std_logic := '0';  

  signal reg_rx_error   : std_logic := '0';  

  signal reg_rx_full    : std_logic;
  signal rx_soh_dly     : std_logic := '0';
  signal rx_eot_dly     : std_logic := '0';
  
  signal n_rx_state     : rx_decode_state;
  signal n_reg_rx_error : std_logic;  
  signal n_reg_rx_data  : std_logic_vector( 7 downto 0 );
  signal n_reg_rx_soh   : std_logic;
  signal n_reg_rx_eot   : std_logic;
  signal n_reg_rx_wr    : std_logic := '0';
  signal n_uart_rx_rd   : std_logic;
  signal n_rx_soh_dly   : std_logic;  
  signal n_reg_rx_wr_dly : std_logic;  
  
  signal uart_rx_data   : std_logic_vector( 7 downto 0 );
  signal uart_rx_rd     : std_logic := '0';
  signal uart_rx_full   : std_logic;
  signal uart_rx_empty  : std_logic;
  
  signal reg_rx_data_z1_ext : std_logic_vector( 31 downto 0 );
  signal rx_data_ext : std_logic_vector( 31 downto 0 );


  attribute mark_debug : string;
  attribute mark_debug of ser_rx: signal is "true";
  attribute mark_debug of ser_tx: signal is "true";
  attribute mark_debug of ser_ck: signal is "true";
  attribute mark_debug of uart_tx_d: signal is "true";
  attribute mark_debug of uart_tx_w : signal is "true";
  attribute mark_debug of uart_tx_f : signal is "true";
  attribute mark_debug of uart_rx_data: signal is "true";
  attribute mark_debug of uart_rx_rd : signal is "true";
  attribute mark_debug of uart_rx_full : signal is "true";
  attribute mark_debug of uart_rx_empty : signal is "true";

  
begin

  tx_err <= tx_err_stat;
  rx_err <= reg_rx_error;

  rx_full  <= reg_rx_full;
  
  tx_rst_or_rst <= tx_rst or rst or tx_por;
  rx_rst_or_rst <= rx_rst or rst or rx_por;
  
  p_tx_por : process( clk )
  begin
    if rising_edge( clk ) then
      if tx_rst = '1' or rst = '1' then
        tx_por <= '1';
        tx_por_vec <= ( others => '1' );
      else
        if tx_por_vec = 0 then
          tx_por <= '0';
        else
          tx_por_vec <= tx_por_vec - 1;
        end if;
      end if;
    end if;
  end process;
  
  p_rx_por : process( clk )
  begin
    if rising_edge( clk ) then
      if rx_rst = '1' or rst = '1' then
        rx_por <= '1';
        rx_por_vec <= ( others => '1' );
      else
        if rx_por_vec = 0 then
          rx_por <= '0';
        else
          rx_por_vec <= rx_por_vec - 1;
        end if;
      end if;
    end if;
  end process;  
  
  tx_data_ext( 7 downto 0 ) <= tx_data;
  tx_data_ext( tx_data_ext'left downto 8 ) <= ( 8 => tx_soh, 9 => tx_eot, others => '-' );
  
  tx_full <= o_tx_full or tx_rst_or_rst;
  tx_empty <= txu_empty and not tx_rst_or_rst;

  i_tx_buffer : entity work.dpram_sfifo_level
      generic map (
        width => 10)
      port map (
        clk    => clk,
        rst    => tx_rst_or_rst,
        wadv   => tx_wr,
        wdata  => tx_data_ext(9 downto 0),
        nfull  => o_tx_full,
        radv   => txu_adv,
        rempty => txu_empty,
        rdata  => txu_data_ext(9 downto 0),
        rlevel => open);
  
  --i_tx_buffer : unisim.vcomponents.fifo18e1 generic map (
  --  almost_empty_offset => X"0080",   -- Sets the almost empty threshold
  --  almost_full_offset => X"0080",    -- Sets almost full threshold
  --  data_width => 18,                 -- Sets data width to 4-36
  --  do_reg => 1,                      -- Enable output register (1-0) Must be 1 if EN_SYN = FALSE
  --  en_syn => FALSE ,                  -- Specifies FIFO as dual-clock (FALSE) or Synchronous (TRUE)
  --  fifo_mode => "FIFO18",            -- Sets mode to FIFO18 or FIFO18_36
  --  first_word_fall_through => TRUE,  -- Sets the FIFO FWFT to FALSE, TRUE
  --  init => x"000000000",             -- Initial values on output tx_port
  --  sim_device => "7SERIES",          -- Must be set to "7SERIES" for simulation behavior
  --  srval => x"000000000"             -- Set/Reset value for output tx_port
  --) port map (
  --  -- Read Data: 32-bit (each) output: Read output data
  --  do => txu_data_ext,         -- 32-bit output: Data output
  --  dop => open,                -- 4-bit output: Parity data output
  --  -- Status: 1-bit (each) output: Flags and other FIFO status outputs
  --  almostempty => open,        -- 1-bit output: Almost empty flag
  --  almostfull => o_tx_full,      -- 1-bit output: Almost full flag
  --  empty => txu_empty,         -- 1-bit output: Empty flag
  --  full => open,               -- 1-bit output: Full flag
  --  rdcount => open,            -- 12-bit output: Read count
  --  rderr => open,              -- 1-bit output: Read error
  --  wrcount => open,            -- 12-bit output: Write count
  --  wrerr => open,              -- 1-bit output: Write error
  --  -- Read Control Signals: 1-bit (each) input: Read clock, enable and reset input signals
  --  rdclk => clk,               -- 1-bit input: Read clock
  --  rden => txu_adv,            -- 1-bit input: Read enable
  --  regce => '1',               -- 1-bit input: Clock enable
  --  rst => tx_rst_or_rst,       -- 1-bit input: Asynchronous Reset
  --  rstreg => tx_rst_or_rst,    -- 1-bit input: Output register set/reset
  --  -- Write Control Signals: 1-bit (each) input: Write clock and enable input signals
  --  wrclk => clk,               -- 1-bit input: Write clock
  --  wren => tx_wr,              -- 1-bit input: Write enable
  --  -- Write Data: 32-bit (each) input: Write input data
  --  di => tx_data_ext,          -- 32-bit input: Data input
  --  dip => ( others => '0' )    -- 4-bit input: Parity input
  --);  
  
  txu_data <= txu_data_ext( 7 downto 0 );
  txu_soh <= txu_data_ext( 8 );
  txu_eot <= txu_data_ext( 9 );
  tx_err <= tx_err_stat;
 
  p_tx_state_sync : process( clk )
  begin
    if rising_edge( clk ) then
      if tx_rst_or_rst = '1' then
        uart_tx_w    <= '0';
        uart_tx_d    <= ( others => '-' );
        txu_adv      <= '0';
        tx_inframe   <= '0';
        tx_err_stat  <= '0';
        tx_state     <= fifo_read;
      else
        uart_tx_w    <= n_uart_tx_w;
        uart_tx_d    <= n_uart_tx_d;
        txu_adv      <= n_txu_adv;
        tx_inframe   <= n_tx_inframe;
        tx_err_stat  <= n_tx_err_stat;
        tx_state     <= n_tx_state;
      end if;
    end if;
  end process;
  
  p_tx_state_async : process( tx_state, uart_tx_f, txu_soh, txu_data, txu_empty, txu_eot, tx_inframe, tx_err_stat, txu_adv )
  begin
    case tx_state is
      when fifo_read => 
        if txu_empty = '0' and uart_tx_f = '0' and txu_adv = '0' then
          if tx_inframe = '0' then
            if txu_soh = '1' then
              -- Send SOH
              n_uart_tx_w <= '1';
              n_uart_tx_d <= X"01";
              n_txu_adv <= '0';
              n_tx_inframe <= not txu_eot;
              n_tx_err_stat <= tx_err_stat;
              if needs_encode( txu_data ) then
                n_tx_state <= fifo_send_encode;
              else
                n_tx_state <= fifo_send_dly;
              end if;
            else
              -- Dump expected data
              n_uart_tx_w <= '0';
              n_uart_tx_d <= ( others => '-' );
              n_txu_adv <= '1';
              n_tx_inframe <= '0';
              n_tx_err_stat <= '1';
              n_tx_state <= fifo_read;
            end if;
          else -- In frame
            n_uart_tx_w <= '1';
            n_tx_inframe <= not txu_eot;
            n_tx_err_stat <= tx_err_stat or txu_soh;
            if needs_encode( txu_data ) then
              n_txu_adv <= '0';
              n_uart_tx_d <= X"10";
              n_tx_state <= fifo_send_dly_encode;
            else
              n_txu_adv <= '1';
              n_uart_tx_d <= txu_data;
              if txu_eot = '1' then
                n_tx_state <= fifo_send_eot;
              else
                n_tx_state <= fifo_read;
              end if;
            end if;
          end if;
        else -- No data of FIFO full
          n_uart_tx_w <= '0';
          n_uart_tx_d <= ( others => '-' );
          n_txu_adv <= '0';
          n_tx_inframe <= tx_inframe;
          n_tx_state <= fifo_read;
          n_tx_err_stat <= tx_err_stat;
        end if;
      when fifo_send_dly =>
        -- Send the next word
        n_uart_tx_w <= '1';
        n_uart_tx_d <= txu_data;
        n_txu_adv <= '1';
        n_tx_inframe <= tx_inframe;
        n_tx_err_stat <= tx_err_stat;
        if txu_eot = '1' then
          n_tx_state <= fifo_send_eot;
        else
          n_tx_state <= fifo_read;
        end if;
      when fifo_send_dly_encode =>
        -- Invert data for escaped charactors
        n_uart_tx_w <= '1';
        n_uart_tx_d <= not txu_data;
        n_txu_adv <= '1';
        n_tx_inframe <= tx_inframe;      
        n_tx_err_stat <= tx_err_stat;
        if txu_eot = '1' then
          n_tx_state <= fifo_send_eot;
        else
          n_tx_state <= fifo_read;
        end if;
      when fifo_send_encode =>
        -- Invert data for escaped charactors
        n_uart_tx_w <= '1';
        n_uart_tx_d <= X"10";
        n_txu_adv <= '0';
        n_tx_inframe <= tx_inframe;      
        n_tx_err_stat <= tx_err_stat;
        n_tx_state <= fifo_send_dly_encode;
      when fifo_send_eot =>
        -- Send EOT
        n_uart_tx_w <= '1';
        n_uart_tx_d <= X"04";
        n_txu_adv <= '0';
        n_tx_inframe <= '0';
        n_tx_state <= fifo_read;
        n_tx_err_stat <= tx_err_stat;
      when others =>
        -- Safty, should not be reached, causes "reset", but sets tx_error bit
        n_uart_tx_w <= '0';
        n_uart_tx_d <= ( others => '-' );
        n_txu_adv <= '0';
        n_tx_inframe <= '0';      
        n_tx_state <= fifo_read;
        n_tx_err_stat <= '1';
    end case;
  end process;


 
  i_uart_sync : entity work.uart_sync generic map (
    data_size => 8,
    msb_first => false
  ) port map (
    -- External Interface
    ser_rx   => ser_rx,
    ser_tx   => ser_tx,
    ser_ck   => ser_ck,
    -- Internal Interface
    clk      => clk,
    tx_rst   => tx_rst_or_rst,
    tx_data  => uart_tx_d,
    tx_wr    => uart_tx_w,
    tx_full  => uart_tx_f,
    tx_empty => open,
    rx_rst   => rx_rst_or_rst,
    rx_data  => uart_rx_data,
    rx_rd    => uart_rx_rd,
    rx_full  => uart_rx_full,
    rx_empty => uart_rx_empty
  );
  
  p_rx_state_sync : process( clk )
  begin
    if rising_edge( clk ) then
      if rx_rst_or_rst = '1' then
        rx_soh_dly    <= '0';
        reg_rx_data   <= ( others => '-' );
        reg_rx_data_z1<= ( others => '-' );
        reg_rx_soh    <= '0';
        reg_rx_eot    <= '0';
        reg_rx_wr     <= '0';
        reg_rx_wr_dly <= '0';
        uart_rx_rd    <= '0';
        rx_state      <= fifo_read_soh;
        reg_rx_error  <= '0';
      else
        rx_soh_dly    <= n_rx_soh_dly;
        reg_rx_data   <= n_reg_rx_data;
        reg_rx_data_z1<= reg_rx_data;
        reg_rx_soh    <= n_reg_rx_soh; 
        reg_rx_eot    <= n_reg_rx_eot; 
        reg_rx_wr     <= n_reg_rx_wr;  
        reg_rx_wr_dly <= n_reg_rx_wr_dly;
        uart_rx_rd    <= n_uart_rx_rd;
        rx_state      <= n_rx_state;
        reg_rx_error  <= n_reg_rx_error;
      end if;
    end if;
  end process;
  
  p_rx_state_async : process( rx_state, uart_rx_data, uart_rx_empty, reg_rx_wr_dly, reg_rx_full, reg_rx_data, reg_rx_error, uart_rx_rd, rx_soh_dly )
  begin
    case rx_state is
      when fifo_read_soh => 
        -- Ignore data till soh is seen
        if uart_rx_empty = '0' and reg_rx_full = '0' and uart_rx_rd = '0' then
          if uart_rx_data = X"01" then
            n_rx_state <= fifo_read_eot;
            n_rx_soh_dly <= '1';
          else
            n_rx_state <= fifo_read_soh;
            n_rx_soh_dly <= '0';
          end if;
          n_uart_rx_rd <= '1';
        else
          n_rx_state <= fifo_read_soh;
          n_rx_soh_dly <= '0';
          n_uart_rx_rd <= '0';
        end if;
        n_reg_rx_wr_dly <= '0';
        n_reg_rx_soh <= '0';
        n_reg_rx_eot <= '0';
        n_reg_rx_data <= ( others => '-' );
        n_reg_rx_wr <= '0';
        n_reg_rx_error <= reg_rx_error;
      when fifo_read_eot => 
        -- Read charactors or eot 
        if uart_rx_empty = '0' and reg_rx_full = '0' and uart_rx_rd = '0' then
          if uart_rx_data = X"04" then
            -- End of Transfer
            n_rx_state <= fifo_read_soh;
            n_reg_rx_wr_dly <= '0';
            n_reg_rx_soh <= rx_soh_dly and reg_rx_wr_dly;
            n_reg_rx_eot <= '1';
            n_rx_soh_dly <= '0';
            n_reg_rx_data <= reg_rx_data;
            n_reg_rx_wr <= reg_rx_wr_dly;
            n_reg_rx_error <= reg_rx_error;
          elsif uart_rx_data = X"10" then
            -- Next char is encoded
            n_rx_state <= fifo_read_decode;
            n_reg_rx_wr_dly <= '0';
            n_reg_rx_soh <= rx_soh_dly and reg_rx_wr_dly;
            n_reg_rx_eot <= '0';
            n_rx_soh_dly <= rx_soh_dly and not reg_rx_wr_dly;
            n_reg_rx_data <= reg_rx_data;
            n_reg_rx_wr <= reg_rx_wr_dly;
            n_reg_rx_error <= reg_rx_error;
          elsif uart_rx_data = X"01" then
            -- Start! (error)
            n_rx_state <= fifo_read_eot;
            n_reg_rx_soh <= rx_soh_dly and reg_rx_wr_dly;
            n_reg_rx_eot <= '0';
            n_reg_rx_data <= uart_rx_data;
            n_reg_rx_wr_dly <= '1';
            n_rx_soh_dly <= rx_soh_dly and not reg_rx_wr_dly;
            n_reg_rx_wr <= reg_rx_wr_dly;
            n_reg_rx_error <= '1';
          else
            -- Next char is normal
            n_rx_state <= fifo_read_eot;
            n_reg_rx_soh <= rx_soh_dly and reg_rx_wr_dly;
            n_reg_rx_eot <= '0';
            n_reg_rx_data <= uart_rx_data;
            n_reg_rx_wr_dly <= '1';
            n_rx_soh_dly <= rx_soh_dly and not reg_rx_wr_dly;
            n_reg_rx_wr <= reg_rx_wr_dly;
            n_reg_rx_error <= reg_rx_error;
          end if;
          n_uart_rx_rd <= '1';
        else
          n_rx_state <= fifo_read_eot;
          n_uart_rx_rd <= '0';
          n_reg_rx_data <= reg_rx_data;
          n_reg_rx_soh <= '0';
          n_reg_rx_wr_dly <= reg_rx_wr_dly;
          n_reg_rx_eot <= '0';
          n_rx_soh_dly <= rx_soh_dly;
          n_reg_rx_wr <= '0';
          n_reg_rx_error <= reg_rx_error;
        end if;
      when fifo_read_decode => 
        n_reg_rx_eot <= '0';
        if uart_rx_empty = '0' and reg_rx_full = '0' and uart_rx_rd = '0' then
          n_rx_state <= fifo_read_eot;
          n_uart_rx_rd <= '1';
          n_reg_rx_data <= not uart_rx_data;
          n_reg_rx_soh <= rx_soh_dly and reg_rx_wr_dly;
          n_reg_rx_wr_dly <= '1';
          n_reg_rx_eot <= '0';
          n_rx_soh_dly <= rx_soh_dly and not reg_rx_wr_dly;
          n_reg_rx_wr <= reg_rx_wr_dly;
          if ( uart_rx_data = X"01" ) or ( uart_rx_data = X"04" ) or ( uart_rx_data = X"10" ) then
            n_reg_rx_error <= '1';
          else
            n_reg_rx_error <= reg_rx_error;
          end if;
        else
          n_rx_state <= fifo_read_decode;
          n_uart_rx_rd <= '0';
          n_reg_rx_data <= reg_rx_data;
          n_reg_rx_soh <= '0';
          n_reg_rx_wr_dly <= '0';
          n_reg_rx_eot <= '0';
          n_rx_soh_dly <= rx_soh_dly;
          n_reg_rx_wr <= '0';
          n_reg_rx_error <= reg_rx_error;
        end if;
      when others =>
        n_rx_soh_dly    <= '0';
        n_reg_rx_data   <= ( others => '-' );
        n_reg_rx_soh    <= '0';
        n_reg_rx_eot    <= '0';
        n_reg_rx_wr     <= '0';
        n_reg_rx_wr_dly <= '0';
        n_uart_rx_rd    <= '0';
        n_rx_state      <= fifo_read_soh;
        n_reg_rx_error  <= '1';
    end case;
  end process;
  
  --i_rx_buffer : entity work.small_sfifo32 generic map (
  --  width => 10 
  --) port map (
  --  clk                 => clk,
  --  rst                 => rx_rst,
  --  wadv                => reg_rx_wr,
  --  wdata( 7 downto 0 ) => reg_rx_data_z1,
  --  wdata( 8 )          => reg_rx_soh,
  --  wdata( 9 )          => reg_rx_eot,
  --  nfull               => reg_rx_full,
  --  radv                => rx_rd,
  --  rempty              => rx_empty,
  --  rdata( 7 downto 0 ) => rx_data,
  --  rdata( 8 )          => rx_soh,
  --  rdata( 9 )          => rx_eot
  --);  
  
  reg_rx_data_z1_ext( 7 downto 0 ) <= reg_rx_data_z1;
  reg_rx_data_z1_ext( reg_rx_data_z1_ext'left downto 8 ) <= (
    8 => reg_rx_soh,
    9 => reg_rx_eot,
    others => '-'
  );
  
  --i_rx_buffer : unisim.vcomponents.fifo18e1 generic map (
  --  almost_empty_offset => X"0080",   -- Sets the almost empty threshold
  --  almost_full_offset => X"0080",    -- Sets almost full threshold
  --  data_width => 18,                 -- Sets data width to 4-36
  --  do_reg => 1,                      -- Enable output register (1-0) Must be 1 if EN_SYN = FALSE
  --  en_syn => FALSE ,                  -- Specifies FIFO as dual-clock (FALSE) or Synchronous (TRUE)
  --  fifo_mode => "FIFO18",            -- Sets mode to FIFO18 or FIFO18_36
  --  first_word_fall_through => TRUE,  -- Sets the FIFO FWFT to FALSE, TRUE
  --  init => x"000000000",             -- Initial values on output tx_port
  --  sim_device => "7SERIES",          -- Must be set to "7SERIES" for simulation behavior
  --  srval => x"000000000"             -- Set/Reset value for output tx_port
  --) port map (
  --  -- Read Data: 32-bit (each) output: Read output data
  --  do => rx_data_ext,         -- 32-bit output: Data output
  --  dop => open,                -- 4-bit output: Parity data output
  --  -- Status: 1-bit (each) output: Flags and other FIFO status outputs
  --  almostempty => open,        -- 1-bit output: Almost empty flag
  --  almostfull => reg_rx_full,      -- 1-bit output: Almost full flag
  --  empty => rx_empty,         -- 1-bit output: Empty flag
  --  full => open,               -- 1-bit output: Full flag
  --  rdcount => open,            -- 12-bit output: Read count
  --  rderr => open,              -- 1-bit output: Read error
  --  wrcount => open,            -- 12-bit output: Write count
  --  wrerr => open,              -- 1-bit output: Write error
  --  -- Read Control Signals: 1-bit (each) input: Read clock, enable and reset input signals
  --  rdclk => clk,               -- 1-bit input: Read clock
  --  rden => rx_rd,            -- 1-bit input: Read enable
  --  regce => '1',               -- 1-bit input: Clock enable
  --  rst => rx_rst_or_rst,       -- 1-bit input: Asynchronous Reset
  --  rstreg => rx_rst_or_rst,    -- 1-bit input: Output register set/reset
  --  -- Write Control Signals: 1-bit (each) input: Write clock and enable input signals
  --  wrclk => clk,               -- 1-bit input: Write clock
  --  wren => reg_rx_wr,              -- 1-bit input: Write enable
  --  -- Write Data: 32-bit (each) input: Write input data
  --  di => reg_rx_data_z1_ext,          -- 32-bit input: Data input
  --  dip => ( others => '0' )    -- 4-bit input: Parity input
  --);    

  i_rx_buffer : entity work.small_sfifo32
    generic map (
      width => 10)
    port map (
        clk    => clk,
        rst    => rx_rst_or_rst,
        wadv   => reg_rx_wr,
        wdata  => reg_rx_data_z1_ext(9 downto 0),
        nfull  => reg_rx_full,
        radv   => rx_rd,
        rempty => rx_empty,
        rdata  => rx_data_ext(9 downto 0));

  
  rx_data <= rx_data_ext( 7 downto 0 );
  rx_soh <= rx_data_ext( 8 );
  rx_eot <= rx_data_ext( 9 );

end architecture;
