--
-- adb3_ocp_reorder_dbuf_mf.vhd
--
-- Multi-slot buffering for adb3_ocp_reorder component.
--
-- This Implementation uses multiple FIFOs each large enough for 2**slot_order
-- OCP data words. Each slot has one FIFO, and at a given moment a particular
-- slot's FIFO is chosen (by read_slot_1hot) to output its data. As long as
-- read_slot_1hot is driven correctly, this results in the required data order.
-- That is, read responses are returned in same order as the corresponding read
-- commands were issued on the in-order side.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;

entity adb3_ocp_reorder_dbuf_mf is
  generic(
    data_width      : in natural;   -- OCP data width, in bits
    num_slot        : in natural;   -- No. of slots
    slot_order      : in natural;   -- Log2 of no. of OCP data words that each slot must be able to hold
    output_pipe     : in boolean);  -- true => extra pipeline stage to mitigate multiplexor complexity
  port(
    rst             : in  std_logic;
    sr              : in  std_logic;
    clk             : in  std_logic;
    -- This is a subset of unordered side OCP signals
    uo_resp         : in  ocp_RespT;
    uo_data         : in  std_logic_vector(data_width - 1 downto 0);
    uo_tag          : in  std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
    uo_resp_accept  : out std_logic;
    -- Read side
    read_slot_1hot  : in  std_logic_vector(num_slot - 1 downto 0);
    read_slot_valid : in  std_logic;
    read            : in  std_logic;
    read_data       : out std_logic_vector(data_width - 1 downto 0);
    read_valid      : out std_logic;
    -- Slot management
    alloc           : in  std_logic;
    alloc_tag       : in  std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
    alloc_length    : in  std_logic_vector(ADB3_OCP_BURST_WIDTH - 1 downto 0);  -- Ignored
    alloc_slot_1hot : in  std_logic_vector(num_slot - 1 downto 0);
    free            : in  std_logic;
    free_slot_1hot  : in  std_logic_vector(num_slot - 1 downto 0);
    error_pulse     : out std_logic_vector(7 downto 0));
end adb3_ocp_reorder_dbuf_mf;

architecture rtl of adb3_ocp_reorder_dbuf_mf is

  -- Return a binary representation of a std_logic_vector value.
  function conv_string_binary(
    constant v : std_logic_vector)
  return string is
    variable tmp : std_logic_vector(1 to v'length);
    variable str : string(1 to v'length);
  begin
    tmp := v;
    for i in str'range loop
      str(i) := std_logic'image(tmp(i))(1);
    end loop;
    return str;
  end function;
  
  -- Return a hexadecimal string representation of a std_logic_vector value.
  function conv_string_hex(
    constant v : std_logic_vector)
  return string is
    variable n : natural := ((v'length + 3) / 4) * 4;
    variable ndigit : natural := n / 4;
    variable tmp : std_logic_vector(n - 1 downto 0);
    variable digit : std_logic_vector(3 downto 0);
    variable str : string(1 to ndigit);
  begin
    tmp(v'length - 1 downto 0) := v;
    if n > v'length then
      tmp(n - 1 downto v'length) := (others => '0');
    end if;
    for i in str'range loop
      digit := tmp(4 * i - 1 downto 4 * (i - 1));
      case digit is
        when "0000" =>
          str(ndigit - i + 1) := '0';

        when "0001" =>
          str(ndigit - i + 1) := '1';

        when "0010" =>
          str(ndigit - i + 1) := '2';

        when "0011" =>
          str(ndigit - i + 1) := '3';

        when "0100" =>
          str(ndigit - i + 1) := '4';

        when "0101" =>
          str(ndigit - i + 1) := '5';

        when "0110" =>
          str(ndigit - i + 1) := '6';

        when "0111" =>
          str(ndigit - i + 1) := '7';

        when "1000" =>
          str(ndigit - i + 1) := '8';

        when "1001" =>
          str(ndigit - i + 1) := '9';

        when "1010" =>
          str(ndigit - i + 1) := 'A';

        when "1011" =>
          str(ndigit - i + 1) := 'B';

        when "1100" =>
          str(ndigit - i + 1) := 'C';

        when "1101" =>
          str(ndigit - i + 1) := 'D';

        when "1110" =>
          str(ndigit - i + 1) := 'E';

        when "1111" =>
          str(ndigit - i + 1) := 'F';

        when others =>
          str(ndigit - i + 1) := 'X';
      end case;
    end loop;
    return str;
  end function;

  component adb3_ocp_reorder_sfifo
    generic(
      ORDER   : in natural range 2 to 9;
      WIDTH   : in natural;
      FAST_SR : in boolean := false);
    port(
      rst    : in  std_logic;
      clk    : in  std_logic;
      sr     : in  std_logic;
      w      : in  std_logic;
      wd     : in  std_logic_vector(WIDTH - 1 downto 0);
      wfull  : out std_logic;
      r      : in  std_logic;
      rvalid : out std_logic;
      rq     : out std_logic_vector(WIDTH - 1 downto 0));
  end component;

  constant slot_size : natural := 2**slot_order;

  -- -------------------------------------------------------------------------
  -- Slot management (SM) signals.
  -- -------------------------------------------------------------------------

  type read_command_slot_t is record
    -- Tag value for command
    tag : std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
  end record;
  type read_command_slot_array_t is array(natural range <>) of read_command_slot_t;
  constant UNOCCUPIED_SLOT : read_command_slot_t := (
    tag => (others => '-'));

  signal sm_slots          : read_command_slot_array_t(0 to num_slot - 1) := (others => UNOCCUPIED_SLOT);
  signal sm_slots_occupied : std_logic_vector(num_slot - 1 downto 0)      := (others => '0');

  -- -------------------------------------------------------------------------
  -- Matching incoming tag to a slot
  -- -------------------------------------------------------------------------

  signal tag_match       : std_logic;
  signal tag_match_valid : std_logic := '0';
  signal tag_q           : std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);

  -- -------------------------------------------------------------------------
  -- Per-slot data FIFO signals
  -- -------------------------------------------------------------------------

  type   fifo_data_array_t is array(natural range <>) of std_logic_vector(data_width - 1 downto 0);
  signal fifo_w      : std_logic_vector(num_slot - 1 downto 0);
  signal fifo_wd     : std_logic_vector(data_width - 1 downto 0);
  signal fifo_wfull  : std_logic_vector(num_slot - 1 downto 0);
  signal fifo_r      : std_logic_vector(num_slot - 1 downto 0);
  signal fifo_rvalid : std_logic_vector(num_slot - 1 downto 0);
  signal fifo_rq     : fifo_data_array_t(num_slot - 1 downto 0);

  -- -------------------------------------------------------------------------
  -- Multiplexed data FIFO signals
  -- -------------------------------------------------------------------------
  signal sel_fifo_rvalid : std_logic;
  signal sel_fifo_rq : std_logic_vector(data_width - 1 downto 0);
  
  -- -------------------------------------------------------------------------
  -- Optional read data pipelining (used when output_pipe is true)
  -- -------------------------------------------------------------------------
  signal read_pipe_en : std_logic;
  signal read_pipe_data : std_logic_vector(data_width - 1 downto 0);
  signal read_pipe_valid : std_logic;
  
  -- -------------------------------------------------------------------------
  -- Debug/status
  -- -------------------------------------------------------------------------

  signal i_error_pulse : std_logic_vector(7 downto 0) := (others => '0');
  
begin

  error_pulse <= i_error_pulse;

  uo_resp_accept <= '1';  -- OK to assume that in-order side master won't request more data than we can buffer?

  implement_slots : process(rst, clk)
    variable case_tmp : std_logic_vector(1 downto 0);
  begin
    if rst = '1' then
      sm_slots          <= (others => UNOCCUPIED_SLOT);
      sm_slots_occupied <= (others => '0');
    elsif rising_edge(clk) then
      for i in 0 to num_slot - 1 loop
        if sr = '1' then
          sm_slots(i)          <= UNOCCUPIED_SLOT;
          sm_slots_occupied(i) <= '0';
        else
          case_tmp(0) := alloc and alloc_slot_1hot(i);
          case_tmp(1) := free and free_slot_1hot(i);
          case case_tmp is
            when "01" =>
              -- synthesis translate_off
              assert sm_slots_occupied(i) = '0'
                report "adb3_ocp_reorder_dbuf_mf: Attempt to allocate already occupied slot, alloc_slot_1h=" & conv_string_binary(alloc_slot_1hot)
                severity error;
              -- synthesis translate_on
              sm_slots_occupied(i) <= '1';
              sm_slots(i).tag      <= alloc_tag;
              
            when "10" =>
              -- synthesis translate_off
              assert sm_slots_occupied(i) = '1'
                report "adb3_ocp_reorder_dbuf_mf: Attempt to free unoccupied slot, alloc_slot_1h=" & conv_string_binary(free_slot_1hot)
                severity error;
              --synthesis translate_on
              sm_slots_occupied(i) <= '0';
              
            when "00" =>
              null;

            when "11" =>
              -- Illegal, should never happen
              -- synthesis translate_off
              assert false report "adb3_ocp_reorder_dbuf_mf: Slot " & natural'image(i) & ": attempt to occupy and free at the same time" severity error;
              -- synthesis translate_on
              sm_slots_occupied(i) <= '-';
              sm_slots(i).tag      <= (others => '-');
              
            when others =>
              -- Illegal, should never happen
              sm_slots_occupied(i) <= '-';
              sm_slots(i).tag      <= (others => '-');
          end case;
        end if;
      end loop;
    end if;
  end process;


  -- -------------------------------------------------------------------------
  -- Match the tag in the response from the unordered side to a particular
  -- slot and generate 1-hot slot FIFO write signal from it.
  -- -------------------------------------------------------------------------

  match_tag_to_slot : process(clk,rst)
    variable matched : std_logic;
  begin
    if rst = '1' then
      tag_match       <= '-';
      tag_match_valid <= '0';
      tag_q           <= (others => '-');
    elsif rising_edge(clk) then
      if sr = '1' then
        tag_match       <= '-';
        tag_match_valid <= '0';
        tag_q           <= (others => '-');
      else
        if uo_resp = OCP_RESP_VALID then
          matched := '0';
          for i in 0 to num_slot - 1 loop
            if sm_slots_occupied(i) = '1' and sm_slots(i).tag = uo_tag then
              matched := matched or '1';
            end if;
          end loop;
          tag_match       <= matched;
          tag_match_valid <= '1';
          tag_q           <= uo_tag;
        else
          tag_match       <= '-';
          tag_match_valid <= '0';
          tag_q           <= (others => '-');
        end if;
      end if;
    end if;
  end process;

  process (clk) is
  begin  -- process
    if rising_edge(clk) then
      for i in 0 to num_slot - 1 loop
        if uo_resp = OCP_RESP_VALID and sm_slots_occupied(i) = '1' and sm_slots(i).tag = uo_tag then
          fifo_w(i) <= '1';
        else
          fifo_w(i) <= '0';
        end if;
      end loop;
      fifo_wd <= uo_data;


    end if;
  end process;


  -- -------------------------------------------------------------------------
  -- Slot data FIFOs
  -- -------------------------------------------------------------------------

  -- synthesis translate_off
  -- Some sanity checks
  read_sanity_check : process(clk)
  begin
    if rising_edge(clk) then
      if read = '1' then
        assert read_slot_valid = '1'
          report "adb3_ocp_reorder_dbuf_mf: 'read' asserted without 'read_slot_valid'"
          severity error;
        if read_slot_valid = '1' then
          assert NOT is_x(read_slot_1hot)
            report "adb3_ocp_reorder_dbuf_mf: 'read_slot_valid' asserted with 'read_slot_1hot' invalid"
            severity error;
        end if;
      end if;
    end if;
  end process;
  -- synthesis translate_on

  sanity_check : process(clk)
  begin
    if rising_edge(clk) then
      -- Sanity check
      if tag_match_valid = '1' and tag_match = '0' then
        -- synthesis translate_off
        assert false
          report "adb3_ocp_reorder_dbuf_ur: Unordered side response with tag 0x" & conv_string_hex(tag_q) & " not matched to any slot"
          severity error;
        -- synthesis translate_on
        i_error_pulse(0) <= '1';
      else
        i_error_pulse(1) <= '0';
      end if;

      for i in 0 to num_slot-1 loop
        if fifo_w(i) = '1' and fifo_wfull(i) = '1' then
          i_error_pulse(1) <= '1';
          -- synthesis translate_off
          assert false
            report "adb3_ocp_reorder_dbuf_mf: Slot " & natural'image(i) & " FIFO overflow detected"
            severity error;
          -- synthesis translate_on
        else
          i_error_pulse(1) <= '0';
        end if;

        if fifo_r(i) = '1' and fifo_rvalid(i) = '0' then
          i_error_pulse(2) <= '1';
          -- synthesis translate_off
          assert false
            report "adb3_ocp_reorder_dbuf_mf: Slot " & natural'image(i) & " FIFO underflow detected"
            severity error;
          -- synthesis translate_on
        else
          i_error_pulse(2) <= '0';
        end if;
      end loop;  -- i
    end if;
  end process;

  gen_slot_fifos : for i in 0 to num_slot - 1 generate
    
    fifo_r(i) <= fifo_rvalid(i) AND (read OR NOT read_pipe_valid) AND read_slot_1hot(i) when output_pipe else
                 read AND read_slot_1hot(i);

    U0 : adb3_ocp_reorder_sfifo
      generic map(
        width => data_width,
        order => slot_order)
      port map(
        rst    => rst,
        sr     => sr,
        clk    => clk,
        wd     => fifo_wd,
        w      => fifo_w(i),
        wfull  => fifo_wfull(i),
        r      => fifo_r(i),
        rq     => fifo_rq(i),
        rvalid => fifo_rvalid(i));

  end generate;


  -- -------------------------------------------------------------------------
  -- Multiplex FIFO outputs to get selected slot FIFO signals
  -- -------------------------------------------------------------------------
  
  multiplex_slot_fifos : process(
    read_slot_1hot,
    fifo_rq,
    fifo_rvalid,
    read_slot_valid)
    variable tmp_rq     : std_logic_vector(data_width - 1 downto 0);
    variable tmp_rvalid : std_logic;
  begin
    tmp_rq     := (others => '0');
    tmp_rvalid := '0';
    for i in 0 to num_slot - 1 loop
      if read_slot_valid = '1' and read_slot_1hot(i) = '1' then
        tmp_rq     := tmp_rq or fifo_rq(i);
        tmp_rvalid := tmp_rvalid or fifo_rvalid(i);
      end if;
    end loop;
    sel_fifo_rq  <= tmp_rq;
    sel_fifo_rvalid <= tmp_rvalid;
  end process;

  -- -------------------------------------------------------------------------
  -- Multiplex FIFO outputs to get selected slot FIFO signals
  -- -------------------------------------------------------------------------
  
  gen_output_pipe : if output_pipe generate
  
    read_pipe_en <= read OR NOT read_pipe_valid;
    
    read_pipe : process(rst, clk)
    begin
      if rst = '1' then
        read_pipe_valid <= '0';
      elsif rising_edge(clk) then
        if sr = '1' then
          read_pipe_valid <= '0';
        else
          if read_pipe_en = '1' then
            read_pipe_valid <= sel_fifo_rvalid;
          end if;
        end if;
      end if;
      
      if rising_edge(clk) then
        if read_pipe_en = '1' then
          read_pipe_data <= sel_fifo_rq;
        end if;
      end if;
    end process;
    
    read_data <= read_pipe_data;
    read_valid <= read_pipe_valid;
  end generate;
  
  gen_no_output_pipe : if NOT output_pipe generate
  
    read_data <= sel_fifo_rq;
    read_valid <= sel_fifo_rvalid;
    
  end generate;
  
end rtl;
