--==============================================================================
--
--     File Name: adb3_ocp_dm_adapter_tmap.vhd
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: adb3_ocp_dm_adapter
--
--   Description: Submodule of adb3_ocp_dm_adapter responsible for mapping
--                tags in responses from Direct Master OCP channel to
--                the original tags of the OCP master.
--
--   Limitations: 
--
--         Notes: 
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2013 Alpha Data, All rights reserved
--
--     Revisions: 0.1, TW, 26 Nov 2013: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;

entity adb3_ocp_dm_adapter_tmap is
generic(
  -- Number of bits used for generating tags for chopped commands
  DM_TAG_BITS     : natural range 1 to ADB3_OCP_TAG_WIDTH := 6;
  -- OCP data width
  DATA_WIDTH      : in    natural);
port(
  ocp_rst         : in    std_logic;
  ocp_clk         : in    std_logic;
  -- Commands in from Chopper
  chop_cmd        : in    ocp_CmdT;
  chop_addr       : in    adb3_ocp_addr_s;
  chop_dm_tag     : in    adb3_ocp_tag_s;
  chop_ocp_tag    : in    adb3_ocp_tag_s;
  chop_blen       : in    adb3_ocp_burst_s;
  chop_cmd_valid  : in    std_logic;
  -- Response from Direct Master channel
  dm_resp_valid   : in    std_logic;
  dm_resp         : in    ocp_RespT;
  dm_data         : in    std_logic_vector(DATA_WIDTH - 1 downto 0);
  dm_tag          : in    adb3_ocp_tag_s;
  -- Tag-mapped response to response FIFO
  tm_resp_valid   : out   std_logic;
  tm_resp         : out   ocp_RespT;
  tm_data         : out   std_logic_vector(DATA_WIDTH - 1 downto 0);
  tm_tag          : out   adb3_ocp_tag_s);
end;


architecture syn of adb3_ocp_dm_adapter_tmap is

  constant RAM_DEPTH : natural := 2**DM_TAG_BITS;
  subtype ram_data_t is adb3_ocp_tag_s;
  type ram_array_t is array(0 to RAM_DEPTH - 1) of ram_data_t;
  signal ram : ram_array_t;
  attribute RAM_STYLE : string;
  attribute RAM_STYLE of ram : signal is "distributed";

begin
  
  -- Store the original OCP tag in the position in the mapping array indicated
  -- by the Direct Master tag.
  make_mapping : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      if chop_cmd_valid = '1' AND chop_cmd = OCP_CMD_READ then
        ram(to_integer(unsigned(chop_dm_tag(DM_TAG_BITS - 1 downto 0)))) <= chop_ocp_tag;
      end if;
    end if;
  end process;
  
  
  -- Map tag in Direct Master response to its original OCP tag.
  -- Register the mapped tag, for ease of timing closure, and pipeline related
  -- signals.
  gen_tm_tag : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      if dm_resp_valid = '1' then
        tm_tag <= ram(to_integer(unsigned(dm_tag(DM_TAG_BITS - 1 downto 0))));
      else
        tm_tag <= (others => '-');
      end if;
      tm_resp_valid <= dm_resp_valid;
      tm_resp <= dm_resp;
      tm_data <= dm_data;
    end if;
  end process;
  
end;
