--==============================================================================
--
--     File Name: adb3_ocp_dm_int_pkg.vhd
--         Model: Synthesisable
--
--  Dependencies: 
--
--       Company: Alpha Data
--        Design: Package adb3_ocp_dm_int
--
--   Description: Package of definitions used internally in the
--                adb3_ocp_dm_adapter module. These definitions must be
--                considered private to adb3_ocp_dm_adapter, and are subject
--                to change without notice.
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2013 Alpha Data, All rights reserved
--
--     Revisions: 0.1, TW, 16 Dec 2013: Initial version
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;

package adb3_ocp_dm_int is

  -- Return a hexadecimal string representation of a std_logic_vector value.
  function conv_string_hex(
    constant v : std_logic_vector)
  return string;

  -- synthesis translate_off
  -- Function for sanity-checking center classification code.
  -- This function should not be considered synthesizable.
  function is_center_classification(
    constant be : in std_logic_vector)
  return std_logic;
  -- synthesis translate_on

  -- A synchronous FIFO, used by various submodules of adb3_ocp_dm_adapter.
  component adb3_ocp_dm_adapter_sfifo
  generic(
    ORDER       : in    natural range 4 to 10;
    WIDTH       : in    natural;
    EMPTY_VAL   : in    std_logic := '0';
    EMPTY_LO    : in    natural := 0;
    EMPTY_HI    : in    natural := 0);
  port(
    rst         : in    std_logic;  -- May be wired to 0 if not used
    sr          : in    std_logic;  -- May be wired to 0 if not used
    clk         : in    std_logic;
    -- Write side
    wen         : in    std_logic;
    wd          : in    std_logic_vector(WIDTH - 1 downto 0);
    wafull      : out   std_logic;
    wnafull     : out   std_logic;
    -- Read side
    radv        : in    std_logic;
    rvalid      : out   std_logic;
    rq          : out   std_logic_vector(WIDTH - 1 downto 0);
    -- Debug
    debug_info  : out   std_logic_vector(11 downto 0));
  end component;

  -- OCP-side command (OCFIFO)
  -- o Inputs OCP commands from general OCP port.
  -- o Outputs OCP commands to Chopper.
  component adb3_ocp_dm_adapter_ocfifo is
  generic(
    ORDER           : in    natural);
  port(
    ocp_rst         : in    std_logic;
    ocp_clk         : in    std_logic;
    -- Commands in from general OCP port
    ocp_cmd         : in    ocp_CmdT;
    ocp_addr        : in    adb3_ocp_addr_s;
    ocp_blen        : in    adb3_ocp_burst_s;
    ocp_tag         : in    adb3_ocp_tag_s;
    ocp_cmd_acc     : out   std_logic;
    -- Commands out to Chopper
    ocf_cmd_adv     : in    std_logic;
    ocf_cmd         : out   ocp_CmdT;
    ocf_addr        : out   adb3_ocp_addr_s;
    ocf_blen        : out   adb3_ocp_burst_s;
    ocf_tag         : out   adb3_ocp_tag_s;
    ocf_cmd_valid   : out   std_logic;
    -- Debug
    debug_info      : out   std_logic_vector(11 downto 0));
  end component;

  -- OCP-side data & byte enable FIFO (OWFIFO)
  -- o Inputs write data & byte enables from general OCP port.
  -- o Outputs write data & byte enables to Chopper.
  component adb3_ocp_dm_adapter_owfifo is
  generic(
    ORDER           : in    natural;
    DATA_WIDTH      : in    natural);
  port(
    ocp_rst         : in    std_logic;
    ocp_clk         : in    std_logic;
    -- Data & byte enables from general OCP port
    ocp_data        : in    std_logic_vector(DATA_WIDTH - 1 downto 0);
    ocp_be          : in    std_logic_vector(DATA_WIDTH / 8 - 1 downto 0);
    ocp_data_valid  : in    std_logic;
    ocp_data_acc    : out   std_logic;
    -- Data & byte enables to Chopper
    owf_data_adv    : in    std_logic;
    owf_data        : out   std_logic_vector(DATA_WIDTH - 1 downto 0);
    owf_be          : out   std_logic_vector(DATA_WIDTH / 8 - 1 downto 0);
    owf_data_valid  : out   std_logic;
    -- Debug
    debug_info      : out   std_logic_vector(11 downto 0));
  end component;

  -- Chopper
  -- o Inputs commands from OCP-side command FIFO and write data & byte enables
  --   from OCP-side data / byte enable FIFO.
  -- o Outputs commands to DM-side command FIFO and write data & byte enables
  --   to DM side data & byte enable FIFO.
  component adb3_ocp_dm_adapter_chop
  generic(
    -- This parameter overrides 'ctl_max_payload' if 'ctl_max_payload' represents
    -- a number of OCP words larger than 2**MAX_WR_BURST_ORDER. This ensures
    -- that the DM data & byte enable FIFO is large enough to hold an entire
    -- burst whilst its byte enables are being analyzed.
    MAX_WR_BURST_ORDER : in    natural range 3 to 9;
    -- Number of bits used for generating tags for chopped commands
    DM_TAG_BITS        : in    natural range 1 to ADB3_OCP_TAG_WIDTH := 6;
    -- true  => Generated write bursts are also compatible with MPTL payload.
    -- false => Generated write bursts in general are not compatible with MPTL
    --          payload.
    MPTL_COMPATIBLE    : in    boolean;
    -- true  => For write bursts, non-OCP word address bits of master_m2s_addr
    --          indicate the first valid byte lane of the first word of the burst
    --          and master_m2s_tag (normally ignored by slave) indicates the last
    --          valid byte of the the last word of the burst.
    --
    --          For read bursts, non-OCP word address bits of master_m2s_addr
    --          are 0 and master_m2s.Tag functions normally.
    --
    --          NOTE: See adb3_ocp_dm_adapter / adb3_ocp_w_dm_adapter for
    --          detailed Addr / Tag usage, because it depends upon DATA_WIDTH.
    --
    -- false => Non-OCP word address bits bits of master_m2s_addr are zero and
    --          master_m2s_tag must be considered invalid unless the burst is a
    --          read.
    PCIE_HINTS         : in    boolean;
    -- OCP data width
    DATA_WIDTH         : in    natural);
  port(
    ocp_rst            : in    std_logic;
    ocp_clk            : in    std_logic;
    -- Commands in from OCP-side command FIFO
    ocf_cmd            : in    ocp_CmdT;
    ocf_addr           : in    adb3_ocp_addr_s;
    ocf_blen           : in    adb3_ocp_burst_s;
    ocf_tag            : in    adb3_ocp_tag_s;
    ocf_cmd_valid      : in    std_logic;
    ocf_cmd_adv        : out   std_logic;
    -- Data in from OCP-side data / byte enable FIFO
    owf_data           : in    std_logic_vector(DATA_WIDTH - 1 downto 0);
    owf_be             : in    std_logic_vector(DATA_WIDTH / 8 - 1 downto 0);
    owf_data_valid     : in    std_logic;
    owf_data_adv       : out   std_logic;
    -- Chopped commands to DM command FIFO
    chop_cmd           : out   ocp_CmdT;
    chop_addr          : out   adb3_ocp_addr_s;
    chop_blen          : out   adb3_ocp_burst_s;
    chop_ocp_tag       : out   adb3_ocp_tag_s;
    chop_dm_tag        : out   adb3_ocp_tag_s;
    chop_cmd_valid     : out   std_logic;
    -- Chopped data & byte enables to DM data FIFO
    chop_data          : out   std_logic_vector(DATA_WIDTH - 1 downto 0);
    chop_be            : out   std_logic_vector(DATA_WIDTH / 8 - 1 downto 0);
    chop_data_valid    : out   std_logic;
    -- Control
    ctl_cmd_enable     : in    std_logic;
    ctl_data_enable    : in    std_logic;
    ctl_max_rd_req     : in    std_logic_vector(2 downto 0);
    ctl_max_payload    : in    std_logic_vector(2 downto 0);
    debug_info         : out   std_logic_vector(63 downto 0));
  end component;

  -- DM-side command FIFO (DCFIFO)
  -- o Inputs commands from Chopper.
  -- o Outputs commands to DM OCP port.
  component adb3_ocp_dm_adapter_dcfifo
  generic(
    ORDER           : in    natural);
  port(
    ocp_rst         : in    std_logic;
    ocp_clk         : in    std_logic;
    -- Commands from Chopper.
    chop_cmd        : in    ocp_CmdT;
    chop_addr       : in    adb3_ocp_addr_s;
    chop_blen       : in    adb3_ocp_burst_s;
    chop_dm_tag     : in    adb3_ocp_tag_s;
    chop_cmd_valid  : in    std_logic;
    -- To DM OCP port
    dm_cmd          : out   ocp_CmdT;
    dm_addr         : out   adb3_ocp_addr_s;
    dm_blen         : out   adb3_ocp_burst_s;
    dm_tag          : out   adb3_ocp_tag_s;
    dm_cmd_acc      : in    std_logic;
    -- Flow control
    nafull          : out   std_logic;
    -- Debug
    debug_info      : out   std_logic_vector(11 downto 0));
  end component;
  
  -- DM-side command FIFO (DWFIFO)
  -- o Inputs write data & byte enables from Chopper.
  -- o Outputs write data & byte enables to DM OCP port.
  component adb3_ocp_dm_adapter_dwfifo is
  generic(
    ORDER           : in    natural;
    DATA_WIDTH      : in    natural);
  port(
    ocp_rst         : in    std_logic;
    ocp_clk         : in    std_logic;
    -- Data & byte enables from Chopper (writes only)
    chop_data       : in    std_logic_vector(DATA_WIDTH - 1 downto 0);
    chop_be         : in    std_logic_vector(DATA_WIDTH / 8 - 1 downto 0);
    chop_data_valid : in    std_logic;
    -- To DM OCP port
    dm_data_acc     : in    std_logic;
    dm_data         : out   std_logic_vector(DATA_WIDTH - 1 downto 0);
    dm_be           : out   std_logic_vector(DATA_WIDTH / 8 - 1 downto 0);
    dm_data_valid   : out   std_logic;
    -- Flow control
    nafull          : out   std_logic;
    -- Debug
    debug_info      : out   std_logic_vector(11 downto 0));
  end component;

  -- Tag mapper (TMAP)
  -- o Establishes mappings between OCP-side tags and DM-side tags as dictated
  --   by Chopper.
  -- o Inputs responses (DM-side tag, resp, data) from DM OCP port.
  -- o Outputs mapped responses (OCP-side tag, resp, data) to response FIFO,
  --   where DM-side tags are mapped to OCP-side tags according to established
  --   mappings.
  component adb3_ocp_dm_adapter_tmap
  generic(
    -- Number of bits used for generating tags for chopped commands
    DM_TAG_BITS     : in    natural range 1 to ADB3_OCP_TAG_WIDTH := 6;
    -- OCP data width
    DATA_WIDTH      : in    natural);
  port(
    ocp_rst         : in    std_logic;
    ocp_clk         : in    std_logic;
    -- Commands in from Chopper
    chop_cmd        : in    ocp_CmdT;
    chop_addr       : in    adb3_ocp_addr_s;
    chop_dm_tag     : in    adb3_ocp_tag_s;
    chop_ocp_tag    : in    adb3_ocp_tag_s;
    chop_blen       : in    adb3_ocp_burst_s;
    chop_cmd_valid  : in    std_logic;
    -- Response from Direct Master channel
    dm_resp_valid   : in    std_logic;
    dm_resp         : in    ocp_RespT;
    dm_data         : in    std_logic_vector(DATA_WIDTH - 1 downto 0);
    dm_tag          : in    adb3_ocp_tag_s;
    -- Tag-mapped response to response FIFO
    tm_resp_valid   : out   std_logic;
    tm_resp         : out   ocp_RespT;
    tm_data         : out   std_logic_vector(DATA_WIDTH - 1 downto 0);
    tm_tag          : out   adb3_ocp_tag_s);
  end component;

  -- Response FIFO (RFIFO)
  -- o Inputs mapped responses (OCP-side tag, resp, data) from tag mapper.
  -- o Outputs mapped responses (OCP-side tag, resp, data) to general OCP port.
  component adb3_ocp_dm_adapter_rfifo
  generic(
    DATA_WIDTH      : in    natural);
  port(
    ocp_rst         : in    std_logic;
    ocp_clk         : in    std_logic;
    -- Tag-mapped response to response FIFO
    tm_resp_valid   : in    std_logic;
    tm_resp         : in    ocp_RespT;
    tm_data         : in    std_logic_vector(DATA_WIDTH - 1 downto 0);
    tm_tag          : in    adb3_ocp_tag_s;
    tm_resp_acc     : out   std_logic;
    -- Response to OCP master
    ocp_resp_acc    : in    std_logic;
    ocp_resp        : out   ocp_RespT;
    ocp_data        : out   std_logic_vector(DATA_WIDTH - 1 downto 0);
    ocp_tag         : out   adb3_ocp_tag_s;
    -- Debug
    debug_info      : out   std_logic_vector(11 downto 0));
  end component;

  component adb3_ocp_dm_adapter_base
  generic(
    -- log2(Size of the DM-side command FIFO)
    DCFIFO_ORDER    : in    natural range 4 to 8;
    -- log2(Size of the OCP-side command FIFO)
    OCFIFO_ORDER    : in    natural range 4 to 8;
    -- log2(Size of the DM-side data & byte enable FIFO)
    DWFIFO_ORDER    : in    natural range 4 to 10;
    -- log2(Size of the OCP-side data & byte enable FIFO)
    OWFIFO_ORDER    : in    natural range 4 to 8;
    -- true  => Generated write bursts are also compatible with MPTL payload.
    -- false => Generated write bursts in general are not compatible with MPTL
    --          payload.
    MPTL_COMPATIBLE : in    boolean;
    -- true  => For write bursts, low bits of master_m2s_addr indicate the
    --          first valid byte lane of the first word of the burst, and
    --          master_m2s_tag (normally ignored by slave) indicates the last
    --          valid byte of the the last word of the burst.
    -- false => Low bits of master_m2s_addr are zero and master_m2s_tag must be
    --          considered invalid unless the burst is a read.
    PCIE_HINTS      : in    boolean;
    -- OCP data width
    DATA_WIDTH      : in    natural);
  port(
    ocp_rst         : in    std_logic;
    ocp_clk         : in    std_logic;
    -- OCP slave port
    slave_m2s_cmd   : in    ocp_CmdT;
    slave_m2s_addr  : in    adb3_ocp_addr_s;
    slave_m2s_blen  : in    adb3_ocp_burst_s;
    slave_m2s_tag   : in    adb3_ocp_tag_s;
    slave_m2s_dval  : in    std_logic;
    slave_m2s_data  : in    std_logic_vector(DATA_WIDTH - 1 downto 0);
    slave_m2s_be    : in    std_logic_vector(DATA_WIDTH / 8 - 1 downto 0);
    slave_m2s_racc  : in    std_logic;
    slave_s2m_cacc  : out   std_logic;
    slave_s2m_dacc  : out   std_logic;
    slave_s2m_resp  : out   ocp_RespT;
    slave_s2m_data  : out   std_logic_vector(DATA_WIDTH - 1 downto 0);
    slave_s2m_tag   : out   adb3_ocp_tag_s;
    -- OCP master port
    master_m2s_cmd  : out   ocp_CmdT;
    master_m2s_addr : out   adb3_ocp_addr_s;
    master_m2s_blen : out   adb3_ocp_burst_s;
    master_m2s_tag  : out   adb3_ocp_tag_s;
    master_m2s_dval : out   std_logic;
    master_m2s_data : out   std_logic_vector(DATA_WIDTH - 1 downto 0);
    master_m2s_be   : out   std_logic_vector(DATA_WIDTH / 8 - 1 downto 0);
    master_m2s_racc : out   std_logic;
    master_s2m_cacc : in    std_logic;
    master_s2m_dacc : in    std_logic;
    master_s2m_resp : in    ocp_RespT;
    master_s2m_data : in    std_logic_vector(DATA_WIDTH - 1 downto 0);
    master_s2m_tag  : in    adb3_ocp_tag_s;
    -- Control ports
    ctl_max_rd_req  : in    std_logic_vector(2 downto 0); -- Maximum read request, encoded as per PCI Express specification
    ctl_max_payload : in    std_logic_vector(2 downto 0); -- Maximum payload, encoded as per PCI Express specification
    -- Debug information
    --
    --    [63:0] CHOP
    --         [0] data_enable_q
    --         [1] cmd_enable_q
    --       [4:2] max_rd_req_q
    --       [7:5] max_payload_q
    --      [11:8] cp_state:
    --             0000 => idle
    --             0001 => read1
    --             0010 => read2
    --             0011 => write_normal
    --             0100 => write_general
    --     [23:12] cp_remaining;
    --     [31:24] cp_addr;
    --     [43:32] cp_start_addr;
    --     [52:44] cp_bcount;
    --        [53] cp_bfirst;
    --        [54] cp_blast;
    --        [55] cp_bboundary;
    --     [63:56] cp_dm_tag;
    --
    --   [75:64] OWFIFO:
    --     [73:64] rn
    --        [74] qvalid
    --        [75] nafull
    --
    --   [87:76] OCFIFO:
    --     [85:76] rn
    --        [86] qvalid
    --        [87] nafull
    --
    --   [99:88] RFIFO:
    --     [97:88] rn
    --        [98] qvalid
    --        [99] nafull
    --
    -- [111:100] DCFIFO:
    --    [99:100] rn
    --       [110] qvalid
    --       [111] nafull
    --
    -- [123:112] DWFIFO:
    --    [121:112] rn
    --       [122] qvalid
    --       [123] nafull
    --   
    debug_info      : out   std_logic_vector(127 downto 0));
  end component;
  
end;

package body adb3_ocp_dm_int is

  function conv_string_hex(
    constant v : std_logic_vector)
  return string is
    variable n : natural := ((v'length + 3) / 4) * 4;
    variable ndigit : natural := n / 4;
    variable tmp : std_logic_vector(n - 1 downto 0);
    variable digit : std_logic_vector(3 downto 0);
    variable str : string(1 to ndigit);
  begin
    tmp(v'length - 1 downto 0) := v;
    if n > v'length then
      tmp(n - 1 downto v'length) := (others => '0');
    end if;
    for i in str'range loop
      digit := tmp(4 * i - 1 downto 4 * (i - 1));
      case digit is
        when "0000" =>
          str(ndigit - i + 1) := '0';

        when "0001" =>
          str(ndigit - i + 1) := '1';

        when "0010" =>
          str(ndigit - i + 1) := '2';

        when "0011" =>
          str(ndigit - i + 1) := '3';

        when "0100" =>
          str(ndigit - i + 1) := '4';

        when "0101" =>
          str(ndigit - i + 1) := '5';

        when "0110" =>
          str(ndigit - i + 1) := '6';

        when "0111" =>
          str(ndigit - i + 1) := '7';

        when "1000" =>
          str(ndigit - i + 1) := '8';

        when "1001" =>
          str(ndigit - i + 1) := '9';

        when "1010" =>
          str(ndigit - i + 1) := 'A';

        when "1011" =>
          str(ndigit - i + 1) := 'B';

        when "1100" =>
          str(ndigit - i + 1) := 'C';

        when "1101" =>
          str(ndigit - i + 1) := 'D';

        when "1110" =>
          str(ndigit - i + 1) := 'E';

        when "1111" =>
          str(ndigit - i + 1) := 'F';

        when others =>
          str(ndigit - i + 1) := 'X';
      end case;
    end loop;
    return str;
  end function;

  -- synthesis translate_off
  function is_center_classification(
    constant be : in std_logic_vector)
  return std_logic is
    variable n : natural := be'length;
    variable curr : std_logic;
    type state_t is (search_1, search_0, check);
    variable state : state_t;
  begin
    if be(0) /= '0' OR be(n - 1) /= '0' then
      -- Not center classification
      return '0';
    end if;
    state := search_1;
    for i in 1 to n - 1 loop
      curr := be(i);
      case state is
        when search_1 =>
          if curr = '1' then
            state := search_0;
          end if;

        when search_0 =>
          if curr = '0' then
            state := check;
          end if;

        when check =>
          if curr = '1' then
            -- Not center classification
            return '0';
          end if;
      end case;
    end loop;
    if state = check then
      return '1'; -- Center classification
    else
      return '0'; -- Not center classification
    end if;
  end;
  -- synthesis translate_on

end;
