--
-- ADB3 Target PCIe Bridge
--
-- Embeddable Target PCIe Bridge
-- Supports Direct Slave and 4 channel DMA transfers only 
--

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_ARITH.all;
use IEEE.STD_LOGIC_UNSIGNED.all;
use IEEE.STD_LOGIC_MISC.all;

library work;
use work.adb3_ocp.all;
use work.adb3_dma.all;
use work.adb3_model_defs.all;

entity adb3_admxrcku1b is
  generic (
    no_dma_engines    : integer := 4;
    pcie_direct_masters : integer := 1;
    reg_direct_masters : integer := 0;
    dma_engine_config : std_logic_vector(15 downto 0) := X"0000";
    bar2_size : natural range 15 to 21 := 15);
  port (
    -- PCIE
    sys_reset_n          : in std_logic;
    sys_clk              : in std_logic;
    sys_clk_gt              : in std_logic;
    ocpclkout            : out std_logic;
    pci_exp_txn          : out std_logic_vector(7 downto 0);
    pci_exp_txp          : out std_logic_vector(7 downto 0);
    pci_exp_rxn          : in std_logic_vector(7 downto 0);
    pci_exp_rxp          : in std_logic_vector(7 downto 0);
    -- Bridge to Target Sideband Signals
    mptl_target_configured_n            : out std_logic;
    mptl_target_gtp_online_n            : out std_logic;
    mptl_bridge_gtp_online_n            : in  std_logic;
    -- OCP interface to User Logic
    direct_slave_m2s       : out adb3w_ocp_m2sT;
    direct_slave_s2m       : in  adb3w_ocp_s2mT;
    dma_channels_m2s       : out adb3w_ocp_m2sT_array(0 to no_dma_engines-1);
    dma_channels_s2m       : in  adb3w_ocp_s2mT_array(0 to no_dma_engines-1);
    direct_master_pcie_m2s : in  adb3w_ocp_m2sT;
    direct_master_pcie_s2m : out adb3w_ocp_s2mT;
    direct_master_regs_m2s : in  adb3l_ocp_m2sT;
    direct_master_regs_s2m : out adb3l_ocp_s2mT;
    interrupt              : in std_logic;
    dma_abort              : in std_logic_vector(no_dma_engines-1 downto 0);
    vpd_in                 : in  std_logic_vector(47 downto 0);
    core_status            : out std_logic_vector(63 downto 0);
    core_rst_out           : out std_logic);

end adb3_admxrcku1b;


architecture synth of adb3_admxrcku1b is

  constant pcie_no_lanes : integer := 8;
 

   component adb3_core_bridge_gen3w_ku1
    generic (
      no_dma_engines     : integer := 4;
      dma_config         : dma_engine_configuration_array(0 to 7) := (others =>  DMA_ENGINE_CONFIGURATION_DEFAULT);
      register_dma_regs  : boolean := false;
      bar_mask_init_val  : std_logic_vector(31 downto 0) := X"003FFFFF";
      pif_init_val       : std_logic_vector(7 downto 0)  := X"02";
      direct_master_channels : integer := 1;
      reg_channels : integer := 0;
      use_icap : boolean := true
      );
    port (
      sys_reset_n            : in  std_logic;
      sys_clk                : in  std_logic;
      sys_clk_gt          : in    std_logic;
      ocpclkout              : out std_logic;
      pci_exp_txn            : out std_logic_vector(pcie_no_lanes-1 downto 0);
      pci_exp_txp            : out std_logic_vector(pcie_no_lanes-1 downto 0);
      pci_exp_rxn            : in  std_logic_vector(pcie_no_lanes-1 downto 0);
      pci_exp_rxp            : in  std_logic_vector(pcie_no_lanes-1 downto 0);
      model_regs_m2s         : out adb3_ocp_m2sT;
      model_regs_s2m         : in  adb3_ocp_s2mT;
      model_regs_irq         : in  std_logic;
      model_regs_reset       : out std_logic;
      model_clk              : in  std_Logic;
      model_clk_locked       : in  std_Logic;
      vpd_valid              : in  std_logic;
      direct_slave_m2s       : out adb3w_ocp_m2sT;
      direct_slave_s2m       : in  adb3w_ocp_s2mT;
      dma_channels_m2s       : out adb3w_ocp_m2sT_array(0 to no_dma_engines-1);
      dma_channels_s2m       : in  adb3w_ocp_s2mT_array(0 to no_dma_engines-1);
      direct_master_pcie_m2s : in  adb3w_ocp_m2sT;
      direct_master_pcie_s2m : out adb3w_ocp_s2mT;
      direct_master_regs_m2s : in  adb3l_ocp_m2sT;
      direct_master_regs_s2m : out adb3l_ocp_s2mT;
      target_irq             : in  std_logic;
      target_dma_abort       : in  std_logic_vector(no_dma_engines-1 downto 0);
      core_status            : out std_logic_vector(63 downto 0);
      core_rst_out           : out std_logic);
  end component;
  
 
  signal model_regs_m2s         : adb3_ocp_m2sT;
  signal model_regs_s2m         : adb3_ocp_s2mT;
  signal model_regs_irq         : std_logic;
  signal model_regs_reset : std_logic;
  

  signal dma_channels_m2s_i       : adb3w_ocp_m2sT_array(0 to 1);
  signal dma_channels_s2m_i       : adb3w_ocp_s2mT_array(0 to 1);

  signal ocp_debug_data  : std_logic_vector(328*2-1 downto 0);
  signal ocp_debug_clk   : std_logic;

  signal target_dma_abort : std_logic_vector(no_dma_engines-1 downto 0) := (others => '0');
  signal linti_n : std_logic;

  signal serial_valid : std_logic;
  signal sys_reset_q_n : std_logic;

  signal ocp_clk : std_logic;
  signal vpd_valid : std_logic;
  signal vpd_valid1 : std_logic;

  signal slaves_m2s : adb3l_ocp_m2sT_array(0 to 1);
  signal slaves_s2m : adb3l_ocp_s2mT_array(0 to 1); 

  signal model_regs_m2s_1         : adb3_ocp_m2sT;
  signal model_regs_s2m_1         : adb3_ocp_s2mT;

  signal model_regs_m2s_fpga         : adb3_ocp_m2sT;
  signal model_regs_s2m_fpga         : adb3_ocp_s2mT;

  signal model_clk,model_clk_locked : std_logic;
  
begin  -- synth

  sys_reset_q_n <= sys_reset_n;-- and serial_valid;
  ocpclkout <= ocp_clk;
  
  adb3_core_bridge_1 : adb3_core_bridge_gen3w_ku1
    generic map (
      no_dma_engines => no_dma_engines,
      dma_config => (0 => (registering_for_speed => true, fpga_descriptors => true, fpga_to_pci_only => (dma_engine_config(0)= '1'), pci_to_fpga_only =>  (dma_engine_config(1) = '1')),
                     1 => (registering_for_speed => true, fpga_descriptors => true, fpga_to_pci_only => (dma_engine_config(2)= '1'), pci_to_fpga_only => (dma_engine_config(3)= '1')),
                     2 => (registering_for_speed => true, fpga_descriptors => true, fpga_to_pci_only => (dma_engine_config(4)= '1'), pci_to_fpga_only => (dma_engine_config(5)= '1')),
                     3 => (registering_for_speed => true, fpga_descriptors => true, fpga_to_pci_only => (dma_engine_config(6)= '1'), pci_to_fpga_only => (dma_engine_config(7)= '1')),
                     others => DMA_ENGINE_CONFIGURATION_DEFAULT),
      register_dma_regs => true,
      pif_init_val   => X"A5",
      direct_master_channels => pcie_direct_masters,
      reg_channels => reg_direct_masters,
      use_icap => false
      )
    port map (
      sys_reset_n            => sys_reset_q_n,
      sys_clk                => sys_clk,
      sys_clk_gt                => sys_clk_gt,
      ocpclkout              => ocp_clk,
      pci_exp_txn            => pci_exp_txn,
      pci_exp_txp            => pci_exp_txp,
      pci_exp_rxn            => pci_exp_rxn,
      pci_exp_rxp            => pci_exp_rxp,
      model_regs_m2s         => model_regs_m2s,
      model_regs_s2m         => model_regs_s2m,
      model_regs_irq         => model_regs_irq,
      model_regs_reset       => model_regs_reset,
      model_clk    => model_clk,
      model_clk_locked => model_clk_locked,
      vpd_valid              => '1',
      direct_slave_m2s       => direct_slave_m2s,
      direct_slave_s2m       => direct_slave_s2m,
      dma_channels_m2s       => dma_channels_m2s(0 to no_dma_engines-1),
      dma_channels_s2m       => dma_channels_s2m(0 to no_dma_engines-1),
      direct_master_pcie_m2s => direct_master_pcie_m2s,
      direct_master_pcie_s2m => direct_master_pcie_s2m,
      direct_master_regs_m2s => slaves_m2s(0),
      direct_master_regs_s2m => slaves_s2m(0),
      target_irq             => '0',
      target_dma_abort       => dma_abort,
      core_status      => core_status,
      core_rst_out     => core_rst_out);


  gen_reg_access: if reg_direct_masters = 1 generate
    
  adb3l_ocp_demux_1: entity work.adb3l_ocp_demux
    generic map (
      number_of_slaves  => 2,
      demux_address_lsb => 12)
    port map (
      clk        => ocp_clk,
      master_m2s => direct_master_regs_m2s,
      master_s2m => direct_master_regs_s2m,
      slaves_m2s => slaves_m2s,
      slaves_s2m => slaves_s2m);
  
  adb3l_to_adb3_1: entity work.adb3l_to_adb3
    port map (
      pri_m2s => slaves_m2s(1),
      pri_s2m => slaves_s2m(1),
      sec_m2s => model_regs_m2s_fpga,
      sec_s2m => model_regs_s2m_fpga);

  adb3_ocp_multiport2_sc_1: entity work.adb3_ocp_multiport2_sc
    generic map (
      round_robin_arb => false,
      use_timeout     => false,
      timeout_bits    => 16)
    port map (
      ocp_m2s_memif => model_regs_m2s_1,
      ocp_s2m_memif => model_regs_s2m_1,
      ocp_m2s0      => model_regs_m2s,
      ocp_s2m0      => model_regs_s2m,
      ocp_m2s1      => model_regs_m2s_fpga,
      ocp_s2m1      => model_regs_s2m_fpga,
      status        => open,
      Clk           => ocp_clk);
  
  
  end generate gen_reg_access;


  gen_no_reg_access: if reg_direct_masters /= 1 generate
    model_regs_m2s_1 <= model_regs_m2s;
    model_regs_s2m <= model_regs_s2m_1;
  end generate gen_no_reg_access;
  
  adm_xrc_ku1_model_1: entity work.adm_xrc_ku1b_model
    port map (
      ocp_clk          => ocp_clk,
      model_regs_m2s   => model_regs_m2s_1,
      model_regs_s2m   => model_regs_s2m_1,
      model_regs_irq   => model_regs_irq,
      model_regs_reset => model_regs_reset,
      model_clk_out    => model_clk,
      model_clk_locked => model_clk_locked,
      linti_n          => linti_n,
      vpd_in           => vpd_in);


  linti_n <= not interrupt;

 
  
end synth;
