--
-- A parameterisable synchronous FIFO.
--
-- Parameters
-- ==========
--
-- 'AFULL_LEVEL' is number of words in FIFO (from writer's perspective) at
-- which 'wafull' flag is asserted.
--
-- 'CONSISTENT_RLEVEL' determines the degree of consistency between 'rempty'
-- and 'rlevel':
--
-- o If 'CONSISTENT_RLEVEL' is false, 'rlevel' indicates the number of words
--   in the FIFO but 'rlevel' and 'rempty' are only weakly consistent. In other
--   words, if 'rlevel' is N, then N words can be read from the FIFO in bounded
--   time, but not necessarily in exactly N cycles. When reading data out of
--   the FIFO, 'rempty' must still be observed regardless of 'rlevel', and
--   'radv' deasserted when 'rempty' is asserted.
--
-- o If 'CONSISTENT_RLEVEL' is true, 'rlevel' indicates the number of words
--   in the FIFO and there is strong consistency between 'rlevel' and 'rempty'.
--   If 'rlevel' is N, then N words can be read from the FIFO in N cycles,
--   and 'rempty' is guaranteed to remain deasserted until all N words have
--   been read out.
--
-- 'MULTICYCLE_RESET' determines whether or not reset logic is included for
-- non-key state:
--
-- o If 'MULTICYCLE_RESET' is true, omits reset logic for non-key state,
--   requiring that reset pulse is at least 2 consecutive cycles.
--
-- o If 'MULTICYCLE_RESET' is false, assumes that reset pulse is as short as
--   one cycle and therefore includes reset logic both for key and non-key
--   state.
--
-- 'ORDER' is log2 of the FIFO depth, e.g. ORDER = 4 => depth is 16 locations.
--
-- 'RAM_TYPE' can be "AUTO", "BLOCK" or "DISTRIBUTED". If "AUTO", style of RAM
-- is determined automatically according to 'ORDER'.
--
-- 'WIDTH' is the width of FIFO locations, in bits.
--
-- Latencies, in cycles
-- ====================
--
-- A latency of 1 means change in next clock cycle.
--
-- 'sr' to all outputs - 1      if NOT MULTICYCLE_RESET
--                       1 or 2 if MULTICYCLE_RESET
--
-- 'wadv' to 'wfull'   - 1
--
-- 'wadv' to 'wafull'  - 1
--
-- 'wadv' to 'wlevel'  - 1
--
-- 'wadv' to 'rempty'  - 3      if actual RAM style is distributed
--                       4      if actual RAM style is block
--
-- 'wadv' to 'rlevel'  - 2      if NOT CONSISTENT_RLEVEL
--                       3      if CONSISTENT_RLEVEL and actual RAM style is distributed
--                       4      if CONSISTENT_RLEVEL and actual RAM style is block
--
-- 'radv' to 'rempty'  - 1
--
-- 'radv' to 'rlevel'  - 1
--
-- 'radv' to 'wfull'   - 2
--
-- 'radv' to 'wlevel'  - 2
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

entity packet_sfifo is
generic(
  AFULL_LEVEL        : in    natural;                              -- 'wafull' asserted when number of occuplied words >= AFULL_LEVEL
  CONSISTENT_RLEVEL  : in    boolean := true;                      -- See notes above.
  MULTICYCLE_RESET   : in    boolean := true;                       -- See notes above.
  ORDER              : in    natural range 1 to 15;                -- Log2 of FIFO depth, in words
  RAM_TYPE           : in    string := "AUTO";                     -- "AUTO", "BLOCK" or "DISTRIBUTED"
  WIDTH              : in    natural);                             -- Width of FIFO, in bits
port(
  rst                : in    std_logic;                            -- Asynchronous reset
  sr                 : in    std_logic;                            -- Synchronous reset
  clk                : in    std_logic;                            -- Clock
  wen                : in    std_logic;                            -- '1' => write value on 'wd' into FIFO
  wd                 : in    std_logic_vector(WIDTH - 1 downto 0); -- Item to write into FIFO, qualified by 'wen'
  wfull              : out   std_logic;                            -- '1' => full
  wafull             : out   std_logic;                            -- '1' => almost full
  wlevel             : out   std_logic_vector(ORDER downto 0);     -- No. of items in FIFO from writer P.O.V.
  radv               : in    std_logic;                            -- '1' => present next item on 'rq'
  rq                 : out   std_logic_vector(WIDTH - 1 downto 0); -- Current item (invalid if 'rempty' asserted)
  rlevel             : out   std_logic_vector(ORDER downto 0);     -- No. of items in FIFO from reader P.O.V. (invalid if 'rempty' asserted)
  rempty             : out   std_logic;                            -- '1' => FIFO is empty, 0 => 'rq' & 'rlevel' are valid
  -- Debug information:
  --   [11:0] rn
  --     [12] not_valid0
  --     [13] not_valid1
  --     [14] ovf
  --     [15] udf
  debug_info         : out   std_logic_vector(15 downto 0));
end entity;

architecture syn of packet_sfifo is

  -- synthesis translate_off
  function conv_string_hex(
    constant v : std_logic_vector)
  return string is
    variable n : natural := ((v'length + 3) / 4) * 4;
    variable ndigit : natural := n / 4;
    variable tmp : std_logic_vector(n - 1 downto 0);
    variable digit : std_logic_vector(3 downto 0);
    variable str : string(1 to ndigit);
  begin
    tmp(v'length - 1 downto 0) := v;
    if n > v'length then
      tmp(n - 1 downto v'length) := (others => '0');
    end if;
    for i in str'range loop
      digit := tmp(4 * i - 1 downto 4 * (i - 1));
      case digit is
        when "0000" =>
          str(ndigit - i + 1) := '0';

        when "0001" =>
          str(ndigit - i + 1) := '1';

        when "0010" =>
          str(ndigit - i + 1) := '2';

        when "0011" =>
          str(ndigit - i + 1) := '3';

        when "0100" =>
          str(ndigit - i + 1) := '4';

        when "0101" =>
          str(ndigit - i + 1) := '5';

        when "0110" =>
          str(ndigit - i + 1) := '6';

        when "0111" =>
          str(ndigit - i + 1) := '7';

        when "1000" =>
          str(ndigit - i + 1) := '8';

        when "1001" =>
          str(ndigit - i + 1) := '9';

        when "1010" =>
          str(ndigit - i + 1) := 'A';

        when "1011" =>
          str(ndigit - i + 1) := 'B';

        when "1100" =>
          str(ndigit - i + 1) := 'C';

        when "1101" =>
          str(ndigit - i + 1) := 'D';

        when "1110" =>
          str(ndigit - i + 1) := 'E';

        when "1111" =>
          str(ndigit - i + 1) := 'F';

        when others =>
          str(ndigit - i + 1) := 'X';
      end case;
    end loop;
    return str;
  end function;
  -- synthesis translate_on

  function determine_ram_style(
    constant RAM_TYPE : in string;
    constant order    : in natural)
  return string is
  begin
    if RAM_TYPE = "AUTO" OR RAM_TYPE = "auto" then
      if order > 6 then
        return "BLOCK";
      else
        return "DISTRIBUTED";
      end if;
    else
      return RAM_TYPE;
    end if;
  end;

  constant ACTUAL_RAM_STYLE : string := determine_ram_style(RAM_TYPE, ORDER);
  constant ACTUAL_RAM_STYLE_BLOCK : boolean := ACTUAL_RAM_STYLE = "BLOCK" OR ACTUAL_RAM_STYLE = "block";
  constant DEPTH : natural := 2**ORDER;
  constant FULL_LEVEL : natural := DEPTH;

  subtype ram_data_t is std_logic_vector(WIDTH - 1 downto 0);
  type ram_array_t is array(0 to DEPTH - 1) of ram_data_t;

  signal rst_nonkey : std_logic; -- Async. reset for non-key state
  signal sr_nonkey : std_logic; -- Sync. reset for non-key state

  signal wptr : std_logic_vector(ORDER downto 0) := (others => '0'); -- Position in RAM where next item is written
  signal wptr_q1 : std_logic_vector(ORDER downto 0) := (others => '0');
  signal wptr_q2 : std_logic_vector(ORDER downto 0) := (others => '0');
  signal i_wlevel : std_logic_vector(ORDER downto 0) := (others => '0'); -- Number of locations occupied, from writer's P.O.V.
  signal i_wfull : std_logic := '0';
  signal i_wafull : std_logic := '0';
  signal ovf : std_logic := '0';

  -- Read logic
  signal i_rlevel : std_logic_vector(ORDER downto 0) := (others => '0');
  signal i_rempty : std_logic := '1';
  signal rptr : std_logic_vector(ORDER downto 0) := (others => '0');
  signal rindex : std_logic_vector(ORDER downto 0) := (others => '0');
  signal udf : std_logic := '0';
  signal rn : std_logic_vector(ORDER - 1 downto 0) := (others => '0');
  signal valid0 : std_logic := '0';
  signal not_valid0 : std_logic := '1';
  signal inc0, rd0 : std_logic;
  signal ram_q0 : std_logic_vector(WIDTH - 1 downto 0);
  signal valid1 : std_logic := '0';
  signal not_valid1 : std_logic := '1';
  signal rd1 : std_logic;
  signal ram_q1 : std_logic_vector(WIDTH - 1 downto 0);

  signal ram : ram_array_t;
  attribute RAM_STYLE : string;
  attribute RAM_STYLE of ram : signal is ACTUAL_RAM_STYLE;
  signal sel_ram : ram_data_t;

  signal rst_int : std_logic := '0';
  signal sr_int : std_logic := '0';

  signal local_rst,local_sync_rst : std_logic := '1';
  attribute keep : string;
  attribute max_fanout : string;
  attribute equivalent_register_removal : string;
  attribute keep of local_rst : signal is "TRUE";
  attribute equivalent_register_removal of local_rst : signal is "no";
  attribute max_fanout of local_rst : signal is "1";
  attribute keep of local_sync_rst : signal is "TRUE";
  attribute equivalent_register_removal of local_sync_rst : signal is "no";
  signal local_rst_r1: std_logic := '1';
  attribute shreg_extract : string;
  attribute keep of local_rst_r1 : signal is "TRUE";  
  attribute equivalent_register_removal of local_rst_r1 : signal is "no";
  attribute max_fanout of local_sync_rst : signal is "16";
  attribute shreg_extract of local_rst_r1 : signal is "no";
  attribute shreg_extract of local_sync_rst : signal is "no";

  signal radv_q : std_logic;
  
begin

  
  process (clk,rst) is
  begin  -- process
    if rst = '1' then
      local_rst <= '1';
    elsif rising_edge(clk) then            -- rising clock edge
      local_rst <= '0';
    end if;
  end process;
  
  process (clk) is
  begin  -- process
    if rising_edge(clk) then       -- rising clock edge
      local_rst_r1 <= local_rst;
      local_sync_rst <= local_rst_r1;
    end if;
  end process;

  -- Only allow local SYNC Reset
  -- SYNC and ASYNC Resets on same flip-flop breaks Xilinx guidelines
  sr_int <= local_sync_rst;
  rst_int <= '0';                     
  
 
  assert RAM_TYPE = "auto" OR RAM_TYPE = "AUTO" OR
         RAM_TYPE = "block" OR RAM_TYPE = "BLOCK" OR
         RAM_TYPE = "distributed" OR RAM_TYPE = "DISTRIBUTED"
    report "*** packet_sfifo: Invalid RAM_TYPE (" & RAM_TYPE & ")"
    severity failure;

  assert AFULL_LEVEL <= DEPTH
    report "*** packet_sfifo: AFULL_LEVEL (" & natural'image(AFULL_LEVEL) & ") larger than DEPTH (" & natural'image(DEPTH) & ")"
    severity failure;

  assert AFULL_LEVEL > 0
    report "*** packet_sfifo: AFULL_LEVEL is zero"
    severity failure;
 
  -- If reset is asserted for multiple cycles, don't need to reset non-key state.
  rst_nonkey <= '0' when MULTICYCLE_RESET else rst_int;
  sr_nonkey <= '0' when MULTICYCLE_RESET else sr_int;

  debug_state : process(rst, clk)
    variable calced_level : natural;
  begin
    if rising_edge(clk) then
      debug_info(11 downto 0) <= std_logic_vector(resize(unsigned(rn), 12));
      debug_info(12) <= not_valid0;
      debug_info(13) <= not_valid1;
      debug_info(14) <= ovf;
      debug_info(15) <= udf;
    end if;

    -- synthesis translate_off
    -- Sanity check for 'rlevel'
    if rst_int = '1' then
      null;
    else
      if sr_int= '1' then
        null;
      else
        if NOT CONSISTENT_RLEVEL then
          -- Can only do this sanity check if 'rlevel' is not calculated for consistency with 'rempty'
          calced_level := to_integer(unsigned(rn));
          if valid0 = '1' then
            calced_level := calced_level + 1;
          end if;
          if valid1 = '1' then
            calced_level := calced_level + 1;
          end if;
          assert calced_level = to_integer(unsigned(i_rlevel))
            report "*** packet_sfifo: rlevel (" & natural'image(to_integer(unsigned(i_rlevel))) & ") doesn't match calculated level (" & natural'image(calced_level) & ")"
            severity error;
        end if;
      end if;
    end if;
    -- synthesis translate_on
  end process;

  radv_q <= radv and not i_rempty;
  
  rempty <= i_rempty;
  rlevel <= i_rlevel;
  wfull  <= i_wfull;
  wafull <= i_wafull;
  wlevel <= i_wlevel;

  write_logic_key : process( clk)
  begin
    if rising_edge(clk) then
      if sr_int= '1' then
        wptr <= (others => '0');
        ovf <= '0';
      else
        -- Position in RAM where next item is written
        if wen = '1' then
          wptr <= wptr + 1;
        end if;

        if wen = '1' and i_wfull = '1' then
          -- synthesis translate_off
          --assert false
          --  report "*** packet_sfifo: FIFO overflow"
          --  severity error;
          -- synthesis translate_on
          ovf <= '1';
        end if;
      end if;
    end if;
  end process;

  write_logic_nonkey : process(rst_nonkey, clk)
    variable next_wlevel : std_logic_vector(ORDER downto 0);
  begin
    if rst_nonkey = '1' then
      wptr_q1 <= (others => '0');
      wptr_q2 <= (others => '0');
      i_wlevel <= (others => '0');
      i_wfull <= '0';
      i_wafull <= '0';
    elsif rising_edge(clk) then
      if sr_nonkey = '1' then
        wptr_q1 <= (others => '0');
        wptr_q2 <= (others => '0');
        i_wlevel <= (others => '0');
        i_wfull <= '0';
        i_wafull <= '0';
      else
        wptr_q1 <= wptr;
        wptr_q2 <= wptr_q1;

        if wen = '1' then
          next_wlevel := wptr - rindex + 1;
        else
          next_wlevel := wptr - rindex;
        end if;

        -- Number of items in FIFO, from writer's perspective
        i_wlevel <= next_wlevel;

        -- Full & almost full flags generated from level, from writer's perspective
        if next_wlevel >= FULL_LEVEL then
          i_wfull <= '1';
        else
          i_wfull <= '0';
        end if;
        if next_wlevel >= AFULL_LEVEL then
          i_wafull <= '1';
        else
          i_wafull <= '0';
        end if;

      end if;
    end if;

    -- Store write data in the RAM
    if rising_edge(clk) then
      if wen = '1' then
        ram(conv_integer(wptr(ORDER - 1 downto 0))) <= wd;
      end if;
    end if;
  end process;

  gen_read_index_key : process( clk)
  begin
    if rising_edge(clk) then
      if sr_int= '1' then
        rindex <= (others => '0');
        udf <= '0';
      else
        if radv_q = '1' then
          rindex <= rindex + 1;

          if i_rempty = '1' then
            udf <= '1';
            -- synthesis translate_off
            assert false
              report "*** packet_sfifo: FIFO underflow"
              severity error;
            -- synthesis translate_on
          end if;
        end if;
      end if;
    end if;
  end process;

  read_logic_nonkey : process(rst_nonkey, clk)
  begin
    if rst_nonkey = '1' then
      i_rlevel <= (others => '0');
    elsif rising_edge(clk) then
      if sr_nonkey = '1' then
        i_rlevel <= (others => '0');
      else
        if CONSISTENT_RLEVEL then
          -- When 'rlevel' must be consistent with 'rempty', use a delayed
          -- version of 'wptr' to compensate for pipeline delay of 'rempty'.
          if ACTUAL_RAM_STYLE_BLOCK then
            if radv_q = '1' then
              i_rlevel <= wptr_q2 - rindex - 1;
            else
              i_rlevel <= wptr_q2 - rindex;
            end if;
          else
            if radv_q = '1' then
              i_rlevel <= wptr_q1 - rindex - 1;
            else
              i_rlevel <= wptr_q1 - rindex;
            end if;
          end if;
        else
          if radv_q = '1' then
            i_rlevel <= wptr - rindex - 1;
          else
            i_rlevel <= wptr - rindex;
          end if;
        end if;
      end if;
    end if;
  end process;

  sel_ram <= ram(conv_integer(rptr(ORDER - 1 downto 0)));

  -- Pipelining strategy is 2 output registers in a pipeline, because
  -- BlockRAM without pipelining option set has Tco.
  gen_output_blockram : if ACTUAL_RAM_STYLE_BLOCK generate

    rd0 <= radv_q OR not_valid0 OR not_valid1;
    inc0 <= (radv_q OR not_valid0 OR not_valid1) AND OR_reduce(rn);

    read_stage_0_key : process(clk)
    begin
      if rising_edge(clk) then
        if sr_int= '1' then
          rptr <= (others => '0');
          valid0 <= '0';
          not_valid0 <= '1';
        else
          if inc0 = '1' then
            rptr <= rptr + 1;
          end if;

          if rd0 = '1' then
            valid0 <= OR_reduce(rn);
            not_valid0 <= NOT OR_reduce(rn);
          end if;
        end if;
      end if;
    end process;

    read_stage_0_nonkey : process( clk)
      variable tmp_rn : std_logic_vector(ORDER downto 0);
    begin
      if rising_edge(clk) then
        if sr_int = '1' then
          rn <= (others => '0');
        else
          if inc0 = '1' then
            tmp_rn := wptr - rptr - 1;
          else
            tmp_rn := wptr - rptr;
          end if;
          -- synthesis translate_off
          -- Sanity check
          assert tmp_rn < DEPTH
            report "*** packet_sfifo: rn value (" & conv_string_hex(tmp_rn) & ") overflows ORDER bits"
            severity error;
          -- synthesis translate_on
          rn <= std_logic_vector(resize(unsigned(tmp_rn), ORDER));
        end if;
      end if;

      if rising_edge(clk) then
        if rd0 = '1' then
          ram_q0 <= sel_ram;
        end if;
      end if;
    end process;

    rd1 <= radv_q OR NOT valid1;

    read_stage_1 : process( clk)
    begin
      if rising_edge(clk) then
        if sr_int= '1' then
          valid1 <= '0';
          not_valid1 <= '1';
        else
          if rd1 = '1' then
            valid1 <= valid0;
            not_valid1 <= not_valid0;
          end if;
        end if;
      end if;
    end process;

    read_stage_1_nonkey : process(clk)
    begin
      if rising_edge(clk) then
        if rd1 = '1' then
          ram_q1 <= ram_q0;
        end if;
      end if;
    end process;

    i_rempty <= not_valid1;
    rq <= ram_q1;

  end generate;

  -- Pipelining strategy is 1 output register, because distributed RAM doesn't
  -- have high Tco.
  gen_output_distributedram : if NOT ACTUAL_RAM_STYLE_BLOCK generate

    rd0 <= radv_q OR not_valid0;
    inc0 <= (radv_q OR not_valid0) AND OR_reduce(rn);

    read_stage_0 : process( clk)
      variable tmp_rn : std_logic_vector(ORDER downto 0);
    begin
      if rising_edge(clk) then
        if sr_int= '1' then
          rptr <= (others => '0');
          valid0 <= '0';
          not_valid0 <= '1';
        else
          if inc0 = '1' then
            rptr <= rptr + 1;
          end if;

          if rd0 = '1' then
            valid0 <= OR_reduce(rn);
            not_valid0 <= NOT OR_reduce(rn);
          end if;
        end if;
      end if;
    end process;

    read_stage_0_nonkey : process(clk)
      variable tmp_rn : std_logic_vector(ORDER downto 0);
    begin
      if rising_edge(clk) then
        if sr_int = '1' then
          rn <= (others => '0');
        else
          if inc0 = '1' then
            tmp_rn := wptr - rptr - 1;
          else
            tmp_rn := wptr - rptr;
          end if;
          -- synthesis translate_off
          -- Sanity check
          assert tmp_rn < DEPTH
            report "*** packet_sfifo: rn value (" & conv_string_hex(tmp_rn) & ") overflows ORDER bits"
            severity error;
          -- synthesis translate_on
          rn <= std_logic_vector(resize(unsigned(tmp_rn), ORDER));
        end if;
      end if;

      if rising_edge(clk) then
        if rd0 = '1' then
          ram_q0 <= sel_ram;
        end if;
      end if;
    end process;

    i_rempty <= not_valid0;
    rq <= ram_q0;

  end generate;

end;
