--------------------------------------------------------------------------------
-- Company:         Alpha-Data Parallel Systems
-- Engineer:        Alan Crawford
--
-- Create Date:
-- Design Name:
-- Module Name:     pcieg3_cc_interface
-- Project Name:    ADB3G3
-- Target Device:   V7
-- Tool versions:
-- Description:     Interfaces to the Gen3 Core BAR I/F and provides the
--                  required number of OCP ports
--
-- Dependencies:    PCIe/OCP Packages for signal records
--
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
--
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

library work;
use work.pcie.ALL;
use work.adb3_ocp.ALL;
use work.pciegen3_pkg.all;
use work.pcie_config.all;

entity pcieg3_cc_interface is
  generic(
    datawidthTX : integer := 128
    );
  port(
    pcie_clk              : in  std_logic;
    pcie_rst              : in  std_logic;

    rx_bar_data_out       : in  std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
    rx_bar_data_push      : in  std_logic := '0';
    rx_bar_data_space     : out std_logic;
    
    -- BAR Completion I/F
    cc_b2c                : out cc_b2c_t;
    cc_c2b                : in  cc_c2b_t
    
    );
end pcieg3_cc_interface;

architecture synth of pcieg3_cc_interface is

signal cc_empty : std_logic;
signal pop : std_logic := '0';
signal cc_pkt_avail : std_logic;
signal cc_tdata : std_logic_vector( datawidthTX+datawidthTX/32+1 downto 0);
signal i_rx_bar_data_space  : std_logic;
signal pop_mask  : std_logic;

signal rx_bar_data_push_r1 : std_logic := '0';
signal rx_bar_data_out_r1 : std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0) := (others => '0');


begin

  reg_push: process (pcie_clk) is
  begin  -- process reg_push
    if rising_edge(pcie_clk) then            -- rising clock edge
      rx_bar_data_push_r1 <= rx_bar_data_push;
      rx_bar_data_out_r1 <= rx_bar_data_out;
    end if;
  end process reg_push;

  pop_mask <=  pop and not cc_empty;
    bar_if_cpld_fifo : packet_sfifo
    generic map(
          RAM_TYPE    => "AUTO",
          ORDER       => 5,
          WIDTH       => datawidthTX+datawidthTX/32+2,
          AFULL_LEVEL => 20
    )
    port map(
      rst         => pcie_rst,
      sr          => '0',--pcie_rst,
      clk         => pcie_clk,
      wen         => rx_bar_data_push,
      wd          => rx_bar_data_out,
      wfull       => open,
      wafull      => i_rx_bar_data_space,
      radv        => pop_mask,
      rq          => cc_tdata, 
      rlevel      => open,
      rempty      => cc_empty,
      debug_info  => open);

  rx_bar_data_space <= not i_rx_bar_data_space;
-- bar_if_cpld_fifo : pcie_pkt_fifo6 --5
--   generic map(
--     fifo_depth          => 5,
--     fifo_width          => datawidthTX+datawidthTX/32+2
--   )
--   port map(
--     clk                 => pcie_clk,
--     sr                  => pcie_rst,
--     max_pkt_size        => "01000",
--     din                 => rx_bar_data_out,
--     dout                => cc_tdata,
--     push                => rx_bar_data_push,
--     pop                 => pop,
--     full                => open,
--     empty               => cc_empty,
-- --     last                => open,
--     free_pkt            => rx_bar_data_space
-- --     pkt_available       => open
-- );

  cc_pkt_avail <= not cc_empty;

  process(cc_tdata,cc_empty)
  begin
    if(cc_empty = '0')then
      cc_b2c.s_axis_cc_tkeep    <= cc_tdata(datawidthTX+datawidthTX/32-1 downto datawidthTX);
      cc_b2c.s_axis_cc_tdata    <= cc_tdata(datawidthTX-1 downto 0);
    else
      cc_b2c.s_axis_cc_tkeep    <= (others => '0');
      cc_b2c.s_axis_cc_tdata    <= (others => '0');
    end if;
  end process;
--   cc_b2c.s_axis_cc_tparity  <= (others => '0'); 
  cc_b2c.s_axis_cc_tvalid   <= cc_pkt_avail ;
  cc_b2c.s_axis_cc_tuser    <= (others => '0');
  cc_b2c.s_axis_cc_tlast    <= cc_tdata(datawidthTX+datawidthTX/32) and not cc_empty;
  pop <= cc_c2b.s_axis_cc_tready;
  
  
  parity_blocks: for index in 1 to 32 generate
  begin
    pg_blk : pargen
      port map(
        data_in =>  cc_tdata(index*8-1 downto (index-1)*8),
        parity_odd => cc_b2c.s_axis_cc_tparity(index-1)
      );
  end generate parity_blocks;
  

end architecture;
