  --------------------------------------------------------------------------------
-- Company:         Alpha-Data Parallel Systems
-- Engineer:        Alan Crawford
--
-- Create Date:
-- Design Name:
-- Module Name:     pcieg3_bar_rx_if
-- Project Name:    ADB3G3
-- Target Device:   V7
-- Tool versions:
-- Description:     Interfaces to the Gen3 Core BAR I/F and provides the
--                  required number of OCP ports
--
-- Dependencies:    PCIe/OCP Packages for signal records
--
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
--
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_ARITH.all;
use IEEE.STD_LOGIC_UNSIGNED.all;

library work;
use work.pcie.all;
use work.adb3_ocp.all;
use work.pciegen3_pkg.all;
use work.pcie_config.all;

entity pcieg3_bar_rx_if is
  generic(
    datawidthTX             : integer := 128
    );
  port(
    pcie_clk                : in  std_logic;
    pcie_rst                : in  std_logic;
    m_m2s_o                 : out adb3_ocp_m2sT;
    m_s2m_i                 : in  adb3_ocp_s2mT;

    rx_data_out             : out std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
    rx_data_pop             : in  std_logic;
    rx_data_pkt_rdy         : out std_logic;

    rx_tag_value            : out std_logic_vector(7 downto 0);
    rx_tag_clear            : out std_logic;
    rx_tag_info             : in  std_logic_vector(71 downto 0);
    rx_tag_valid            : in  std_logic

  );
end pcieg3_bar_rx_if;

architecture synth of pcieg3_bar_rx_if is

  signal valid_rx_data    : std_logic := '0';
  signal rx_data_tag      : std_logic_vector(135 downto 0);
  signal rx_data_tag_fifo : std_logic_vector(135 downto 0);
  signal rx_push          : std_logic := '0';
  signal rx_full          : std_logic := '0';
  signal rx_pop           : std_logic := '0';
  signal rx_pop_int       : std_logic := '0';
  signal rx_tag_fifo      : std_logic_vector(7 downto 0);
  signal rx_empty         : std_logic := '1';

  signal rx_sequencer     : std_logic_vector(4 downto 0);  -- 7 downto 0
  signal rx_descriptor    : std_logic_vector(127 downto 0);
  signal push_descriptor  : std_logic := '0';
  signal cpld_data_in     : std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
  signal push_cpld        : std_logic := '0';
  signal rx_cpld_empty    : std_logic := '1';
  signal rx_cpld_full     : std_logic := '0';
  signal rx_count         : std_logic_vector(7 downto 0);
  signal data_available   : std_logic_vector(5 downto 0);

  signal end_address_dw   : std_logic_vector(2 downto 0);

  signal store_data       : std_logic_vector(127 downto 0);
  signal start_half       : std_logic := '0';
  signal last_transfer    : std_logic := '0';
  signal i_rx_data_out    : std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
  signal i_rx_data_pop         : std_logic := '0';

  signal local_rst,local_sync_rst : std_logic := '1';
  attribute keep : string;
  attribute max_fanout : string;
  attribute equivalent_register_removal : string;
  attribute keep of local_rst : signal is "TRUE";
  attribute equivalent_register_removal of local_rst : signal is "no";
  attribute max_fanout of local_rst : signal is "1";
  attribute keep of local_sync_rst : signal is "TRUE";
  attribute equivalent_register_removal of local_sync_rst : signal is "no";
  signal local_rst_r1: std_logic := '1';
  attribute shreg_extract : string;
  attribute keep of local_rst_r1 : signal is "TRUE";  
  attribute equivalent_register_removal of local_rst_r1 : signal is "no";
  attribute max_fanout of local_sync_rst : signal is "16";
  attribute shreg_extract of local_rst_r1 : signal is "no";
  attribute shreg_extract of local_sync_rst : signal is "no";
begin

  process (pcie_clk,pcie_rst) is
  begin  -- process
    if pcie_rst = '1' then
      local_rst <= '1';
    elsif rising_edge(pcie_clk) then            -- rising clock edge
      local_rst <= '0';
    end if;
  end process;
  
  process (pcie_clk) is
  begin  -- process
    if rising_edge(pcie_clk) then       -- rising clock edge
      local_rst_r1 <= local_rst;
      local_sync_rst <= local_rst_r1;
    end if;
  end process;
  
  rx_data_tag <= m_s2m_i.tag & m_s2m_i.data;

     bar_if_cmd_fifo : packet_sfifo
    generic map(
          RAM_TYPE    => "AUTO",
          ORDER       => 5,
          WIDTH       => 128+8,
          AFULL_LEVEL => 32
    )
    port map(
      rst         => pcie_rst,
      sr          => '0',
      clk         => pcie_clk,
      wen         => rx_push,
      wd          => rx_data_tag,
      wfull       => open,
      wafull      => rx_full,
      radv        => rx_pop,
      rq          => rx_data_tag_fifo, 
      rlevel      => data_available,
      rempty      => rx_empty,
      debug_info  => open);


--   bar_if_cmd_fifo : pcie_pkt_fifo4
--     generic map(
--       fifo_depth => 5,
--       fifo_width => 128+8
--       )
--     port map(
--       clk            => pcie_clk,
--       sr             => pcie_rst,
--       max_pkt_size   => "01000",
--       din            => rx_data_tag,
--       dout           => rx_data_tag_fifo,
--       push           => rx_push,
--       pop            => rx_pop ,
--       full           => open,
--       empty          => rx_empty,
--       last           => open,
--       free_pkt       => rx_full,
--       data_available => data_available
--     );

  m_m2s_o.respaccept  <= not rx_full;
  m_m2s_o.Cmd         <= (others => '0');
  m_m2s_o.Addr        <= (others => '0');
  m_m2s_o.BurstLength <= (others => '0');
  m_m2s_o.Data        <= (others => '0');
  m_m2s_o.DataByteEn  <= (others => '0');
  m_m2s_o.Tag         <= (others => '0');
  m_m2s_o.DataValid   <= '0'; 


  gen_push : process(m_s2m_i, rx_full)
  begin
    if(rx_full = '0')and(m_s2m_i.resp = "01")then
      rx_push <= '1';
    else
      rx_push <= '0';
    end if;
  end process gen_push;

--    rx_tag_value <= rx_data_tag_fifo(135 downto 128);
  process (rx_data_tag_fifo,rx_empty)
  begin
    if(rx_empty = '1')then
      rx_tag_value <= (others => '0');
    else
      rx_tag_value <= rx_data_tag_fifo(135 downto 128);
    end if;
  end process;

  rx_descriptor <= x"0000_0000" &                     -- data if in DW aligned mode              32  127:96
                   '0' &                              -- Force ECRC                               1      95
                   rx_tag_info(60 downto 58) &        -- Packet attributes                        3   94:92
                   rx_tag_info(57 downto 55) &        -- TC                                       3   91:89
                   '0' &                              -- use Completer ID in core                 1      88
                   x"0000" &                          -- unused when previous bit is '0'         16   87:72
                   rx_tag_info(68 downto 61) &        -- PCIE TAG                                 8   71:64
                   rx_tag_info(54 downto 39) &        -- Requester ID                            16   63:48
                   '0' &                              -- reserved                                 1      47 
                   '0' &                              -- Poison TLP Not used                      1      46
                   "000" &                            -- completer status (Successful)            3   45:43
                   '0'& rx_tag_info(38 downto 29) &   -- Length in DW of completion              11   42:32
                   "00" &                             -- reserved                                 2   31:30
                   '0' &                              -- locked read (NOT used)                   1      29
                   rx_tag_info(20 downto 8) &         -- Byte Count                              13   28:16
                   "000000" &                         -- reserved                                 6   15:10
                   "00" & --rx_tag_info(56 downto 55) &        -- address type                             2    9: 8 
                   '0' &                              -- reserved                                 1       7
                   rx_tag_info(28 downto 22);         -- low address                              7    6: 0
  
  end_address_dw <= rx_tag_info(71 downto 69);
  
  seq_rx : process(pcie_clk)
  begin
    if pcie_clk'event and pcie_clk = '1' then
      if(local_sync_rst = '1')then
        rx_sequencer  <= (others => '0');
        rx_count      <= (others => '0');
        last_transfer <= '0';
      else
        last_transfer <= '0';
        rx_sequencer  <= rx_sequencer(3 downto 0)& '0';      -- 6 downto 0
        if((rx_empty = '0')and(rx_sequencer = "00000"))then  -- 0000000
          rx_sequencer(0) <= '1';
        end if;
        rx_pop_int      <= '0';
        rx_tag_clear    <= rx_sequencer(0);                  -- 1
        push_descriptor <= rx_sequencer(0);                  -- 1
        if(rx_sequencer(0) = '1')then                        --1
          rx_count <= rx_tag_info(7 downto 0);
        end if;

        if(rx_sequencer(1) = '1')then          -- 2
          if((rx_count > data_available)or(rx_cpld_full = '1'))then
            rx_sequencer(2 downto 1) <= "01";  --2 downto 1
          end if;
        end if;
        if(rx_sequencer(2) = '1')then          -- 3
          if((rx_count > "00000001"))then
            rx_sequencer(3 downto 2) <= "01";  -- 4 downto 3
            rx_count                 <= rx_count - 1;
            rx_pop_int               <= '1';
          elsif((rx_count = "00000001"))then
            rx_sequencer(3 downto 2) <= "10";  -- 4 downto 3
            rx_count                 <= rx_count - 1;
            rx_pop_int               <= '1';
            last_transfer            <= '1';
          end if;
        end if;
        if(rx_pop = '1')then
        end if;
      end if;
    end if;
  end process seq_rx;

  process(rx_pop_int,rx_empty)
  begin
    if(rx_empty = '0')then
      rx_pop <= rx_pop_int;
    else
      rx_pop <= '0';
    end if;
  end process;

  reg_data_half : process(pcie_clk)
  begin
    if pcie_clk'event and pcie_clk = '1' then
      if(local_sync_rst = '1')then
        store_data <= (others => '0');
        start_half <= '0';
      else
        if(rx_sequencer(0) = '1')then
          start_half <= rx_tag_info(26);
          
        end if;

        if(rx_pop = '1')then
          start_half <= not start_half;
          store_data <= rx_data_tag_fifo(127 downto 0);
        end if;
      end if;
    end if;
  end process reg_data_half;

  push_cpld <= (rx_pop and start_half) or push_descriptor or (rx_pop and last_transfer);

  sel_op_data256 : process(push_descriptor, rx_descriptor, rx_data_tag_fifo, store_data, last_transfer, start_half)
  begin
    if(push_descriptor = '1')then
      cpld_data_in(255 downto 0) <= x"0000_0000_0000_0000_0000_0000_0000_0000" & rx_descriptor;
    else
      if(last_transfer = '1')and(start_half = '0')then
        cpld_data_in(255 downto 0) <= x"0000_0000_0000_0000_0000_0000_0000_0000" & rx_data_tag_fifo(127 downto 0);
      else
        cpld_data_in(255 downto 0) <= rx_data_tag_fifo(127 downto 0) & store_data;
      end if;
    end if;
  end process sel_op_data256;

  sel_op_flags256 : process(push_descriptor, rx_descriptor, rx_data_tag_fifo, rx_pop, rx_count, end_address_dw)
  begin
    if(rx_count > "000000000")then
      if(push_descriptor = '1')then
        cpld_data_in(265 downto 256) <= "10" & X"0F";
      else
        cpld_data_in(265 downto 256) <= "00" & X"FF";
      end if;
    else
      case (end_address_dw(2 downto 0)) is
        when "000"  => cpld_data_in(265 downto 256) <= "01" & X"01";
        when "001"  => cpld_data_in(265 downto 256) <= "01" & X"03";
        when "010"  => cpld_data_in(265 downto 256) <= "01" & X"07";
        when "011"  => cpld_data_in(265 downto 256) <= "01" & X"0F";
        when "100"  => cpld_data_in(265 downto 256) <= "01" & X"1F";
        when "101"  => cpld_data_in(265 downto 256) <= "01" & X"3F";
        when "110"  => cpld_data_in(265 downto 256) <= "01" & X"7F";
        when others => cpld_data_in(265 downto 256) <= "01" & X"FF";
      end case;
    end if;
  end process sel_op_flags256;

    bar_if_cpld_fifo : packet_sfifo
    generic map(
          RAM_TYPE    => "AUTO",
          ORDER       => 5,
          WIDTH       => datawidthTX+datawidthTX/32+2,
          AFULL_LEVEL => 16
    )
    port map(
      rst         => pcie_rst,
      sr          => '0',
      clk         => pcie_clk,
      wen         => push_cpld,
      wd          => cpld_data_in,
      wfull       => open,
      wafull      => rx_cpld_full,
      radv        => i_rx_data_pop,
      rq          => i_rx_data_out, 
      rlevel      => open,
      rempty      => rx_cpld_empty,
      debug_info  => open);
   process(rx_data_pop,rx_cpld_empty)
   begin
    if(rx_cpld_empty = '1')then
      i_rx_data_pop <= '0';
    else
      i_rx_data_pop <= rx_data_pop;
    end if;
  end process;
      
    test : pcie_pkt_counter 
      port map(
        clk            => pcie_clk, 
        sr             => pcie_rst,
        eop_in         => cpld_data_in(datawidthTX+datawidthTX/32),
        eop_out        => i_rx_data_out(datawidthTX+datawidthTX/32),
        push           => push_cpld,
        pop            => i_rx_data_pop,
        pkt_available  => rx_data_pkt_rdy
      ); 

      
      
      rx_data_out <= i_rx_data_out;
      
--   bar_if_cpld_fifo : pcie_pkt_fifo5  --- not 3
--     generic map(
--       fifo_depth => 5,
--       fifo_width => datawidthTX+datawidthTX/32+2
--     )
--     port map(
--       clk           => pcie_clk,
--       sr            => pcie_rst,
--       max_pkt_size  => "01000",
--       din           => cpld_data_in,
--       dout          => rx_data_out,
--       push          => push_cpld,
--       pop           => rx_data_pop,
--       full          => open,
--       empty         => rx_cpld_empty,
--       last          => open,
--       free_pkt      => rx_cpld_full,
--       pkt_available => rx_data_pkt_rdy
--     );

end architecture;
