--
-- pcie_pkg.vhd - definitions file for pcie bridge signals in adb3
--
--
-- (c) copyright alpha data ltd. 2009
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.ALL;
use work.pcie_config.ALL;

package pcie is

constant MAX_TIMEOUT : std_logic_vector(15 downto 0) := X"FFF0";


constant no_masters     : natural := cfg_no_masters;
constant no_slaves      : natural := cfg_no_slaves;
constant source_size    : natural := cfg_source_size;
constant no_32bit_bars  : natural := cfg_no_32bit_bars;

constant max_data_width : natural := cfg_max_data_width;
constant no_control_bits : natural := cfg_no_control_bits;
constant pcie_data_size : natural := (max_data_width + no_control_bits);

constant pcie_sop_flag   : natural :=pcie_data_size-1;
constant pcie_eop_flag   : natural :=pcie_data_size-2;
constant pcie_emod_flag   : natural :=pcie_data_size-3;

constant p2o_master_pcie_pkt_fifo_size : natural := 5;    -- Master fifo contains PCIE packets for decoding
constant p2o_cmd_fifo_size : natural := 2;                -- holds read/write commands for OCP master port
constant p2o_data_fifo_size : natural := 5;               -- holds write data for OCP master port
constant p2o_slave_pcie_pkt_fifo_size : natural := 8;     -- holds converted OCP reads/writes for OCP slave Port
constant p2o_cpld_fifo_size : natural := 6;               -- holds completion data for OCP slave port
constant o2p_pcie_pkt_fifo_size : natural := 5;           -- holds PCIE completions prior to decode in OCP Slave port
constant pcie_pkt_fifo_size : natural := 5;               -- PCIE packet fifo for transferring to CORE


-- -----------------------------------------------------------------------------
-- pcie parameter constants
-- data width of 128 bits + 3 control (SOP,EOP,EMOD)
-- -----------------------------------------------------------------------------
constant pcie_core_width    : integer := cfg_pcie_core_width;
constant pcie_addr_width    : integer := cfg_pcie_addr_width;
constant pcie_data_width    : integer := pcie_data_size;
-- constant pcie_no_lanes      : integer := pcie_no_lanes;

subtype bar_dataT is std_logic_vector((pcie_data_size-1) downto 0);
type  bar_data_array is array (0 to no_masters-1) of bar_dataT;
type  slave_data_array is array (0 to no_slaves-1) of bar_dataT;
type  all_data_array is array (0 to no_slaves+no_masters-1) of bar_dataT;

type  tag_limits_t is array(0 to 4) of std_logic_vector(7 downto 0);

constant tag_value_lower : tag_limits_t := (0=>"00011000",
                                            1=>"00000000",
                                            2=>"00000110",
                                            3=>"00001100",
                                            4=>"00010010");

constant tag_value_upper : tag_limits_t := (0=>"00011111",
                                            1=>"00000101",
                                            2=>"00001011",
                                            3=>"00010001",
                                            4=>"00010111");

--type  tag_limits_t is array(0 to 4) of std_logic_vector(7 downto 0);
--constant tag_value_lower : tag_limits_t := (0=>"00000000",
--                                            1=>"00000100",
--                                            2=>"00001100",
--                                            3=>"00010100",
--                                            4=>"00011100");
--constant tag_value_upper : tag_limits_t := (0=>"00000011",
--                                            1=>"00001011",
--                                            2=>"00010011",
--                                            3=>"00011011",
--                                            4=>"00011111");

subtype data_stripe is std_logic_vector(pcie_data_size-1 downto 0);

subtype dma_flags is std_logic_vector(no_slaves-1 downto 0);
subtype bar_flags is std_logic_vector(no_masters-1 downto 0);
subtype comb_flags is std_logic_vector(no_slaves+no_masters-1 downto 0);
subtype comb_count is std_logic_vector(source_size-1 downto 0);

subtype array_bar_stripes is std_logic_vector((pcie_data_size*(no_masters)-1) downto 0);
subtype array_dma_stripes is std_logic_vector((pcie_data_size*(no_slaves)-1) downto 0);
subtype array_all_stripes is std_logic_vector((pcie_data_size*(no_masters+no_slaves)-1) downto 0);


subtype  pcie_data_t        is std_logic_vector(pcie_data_width-1 downto 0);
constant pcie_data_fifo_depth      : integer := 64;
constant pcie_data_fifo_depth_bits : integer := 6;

subtype  pcie_data_fifo_ptr_t is std_logic_vector(pcie_data_fifo_depth_bits-1 downto 0);
type     pcie_data_fifo_t is array(pcie_data_fifo_depth-1 downto 0) of pcie_data_t;


constant idle_ocp_master : adb3_ocp_m2sT := (
        Cmd           => "000",
        Addr          => (others => '0'),
        BurstLength   => (others => '0'),
        Data          => (others => '0'),
        DataByteEn    => (others => '0'),
        DataValid     => '0',
        RespAccept    => '0',
        Tag           => (others => '0'));
constant idle_ocp_slave : adb3_ocp_s2mT := (
        CmdAccept     => '0',
        DataAccept    => '0',
        Data          => (others => '0'),
        Resp          => "00",
        Tag           => (others => '0'));

-- -----------------------------------------------------------------------------
-- tx pkt fifo sizes
-- -----------------------------------------------------------------------------
constant pcie_read_pkt_fifo_depth    : integer := 4;
constant pcie_write_pkt_fifo_depth   : integer := 64;
constant pcie_cpld_pkt_fifo_depth    : integer := 4;

subtype  pcie_read_pkt_fifo_depth_t  is std_logic_vector(pcie_read_pkt_fifo_depth-1 downto 0);
subtype  pcie_write_pkt_fifo_depth_t is std_logic_vector(pcie_write_pkt_fifo_depth-1 downto 0);
subtype  pcie_cpld_pkt_fifo_depth_t  is std_logic_vector(pcie_cpld_pkt_fifo_depth-1 downto 0);

-- -----------------------------------------------------------------------------
-- slave tag stores
-- -----------------------------------------------------------------------------
constant pcie_tag_depth              : integer := 32;
subtype  ocp_tag_entry_t             is std_logic_vector(adb3_ocp_tag_width-1 downto 0);
type     pcie_tag_store_t            is array(pcie_tag_depth-1 downto 0) of ocp_tag_entry_t;
subtype  pcie_tag_t                  is std_logic_vector(7 downto 0);

-- -----------------------------------------------------------------------------
-- declare pcie format and type info
-- -----------------------------------------------------------------------------
constant pcie_fmt_mem32_read  : std_logic_vector(1 downto 0) := "00";
constant pcie_fmt_mem32_write : std_logic_vector(1 downto 0) := "10";
constant pcie_fmt_mem64_read  : std_logic_vector(1 downto 0) := "01";
constant pcie_fmt_mem64_write : std_logic_vector(1 downto 0) := "11";
constant pcie_fmt_cpl_no_data : std_logic_vector(1 downto 0) := "00";
constant pcie_fmt_cpl_data    : std_logic_vector(1 downto 0) := "10";

constant pcie_type_mem32_read : std_logic_vector(4 downto 0) := "00000";
constant pcie_type_mem32_write: std_logic_vector(4 downto 0) := "00000";
constant pcie_type_mem64_read : std_logic_vector(4 downto 0) := "00000";
constant pcie_type_mem64_write: std_logic_vector(4 downto 0) := "00000";

constant pcie_type_cpl_no_data: std_logic_vector(4 downto 0) := "01010";
constant pcie_type_cpl_data   : std_logic_vector(4 downto 0) := "01010";

constant pcie_mem32_read      : std_logic_vector(7 downto 0) := "0" & pcie_fmt_mem32_read  & pcie_type_mem32_read;
constant pcie_mem64_read      : std_logic_vector(7 downto 0) := "0" & pcie_fmt_mem64_read  & pcie_type_mem64_read;
constant pcie_mem32_write     : std_logic_vector(7 downto 0) := "0" & pcie_fmt_mem32_write & pcie_type_mem32_write;
constant pcie_mem64_write     : std_logic_vector(7 downto 0) := "0" & pcie_fmt_mem64_write & pcie_type_mem64_write;
constant pcie_cpl_data        : std_logic_vector(7 downto 0) := "0" & pcie_fmt_cpl_data    & pcie_type_cpl_data;
constant pcie_cpl_no_data     : std_logic_vector(7 downto 0) := "0" & pcie_fmt_cpl_no_data & pcie_type_cpl_no_data;

-- Declare PCIe completion codes

constant PCIE_CPL_SC          : std_logic_vector(2 downto 0) := "000";  -- successful completion
constant PCIE_CPL_UR          : std_logic_vector(2 downto 0) := "001";  -- unsupported request
constant PCIE_CPL_CRS         : std_logic_vector(2 downto 0) := "010";  -- Config Request Retry Status
constant PCIE_CPL_CA          : std_logic_vector(2 downto 0) := "100";  -- Completer Abort


type pcie_system_info_t is record  -- system info for packet formatting
  bus_number        : std_logic_vector(7 downto 0);
  device_number     : std_logic_vector(4 downto 0);
  function_number   : std_logic_vector(2 downto 0);
end record pcie_system_info_t;


constant  pcie_dword_count_1  : std_logic_vector(9 downto 0) := "0000000001";
constant  pcie_dword_count_2  : std_logic_vector(9 downto 0) := "0000000010";
constant  pcie_dword_count_3  : std_logic_vector(9 downto 0) := "0000000011";
constant  pcie_dword_count_4  : std_logic_vector(9 downto 0) := "0000000100";


-- type ocp_cmd_fifo_data_t is record
--     Cmd         : ocp_CmdT;   -- Command b'000=Idle b'010=Read b'001=Write
--     Addr        : std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
--     BurstLength : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
--     Tag         : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
-- end record ocp_cmd_fifo_data_t;
subtype ocp_cmd_fifo_data_t is  std_logic_vector(ADB3_OCP_ADDR_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_TAG_WIDTH+2 downto 0);

constant  ocp_cmd_fifo_depth : integer := 2;
constant  ocp_data_fifo_depth : integer := 6;
constant  ocp_data_fifo_width : integer := max_data_width;
subtype   ocp_data_fifo_data_t is std_logic_vector(ocp_data_fifo_width-1 downto 0); -------------------FIX

type pcie_bar_mask_t is record -- Masks and page info for each bar
  bar0_mask             : std_logic_vector(pcie_addr_width-1 downto 0);
  bar0_page             : std_logic_vector(pcie_addr_width-1 downto 0);
  bar1_mask             : std_logic_vector(pcie_addr_width-1 downto 0);
  bar1_page             : std_logic_vector(pcie_addr_width-1 downto 0);
  bar23_mask            : std_logic_vector(pcie_addr_width-1 downto 0);
  bar23_page            : std_logic_vector(pcie_addr_width-1 downto 0);
  bar45_mask            : std_logic_vector(pcie_addr_width-1 downto 0);
  bar45_page            : std_logic_vector(pcie_addr_width-1 downto 0);
end record pcie_bar_mask_t;

type pcie_core_rd_o_t is record --V5 signal definitions as per datasheet (hence _n for act low)
  trn_rsof_n            : std_logic;
  trn_reof_n            : std_logic;
  trn_rd                : std_logic_vector(pcie_core_width-1 downto 0);
  trn_rrem_n            : std_logic_vector((pcie_core_width/8)-1 downto 0);
  trn_rerrfwd_n         : std_logic;
  trn_rsrc_rdy_n        : std_logic;
  trn_rsrc_dsc_n        : std_logic;
  trn_rbar_hit_n        : std_logic_vector(6 downto 0);
end record pcie_core_rd_o_t;

type pcie_core_rd_i_t is record --V5 signal definitions as per datasheet (hence _n for act low)
  trn_rdst_rdy_n        : std_logic;
  trn_rdst_dsc_n        : std_logic;
  dbg_rdst_rdy_n_sticky : std_Logic;
  dbg_rdst_fifo_afull : std_Logic;
  dbg_rdst_fifo_afull_sticky : std_Logic;
end record pcie_core_rd_i_t;

type pcie_core_td_o_t is record --V5 signal definitions as per datasheet (hence _n for act low)
  trn_tdst_rdy_n        : std_logic;
  trn_tdst_dsc_n        : std_logic;
  trn_tbuf_av           : std_logic_vector(5 downto 0);
  ltssm_state           : std_logic_vector(5 downto 0);
end record pcie_core_td_o_t;

type pcie_core_td_i_t is record --V5 signal definitions as per datasheet (hence _n for act low)
  trn_tsof_n            : std_logic;
  trn_teof_n            : std_logic;
  trn_td                : std_logic_vector(pcie_core_width-1 downto 0);
  trn_trem_n            : std_logic_vector((pcie_core_width/8)-1 downto 0);
  trn_tsrc_rdy_n        : std_logic;
  trn_tsrc_dsc_n        : std_logic;
  trn_rcpl_streaming_n  : std_logic;
end record pcie_core_td_i_t;

type pcie_core_status_t is record --V5 signal definitions as per datasheet (hence _n for act low)
  cfg_do                  : std_logic_vector(31 downto 0);
  cfg_rd_wr_done_n        : std_logic;
  cfg_interrupt_rdy_n     : std_logic;
  cfg_interrupt_mmenable  : std_logic_vector(2 downto 0);
  cfg_interrupt_msienable : std_logic;
  cfg_interrupt_do        : std_logic_vector(7 downto 0);
  cfg_to_turnoff_n        : std_logic;
  cfg_bus_number          : std_logic_vector(7 downto 0);
  cfg_device_number       : std_logic_vector(4 downto 0);
  cfg_function_number     : std_logic_vector(2 downto 0);
  cfg_status              : std_logic_vector(15 downto 0);
  cfg_command             : std_logic_vector(15 downto 0);
  cfg_dstatus             : std_logic_vector(15 downto 0);
  cfg_dcommand            : std_logic_vector(15 downto 0);
  cfg_lstatus             : std_logic_vector(15 downto 0);
  cfg_lcommand            : std_logic_vector(15 downto 0);
  cfg_pcie_link_state_n   : std_logic_vector(2 downto 0);
  cfg_err_cpl_rdy_n       : std_logic;
  cfg_debug               : std_logic_vector(63 downto 0);
end record pcie_core_status_t;

type pcie_core_id_t is record --bus/device/function
  cfg_bus_number          : std_logic_vector(7 downto 0);
  cfg_device_number       : std_logic_vector(4 downto 0);
  cfg_function_number     : std_logic_vector(2 downto 0);
end record pcie_core_id_t;

type pcie_core_ctrl_t is record --V5 signal definitions as per datasheet (hence _n for act low)
  cfg_di                  : std_logic_vector(31 downto 0);
  cfg_dwaddr              : std_logic_vector(9 downto 0);
  cfg_wr_en_n             : std_logic;
  cfg_rd_en_n             : std_logic;
  cfg_interrupt_n         : std_logic;
  cfg_interrupt_di        : std_logic_vector(7 downto 0);
  cfg_interrupt_assert_n  : std_logic;
  cfg_byte_en_n           : std_logic_vector(3 downto 0);
  cfg_pm_wake_n           : std_logic;
  cfg_trn_pending_n       : std_logic;
  cfg_dsn                 : std_logic_vector(63 downto 0);
  pl_directed_link_auton  : std_logic;
  pl_directed_link_change : std_logic_vector(1 downto 0);
  pl_directed_link_speed  : std_logic;
  pl_directed_link_width  : std_logic_vector(1 downto 0);
  pl_upstream_prefer_deemph  : std_logic;
end record pcie_core_ctrl_t;

type pcie_core_err_status_t is record --V5 signal definitions as per datasheet (hence _n for act low)
  cfg_err_cpl_rdy_n       : std_logic;
end record pcie_core_err_status_t;

type pcie_core_err_ctrl_t is record --V5 signal definitions as per datasheet (hence _n for act low)
  cfg_err_ecrc_n          : std_logic;
  cfg_err_ur_n            : std_logic;
  cfg_err_cpl_timeout_n   : std_logic;
  cfg_err_cpl_unexpect_n  : std_logic;
  cfg_err_cpl_abort_n     : std_logic;
  cfg_err_posted_n        : std_logic;
  cfg_err_cor_n           : std_logic;
  cfg_err_tlp_cpl_header  : std_logic_vector(47 downto 0);
  cfg_err_locked_n        : std_logic;
end record pcie_core_err_ctrl_t;

type pcie_reg_ctrl_t is record
  tx_pkt_size             : std_logic_vector(9 downto 0);
  rx_pkt_size             : std_logic_vector(9 downto 0);
  rx_pkts_in_flight       : std_logic_vector(3 downto 0);
  total_tags_inflight     : std_logic_vector(5 downto 0);
  interrupt_en            : std_logic;
  dma_to_clears           : std_logic_vector(no_slaves-1 downto 0);
  pl_directed_link_auton  : std_logic;
  pl_directed_link_change : std_logic_vector(1 downto 0);
  pl_directed_link_speed  : std_logic;
  pl_directed_link_width  : std_logic_vector(1 downto 0);
  pl_upstream_prefer_deemph : std_logic;
end record pcie_reg_ctrl_t;

type pcie_reg_status_t is record
  maxpayloadsize          : std_logic_vector(2 downto 0);
  act_link_width          : std_logic_vector(5 downto 0);
  link_speed              : std_logic_vector(1 downto 0);
  link_up                 : std_logic;
  extended_tag            : std_logic;
  interrupt_en            : std_logic;
  bus_master_en           : std_logic;
  maxreadreqsize          : std_logic_vector(2 downto 0);
  cfg_bus_number          : std_logic_vector(7 downto 0);
  cfg_device_number       : std_logic_vector(4 downto 0);
  trn_fc_info             : std_logic_vector(63 downto 0);
end record pcie_reg_status_t;

type pcie_bridge_status_t is record
  link_up_n               : std_logic;
  rx_bar_not_full         : std_logic_vector(3 downto 0);
  rx_cpld_not_full        : std_logic_vector(4 downto 0);

  tx_fsm                  : std_logic_vector(2 downto 0);
  tx_tags_active          : std_logic_vector(4 downto 0);


  interrupt               : std_logic;

  trn_tx_debug            : std_logic_vector(63 downto 0);
  trn_rx_debug            : std_Logic_vector(63 downto 0);
  dma_to_errors           : std_logic_vector(no_slaves-1 downto 0);
  debug                   : std_logic_vector(63 downto 0);
end record pcie_bridge_status_t;


type pcie_fifo_out_data_t is record
  data                    : pcie_data_t;
  data_valid              : std_logic;
  eop                     : std_logic;
  pkt_available           : std_logic;
end record pcie_fifo_out_data_t;

type pcie_fifo_out_ctrl_t is record
  pop                     : std_logic;
end record pcie_fifo_out_ctrl_t;

type pcie_fifo_in_data_t is record
  data                    : pcie_data_t;
  push                    : std_logic;
  eop                     : std_logic;
end record pcie_fifo_in_data_t;

type pcie_fifo_in_ctrl_t is record
  status                  : std_logic;
end record pcie_fifo_in_ctrl_t;


-- -----------------------------------------------------------------------------
-- FSM State declarations
-- -----------------------------------------------------------------------------

type PCIE_M_FSM is (M_FSM_IDLE, M_FSM_HEADER, M_FSM_HEADER2, M_FSM_READ1,M_FSM_READ2,M_FSM_READ3,M_FSM_WRITE1,M_FSM_WRITE2,M_FSM_WRITE3,M_FSM_WRITE4,M_FSM_WRITE5,M_FSM_END,M_FSM_FLUSH_CMD);
type PCIE_S_FSM is (S_FSM_IDLE, S_FSM_HEADER1, S_FSM_HEADER2, S_FSM_HEADER3, S_FSM_HEADER4, S_FSM_READ1,S_FSM_READ2,S_FSM_READ3,
     S_FSM_ERROR_RESP1, S_FSM_ERROR_RESP2, S_FSM_ERROR_RESP3, S_FSM_ERROR_RESP4, S_FSM_ERROR_RESP5);
type o2p_slave_fsm_t is( SLAVE_IDLE,  SLAVE_CMD_WAIT,SLAVE_CMD,SLAVE_SEND_READ,SLAVE_SEND_WRITE1,SLAVE_SEND_WRITE2, SLAVE_SEND_WRITE3, SLAVE_SEND_WRITE4, SLAVE_SEND_WRITE5);

type CPLD_S_FSM_T is (S_FSM_IDLE, S_FSM_CMD, S_FSM_TAG1, S_FSM_TAG2, S_FSM_SEND_DATA1, S_FSM_SEND_DATA2, S_FSM_SEND_DATA3);

type TD_FSM_T is (FSM_IDLE, FSM_HDR, FSM_HDR2, FSM_HDR3, FSM_HDR4, FSM_STRIPE1, FSM_STRIPE2, FSM_END);

component pcie_core_wrapper
  port(
    sys_reset_n       : in      std_logic;
    sys_clk           : in      std_logic;
    refclkout         : out     std_logic;

    -- physical PCIe I/O
    pci_exp_txn       : out     std_logic_vector(pcie_no_lanes-1 downto 0);
    pci_exp_txp       : out     std_logic_vector(pcie_no_lanes-1 downto 0);
    pci_exp_rxn       : in      std_logic_vector(pcie_no_lanes-1 downto 0);
    pci_exp_rxp       : in      std_logic_vector(pcie_no_lanes-1 downto 0);

    -- Transport Layer Packet Interface
    trn_clk           : out     std_logic;
    trn_reset_n       : out     std_logic;

    pcie_core_td_o    : out     pcie_core_td_o_t;
    pcie_core_td_i    : in      pcie_core_td_i_t;
    pcie_core_rd_o    : out     pcie_core_rd_o_t;
    pcie_core_rd_i    : in      pcie_core_rd_i_t;

    -- configuration Interface
    pcie_core_ctrl    : in      pcie_core_ctrl_t;
    pcie_core_status  : out     pcie_core_status_t;

    -- error interface
    pcie_core_err_status  : out pcie_core_err_status_t;
    pcie_core_err_ctrl    : in  pcie_core_err_ctrl_t;
    pcie_reg_status       : out pcie_reg_status_t);
end component;

component pcie_core_wrapper2
  port(
    sys_reset_n       : in      std_logic;
    sys_clk           : in      std_logic;
    refclkout         : out     std_logic;

    -- physical PCIe I/O
    pci_exp_txn       : out     std_logic_vector(3 downto 0);
    pci_exp_txp       : out     std_logic_vector(3 downto 0);
    pci_exp_rxn       : in      std_logic_vector(3 downto 0);
    pci_exp_rxp       : in      std_logic_vector(3 downto 0);

    -- Transport Layer Packet Interface
    trn_clk           : out     std_logic;
    trn_reset_n       : out     std_logic;

    pcie_core_td_o    : out     pcie_core_td_o_t;
    pcie_core_td_i    : in      pcie_core_td_i_t;
    pcie_core_rd_o    : out     pcie_core_rd_o_t;
    pcie_core_rd_i    : in      pcie_core_rd_i_t;

    -- configuration Interface
    pcie_core_ctrl    : in      pcie_core_ctrl_t;
    pcie_core_status  : out     pcie_core_status_t;

    -- error interface
    pcie_core_err_status  : out pcie_core_err_status_t;
    pcie_core_err_ctrl    : in  pcie_core_err_ctrl_t;
    pcie_reg_status       : out pcie_reg_status_t);
end component;

component endpoint_blk_plus_v1_9

  port(
    refclkout                   : out std_logic;
    trn_rcpl_streaming_n        : in  std_logic;
    cfg_to_turnoff_n            : out std_logic;
    trn_rdst_rdy_n              : in  std_logic;
    trn_tdst_dsc_n              : out std_logic;
    cfg_err_posted_n            : in  std_logic;
    trn_tdst_rdy_n              : out std_logic;
    cfg_trn_pending_n           : in  std_logic;
    cfg_err_cpl_timeout_n       : in  std_logic;
    trn_terrfwd_n               : in  std_logic;
    trn_reof_n                  : out std_logic;
    trn_rnp_ok_n                : in  std_logic;
    cfg_interrupt_n             : in  std_logic;
    cfg_interrupt_rdy_n         : out std_logic;
    cfg_rd_en_n                 : in  std_logic;
    cfg_err_ur_n                : in  std_logic;
    trn_reset_n                 : out std_logic;
    cfg_err_ecrc_n              : in  std_logic;
    cfg_err_cpl_abort_n         : in  std_logic;
    trn_rsrc_dsc_n              : out std_logic;
    trn_clk                     : out std_logic;
    cfg_err_locked_n            : in  std_logic;
    cfg_interrupt_assert_n      : in  std_logic;
    trn_rerrfwd_n               : out std_logic;
    cfg_wr_en_n                 : in  std_logic;
    trn_lnk_up_n                : out std_logic;
    cfg_interrupt_msienable     : out std_logic;
    trn_teof_n                  : in  std_logic;
    trn_rsof_n                  : out std_logic;
    cfg_rd_wr_done_n            : out std_logic;
    trn_tsrc_dsc_n              : in  std_logic;
    trn_rsrc_rdy_n              : out std_logic;
    fast_train_simulation_only  : in  std_logic;
    cfg_err_cor_n               : in  std_logic;
    sys_reset_n                 : in  std_logic;
    cfg_err_cpl_rdy_n           : out std_logic;
    trn_tsrc_rdy_n              : in  std_logic;
    sys_clk                     : in  std_logic;
    cfg_err_cpl_unexpect_n      : in  std_logic;
    trn_tsof_n                  : in  std_logic;
    cfg_pm_wake_n               : in  std_logic;
    cfg_byte_en_n               : in  std_logic_vector ( 3 downto 0 );
    cfg_interrupt_do            : out std_logic_vector ( 7 downto 0 );
    trn_rfc_ph_av               : out std_logic_vector ( 7 downto 0 );
    trn_rd                      : out std_logic_vector ( 63 downto 0 );
    trn_td                      : in  std_logic_vector ( 63 downto 0 );
    cfg_err_tlp_cpl_header      : in  std_logic_vector ( 47 downto 0 );
    trn_rbar_hit_n              : out std_logic_vector ( 6 downto 0 );
    cfg_dstatus                 : out std_logic_vector ( 15 downto 0 );
    cfg_lcommand                : out std_logic_vector ( 15 downto 0 );
    cfg_dsn                     : in  std_logic_vector ( 63 downto 0 );
    trn_rrem_n                  : out std_logic_vector ( 7 downto 0 );
    cfg_status                  : out std_logic_vector ( 15 downto 0 );
    cfg_interrupt_mmenable      : out std_logic_vector ( 2 downto 0 );
    pci_exp_txn                 : out std_logic_vector ( 7 downto 0 );
    cfg_command                 : out std_logic_vector ( 15 downto 0 );
    pci_exp_txp                 : out std_logic_vector ( 7 downto 0 );
    trn_trem_n                  : in  std_logic_vector ( 7 downto 0 );
    cfg_di                      : in  std_logic_vector ( 31 downto 0 );
    cfg_do                      : out std_logic_vector ( 31 downto 0 );
    cfg_bus_number              : out std_logic_vector ( 7 downto 0 );
    cfg_device_number           : out std_logic_vector ( 4 downto 0 );
    cfg_dwaddr                  : in  std_logic_vector ( 9 downto 0 );
    trn_rfc_npd_av              : out std_logic_vector ( 11 downto 0 );
    cfg_dcommand                : out std_logic_vector ( 15 downto 0 );
    trn_rfc_nph_av              : out std_logic_vector ( 7 downto 0 );
    cfg_pcie_link_state_n       : out std_logic_vector ( 2 downto 0 );
    cfg_function_number         : out std_logic_vector ( 2 downto 0 );
    cfg_lstatus                 : out std_logic_vector ( 15 downto 0 );
    pci_exp_rxn                 : in  std_logic_vector ( 7 downto 0 );
    pci_exp_rxp                 : in  std_logic_vector ( 7 downto 0 );
    trn_rfc_pd_av               : out std_logic_vector ( 11 downto 0 );
    trn_tbuf_av                 : out std_logic_vector ( 3 downto 0 );
    cfg_interrupt_di            : in  std_logic_vector ( 7 downto 0 ));
end component;

component endpoint_blk_plus_v1_11
  port (
  pci_exp_txp                   : out std_logic_vector(7 downto 0);
  pci_exp_txn                   : out std_logic_vector(7 downto 0);
  pci_exp_rxp                   : in  std_logic_vector(7 downto 0);
  pci_exp_rxn                   : in  std_logic_vector(7 downto 0);
  trn_clk                       : out std_logic;
  trn_reset_n                   : out std_logic;
  trn_lnk_up_n                  : out std_logic;
  trn_td                        : in  std_logic_vector(63 downto 0);
  trn_trem_n                    : in  std_logic_vector(7 downto 0);
  trn_tsof_n                    : in  std_logic;
  trn_teof_n                    : in  std_logic;
  trn_tsrc_rdy_n                : in  std_logic;
  trn_tdst_rdy_n                : out std_logic;
  trn_tdst_dsc_n                : out std_logic;
  trn_tsrc_dsc_n                : in  std_logic;
  trn_terrfwd_n                 : in  std_logic;
  trn_tbuf_av                   : out std_logic_vector(3 downto 0);
  trn_rd                        : out std_logic_vector(63 downto 0);
  trn_rrem_n                    : out std_logic_vector(7 downto 0);
  trn_rsof_n                    : out std_logic;
  trn_reof_n                    : out std_logic;
  trn_rsrc_rdy_n                : out std_logic;
  trn_rsrc_dsc_n                : out std_logic;
  trn_rdst_rdy_n                : in  std_logic;
  trn_rerrfwd_n                 : out std_logic;
  trn_rnp_ok_n                  : in  std_logic;
  trn_rbar_hit_n                : out std_logic_vector(6 downto 0);
  trn_rfc_nph_av                : out std_logic_vector(7 downto 0);
  trn_rfc_npd_av                : out std_logic_vector(11 downto 0);
  trn_rfc_ph_av                 : out std_logic_vector(7 downto 0);
  trn_rfc_pd_av                 : out std_logic_vector(11 downto 0);
  trn_rcpl_streaming_n          : in  std_logic;
  cfg_do                        : out std_logic_vector(31 downto 0);
  cfg_rd_wr_done_n              : out std_logic;
  cfg_di                        : in  std_logic_vector(31 downto 0);
  cfg_byte_en_n                 : in  std_logic_vector(3 downto 0);
  cfg_dwaddr                    : in  std_logic_vector(9 downto 0);
  cfg_wr_en_n                   : in  std_logic;
  cfg_rd_en_n                   : in  std_logic;
  cfg_err_cor_n                 : in  std_logic;
  cfg_err_ur_n                  : in  std_logic;
  cfg_err_ecrc_n                : in  std_logic;
  cfg_err_cpl_timeout_n         : in  std_logic;
  cfg_err_cpl_abort_n           : in  std_logic;
  cfg_err_cpl_unexpect_n        : in  std_logic;
  cfg_err_posted_n              : in  std_logic;
  cfg_err_locked_n              : in  std_logic;
  cfg_err_tlp_cpl_header        : in  std_logic_vector(47 downto 0);
  cfg_err_cpl_rdy_n             : out std_logic;
  cfg_interrupt_n               : in  std_logic;
  cfg_interrupt_rdy_n           : out std_logic;
  cfg_interrupt_assert_n        : in  std_logic;
  cfg_interrupt_di              : in  std_logic_vector(7 downto 0);
  cfg_interrupt_do              : out std_logic_vector(7 downto 0);
  cfg_interrupt_mmenable        : out std_logic_vector(2 downto 0);
  cfg_interrupt_msienable       : out std_logic;
  cfg_to_turnoff_n              : out std_logic;
  cfg_pm_wake_n                 : in  std_logic;
  cfg_pcie_link_state_n         : out std_logic_vector(2 downto 0);
  cfg_trn_pending_n             : in  std_logic;
  cfg_bus_number                : out std_logic_vector(7 downto 0);
  cfg_device_number             : out std_logic_vector(4 downto 0);
  cfg_function_number           : out std_logic_vector(2 downto 0);
  cfg_dsn                       : in  std_logic_vector(63 downto 0);
  cfg_status                    : out std_logic_vector(15 downto 0);
  cfg_command                   : out std_logic_vector(15 downto 0);
  cfg_dstatus                   : out std_logic_vector(15 downto 0);
  cfg_dcommand                  : out std_logic_vector(15 downto 0);
  cfg_lstatus                   : out std_logic_vector(15 downto 0);
  cfg_lcommand                  : out std_logic_vector(15 downto 0);
  fast_train_simulation_only    : in  std_logic;
  sys_clk                       : in  std_logic;
  refclkout                     : out std_logic;
  sys_reset_n                   : in  std_logic);
end component;

component endpoint_blk_plus_v1_11_4lane
  port (
  pci_exp_txp                   : out std_logic_vector(3 downto 0);
  pci_exp_txn                   : out std_logic_vector(3 downto 0);
  pci_exp_rxp                   : in  std_logic_vector(3 downto 0);
  pci_exp_rxn                   : in  std_logic_vector(3 downto 0);
  trn_clk                       : out std_logic;
  trn_reset_n                   : out std_logic;
  trn_lnk_up_n                  : out std_logic;
  trn_td                        : in  std_logic_vector(63 downto 0);
  trn_trem_n                    : in  std_logic_vector(7 downto 0);
  trn_tsof_n                    : in  std_logic;
  trn_teof_n                    : in  std_logic;
  trn_tsrc_rdy_n                : in  std_logic;
  trn_tdst_rdy_n                : out std_logic;
  trn_tdst_dsc_n                : out std_logic;
  trn_tsrc_dsc_n                : in  std_logic;
  trn_terrfwd_n                 : in  std_logic;
  trn_tbuf_av                   : out std_logic_vector(3 downto 0);
  trn_rd                        : out std_logic_vector(63 downto 0);
  trn_rrem_n                    : out std_logic_vector(7 downto 0);
  trn_rsof_n                    : out std_logic;
  trn_reof_n                    : out std_logic;
  trn_rsrc_rdy_n                : out std_logic;
  trn_rsrc_dsc_n                : out std_logic;
  trn_rdst_rdy_n                : in  std_logic;
  trn_rerrfwd_n                 : out std_logic;
  trn_rnp_ok_n                  : in  std_logic;
  trn_rbar_hit_n                : out std_logic_vector(6 downto 0);
  trn_rfc_nph_av                : out std_logic_vector(7 downto 0);
  trn_rfc_npd_av                : out std_logic_vector(11 downto 0);
  trn_rfc_ph_av                 : out std_logic_vector(7 downto 0);
  trn_rfc_pd_av                 : out std_logic_vector(11 downto 0);
  trn_rcpl_streaming_n          : in  std_logic;
  cfg_do                        : out std_logic_vector(31 downto 0);
  cfg_rd_wr_done_n              : out std_logic;
  cfg_di                        : in  std_logic_vector(31 downto 0);
  cfg_byte_en_n                 : in  std_logic_vector(3 downto 0);
  cfg_dwaddr                    : in  std_logic_vector(9 downto 0);
  cfg_wr_en_n                   : in  std_logic;
  cfg_rd_en_n                   : in  std_logic;
  cfg_err_cor_n                 : in  std_logic;
  cfg_err_ur_n                  : in  std_logic;
  cfg_err_ecrc_n                : in  std_logic;
  cfg_err_cpl_timeout_n         : in  std_logic;
  cfg_err_cpl_abort_n           : in  std_logic;
  cfg_err_cpl_unexpect_n        : in  std_logic;
  cfg_err_posted_n              : in  std_logic;
  cfg_err_locked_n              : in  std_logic;
  cfg_err_tlp_cpl_header        : in  std_logic_vector(47 downto 0);
  cfg_err_cpl_rdy_n             : out std_logic;
  cfg_interrupt_n               : in  std_logic;
  cfg_interrupt_rdy_n           : out std_logic;
  cfg_interrupt_assert_n        : in  std_logic;
  cfg_interrupt_di              : in  std_logic_vector(7 downto 0);
  cfg_interrupt_do              : out std_logic_vector(7 downto 0);
  cfg_interrupt_mmenable        : out std_logic_vector(2 downto 0);
  cfg_interrupt_msienable       : out std_logic;
  cfg_to_turnoff_n              : out std_logic;
  cfg_pm_wake_n                 : in  std_logic;
  cfg_pcie_link_state_n         : out std_logic_vector(2 downto 0);
  cfg_trn_pending_n             : in  std_logic;
  cfg_bus_number                : out std_logic_vector(7 downto 0);
  cfg_device_number             : out std_logic_vector(4 downto 0);
  cfg_function_number           : out std_logic_vector(2 downto 0);
  cfg_dsn                       : in  std_logic_vector(63 downto 0);
  cfg_status                    : out std_logic_vector(15 downto 0);
  cfg_command                   : out std_logic_vector(15 downto 0);
  cfg_dstatus                   : out std_logic_vector(15 downto 0);
  cfg_dcommand                  : out std_logic_vector(15 downto 0);
  cfg_lstatus                   : out std_logic_vector(15 downto 0);
  cfg_lcommand                  : out std_logic_vector(15 downto 0);
  fast_train_simulation_only    : in  std_logic;
  sys_clk                       : in  std_logic;
  refclkout                     : out std_logic;
  sys_reset_n                   : in  std_logic);
end component;

component endpoint_blk_plus_v1_14
  generic (
    REV_ID                      : bit_vector := X"00"
  );
  port (
  pci_exp_txp                   : out std_logic_vector(3 downto 0);
  pci_exp_txn                   : out std_logic_vector(3 downto 0);
  pci_exp_rxp                   : in  std_logic_vector(3 downto 0);
  pci_exp_rxn                   : in  std_logic_vector(3 downto 0);
  trn_clk                       : out std_logic;
  trn_reset_n                   : out std_logic;
  trn_lnk_up_n                  : out std_logic;
  trn_td                        : in  std_logic_vector(63 downto 0);
  trn_trem_n                    : in  std_logic_vector(7 downto 0);
  trn_tsof_n                    : in  std_logic;
  trn_teof_n                    : in  std_logic;
  trn_tsrc_rdy_n                : in  std_logic;
  trn_tdst_rdy_n                : out std_logic;
  trn_tdst_dsc_n                : out std_logic;
  trn_tsrc_dsc_n                : in  std_logic;
  trn_terrfwd_n                 : in  std_logic;
  trn_tbuf_av                   : out std_logic_vector(3 downto 0);
  trn_rd                        : out std_logic_vector(63 downto 0);
  trn_rrem_n                    : out std_logic_vector(7 downto 0);
  trn_rsof_n                    : out std_logic;
  trn_reof_n                    : out std_logic;
  trn_rsrc_rdy_n                : out std_logic;
  trn_rsrc_dsc_n                : out std_logic;
  trn_rdst_rdy_n                : in  std_logic;
  trn_rerrfwd_n                 : out std_logic;
  trn_rnp_ok_n                  : in  std_logic;
  trn_rbar_hit_n                : out std_logic_vector(6 downto 0);
  trn_rfc_nph_av                : out std_logic_vector(7 downto 0);
  trn_rfc_npd_av                : out std_logic_vector(11 downto 0);
  trn_rfc_ph_av                 : out std_logic_vector(7 downto 0);
  trn_rfc_pd_av                 : out std_logic_vector(11 downto 0);
  trn_rcpl_streaming_n          : in  std_logic;
  cfg_do                        : out std_logic_vector(31 downto 0);
  cfg_rd_wr_done_n              : out std_logic;
  cfg_di                        : in  std_logic_vector(31 downto 0);
  cfg_byte_en_n                 : in  std_logic_vector(3 downto 0);
  cfg_dwaddr                    : in  std_logic_vector(9 downto 0);
  cfg_wr_en_n                   : in  std_logic;
  cfg_rd_en_n                   : in  std_logic;
  cfg_err_cor_n                 : in  std_logic;
  cfg_err_ur_n                  : in  std_logic;
  cfg_err_ecrc_n                : in  std_logic;
  cfg_err_cpl_timeout_n         : in  std_logic;
  cfg_err_cpl_abort_n           : in  std_logic;
  cfg_err_cpl_unexpect_n        : in  std_logic;
  cfg_err_posted_n              : in  std_logic;
  cfg_err_locked_n              : in  std_logic;
  cfg_err_tlp_cpl_header        : in  std_logic_vector(47 downto 0);
  cfg_err_cpl_rdy_n             : out std_logic;
  cfg_interrupt_n               : in  std_logic;
  cfg_interrupt_rdy_n           : out std_logic;
  cfg_interrupt_assert_n        : in  std_logic;
  cfg_interrupt_di              : in  std_logic_vector(7 downto 0);
  cfg_interrupt_do              : out std_logic_vector(7 downto 0);
  cfg_interrupt_mmenable        : out std_logic_vector(2 downto 0);
  cfg_interrupt_msienable       : out std_logic;
  cfg_to_turnoff_n              : out std_logic;
  cfg_pm_wake_n                 : in  std_logic;
  cfg_pcie_link_state_n         : out std_logic_vector(2 downto 0);
  cfg_trn_pending_n             : in  std_logic;
  cfg_bus_number                : out std_logic_vector(7 downto 0);
  cfg_device_number             : out std_logic_vector(4 downto 0);
  cfg_function_number           : out std_logic_vector(2 downto 0);
  cfg_dsn                       : in  std_logic_vector(63 downto 0);
  cfg_status                    : out std_logic_vector(15 downto 0);
  cfg_command                   : out std_logic_vector(15 downto 0);
  cfg_dstatus                   : out std_logic_vector(15 downto 0);
  cfg_dcommand                  : out std_logic_vector(15 downto 0);
  cfg_lstatus                   : out std_logic_vector(15 downto 0);
  cfg_lcommand                  : out std_logic_vector(15 downto 0);
  fast_train_simulation_only    : in  std_logic;
  sys_clk                       : in  std_logic;
  refclkout                     : out std_logic;
  sys_reset_n                   : in  std_logic);
end component;


component pcie_pkt_fifo
  generic(
    fifo_depth          : integer;
    fifo_width          : integer;
    ram_style_def       : string  := "distributed");
  port(
    clk                 : in  std_logic;
--     ar                  : in  std_logic;
    sr                  : in  std_logic;
    max_pkt_size        : in  std_logic_vector(fifo_depth-1 downto 0);

    din                 : in  std_logic_vector(fifo_width-1 downto 0);
    dout                : out std_logic_vector(fifo_width-1 downto 0);

    push                : in  std_logic;
    pop                 : in  std_logic;

    full                : out std_logic;
    empty               : out std_logic;
    last                : out std_logic;
    free_pkt            : out std_logic;
    pkt_available       : out std_logic);
  end component;
component pcie_pkt_fifo_nofel
  generic(
    fifo_depth          : integer;
    fifo_width          : integer;
    ram_style_def       : string  := "distributed");
  port(
    clk                 : in  std_logic;
--     ar                  : in  std_logic;
    sr                  : in  std_logic;
    max_pkt_size        : in  std_logic_vector(fifo_depth-1 downto 0);

    din                 : in  std_logic_vector(fifo_width-1 downto 0);
    dout                : out std_logic_vector(fifo_width-1 downto 0);

    push                : in  std_logic;
    pop                 : in  std_logic;

--     full                : out std_logic;
--     empty               : out std_logic;
--     last                : out std_logic;
    free_pkt            : out std_logic;
    pkt_available       : out std_logic);
  end component;

component pcie_pkt_fifo_noflpa
  generic(
    fifo_depth          : integer;
    fifo_width          : integer;
    ram_style_def       : string  := "distributed");
  port(
    clk                 : in  std_logic;
--     ar                  : in  std_logic;
    sr                  : in  std_logic;
    max_pkt_size        : in  std_logic_vector(fifo_depth-1 downto 0);

    din                 : in  std_logic_vector(fifo_width-1 downto 0);
    dout                : out std_logic_vector(fifo_width-1 downto 0);

    push                : in  std_logic;
    pop                 : in  std_logic;

--     full                : out std_logic;
    empty               : out std_logic;
--     last                : out std_logic;
    free_pkt            : out std_logic);
--     pkt_available       : out std_logic);
  end component;
component pcie_pkt_fifo_nofppa
  generic(
    fifo_depth          : integer;
    fifo_width          : integer;
    ram_style_def       : string  := "distributed");
  port(
    clk                 : in  std_logic;
--     ar                  : in  std_logic;
    sr                  : in  std_logic;
--     max_pkt_size        : in  std_logic_vector(fifo_depth-1 downto 0);

    din                 : in  std_logic_vector(fifo_width-1 downto 0);
    dout                : out std_logic_vector(fifo_width-1 downto 0);

    push                : in  std_logic;
    pop                 : in  std_logic;

    full                : out std_logic;
    empty               : out std_logic;
    last                : out std_logic);
--     free_pkt            : out std_logic);
--     pkt_available       : out std_logic);
  end component;

component pcie_master_ocp
--   generic(
--     low_tag_value       : integer;
--     high_tag_value      : integer);
  port(
    clk                 : in  std_logic;
--     ar                  : in  std_logic;
    sr                  : in  std_logic;
    us_strobe           : in  std_logic;

    core_status         : in  pcie_core_id_t;

    m2s_o               : out adb3_ocp_m2sT;
    s2m_i               : in  adb3_ocp_s2mT;
    tag_timeout_error     : out std_logic;
--    tag_timeout_status    : out std_logic_vector(4 downto 0);
    clear_to_error        : in  std_logic;

    -- read/write pkts for conversion to OCP
    master_p2o_data     : in  std_logic_vector(pcie_data_size-1 downto 0);
    master_p2o_push     : in  std_logic;
    master_p2o_notfull  : out std_logic;
    --CPLD packets for transfer over PCIe
    master_o2p_data     : out std_logic_vector(pcie_data_size-1 downto 0);
    master_o2p_push     : out std_logic;
--     master_o2p_eop      : out std_logic;
    master_o2p_notfull  : in  std_logic;
    address_mask        : in  std_logic_vector(63 downto 0);
    address_page        : in  std_logic_vector(63 downto 0);
    debug               : out std_Logic_vector(63 downto 0));
  end component;

component pcie_master_ocp_32
--   generic(
--     low_tag_value       : integer;
--     high_tag_value      : integer);
  port(
    clk                 : in  std_logic;
--     ar                  : in  std_logic;
    sr                  : in  std_logic;
    us_counter          : in  std_logic;

    core_status         : in  pcie_core_id_t;

    m2s_o               : out adb3_ocp_m2sT;
    s2m_i               : in  adb3_ocp_s2mT;
    tag_timeout_error     : out std_logic;
    clear_to_error        : in  std_logic;

    -- read/write pkts for conversion to OCP
    master_p2o_data     : in  std_logic_vector(pcie_data_size-1 downto 0);
    master_p2o_push     : in  std_logic;
    master_p2o_notfull  : out std_logic;
    --CPLD packets for transfer over PCIe
    master_o2p_data     : out std_logic_vector(pcie_data_size-1 downto 0);
    master_o2p_push     : out std_logic;
--     master_o2p_eop      : out std_logic;
    master_o2p_notfull  : in  std_logic;
    address_mask        : in  std_logic_vector(31 downto 0);
    address_page        : in  std_logic_vector(31 downto 0);
    debug               : out std_Logic_vector(63 downto 0));

end component;

-- component pcie_master_m2o_gen
--   port(
--     clk                 : in  std_logic;
--     ar                  : in  std_logic;
--     sr                  : in  std_logic;

--     core_status         : in  pcie_core_status_t;

--     rx_data_in          : in  std_logic_vector(127 downto 0);
--     rx_be_in            : in  std_logic_vector(3 downto 0);
--     rx_tag_in           : in  std_logic_vector(7 downto 0);
--     rx_data_pop         : out std_logic;
--     rx_fifo_empty       : in  std_logic;
--     rx_fifo_last        : in  std_logic;

--     tag_return          : out std_logic;
--     rx_tag_value        : out std_logic_vector(7 downto 0);
--     tag_info            : in  std_logic_vector(127 downto 0);


--     --CPLD packets for transfer over PCIe
--     master_o2p_data     : out std_logic_vector(pcie_data_size-1 downto 0);
--     master_o2p_push     : out std_logic;
--     master_o2p_eop      : out std_logic;
--     master_o2p_notfull     : in  std_logic);


-- end component;

component pcie_master_o2p_gen
  port(
    clk                 : in  std_logic;
--     ar                  : in  std_logic;
    sr                  : in  std_logic;

    core_status         : in  pcie_core_id_t;

    rx_data_in          : in  std_logic_vector(127 downto 0);
--     rx_be_in            : in  std_logic_vector(3 downto 0);
    rx_tag_in           : in  std_logic_vector(2 downto 0);
    rx_data_pop         : out std_logic;
    rx_fifo_empty       : in  std_logic;
    rx_fifo_last        : in  std_logic;

    tag_timed_out       : in  std_logic;
    timeout_tag_value   : in  std_logic_vector(48 downto 0);
    timeout_error_clr   : out std_logic;
    tag_return          : out std_logic;
    rx_tag_value        : out std_logic_vector(2 downto 0);
    tag_info            : in  std_logic_vector(48 downto 0);
    tag_active          : in  std_logic;

    --CPLD packets for transfer over PCIe
    master_o2p_data     : out std_logic_vector(pcie_data_size-1 downto 0);
    master_o2p_push     : out std_logic;
--     master_o2p_eop      : out std_logic;
    master_o2p_notfull     : in  std_logic;
    debug : out std_logic_vector(9 downto 0));


end component;

component pcie_master_o2p_gen_32
  port(
    clk                 : in  std_logic;
--     ar                  : in  std_logic;
    sr                  : in  std_logic;

    core_status         : in  pcie_core_id_t;

    rx_data_in          : in  std_logic_vector(127 downto 0);
--     rx_be_in            : in  std_logic_vector(3 downto 0);
--     rx_tag_in           : in  std_logic_vector(7 downto 0);
    rx_data_pop         : out std_logic;
    rx_fifo_empty       : in  std_logic;
    rx_fifo_last        : in  std_logic;
    tag_timed_out       : in  std_logic;

    tag_return          : out std_logic;
    tag_info            : in  std_logic_vector(52 downto 0);
    tag_active          : in std_logic;
    --CPLD packets for transfer over PCIe
    master_o2p_data     : out std_logic_vector(pcie_data_size-1 downto 0);
    master_o2p_push     : out std_logic;
--     master_o2p_eop      : out std_logic;
    master_o2p_notfull     : in  std_logic;
    debug               : out std_Logic_vector(7 downto 0));


end component;

component pcie_master_p2o_gen
  port(
    clk                 : in  std_logic;
--     ar                  : in  std_logic;
    sr                  : in  std_logic;

--     core_status         : in  pcie_core_status_t;

    tag_request         : out std_logic;
 --   tag_request_valid   : in  std_logic;
    next_tag            : in  std_logic_vector(2 downto 0);
    tag_pcie_header     : out std_logic_vector(127 downto 0);

    -- read/write pkts for conversion to OCP
    tx_pkt_data         : in  std_logic_vector(pcie_data_size-1 downto 0);
    tx_pkt_pop          : out std_logic;
    tx_pkt_empty        : in  std_logic;

    block_pcie          : in  std_logic;
    timeout_error       : in  std_logic;

    cmd_fifo_data       : out std_logic_vector(80 downto 0);
    cmd_fifo_push       : out std_logic;
    cmd_fifo_notfull    : in  std_logic;
    data_fifo_data      : out std_logic_vector(143 downto 0);
    data_fifo_push      : out std_logic;
    data_fifo_notfull   : in  std_logic;
    address_mask        : in  std_logic_vector(63 downto 0);
    address_page        : in  std_logic_vector(63 downto 0));


end component;

component pcie_master_p2o_gen_32
  port(
    clk                 : in  std_logic;
--     ar                  : in  std_logic;
    sr                  : in  std_logic;

--     core_status         : in  pcie_core_status_t;

    tag_request         : out std_logic;
    next_tag            : in  std_logic_vector(2 downto 0);
    tag_pcie_header     : out std_logic_vector(127 downto 0);

    -- read/write pkts for conversion to OCP
    tx_pkt_data         : in  std_logic_vector(pcie_data_size-1 downto 0);
    tx_pkt_pop          : out std_logic;
    tx_pkt_empty        : in  std_logic;

    tag_timeout         : in  std_logic;

    cmd_fifo_data       : out std_logic_vector(43 downto 0);
    cmd_fifo_push       : out std_logic;
    cmd_fifo_notfull    : in  std_logic;
    data_fifo_data      : out std_logic_vector(143 downto 0);
    data_fifo_push      : out std_logic;
    data_fifo_notfull   : in  std_logic;
    address_mask        : in  std_logic_vector(31 downto 0);
    address_page        : in  std_logic_vector(31 downto 0));


end component;

component pcie_master_tag_mgr
  port(
    clk                 : in  std_logic;
--     ar                  : in  std_logic;
    sr                  : in  std_logic;
    us_strobe           : in  std_logic;

    tag_request         : in  std_logic;
--    tag_request_valid   : out std_logic;
    next_tag            : out std_logic_vector(2 downto 0);
    tag_pcie_header     : in  std_logic_vector(127 downto 0);

    tag_return          : in  std_logic;
    rx_tag_value        : in  std_logic_vector(2 downto 0);
    tag_info            : out std_logic_vector(48 downto 0);
    tag_active          : out std_logic;

    timeout_error       : out std_logic := '0';
    clear_to_error      : in  std_logic;
    block_pcie          : out std_logic := '0';


    timeout_error_clr   : in  std_logic := '0';
--    tag_timeout_status  : out std_logic_vector(4 downto 0);
    timeout_tag_value   : out std_logic_vector(48 downto 0);
    tag_timed_out       : out std_logic;
    tag_active_flags_dbg : out std_logic_vector(7 downto 0));
end component;

component pcie_master_tag_mgr_32
  port(
    clk                 : in  std_logic;
--     ar                  : in  std_logic;
    sr                  : in  std_logic;
    us_counter          : in  std_logic;

    tag_request         : in  std_logic;
    next_tag            : out std_logic_vector(2 downto 0);
    tag_pcie_header     : in  std_logic_vector(52 downto 0);

    tag_return          : in  std_logic;
    tag_info            : out std_logic_vector(52 downto 0);
    tag_active          : out std_logic;
    tag_timed_out       : out std_logic);
end component;

component pcie_slave_ocp
  generic(low_tag : std_logic_vector(7 downto 0) := X"00";
          hi_tag  : std_logic_vector(7 downto 0) := X"1F");
  port(
    clk                 : in  std_logic;
--     ar                  : in  std_logic;
    sr                  : in  std_logic;
    us_counter          : in  std_logic_vector(9 downto 0);
    core_status         : in  pcie_core_id_t;
    tag_used            : out std_logic;
    tag_available       : in std_logic;
    tag_clear           : out std_logic;
    to_clear            : in std_logic;
    to_error            : out std_logic;
    max_pif             : in  std_logic_vector(3 downto 0);
    m2s_i               : in  adb3_ocp_m2sT;
    s2m_o               : out adb3_ocp_s2mT;

    slave_o2p_push      : out std_logic;
    slave_o2p_data      : out std_logic_vector(pcie_data_size-1 downto 0);
    slave_o2p_notfull   : in  std_logic;

    slave_p2o_push      : in  std_logic;
    slave_p2o_data      : in  std_logic_vector(pcie_data_size-1 downto 0);
    slave_p2o_notfull   : out std_logic;
    tag_active          : out std_logic;--);
     debug                   : out std_Logic_vector(15 downto 0));

end component;

component pcie_trn_rx_if
  generic(
    async_mode          : in boolean);
  port(
    pcie_clk                 : in  std_logic;
    bridge_clk                 : in  std_logic;
    ar                  : in  std_logic;
--     sr                  : in  std_logic;

    pcie_core_o         : in  pcie_core_rd_o_t;
    pcie_core_i         : out pcie_core_rd_i_t;

    rx_fifo_data        : out std_logic_vector(pcie_data_size-1 downto 0);
    bar_fifo_push       : out std_logic_vector(no_masters-1 downto 0);
    bar_fifo_notfull    : in  std_logic_vector(no_masters-1 downto 0);
    cpld_fifo_push      : out std_logic_vector(no_slaves-1 downto 0);
    cpld_fifo_notfull   : in  std_logic_vector(no_slaves-1 downto 0);

--     core_status         : in  pcie_core_status_t;

    rx_errors           : out std_logic_vector(15 downto 0);
    debug               : out std_logic_vector(63 downto 0));
end component;

component pcie_trn_rx_if2
  port(
    clk                 : in  std_logic;
    ar                  : in  std_logic;
    sr                  : in  std_logic;

    pcie_core_o         : in  pcie_core_rd_o_t;
    pcie_core_i         : out pcie_core_rd_i_t;

    bar0_fifo_data      : out std_logic_vector(pcie_data_size-1 downto 0);
    bar0_fifo_push      : out std_logic;
    bar0_fifo_notfull   : in  std_logic;
    bar1_fifo_data      : out std_logic_vector(pcie_data_size-1 downto 0);
    bar1_fifo_push      : out std_logic;
    bar1_fifo_notfull   : in  std_logic;
    bar23_fifo_data     : out std_logic_vector(pcie_data_size-1 downto 0);
    bar23_fifo_push     : out std_logic;
    bar23_fifo_notfull  : in  std_logic;
    cpld_fifo_data      : out std_logic_vector(pcie_data_size-1 downto 0);
    cpld_fifo_push      : out std_logic_vector(4 downto 0);
    cpld_fifo_notfull   : in  std_logic_vector(4 downto 0);



    core_status         : in  pcie_core_status_t;

    rx_errors           : out std_logic_vector(15 downto 0);
    debug               : out std_logic_vector(63 downto 0));
end component;

component pcie_trn_tx_if is
  generic(
    async_mode          : in boolean);
  port(
    bridge_clk          : in  std_logic;
    pcie_clk            : in  std_logic;
    ar                  : in  std_logic;
    sr                  : in  std_logic;
    pcie_core_o         : in  pcie_core_td_o_t;
    pcie_core_i         : out pcie_core_td_i_t;

    bar_cpld_data       : in  bar_data_array;
    bar_cpld_push       : in  std_logic_vector(no_masters-1 downto 0);
    bar_cpld_notfull    : out std_logic_vector(no_masters-1 downto 0);

    slave_fifo_data     : in  slave_data_array;
    slave_fifo_push     : in  std_logic_vector(no_slaves-1 downto 0);
    slave_fifo_notfull  : out std_logic_vector(no_slaves-1 downto 0);

--     core_status         : in  pcie_core_status_t;
    busmaster_en        : in  std_logic;
    link_up_n           : in  std_logic;

    tx_errors           : out std_logic_vector(15 downto 0);
    debug               : out std_logic_vector(31 downto 0));
end component;

component pcie_trn_tx_if2 is
  port(
    clk                 : in  std_logic;
    ar                  : in  std_logic;
    sr                  : in  std_logic;

    pcie_core_o         : in  pcie_core_td_o_t;
    pcie_core_i         : out pcie_core_td_i_t;

    bar0_cpld_data      : in  std_logic_vector(pcie_data_size-1 downto 0);
    bar0_cpld_push      : in  std_logic;
    bar0_cpld_notfull      : out std_logic;
    bar1_cpld_data      : in  std_logic_vector(pcie_data_size-1 downto 0);
    bar1_cpld_push      : in  std_logic;
    bar1_cpld_notfull      : out std_logic;
    bar23_cpld_data     : in  std_logic_vector(pcie_data_size-1 downto 0);
    bar23_cpld_push     : in  std_logic;
    bar23_cpld_notfull     : out std_logic;

    slave0_fifo_data    : in  std_logic_vector(pcie_data_size-1 downto 0);
    slave0_fifo_push    : in  std_logic;
    slave0_fifo_notfull    : out std_logic;
    slave1_fifo_data    : in  std_logic_vector(pcie_data_size-1 downto 0);
    slave1_fifo_push    : in  std_logic;
    slave1_fifo_notfull    : out std_logic;



    core_status         : in  pcie_core_status_t;
    busmaster_en        : in  std_logic;
    link_up_n           : in  std_logic;


    temp_mux_data1      : out std_logic_vector(pcie_data_size-1 downto 0);
    temp_mux_valid1     : out std_logic;
    temp_mux_data2      : out std_logic_vector(pcie_data_size-1 downto 0);
    temp_mux_valid2     : out std_logic;
    tx_errors           : out std_logic_vector(15 downto 0);
    debug               : out std_logic_vector(31 downto 0));
end component;

component pcie_bridge is
  generic(
    async_mode                : in  boolean := TRUE);
  port( clk                   : in  std_logic;
        pcie_clk              : in  std_logic;
        ar                    : in  std_logic;
        sr                    : in  std_logic;

        -- PCIE Core interface
        pcie_core_td_o        : in  pcie_core_td_o_t;
        pcie_core_td_i        : out pcie_core_td_i_t;
        pcie_core_rd_o        : in  pcie_core_rd_o_t;
        pcie_core_rd_i        : out pcie_core_rd_i_t;
        pcie_core_ctrl        : out pcie_core_ctrl_t;
        pcie_core_status      : in  pcie_core_status_t;
        pcie_core_err_status  : in  pcie_core_err_status_t;
        pcie_core_err_ctrl    : out pcie_core_err_ctrl_t;
        max_pif               : in  std_logic_vector(3 downto 0);

        tag_timeout_error     : out std_logic_vector(no_masters-1 downto 0);
--        tag_timeout_status    : out std_logic_vector(4 downto 0);
        clear_to_error        : in  std_logic_vector(no_masters-1 downto 0);
        -- OCP Masters
        m_m2s_o               : out adb3_ocp_m2sT_array(0 to no_masters-1);
        m_s2m_i               : in  adb3_ocp_s2mT_array(0 to no_masters-1);

        -- OCP Slaves
        s_m2s_i               : in  adb3_ocp_m2sT_array(0 to no_slaves-1);
        s_s2m_o               : out adb3_ocp_s2mT_array(0 to no_slaves-1);

        interrupt             : in  std_logic;

        bar_masks             : in  pcie_bar_mask_t;

        -- status and control to/from register bank
        pcie_reg_ctrl         : in pcie_reg_ctrl_t;
        temp_mux_data1        : out std_logic_vector(pcie_data_size-1 downto 0);
        temp_mux_valid1       : out std_logic;
        temp_mux_data2        : out std_logic_vector(pcie_data_size-1 downto 0);
        temp_mux_valid2       : out std_logic;

        pcie_bridge_status    : out pcie_bridge_status_t;
        pcie_reg_status       : in  pcie_reg_status_t;
        cscope_bus            : out std_logic_vector(135 downto 0));
end component;

component pcie_bridge2 is
  port( clk                   : in  std_logic;
        ar                    : in  std_logic;
        sr                    : in  std_logic;

        -- PCIE Core interface
        pcie_core_td_o        : in  pcie_core_td_o_t;
        pcie_core_td_i        : out pcie_core_td_i_t;
        pcie_core_rd_o        : in  pcie_core_rd_o_t;
        pcie_core_rd_i        : out pcie_core_rd_i_t;
        pcie_core_ctrl        : out pcie_core_ctrl_t;
        pcie_core_status      : in  pcie_core_status_t;
        pcie_core_err_status  : in  pcie_core_err_status_t;
        pcie_core_err_ctrl    : out pcie_core_err_ctrl_t;

        -- OCP Masters
        m0_m2s_o              : out adb3_ocp_m2sT;
        m0_s2m_i              : in  adb3_ocp_s2mT;
        m1_m2s_o              : out adb3_ocp_m2sT;
        m1_s2m_i              : in  adb3_ocp_s2mT;
        m23_m2s_o             : out adb3_ocp_m2sT;
        m23_s2m_i             : in  adb3_ocp_s2mT;

        -- OCP Slaves
        s0_m2s_i              : in  adb3_ocp_m2sT;
        s0_s2m_o              : out adb3_ocp_s2mT;
        s1_m2s_i              : in  adb3_ocp_m2sT;
        s1_s2m_o              : out adb3_ocp_s2mT;

        interrupt             : in  std_logic;

        -- status and control to/from register bank
        pcie_reg_ctrl         : in  pcie_reg_ctrl_t;
        temp_mux_data1        : out std_logic_vector(pcie_data_size-1 downto 0);
        temp_mux_valid1       : out std_logic;
        temp_mux_data2        : out std_logic_vector(pcie_data_size-1 downto 0);
        temp_mux_valid2       : out std_logic;

        pcie_bridge_status    : out pcie_bridge_status_t;
        pcie_reg_status       : in  pcie_reg_status_t;
        bar_masks             : in  pcie_bar_mask_t;
        max_pif               : in  std_logic_vector(3 downto 0) := "0010";
        cscope_bus            : out std_logic_vector(135 downto 0));
end component;

component pcie_core_td_if is
  generic(
    async_mode   : in boolean);
  port(
    pcie_clk                 : in  std_logic;
    bridge_clk                 : in  std_logic;
--     ar                  : in  std_logic;
    ar                  : in  std_logic;

    tx_data             : in  std_logic_vector(pcie_data_size-1 downto 0);
    tx_pop              : out std_logic;
    tx_pkt_avail        : in  std_logic;

    pcie_core_o         : in  pcie_core_td_o_t;
    pcie_core_i         : out pcie_core_td_i_t;
    busmaster_en        : in  std_logic;
    link_up_n           : in  std_logic;
    debug               : out std_logic_vector(7 downto 0));
end component;

component pcie_irpt_ctrl is
   port(
      rst                   : in  std_logic;
      clk                   : in  std_logic;
      irpt_enable           : in  std_logic;
      irpt_msienable        : in  std_logic;
      irpt_rdy_n            : in  std_logic;
      pci_irpt_n            : out std_logic;
      irpt_assert_n         : out std_logic;
      interrupt_i           : in  std_logic
   );
end component;

component pcie_arbit_sequencer is
  port( clk                   : in  std_logic;
        pkt_avail             : in  comb_flags;
        source                : in  comb_count;
        next_source           : out comb_count);
end component;


end package pcie;


package body pcie is
end package body pcie;

