--------------------------------------------------------------------------------
--
-- Module Name      : axi4_lite_to_adb3_ocpl
--
--                    Copyright 2012 by
--                    Alpha Data Parallel Systems Ltd.
--                    All Rights Reserved.
--
-- Original Author  : Andrew McCormick
-- Created          : 
--
-- Modified By      : 
-- Date             :
-- Change Notes     :     
--
-- Description      : Package For Interfacing ADB3 OCP Lite Type Logic With
--                    AXI-4 Lite Compatible IP
--
--
--
-- Disclaimer       : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                    WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                    WARRANTIES IMPLIED OF MERCHANTABILITY, FITNESS FOR A
--                    PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;

entity axi4_lite_to_adb3_ocpl is
  port (
    -- Slave AXI Port
    S_AXI_ACLK                     : in  std_logic;
    S_AXI_ARESETN                  : in  std_logic;
    S_AXI_AWADDR                   : in  std_logic_vector(31 downto 0);
    S_AXI_AWVALID                  : in  std_logic;
    S_AXI_WDATA                    : in  std_logic_vector(31 downto 0);
    S_AXI_WSTRB                    : in  std_logic_vector(3 downto 0);
    S_AXI_WVALID                   : in  std_logic;
    S_AXI_BREADY                   : in  std_logic;
    S_AXI_ARADDR                   : in  std_logic_vector(31 downto 0);
    S_AXI_ARVALID                  : in  std_logic;
    S_AXI_RREADY                   : in  std_logic;
    S_AXI_ARREADY                  : out std_logic;
    S_AXI_RDATA                    : out std_logic_vector(31 downto 0);
    S_AXI_RRESP                    : out std_logic_vector(1 downto 0);
    S_AXI_RVALID                   : out std_logic;
    S_AXI_WREADY                   : out std_logic;
    S_AXI_BRESP                    : out std_logic_vector(1 downto 0);
    S_AXI_BVALID                   : out std_logic;
    S_AXI_AWREADY                  : out std_logic;
    -- OCP Lite Port (as std_logic_1164)
    ocpl_cmd                       : out std_logic_vector(2 downto 0);
    ocpl_addr                      : out std_logic_vector(31 downto 0);
    ocpl_m2sdata                   : out std_logic_vector(31 downto 0);
    ocpl_databyteen                : out std_logic_vector(3 downto 0);
    ocpl_datavalid                 : out std_logic;
    ocpl_respaccept                : out std_logic;
    ocpl_cmdaccept                 : in  std_logic;
    ocpl_dataaccept                : in  std_logic;
    ocpl_resp                      : in  std_logic_vector(1 downto 0);
    ocpl_s2mdata                   : in  std_logic_vector(31 downto 0));
end entity axi4_lite_to_adb3_ocpl;

architecture rtl of axi4_lite_to_adb3_ocpl is

  signal ocpl_cmd_i                       : std_logic_vector(2 downto 0);
  signal ocpl_addr_i                      : std_logic_vector(31 downto 0);
  signal ocpl_m2sdata_i                   : std_logic_vector(31 downto 0);
  signal ocpl_databyteen_i                : std_logic_vector(3 downto 0);

  signal axi_wcmd_latched : std_logic := '0';
  signal axi_rcmd_latched : std_logic := '0';
  signal axi_data_latched : std_logic := '0';
  signal bvalid_i : std_logic := '0';
  signal respvalid_i : std_logic := '0';

  signal axi_rcmd_busy: std_logic := '0';

  attribute mark_debug : string;
  attribute keep : string;

--attribute mark_debug of axi_wcmd_latched : signal is "true";
--  attribute keep of axi_wcmd_latched : signal is "true";
--  attribute mark_debug of axi_rcmd_latched : signal is "true";
--  attribute keep of axi_rcmd_latched : signal is "true";
--  attribute mark_debug of axi_data_latched : signal is "true";
--  attribute keep of axi_data_latched : signal is "true";
--  attribute mark_debug of axi_rcmd_busy : signal is "true";
--  attribute keep of axi_rcmd_busy : signal is "true";
--  attribute mark_debug of respvalid_i: signal is "true";
--  attribute keep of respvalid_i : signal is "true";

  
begin  -- architecture rtl

  latch_data_and_cmd: process (S_AXI_ACLK) is
  begin  -- process latch_data_and_cmd
    if rising_edge(S_AXI_ACLK) then      -- rising clock edge
      if S_AXI_ARESETN = '0' then
        axi_wcmd_latched <= '0';
        axi_rcmd_latched <= '0';
        axi_data_latched <= '0';
        axi_rcmd_busy <= '0';
      else
        if axi_wcmd_latched = '0' and axi_rcmd_latched = '0' and axi_rcmd_busy = '0' then
          if S_AXI_ARVALID = '1' then
            axi_rcmd_latched <= '1';
          elsif S_AXI_AWVALID = '1' then
            axi_wcmd_latched <= '1';
          end if;
        else
          if axi_rcmd_latched = '1' and ocpl_cmdaccept = '1' then
             axi_rcmd_latched <= '0';
             axi_rcmd_busy <= '1';
          end if;
          if axi_wcmd_latched = '1' and axi_data_latched = '1' and ocpl_cmdaccept = '1' then
             axi_wcmd_latched <= '0';
          end if;
        end if;
        if axi_data_latched = '0' then
          if S_AXI_WVALID = '1' then
            axi_data_latched <= '1';
          end if;
        else
          if axi_data_latched = '1' and ocpl_dataaccept = '1' then
             axi_data_latched <= '0';
          end if;
        end if;
      end if;
      if axi_wcmd_latched = '0' and axi_rcmd_latched = '0' and axi_rcmd_busy = '0' then
        if S_AXI_ARVALID = '1' then
          ocpl_addr_i <= S_AXI_ARADDR;
        elsif S_AXI_AWVALID = '1' then
          ocpl_addr_i <= S_AXI_AWADDR;
        end if;
      end if;
      if axi_data_latched = '0' then
        if S_AXI_WVALID = '1' then
          ocpl_m2sdata_i <= S_AXI_WDATA;
          ocpl_databyteen_i <= S_AXI_WSTRB;
        end if;
      end if;

      if S_AXI_ARESETN = '0' then
        bvalid_i <= '0';
      else
        if axi_data_latched = '1' and ocpl_dataaccept = '1' then
          bvalid_i <= '1';
        elsif S_AXI_BREADY = '1' then
          bvalid_i <= '0';
        end if;
      end if;

      if S_AXI_ARESETN = '0' then
        respvalid_i <= '0';
      else
        if ocpl_resp = "01" then
          respvalid_i <= '1';        
        elsif S_AXI_RREADY = '1' then
          respvalid_i <= '0';
        end if;
      end if;
      
      if ocpl_resp = "01" then
        S_AXI_RDATA <= ocpl_s2mdata;
        axi_rcmd_busy <= '0';
      end if;

      
    end if;
  end process latch_data_and_cmd;

  ocpl_cmd_i <= '0' & axi_rcmd_latched & (axi_wcmd_latched and axi_data_latched);

  ocpl_datavalid <= axi_data_latched;
  ocpl_respaccept <= axi_rcmd_busy;
  ocpl_cmd <= ocpl_cmd_i;
  ocpl_addr <= ocpl_addr_i;
  ocpl_m2sdata <= ocpl_m2sdata_i;
  ocpl_databyteen <= ocpl_databyteen_i;

  S_AXI_BVALID <= bvalid_i;
  S_AXI_BRESP <= "00";

  S_AXI_ARREADY <= not (axi_wcmd_latched or axi_rcmd_latched or axi_rcmd_busy); -- or axi_data_latched); 
  S_AXI_AWREADY <= not (axi_wcmd_latched or axi_rcmd_latched or axi_rcmd_busy or S_AXI_ARVALID); -- or axi_data_latched);
  S_AXI_WREADY  <= not axi_data_latched;

  S_AXI_RVALID <= respvalid_i;
  S_AXI_RRESP <= "00";
 
  

end architecture rtl;
