--
-- Behavioural simulation model of ADM-XRC-KU1-MPTL, Target FPGA side
--
-- IMPORTANT: Simulator resolution must be 1 ps or better.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

entity mptl_tgt_ku_axi_bm is
  generic(
    enable_bridge_master  : boolean := true;
    enable_target_master  : boolean := false;
    enable_dma            : boolean := true;
    fabric_reset_in       : boolean := true;
    output_mptlclk        : boolean := true
  );
  port(
    -- MPTL
    mptl_tx_p                : out std_logic_vector(3 downto 0);
    mptl_tx_n                : out std_logic_vector(3 downto 0);
    mptl_rx_p                : in  std_logic_vector(3 downto 0);
    mptl_rx_n                : in  std_logic_vector(3 downto 0);
    mptl_clk_p               : in  std_logic;
    mptl_clk_n               : in  std_logic;
    mptl_clk_out             : out std_logic;
    -- MPTL initialisation
    mptl_target_configured_n : out std_logic;
    mptl_target_gtp_online_n : out std_logic;
    mptl_bridge_gtp_online_n : in  std_logic;

    -- Optional ASYNC reset from Fabric
    async_sys_rst            : in  std_logic := '0';

    -- Signals that are replicated between Target and Host via mptl_{tx,rx}
    -- AXI
    aclk                     : out std_logic;
    aresetn                  : out std_logic;

    ds_axi_awaddr            : out std_logic_vector(63 downto 0);
    ds_axi_awlen             : out std_logic_vector(7 downto 0);
    ds_axi_awsize            : out std_logic_vector(2 downto 0);
    ds_axi_awburst           : out std_logic_vector(1 downto 0);
    ds_axi_awcache           : out std_logic_vector(3 downto 0);
    ds_axi_awprot            : out std_logic_vector(2 downto 0);
    ds_axi_awvalid           : out std_logic;
    ds_axi_wdata             : out std_logic_vector(127 downto 0);
    ds_axi_wstrb             : out std_logic_vector(15 downto 0);
    ds_axi_wlast             : out std_logic;
    ds_axi_wvalid            : out std_logic;
    ds_axi_bready            : out std_logic;
    ds_axi_araddr            : out std_logic_vector(63 downto 0);
    ds_axi_arlen             : out std_logic_vector(7 downto 0);
    ds_axi_arsize            : out std_logic_vector(2 downto 0);
    ds_axi_arburst           : out std_logic_vector(1 downto 0);
    ds_axi_arcache           : out std_logic_vector(3 downto 0);
    ds_axi_arprot            : out std_logic_vector(2 downto 0);
    ds_axi_arvalid           : out std_logic;
    ds_axi_rready            : out std_logic;

    ds_axi_awlock            : out std_logic_vector(0 downto 0);
    ds_axi_awqos             : out std_logic_vector(3 downto 0);
    ds_axi_arlock            : out std_logic_vector(0 downto 0);
    ds_axi_arqos             : out std_logic_vector(3 downto 0);
    ds_axi_awregion          : out std_logic_vector(3 downto 0);
    ds_axi_arregion          : out std_logic_vector(3 downto 0);

    ds_axi_awready           : in  std_logic := '0';
    ds_axi_wready            : in  std_logic := '0';
    ds_axi_bresp             : in  std_logic_vector(1 downto 0) := (others => '0');
    ds_axi_bvalid            : in  std_logic := '0';
    ds_axi_arready           : in  std_logic := '0';
    ds_axi_rdata             : in  std_logic_vector(127 downto 0) := (others => '0');
    ds_axi_rresp             : in  std_logic_vector(1 downto 0) := (others => '0');
    ds_axi_rlast             : in  std_logic := '0';
    ds_axi_rvalid            : in  std_logic := '0';

    dma0_axi_awaddr          : out std_logic_vector(38 downto 0);
    dma0_axi_awlen           : out std_logic_vector(7 downto 0);
    dma0_axi_awsize          : out std_logic_vector(2 downto 0);
    dma0_axi_awburst         : out std_logic_vector(1 downto 0);
    dma0_axi_awcache         : out std_logic_vector(3 downto 0);
    dma0_axi_awprot          : out std_logic_vector(2 downto 0);
    dma0_axi_awvalid         : out std_logic;
    dma0_axi_wdata           : out std_logic_vector(127 downto 0);
    dma0_axi_wstrb           : out std_logic_vector(15 downto 0);
    dma0_axi_wlast           : out std_logic;
    dma0_axi_wvalid          : out std_logic;
    dma0_axi_bready          : out std_logic;
    dma0_axi_araddr          : out std_logic_vector(38 downto 0);
    dma0_axi_arlen           : out std_logic_vector(7 downto 0);
    dma0_axi_arsize          : out std_logic_vector(2 downto 0);
    dma0_axi_arburst         : out std_logic_vector(1 downto 0);
    dma0_axi_arcache         : out std_logic_vector(3 downto 0);
    dma0_axi_arprot          : out std_logic_vector(2 downto 0);
    dma0_axi_arvalid         : out std_logic;
    dma0_axi_rready          : out std_logic;

    dma0_axi_awlock          : out std_logic_vector(0 downto 0);
    dma0_axi_awqos           : out std_logic_vector(3 downto 0);
    dma0_axi_arlock          : out std_logic_vector(0 downto 0);
    dma0_axi_arqos           : out std_logic_vector(3 downto 0);
    dma0_axi_awregion        : out std_logic_vector(3 downto 0);
    dma0_axi_arregion        : out std_logic_vector(3 downto 0);

    dma0_axi_awready         : in  std_logic := '0';
    dma0_axi_wready          : in  std_logic := '0';
    dma0_axi_bresp           : in  std_logic_vector(1 downto 0) := (others => '0');
    dma0_axi_bvalid          : in  std_logic := '0';
    dma0_axi_arready         : in  std_logic := '0';
    dma0_axi_rdata           : in  std_logic_vector(127 downto 0) := (others => '0');
    dma0_axi_rresp           : in  std_logic_vector(1 downto 0) := (others => '0');
    dma0_axi_rlast           : in  std_logic := '0';
    dma0_axi_rvalid          : in  std_logic := '0';

    dma1_axi_awaddr          : out std_logic_vector(38 downto 0);
    dma1_axi_awlen           : out std_logic_vector(7 downto 0);
    dma1_axi_awsize          : out std_logic_vector(2 downto 0);
    dma1_axi_awburst         : out std_logic_vector(1 downto 0);
    dma1_axi_awcache         : out std_logic_vector(3 downto 0);
    dma1_axi_awprot          : out std_logic_vector(2 downto 0);
    dma1_axi_awvalid         : out std_logic;
    dma1_axi_wdata           : out std_logic_vector(127 downto 0);
    dma1_axi_wstrb           : out std_logic_vector(15 downto 0);
    dma1_axi_wlast           : out std_logic;
    dma1_axi_wvalid          : out std_logic;
    dma1_axi_bready          : out std_logic;
    dma1_axi_araddr          : out std_logic_vector(38 downto 0);
    dma1_axi_arlen           : out std_logic_vector(7 downto 0);
    dma1_axi_arsize          : out std_logic_vector(2 downto 0);
    dma1_axi_arburst         : out std_logic_vector(1 downto 0);
    dma1_axi_arcache         : out std_logic_vector(3 downto 0);
    dma1_axi_arprot          : out std_logic_vector(2 downto 0);
    dma1_axi_arvalid         : out std_logic;
    dma1_axi_rready          : out std_logic;

    dma1_axi_awlock          : out std_logic_vector(0 downto 0);
    dma1_axi_awqos           : out std_logic_vector(3 downto 0);
    dma1_axi_arlock          : out std_logic_vector(0 downto 0);
    dma1_axi_arqos           : out std_logic_vector(3 downto 0);
    dma1_axi_awregion        : out std_logic_vector(3 downto 0);
    dma1_axi_arregion        : out std_logic_vector(3 downto 0);

    dma1_axi_awready         : in  std_logic := '0';
    dma1_axi_wready          : in  std_logic := '0';
    dma1_axi_bresp           : in  std_logic_vector(1 downto 0) := (others => '0');
    dma1_axi_bvalid          : in  std_logic := '0';
    dma1_axi_arready         : in  std_logic := '0';
    dma1_axi_rdata           : in  std_logic_vector(127 downto 0) := (others => '0');
    dma1_axi_rresp           : in  std_logic_vector(1 downto 0) := (others => '0');
    dma1_axi_rlast           : in  std_logic := '0';
    dma1_axi_rvalid          : in  std_logic := '0';

    abmi_axi_awaddr          : in  std_logic_vector(11 downto 0) := (others => '0');
    abmi_axi_awvalid         : in  std_logic := '0';
    abmi_axi_wdata           : in  std_logic_vector(31 downto 0) := (others => '0');
    abmi_axi_wstrb           : in  std_logic_vector(3 downto 0) := (others => '0');
    abmi_axi_wvalid          : in  std_logic := '0';
    abmi_axi_bready          : in  std_logic := '0';
    abmi_axi_araddr          : in  std_logic_vector(11 downto 0) := (others => '0');
    abmi_axi_arvalid         : in  std_logic := '0';
    abmi_axi_rready          : in  std_logic := '0';

    abmi_axi_awready         : out std_logic;
    abmi_axi_wready          : out std_logic;
    abmi_axi_bresp           : out std_logic_vector(1 downto 0);
    abmi_axi_bvalid          : out std_logic;
    abmi_axi_arready         : out std_logic;
    abmi_axi_rdata           : out std_logic_vector(31 downto 0);
    abmi_axi_rresp           : out std_logic_vector(1 downto 0);
    abmi_axi_rvalid          : out std_logic;

    -- MPTL sideband
    gpio_t2b                 : in  std_logic_vector(15 downto 0);
    gpio_b2t                 : out std_logic_vector(15 downto 0)
  );
end entity mptl_tgt_ku_axi_bm;


architecture behav of mptl_tgt_ku_axi_bm is

  constant handshake_quantum : time := 1 ps;
  -- If true, handshake protocol delays assignments to handshake signals by 'handshake_quantum'.
  constant handshake_use_quantum : boolean := false;
  
  constant transmit_quantum : time := 1 ps;
  -- If true, transmit process waits for 'transmit_quantum' between each item.
  constant transmit_use_quantum : boolean := true;
  
  -- Receive a header, which may include a single-bit std_logic value along with signal index
  procedure receive_header(
    signal   mptl_exp_rxn : in    std_logic_vector(3 downto 0);
    signal   mptl_exp_rxp : in    std_logic_vector(2 downto 0);
    signal   mptl_exp_txp : out   std_logic_vector(3 downto 3);
    variable index        : out   natural;
    variable val          : out   std_logic)
  is
    variable index_vec : std_logic_vector(10 downto 0);
  begin
    -- Header1, giving ms index of signal & std_logic value
    wait until mptl_exp_rxp(2) = '1';
    val := mptl_exp_rxp(1);
    index_vec(10 downto 10) := mptl_exp_rxp(0 downto 0);
    index_vec(9 downto 6)   := mptl_exp_rxn;
    if handshake_use_quantum then
      mptl_exp_txp(3) <= '1' after handshake_quantum;
      wait until mptl_exp_rxp(2) = '0';
      mptl_exp_txp(3) <= '0' after handshake_quantum;
    else
      mptl_exp_txp(3) <= '1';
      wait until mptl_exp_rxp(2) = '0';
      mptl_exp_txp(3) <= '0';
    end if;

    -- Header2, giving ls index of signal
    wait until mptl_exp_rxp(2) = '1';
    index_vec(5 downto 4) := mptl_exp_rxp(1 downto 0);
    index_vec(3 downto 0) := mptl_exp_rxn;
    index := to_integer(unsigned(index_vec));
    if handshake_use_quantum then
      mptl_exp_txp(3) <= '1' after handshake_quantum;
      wait until mptl_exp_rxp(2) = '0';
      mptl_exp_txp(3) <= '0' after handshake_quantum;
    else
      mptl_exp_txp(3) <= '1';
      wait until mptl_exp_rxp(2) = '0';
      mptl_exp_txp(3) <= '0';
    end if;
  end;

  -- Receive a std_logic_vector value
  procedure receive(
    signal   mptl_exp_rxn : in    std_logic_vector(3 downto 0);
    signal   mptl_exp_rxp : in    std_logic_vector(2 downto 0);
    signal   mptl_exp_txp : out   std_logic_vector(3 downto 3);
    signal   vec          : out   std_logic_vector)
  is
    constant n : natural := vec'length;
    constant nword : natural := n / 6;
    constant remainder : natural := n mod 6;
    variable tmp : std_logic_vector(5 downto 0);
    variable low : natural;
  begin
    -- Payload, giving a series of 6-bit vectors making up the std_logic_vector value
    low := 0;
    for i in 1 to nword loop
      wait until mptl_exp_rxp(2) = '1';
      vec(low + 3 downto low) <= mptl_exp_rxn;
      vec(low + 5 downto low + 4) <= mptl_exp_rxp(1 downto 0);
      if handshake_use_quantum then
        mptl_exp_txp(3) <= '1' after handshake_quantum;
        wait until mptl_exp_rxp(2) = '0';
        mptl_exp_txp(3) <= '0' after handshake_quantum;
      else
        mptl_exp_txp(3) <= '1';
        wait until mptl_exp_rxp(2) = '0';
        mptl_exp_txp(3) <= '0';
      end if;
      low := low + 6;
    end loop;
    if remainder > 0 then
      -- Remainder of payload, i.e. final slice (less than 6 bits long) of std_logic_vector value
      wait until mptl_exp_rxp(2) = '1';
      tmp(3 downto 0) := mptl_exp_rxn;
      tmp(5 downto 4) := mptl_exp_rxp(1 downto 0);
      vec(low + remainder - 1 downto low) <= tmp(remainder - 1 downto 0);
      if handshake_use_quantum then
        mptl_exp_txp(3) <= '1' after handshake_quantum;
        wait until mptl_exp_rxp(2) = '0';
        mptl_exp_txp(3) <= '0' after handshake_quantum;
      else
        mptl_exp_txp(3) <= '1';
        wait until mptl_exp_rxp(2) = '0';
        mptl_exp_txp(3) <= '0';
      end if;
    end if;
  end;

  -- Transmit a std_logic signal
  procedure transmit(
    signal   mptl_exp_txn : out   std_logic_vector(3 downto 0);
    signal   mptl_exp_txp : out   std_logic_vector(2 downto 0);
    signal   mptl_exp_rxp : in    std_logic_vector(3 downto 3);
    constant index        : in    natural;
    signal   val          : in    std_logic)
  is
    variable index_vec : std_logic_vector(10 downto 0);
  begin
    -- Work around to resolve XSIM bug with constants with initial values
    index_vec := std_logic_vector(to_unsigned(index, 11));

    -- Header1, giving ms index of signal & std_logic value
    mptl_exp_txp(1 downto 0) <= val & index_vec(10 downto 10);
    mptl_exp_txn(3 downto 0) <= index_vec(9 downto 6);
    if handshake_use_quantum then
      mptl_exp_txp(2) <= '1' after handshake_quantum;
      wait until mptl_exp_rxp(3) = '1';
      mptl_exp_txp(2) <= '0' after handshake_quantum;
    else
      mptl_exp_txp(2) <= '1';
      wait until mptl_exp_rxp(3) = '1';
      mptl_exp_txp(2) <= '0';
    end if;
    wait until mptl_exp_rxp(3) = '0';

    -- Header2, giving ls index of signal
    mptl_exp_txp(1 downto 0) <= index_vec(5 downto 4);
    mptl_exp_txn(3 downto 0) <= index_vec(3 downto 0);
    if handshake_use_quantum then
      mptl_exp_txp(2) <= '1' after handshake_quantum;
      wait until mptl_exp_rxp(3) = '1';
      mptl_exp_txp(2) <= '0' after handshake_quantum;
    else
      mptl_exp_txp(2) <= '1';
      wait until mptl_exp_rxp(3) = '1';
      mptl_exp_txp(2) <= '0';
    end if;
    wait until mptl_exp_rxp(3) = '0';
  end;

  -- Transmit a std_logic_vector signal, 6 bits at a time
  procedure transmit(
    signal   mptl_exp_txn : out   std_logic_vector(3 downto 0);
    signal   mptl_exp_txp : out   std_logic_vector(2 downto 0);
    signal   mptl_exp_rxp : in    std_logic_vector(3 downto 3);
    constant index        : in    natural;
    signal   vec          : in    std_logic_vector)
  is
    variable index_vec : std_logic_vector(10 downto 0);
    constant n : natural := vec'length;
    constant nword : natural := n / 6;
    constant remainder : natural := n mod 6;
    variable tmp : std_logic_vector(5 downto 0);
    variable low : natural;
  begin
    -- Work around to resolve XSIM bug with constants with initial values
    index_vec := std_logic_vector(to_unsigned(index, 11));

    -- Header1, giving ms index of signal
    mptl_exp_txp(1 downto 0) <= "0" & index_vec(10 downto 10);
    mptl_exp_txn(3 downto 0) <= index_vec(9 downto 6);
    if handshake_use_quantum then
      mptl_exp_txp(2) <= '1' after handshake_quantum;
      wait until mptl_exp_rxp(3) = '1';
      mptl_exp_txp(2) <= '0' after handshake_quantum;
    else
      mptl_exp_txp(2) <= '1';
      wait until mptl_exp_rxp(3) = '1';
      mptl_exp_txp(2) <= '0';
    end if;
    wait until mptl_exp_rxp(3) = '0';

    -- Header2, giving ls index of signal
    mptl_exp_txp(1 downto 0) <= index_vec(5 downto 4);
    mptl_exp_txn(3 downto 0) <= index_vec(3 downto 0);
    if handshake_use_quantum then
      mptl_exp_txp(2) <= '1' after handshake_quantum;
      wait until mptl_exp_rxp(3) = '1';
      mptl_exp_txp(2) <= '0' after handshake_quantum;
    else
      mptl_exp_txp(2) <= '1';
      wait until mptl_exp_rxp(3) = '1';
      mptl_exp_txp(2) <= '0';
    end if;
    wait until mptl_exp_rxp(3) = '0';

    -- Payload, giving a series of 6-bit vectors making up the std_logic_vector value
    low := 0;
    for i in 1 to nword loop
      mptl_exp_txp(1 downto 0) <= vec(low + 5 downto low + 4);
      mptl_exp_txn(3 downto 0) <= vec(low + 3 downto low);
      if handshake_use_quantum then
        mptl_exp_txp(2) <= '1' after handshake_quantum;
        wait until mptl_exp_rxp(3) = '1';
        mptl_exp_txp(2) <= '0' after handshake_quantum;
      else
        mptl_exp_txp(2) <= '1';
        wait until mptl_exp_rxp(3) = '1';
        mptl_exp_txp(2) <= '0';
      end if;
      wait until mptl_exp_rxp(3) = '0';
      low := low + 6;
    end loop;
    if remainder > 0 then
      -- Remainder of payload, i.e. final slice (less than 6 bits long) of std_logic_vector value
      tmp := std_logic_vector(resize(unsigned(vec(low + remainder - 1 downto low)), 6));
      mptl_exp_txp(1 downto 0) <= tmp(5 downto 4);
      mptl_exp_txn(3 downto 0) <= tmp(3 downto 0);
      if handshake_use_quantum then
        mptl_exp_txp(2) <= '1' after handshake_quantum;
        wait until mptl_exp_rxp(3) = '1';
        mptl_exp_txp(2) <= '0' after handshake_quantum;
      else
        mptl_exp_txp(2) <= '1';
        wait until mptl_exp_rxp(3) = '1';
        mptl_exp_txp(2) <= '0';
      end if;
      wait until mptl_exp_rxp(3) = '0';
    end if;
  end;

  type port_in_enum_t is (
    port_ds_axi_awready,
    port_ds_axi_wready,
    port_ds_axi_bresp, 
    port_ds_axi_bvalid, 
    port_ds_axi_arready,
    port_ds_axi_rdata,
    port_ds_axi_rresp, 
    port_ds_axi_rlast, 
    port_ds_axi_rvalid, 
    --
    port_dma0_axi_awready,
    port_dma0_axi_wready,
    port_dma0_axi_bresp,
    port_dma0_axi_bvalid, 
    port_dma0_axi_arready,
    port_dma0_axi_rdata,
    port_dma0_axi_rresp, 
    port_dma0_axi_rlast, 
    port_dma0_axi_rvalid, 
    --
    port_dma1_axi_awready,
    port_dma1_axi_wready,
    port_dma1_axi_bresp,
    port_dma1_axi_bvalid, 
    port_dma1_axi_arready,
    port_dma1_axi_rdata,
    port_dma1_axi_rresp, 
    port_dma1_axi_rlast, 
    port_dma1_axi_rvalid, 
    --
    port_abmi_axi_awaddr,
    port_abmi_axi_awvalid,
    port_abmi_axi_wdata,
    port_abmi_axi_wstrb, 
    port_abmi_axi_wvalid, 
    port_abmi_axi_bready,
    port_abmi_axi_araddr,
    port_abmi_axi_arvalid,
    port_abmi_axi_rready,
    --
    port_gpio_t2b
  );

  type port_out_enum_t is (
    port_aclk,
    port_aresetn,
    --
    port_ds_axi_awaddr,
    port_ds_axi_awlen,
    port_ds_axi_awsize, 
    port_ds_axi_awburst,
    port_ds_axi_awcache,
    port_ds_axi_awprot,
    port_ds_axi_awvalid,
    port_ds_axi_wdata,
    port_ds_axi_wstrb, 
    port_ds_axi_wlast, 
    port_ds_axi_wvalid, 
    port_ds_axi_bready,
    port_ds_axi_araddr,
    port_ds_axi_arlen,
    port_ds_axi_arsize, 
    port_ds_axi_arburst,
    port_ds_axi_arcache,
    port_ds_axi_arprot,
    port_ds_axi_arvalid,
    port_ds_axi_rready,
    --
    port_ds_axi_awlock,
    port_ds_axi_awqos,
    port_ds_axi_arlock,
    port_ds_axi_arqos,
    port_ds_axi_awregion,
    port_ds_axi_arregion,
    --
    port_dma0_axi_awaddr, 
    port_dma0_axi_awlen,  
    port_dma0_axi_awsize, 
    port_dma0_axi_awburst,
    port_dma0_axi_awcache,
    port_dma0_axi_awprot, 
    port_dma0_axi_awvalid,
    port_dma0_axi_wdata,  
    port_dma0_axi_wstrb,  
    port_dma0_axi_wlast,  
    port_dma0_axi_wvalid, 
    port_dma0_axi_bready, 
    port_dma0_axi_araddr, 
    port_dma0_axi_arlen,  
    port_dma0_axi_arsize, 
    port_dma0_axi_arburst,
    port_dma0_axi_arcache,
    port_dma0_axi_arprot, 
    port_dma0_axi_arvalid,
    port_dma0_axi_rready, 
    --
    port_dma0_axi_awlock, 
    port_dma0_axi_awqos,  
    port_dma0_axi_arlock, 
    port_dma0_axi_arqos,  
    port_dma0_axi_awregion,
    port_dma0_axi_arregion,
    --
    port_dma1_axi_awaddr,  
    port_dma1_axi_awlen,   
    port_dma1_axi_awsize,  
    port_dma1_axi_awburst, 
    port_dma1_axi_awcache, 
    port_dma1_axi_awprot, 
    port_dma1_axi_awvalid, 
    port_dma1_axi_wdata,   
    port_dma1_axi_wstrb,   
    port_dma1_axi_wlast,   
    port_dma1_axi_wvalid,  
    port_dma1_axi_bready,  
    port_dma1_axi_araddr,  
    port_dma1_axi_arlen,   
    port_dma1_axi_arsize,  
    port_dma1_axi_arburst, 
    port_dma1_axi_arcache, 
    port_dma1_axi_arprot,  
    port_dma1_axi_arvalid, 
    port_dma1_axi_rready,  
    --
    port_dma1_axi_awlock,  
    port_dma1_axi_awqos,   
    port_dma1_axi_arlock,  
    port_dma1_axi_arqos,   
    port_dma1_axi_awregion,
    port_dma1_axi_arregion,
    --
    port_abmi_axi_awready,
    port_abmi_axi_wready, 
    port_abmi_axi_bresp,
    port_abmi_axi_bvalid, 
    port_abmi_axi_arready,
    port_abmi_axi_rdata,
    port_abmi_axi_rresp, 
    port_abmi_axi_rvalid, 
    --
    port_gpio_b2t
  );

  constant num_port_in  : natural := port_in_enum_t'pos(port_in_enum_t'high) + 1;
  constant num_port_out : natural := port_out_enum_t'pos(port_out_enum_t'high) + 1;

  type boolean_array_t is array(natural range <>) of boolean;

  signal pending_in : boolean_array_t(0 to num_port_in - 1) := (others => false);
  signal ack_in     : boolean_array_t(0 to num_port_in - 1) := (others => false);

  constant inertial_delay : time := 1 ns;


  -- Input ports
  signal i_ds_axi_awready : std_logic;
  signal i_ds_axi_wready  : std_logic;
  signal i_ds_axi_bresp   : std_logic_vector(1 downto 0);
  signal i_ds_axi_bvalid  : std_logic;
  signal i_ds_axi_arready : std_logic;
  signal i_ds_axi_rdata   : std_logic_vector(127 downto 0);
  signal i_ds_axi_rresp   : std_logic_vector(1 downto 0);
  signal i_ds_axi_rlast   : std_logic;
  signal i_ds_axi_rvalid  : std_logic;

  signal i_dma0_axi_awready : std_logic;
  signal i_dma0_axi_wready  : std_logic;
  signal i_dma0_axi_bresp   : std_logic_vector(1 downto 0);
  signal i_dma0_axi_bvalid  : std_logic;
  signal i_dma0_axi_arready : std_logic;
  signal i_dma0_axi_rdata   : std_logic_vector(127 downto 0);
  signal i_dma0_axi_rresp   : std_logic_vector(1 downto 0);
  signal i_dma0_axi_rlast   : std_logic;
  signal i_dma0_axi_rvalid  : std_logic;

  signal i_dma1_axi_awready : std_logic;
  signal i_dma1_axi_wready  : std_logic;
  signal i_dma1_axi_bresp   : std_logic_vector(1 downto 0);
  signal i_dma1_axi_bvalid  : std_logic;
  signal i_dma1_axi_arready : std_logic;
  signal i_dma1_axi_rdata   : std_logic_vector(127 downto 0);
  signal i_dma1_axi_rresp   : std_logic_vector(1 downto 0);
  signal i_dma1_axi_rlast   : std_logic;
  signal i_dma1_axi_rvalid  : std_logic;

  signal i_abmi_axi_awaddr  : std_logic_vector(11 downto 0);
  signal i_abmi_axi_awvalid : std_logic;
  signal i_abmi_axi_wdata   : std_logic_vector(31 downto 0);
  signal i_abmi_axi_wstrb   : std_logic_vector(3 downto 0);
  signal i_abmi_axi_wvalid  : std_logic;
  signal i_abmi_axi_bready  : std_logic;
  signal i_abmi_axi_araddr  : std_logic_vector(11 downto 0);
  signal i_abmi_axi_arvalid : std_logic;
  signal i_abmi_axi_rready  : std_logic;

  signal i_gpio_t2b : std_logic_vector(15 downto 0);

  -- mptl_clk_out 125MHz  
  constant mptl_clk_out_period : time := 8 ns;

  signal mptl_clk_out_i : std_logic := '0';
  signal mptl_target_configured_l : std_logic := '1';

begin

  -- Bridge to Target Sideband Signals
  mptl_target_configured_l <= '0' after 50 us;
  mptl_target_configured_n <= '1' when (fabric_reset_in and (async_sys_rst = '1')) else mptl_target_configured_l;
  mptl_target_gtp_online_n <= '1' when (fabric_reset_in and (async_sys_rst = '1')) else mptl_bridge_gtp_online_n;

  -- GTHE3_CHANNEL_TXOUTCLK[0] (125 MHz)
  mptl_clk_out_i <= not mptl_clk_out_i after mptl_clk_out_period;
  mptl_clk_out <= mptl_clk_out_i when output_mptlclk else '0';


  -- Input ports   
  i_ds_axi_awready <= ds_axi_awready after inertial_delay;
  i_ds_axi_wready  <= ds_axi_wready  after inertial_delay;
  i_ds_axi_bresp   <= ds_axi_bresp   after inertial_delay;
  i_ds_axi_bvalid  <= ds_axi_bvalid  after inertial_delay;
  i_ds_axi_arready <= ds_axi_arready after inertial_delay;
  i_ds_axi_rdata   <= ds_axi_rdata   after inertial_delay;
  i_ds_axi_rresp   <= ds_axi_rresp   after inertial_delay;
  i_ds_axi_rlast   <= ds_axi_rlast   after inertial_delay;
  i_ds_axi_rvalid  <= ds_axi_rvalid  after inertial_delay;
   
  i_dma0_axi_awready <= dma0_axi_awready after inertial_delay;
  i_dma0_axi_wready  <= dma0_axi_wready  after inertial_delay;
  i_dma0_axi_bresp   <= dma0_axi_bresp   after inertial_delay;
  i_dma0_axi_bvalid  <= dma0_axi_bvalid  after inertial_delay;
  i_dma0_axi_arready <= dma0_axi_arready after inertial_delay;
  i_dma0_axi_rdata   <= dma0_axi_rdata   after inertial_delay;
  i_dma0_axi_rresp   <= dma0_axi_rresp   after inertial_delay;
  i_dma0_axi_rlast   <= dma0_axi_rlast   after inertial_delay;
  i_dma0_axi_rvalid  <= dma0_axi_rvalid  after inertial_delay;
 
  i_dma1_axi_awready <= dma1_axi_awready after inertial_delay;
  i_dma1_axi_wready  <= dma1_axi_wready  after inertial_delay;
  i_dma1_axi_bresp   <= dma1_axi_bresp   after inertial_delay;
  i_dma1_axi_bvalid  <= dma1_axi_bvalid  after inertial_delay;
  i_dma1_axi_arready <= dma1_axi_arready after inertial_delay;
  i_dma1_axi_rdata   <= dma1_axi_rdata   after inertial_delay;
  i_dma1_axi_rresp   <= dma1_axi_rresp   after inertial_delay;
  i_dma1_axi_rlast   <= dma1_axi_rlast   after inertial_delay;
  i_dma1_axi_rvalid  <= dma1_axi_rvalid  after inertial_delay;
    
  i_abmi_axi_awaddr  <= abmi_axi_awaddr  after inertial_delay;
  i_abmi_axi_awvalid <= abmi_axi_awvalid after inertial_delay;
  i_abmi_axi_wdata   <= abmi_axi_wdata   after inertial_delay;
  i_abmi_axi_wstrb   <= abmi_axi_wstrb   after inertial_delay;
  i_abmi_axi_wvalid  <= abmi_axi_wvalid  after inertial_delay;
  i_abmi_axi_bready  <= abmi_axi_bready  after inertial_delay;
  i_abmi_axi_araddr  <= abmi_axi_araddr  after inertial_delay;
  i_abmi_axi_arvalid <= abmi_axi_arvalid after inertial_delay;
  i_abmi_axi_rready  <= abmi_axi_rready  after inertial_delay;

  i_gpio_t2b <= gpio_t2b after inertial_delay;


  -- ds_axi
  propagate_ds_axi_awready : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_awready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_awready;
    end loop;
  end process;

  propagate_ds_axi_wready : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_wready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_wready;
    end loop;
  end process;

  propagate_ds_axi_bresp : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_bresp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_bresp;
    end loop;
  end process;

  propagate_ds_axi_bvalid : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_bvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_bvalid;
    end loop;
  end process;

  propagate_ds_axi_arready : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_arready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_arready;
    end loop;
  end process;

  propagate_ds_axi_rdata : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_rdata);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_rdata;
    end loop;
  end process;

  propagate_ds_axi_rresp : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_rresp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_rresp;
    end loop;
  end process;

  propagate_ds_axi_rlast : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_rlast);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_rlast;
    end loop;
  end process;

  propagate_ds_axi_rvalid : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_rvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_rvalid;
    end loop;
  end process;


  -- dma0_axi
  propagate_dma0_axi_awready : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_awready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_awready;
    end loop;
  end process;

  propagate_dma0_axi_wready : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_wready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_wready;
    end loop;
  end process;

  propagate_dma0_axi_bresp : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_bresp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_bresp;
    end loop;
  end process;

  propagate_dma0_axi_bvalid : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_bvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_bvalid;
    end loop;
  end process;

  propagate_dma0_axi_arready : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_arready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_arready;
    end loop;
  end process;

  propagate_dma0_axi_rdata : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_rdata);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_rdata;
    end loop;
  end process;

  propagate_dma0_axi_rresp : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_rresp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_rresp;
    end loop;
  end process;

  propagate_dma0_axi_rlast : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_rlast);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_rlast;
    end loop;
  end process;

  propagate_dma0_axi_rvalid : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_rvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_rvalid;
    end loop;
  end process;

  -- dma1_s2m
  propagate_dma1_axi_awready : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_awready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_awready;
    end loop;
  end process;

  propagate_dma1_axi_wready : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_wready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_wready;
    end loop;
  end process;

  propagate_dma1_axi_bresp : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_bresp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_bresp;
    end loop;
  end process;

  propagate_dma1_axi_bvalid : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_bvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_bvalid;
    end loop;
  end process;

  propagate_dma1_axi_arready : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_arready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_arready;
    end loop;
  end process;

  propagate_dma1_axi_rdata : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_rdata);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_rdata;
    end loop;
  end process;

  propagate_dma1_axi_rresp : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_rresp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_rresp;
    end loop;
  end process;

  propagate_dma1_axi_rlast : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_rlast);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_rlast;
    end loop;
  end process;

  propagate_dma1_axi_rvalid : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_rvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_rvalid;
    end loop;
  end process;

  -- abmi_m2s
  propagate_abmi_axi_awaddr : process
    constant index : natural := port_in_enum_t'pos(port_abmi_axi_awaddr);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_axi_awaddr;
    end loop;
  end process;

  propagate_abmi_axi_awvalid : process
    constant index : natural := port_in_enum_t'pos(port_abmi_axi_awvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_axi_awvalid;
    end loop;
  end process;

  propagate_abmi_axi_wdata : process
    constant index : natural := port_in_enum_t'pos(port_abmi_axi_wdata);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_axi_wdata;
    end loop;
  end process;

  propagate_abmi_axi_wstrb : process
    constant index : natural := port_in_enum_t'pos(port_abmi_axi_wstrb);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_axi_wstrb;
    end loop;
  end process;

  propagate_abmi_axi_wvalid : process
    constant index : natural := port_in_enum_t'pos(port_abmi_axi_wvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_axi_wvalid;
    end loop;
  end process;

  propagate_abmi_axi_bready : process
    constant index : natural := port_in_enum_t'pos(port_abmi_axi_bready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_axi_bready;
    end loop;
  end process;

  propagate_abmi_axi_araddr : process
    constant index : natural := port_in_enum_t'pos(port_abmi_axi_araddr);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_axi_araddr;
    end loop;
  end process;

  propagate_abmi_axi_arvalid : process
    constant index : natural := port_in_enum_t'pos(port_abmi_axi_arvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_axi_arvalid;
    end loop;
  end process;

  propagate_abmi_axi_rready : process
    constant index : natural := port_in_enum_t'pos(port_abmi_axi_rready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_axi_rready;
    end loop;
  end process;

  -- gpio_t2b
  propagate_gpio_t2b : process
    constant index : natural := port_in_enum_t'pos(port_gpio_t2b);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_gpio_t2b;
    end loop;
  end process;


  transmit_signals : process
    variable port_in : port_in_enum_t;
    variable work_done : boolean;
  begin
    loop
      loop
        work_done := false;
        for index in pending_in'range loop
          if pending_in(index) then
            work_done := true;

            -- Send current value of port to other side of link
            port_in := port_in_enum_t'val(index);
--            report "bm transmiting index " & integer'image(index) & " port " & port_in_enum_t'image(port_in);

            case port_in is
              -- 
              when port_ds_axi_awready   => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_ds_axi_awready);
              when port_ds_axi_wready    => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_ds_axi_wready);
              when port_ds_axi_bresp     => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_ds_axi_bresp);
              when port_ds_axi_bvalid    => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_ds_axi_bvalid);
              when port_ds_axi_arready   => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_ds_axi_arready);
              when port_ds_axi_rdata     => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_ds_axi_rdata);
              when port_ds_axi_rresp     => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_ds_axi_rresp);
              when port_ds_axi_rlast     => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_ds_axi_rlast);
              when port_ds_axi_rvalid    => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_ds_axi_rvalid); 
              -- 
              when port_dma0_axi_awready => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma0_axi_awready);
              when port_dma0_axi_wready  => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma0_axi_wready);
              when port_dma0_axi_bresp   => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma0_axi_bresp);
              when port_dma0_axi_bvalid  => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma0_axi_bvalid);
              when port_dma0_axi_arready => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma0_axi_arready);
              when port_dma0_axi_rdata   => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma0_axi_rdata);
              when port_dma0_axi_rresp   => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma0_axi_rresp);
              when port_dma0_axi_rlast   => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma0_axi_rlast);
              when port_dma0_axi_rvalid  => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma0_axi_rvalid);
              --
              when port_dma1_axi_awready => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma1_axi_awready);
              when port_dma1_axi_wready  => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma1_axi_wready);
              when port_dma1_axi_bresp   => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma1_axi_bresp);
              when port_dma1_axi_bvalid  => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma1_axi_bvalid);
              when port_dma1_axi_arready => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma1_axi_arready);
              when port_dma1_axi_rdata   => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma1_axi_rdata);
              when port_dma1_axi_rresp   => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma1_axi_rresp);
              when port_dma1_axi_rlast   => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma1_axi_rlast);
              when port_dma1_axi_rvalid  => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma1_axi_rvalid);
              --
              when port_abmi_axi_awaddr  => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_axi_awaddr);
              when port_abmi_axi_awvalid => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_axi_awvalid);
              when port_abmi_axi_wdata   => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_axi_wdata);
              when port_abmi_axi_wstrb   => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_axi_wstrb);
              when port_abmi_axi_wvalid  => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_axi_wvalid);
              when port_abmi_axi_bready  => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_axi_bready);
              when port_abmi_axi_araddr  => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_axi_araddr);
              when port_abmi_axi_arvalid => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_axi_arvalid);
              when port_abmi_axi_rready  => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_axi_rready);
              --
              when port_gpio_t2b         => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_gpio_t2b);
              --
              when others => report "transmit: case " & port_in_enum_t'image(port_in) & " is not handled.";-- severity failure;
            end case;

            -- Indicate that signal has been updated
            if handshake_use_quantum then
              ack_in(index) <= true after handshake_quantum;
            else
              ack_in(index) <= true;
            end if;
            
            -- Complete the handshake
            wait until not pending_in(index);
            if handshake_use_quantum then
              ack_in(index) <= false after handshake_quantum;
            else
              ack_in(index) <= false;
            end if;

            if transmit_use_quantum then
              -- Avoid excessive numbers of delta cycles at a given simulation time.
              wait for transmit_quantum;
            end if;
          end if;
        end loop;
        exit when not work_done;
      end loop;
      wait on pending_in;
    end loop;
  end process;


  receive_signals : process
    variable index : natural;
    variable val : std_logic;
    variable port_out : port_out_enum_t;
  begin
    loop
      receive_header(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), index, val);
      port_out := port_out_enum_t'val(index);
--      report "bm receiving index " & integer'image(index) & " port " & port_out_enum_t'image(port_out) & " val " & std_logic'image(val);

      case port_out is
        when port_aclk    => aclk    <= val;
        when port_aresetn => aresetn <= val;
        --
        when port_ds_axi_awaddr  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_awaddr);
        when port_ds_axi_awlen   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_awlen);
        when port_ds_axi_awsize  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_awsize);
        when port_ds_axi_awburst => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_awburst);
        when port_ds_axi_awcache => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_awcache);
        when port_ds_axi_awprot  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_awprot);
        when port_ds_axi_awvalid => ds_axi_awvalid <= val;
        when port_ds_axi_wdata   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_wdata);
        when port_ds_axi_wstrb   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_wstrb);
        when port_ds_axi_wlast   => ds_axi_wlast <= val;
        when port_ds_axi_wvalid  => ds_axi_wvalid <= val;
        when port_ds_axi_bready  => ds_axi_bready <= val;
        when port_ds_axi_araddr  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_araddr);
        when port_ds_axi_arlen   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_arlen);
        when port_ds_axi_arsize  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_arsize);
        when port_ds_axi_arburst => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_arburst);
        when port_ds_axi_arcache => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_arcache);
        when port_ds_axi_arprot  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_arprot);
        when port_ds_axi_arvalid => ds_axi_arvalid <= val;
        when port_ds_axi_rready  => ds_axi_rready <= val;
        --
        when port_ds_axi_awlock   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_awlock);
        when port_ds_axi_awqos    => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_awqos);
        when port_ds_axi_arlock   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_arlock);
        when port_ds_axi_arqos    => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_arqos);
        when port_ds_axi_awregion => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_awregion);
        when port_ds_axi_arregion => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), ds_axi_arregion);
        --
        when port_dma0_axi_awaddr  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_awaddr);
        when port_dma0_axi_awlen   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_awlen);
        when port_dma0_axi_awsize  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_awsize);
        when port_dma0_axi_awburst => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_awburst);
        when port_dma0_axi_awcache => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_awcache);
        when port_dma0_axi_awprot  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_awprot);
        when port_dma0_axi_awvalid => dma0_axi_awvalid <= val;
        when port_dma0_axi_wdata   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_wdata);
        when port_dma0_axi_wstrb   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_wstrb);
        when port_dma0_axi_wlast   => dma0_axi_wlast <= val;
        when port_dma0_axi_wvalid  => dma0_axi_wvalid <= val;
        when port_dma0_axi_bready  => dma0_axi_bready <= val;
        when port_dma0_axi_araddr  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_araddr);
        when port_dma0_axi_arlen   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_arlen);
        when port_dma0_axi_arsize  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_arsize);
        when port_dma0_axi_arburst => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_arburst);
        when port_dma0_axi_arcache => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_arcache);
        when port_dma0_axi_arprot  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_arprot);
        when port_dma0_axi_arvalid => dma0_axi_arvalid <= val;
        when port_dma0_axi_rready  => dma0_axi_rready <= val;
        --
        when port_dma0_axi_awlock   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_awlock);
        when port_dma0_axi_awqos    => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_awqos);
        when port_dma0_axi_arlock   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_arlock);
        when port_dma0_axi_arqos    => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_arqos);
        when port_dma0_axi_awregion => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_awregion);
        when port_dma0_axi_arregion => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma0_axi_arregion);
        --
        when port_dma1_axi_awaddr  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_awaddr);
        when port_dma1_axi_awlen   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_awlen);
        when port_dma1_axi_awsize  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_awsize);
        when port_dma1_axi_awburst => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_awburst);
        when port_dma1_axi_awcache => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_awcache);
        when port_dma1_axi_awprot  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_awprot);
        when port_dma1_axi_awvalid => dma1_axi_awvalid <= val;
        when port_dma1_axi_wdata   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_wdata);
        when port_dma1_axi_wstrb   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_wstrb);
        when port_dma1_axi_wlast   => dma1_axi_wlast <= val;
        when port_dma1_axi_wvalid  => dma1_axi_wvalid <= val;
        when port_dma1_axi_bready  => dma1_axi_bready <= val;
        when port_dma1_axi_araddr  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_araddr);
        when port_dma1_axi_arlen   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_arlen);
        when port_dma1_axi_arsize  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_arsize);
        when port_dma1_axi_arburst => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_arburst);
        when port_dma1_axi_arcache => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_arcache);
        when port_dma1_axi_arprot  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_arprot);
        when port_dma1_axi_arvalid => dma1_axi_arvalid <= val;
        when port_dma1_axi_rready  => dma1_axi_rready <= val;
        --
        when port_dma1_axi_awlock   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_awlock);
        when port_dma1_axi_awqos    => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_awqos);
        when port_dma1_axi_arlock   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_arlock);
        when port_dma1_axi_arqos    => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_arqos);
        when port_dma1_axi_awregion => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_awregion);
        when port_dma1_axi_arregion => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), dma1_axi_arregion);
        --
        when port_abmi_axi_awready  => abmi_axi_awready <= val;
        when port_abmi_axi_wready   => abmi_axi_wready <= val;
        when port_abmi_axi_bresp    => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), abmi_axi_bresp);
        when port_abmi_axi_bvalid   => abmi_axi_bvalid <= val;
        when port_abmi_axi_arready  => abmi_axi_arready <= val;
        when port_abmi_axi_rdata    => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), abmi_axi_rdata);
        when port_abmi_axi_rresp    => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), abmi_axi_rresp);
        when port_abmi_axi_rvalid   => abmi_axi_rvalid <= val;
        --
        when port_gpio_b2t          => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), gpio_b2t);
        --
        when others => report "receive: case " & port_out_enum_t'image(port_out) & " is not handled.";-- severity failure;
      end case;
    end loop;
  end process;

end architecture;
