--------------------------------------------------------------------------------
--
-- Module Name      : adb3_ocp_resp_to_mptl128
--
--                    Copyright 2009 by
--                    Alpha Data Parallel Systems Ltd.
--                    All Rights Reserved.
--
-- Original Author  : Andrew McCormick
-- Created          : 
--
-- Modified By      : 
-- Date             :
-- Change Notes     :     
--
-- Description      :  Bridge module for converting ADB3_OCP slave responses into MPTL packets
--
--
-- Dependencies     : adb3_ocp, mptl
--
-- Disclaimer       : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                    WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                    WARRANTIES IMPLIED OF MERCHANTABILITY, FITNESS FOR A
--                    PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.mptl.all;


entity adb3_ocp_resp_to_mptl128 is

  generic (
    channel : integer := 0);
  port (
    -- OCP Interface
    ocp_clk            : in  std_logic;
    ocp_m2s            : out adb3_ocp_m2sT;
    ocp_s2m            : in  adb3_ocp_s2mT;
    -- MPTL interface
    mptl128_data        : out std_logic_vector(127 downto 0);
    mptl128_valid       : out std_logic;
    mptl128_accept      : in  std_logic;
    online             : in  std_logic;
    -- Last packet consumed by
    remote_last_packet : in  std_logic_vector(7 downto 0);
--    current_packet_o : out std_logic_vector(7 downto 0);
    -- Burst Length Info
    local_ocp_read     : in  std_logic;
    local_burst_length : in  std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
    local_tag          : in  std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
    local_ocp_nfull    : out std_logic;  -- stop further OCP reads
    -- Error
    tag_mismatch       : out std_logic);

end adb3_ocp_resp_to_mptl128;

architecture rtl of adb3_ocp_resp_to_mptl128 is

  constant IMPROVE_TIMING_ON_MPTL128 : boolean := true;

  signal current_packet    : std_logic_vector(7 downto 0) := (others => '0');
  signal remote_fifo_level : std_logic_vector(7 downto 0) := (others => '0');
  signal can_accept_data   : std_logic                    := '1';

  type state_type is (idle, send_burst_data);
  signal state        : state_type := idle;
  signal beat_counter : std_logic_vector(3 downto 0);

  signal resp_accept : std_logic;

  signal packet_header, prev_packet_header : std_logic_vector(127 downto 0);
  signal mptl128_hvalid, prev_mptl128_hvalid : std_logic;

  signal mptl128_dvalid : std_logic := '0';

  signal burst_length_left           : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0) := (others => '0');
  signal next_burst_packet_size      : std_logic_vector(4 downto 0)                    := "00001";
  signal beat_limit, beat_limit_calc : std_logic_vector(3 downto 0)                    := "0001";  --(others => '0');
  signal next_burst_last, last_burst : std_logic                                       := '0';


  signal actual_packet_header : std_logic_vector(127 downto 0);

  signal wadv, radv, radv1   : std_logic := '0';
  signal wdata, rdata : std_logic_vector(ADB3_OCP_BURST_WIDTH+ADB3_OCP_TAG_WIDTH-1 downto 0);
  signal rempty       : std_logic;


  signal next_bl_valid     : std_logic := '0';
  signal next_burst_length : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0) := (others => '0');
  signal next_tag          : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0) := (others => '0');


  component small_sfifo
    generic (
      width : integer := 64);
    port (
      clk    : in  std_logic;
      rst    : in  std_logic;
      wadv   : in  std_logic;
      wdata  : in  std_logic_vector(width-1 downto 0);
      nfull  : out std_logic;
      radv   : in  std_logic;
      rempty : out std_logic;
      rdata  : out std_logic_vector(width-1 downto 0));
  end component;

  component small_sfifo32
    generic (
      width : integer := 64);
    port (
      clk    : in  std_logic;
      rst    : in  std_logic;
      wadv   : in  std_logic;
      wdata  : in  std_logic_vector(width-1 downto 0);
      nfull  : out std_logic;
      radv   : in  std_logic;
      rempty : out std_logic;
      rdata  : out std_logic_vector(width-1 downto 0));
  end component;

  component small_sfifo64
    generic (
      width : integer := 64);
    port (
      clk    : in  std_logic;
      rst    : in  std_logic;
      wadv   : in  std_logic;
      wdata  : in  std_logic_vector(width-1 downto 0);
      nfull  : out std_logic;
      radv   : in  std_logic;
      rempty : out std_logic;
      rdata  : out std_logic_vector(width-1 downto 0));
  end component;
  
  signal mptl128_data_i   : std_logic_vector(127 downto 0);
  signal mptl128_valid_i  : std_logic := '0';
  signal mptl128_accept_i : std_logic := '0';
  signal mptl128_accept_q : std_logic := '0';
  
  signal burst_accepted : std_logic := '0';

  signal mptl128_ofifo_d   : std_logic_vector(128 downto 0);
  signal mptl128_ofifo_q   : std_logic_vector(128 downto 0);
  signal mptl128_ofifo_nfull  : std_logic := '0';
  signal mptl128_ofifo_nfull_r1  : std_logic := '0';
  signal mptl128_ofifo_wadv   : std_logic := '0';
  signal mptl128_ofifo_radv   : std_logic := '0';
  signal mptl128_ofifo_empty  : std_logic := '0';

  signal mptl128_pvalid : std_logic := '0';
  signal mptl128_pvalid_q : std_logic := '0';
  signal mptl128_hvalid_q : std_logic := '0';

  signal eop_count : std_logic_vector(1 downto 0) := "00";
  signal eop_fifo_radv : std_logic;
  signal eop_fifo_full : std_logic;
  
  signal online_r1 : std_logic := '0';

  signal fifo_rst : std_logic;
  
begin  -- rtl


  -- Don't drive most of OCP Master Outputs: only driving RespAccept
  ocp_m2s.Cmd         <= (others => 'Z');
  ocp_m2s.Addr        <= (others => 'Z');
  ocp_m2s.BurstLength <= (others => 'Z');
  ocp_m2s.Data        <= (others => 'Z');
  ocp_m2s.DataByteEn  <= (others => 'Z');
  ocp_m2s.DataValid   <= 'Z';
  ocp_m2s.Tag         <= (others => 'Z');

  ocp_m2s.RespAccept <= resp_accept;

  -- Keep track of remote FIFO level
  remote_fifo_level0 : process (ocp_clk)
  begin  -- process remote_fifo_level0
    if ocp_clk'event and ocp_clk = '1' then  -- rising clock edge
      -- Compute number of packets in flight or in remote FIFO
  
      remote_fifo_level <= current_packet-remote_last_packet;
      
      -- Allow up to 16 packets in flight
      can_accept_data   <= (not remote_fifo_level(6)) and online_r1;

      if online_r1 = '0' then
        current_packet <= (others => '0');
      elsif mptl128_hvalid = '1' and mptl128_accept_i = '1' then
        current_packet <= current_packet+1;
      end if;
      --current_packet_o <= current_packet;
      --current_packet_o(0) <= can_accept_data;
      --current_packet_o(1) <= mptl128_ofifo_nfull_r1;
      --current_packet_o(2) <= eop_fifo_full;
      --current_packet_o(3) <= mptl128_hvalid;
      
    end if;
  end process remote_fifo_level0;
  
  fifo_rst <= not online_r1;

  -- Small FIFO for tracking in flight OCP bursts
  wadv  <= local_ocp_read;
  wdata <= local_tag & local_burst_length;

  small_sfifo0 : small_sfifo32
    generic map (
      width => ADB3_OCP_BURST_WIDTH+ADB3_OCP_TAG_WIDTH)
    port map (
      clk    => ocp_clk,
      rst    => fifo_rst,
      wadv   => wadv,
      wdata  => wdata,
      nfull  => local_ocp_nfull,
      radv   => radv1,
      rempty => rempty,
      rdata  => rdata);

  
  
  reg_inflight: process (ocp_clk) is
  begin  -- process reg_inflight
    if rising_edge(ocp_clk) then        -- rising clock edge
      if radv1 = '1'  then
        next_bl_valid     <= '1';
        next_burst_length <= rdata(ADB3_OCP_BURST_WIDTH-1 downto 0);
        next_tag          <= rdata(ADB3_OCP_BURST_WIDTH+ADB3_OCP_TAG_WIDTH-1 downto ADB3_OCP_BURST_WIDTH);
      elsif radv = '1' then
        next_bl_valid     <= '0';
      end if;
      radv1 <= not rempty and not next_bl_valid and not radv1;
    end if;
  end process reg_inflight;
 
  radv <= '1' when state = send_burst_data and mptl128_hvalid = '0' and ocp_s2m.Resp = OCP_RESP_VALID and
          beat_counter = beat_limit and last_burst = '1' else '0';

  -- FSM
  fsm0 : process (ocp_clk)
  begin  -- process fsm0
    if ocp_clk'event and ocp_clk = '1' then  -- rising clock edge
      -- Default Values (overwritten)
      mptl128_hvalid <= '0';
      mptl128_dvalid <= '0';
      mptl128_pvalid <= '0';

      online_r1 <= online;
      
      case state is
        when idle =>
          if can_accept_data = '1' and online_r1 = '1' then
            if ocp_s2m.Resp = OCP_RESP_VALID and next_bl_valid = '1' then
              mptl128_hvalid <= '1';
              state         <= send_burst_data;
            end if;
            beat_counter <= (others => '0');
          end if;
        when send_burst_data =>
          if mptl128_hvalid = '1' then
            if mptl128_accept_i = '0' then
              mptl128_hvalid <= '1';
            else
              mptl128_dvalid <= '1';             
            end if;
          else
            mptl128_dvalid <= '1';
            if ocp_s2m.Resp = OCP_RESP_VALID then
              if beat_counter = beat_limit then
                if last_burst = '1' then
                  state         <= idle;
                  mptl128_dvalid <= '0';
                  mptl128_pvalid <= '1';
                else
                  if can_accept_data = '1' then
                     beat_counter  <= (others => '0');
                     mptl128_hvalid <= '1';
                     mptl128_dvalid <= '0';
                     mptl128_pvalid <= '1';
                  else
                    beat_counter  <= (others => '0');
                    mptl128_hvalid <= '0';
                    mptl128_dvalid <= '0';
                    state <= idle;
                    mptl128_pvalid <= '1';
                  end if;                 
                end if;
              else
                beat_counter <= beat_counter+1;
              end if;
            end if;
          end if;
          if online_r1 = '0' then
            state <= idle;
          end if;
        when others =>
          state <= idle;
      end case;
    end if;
  end process fsm0;

  fsm_outputs0 : process (beat_counter, next_bl_valid, next_tag,
                          ocp_s2m.Resp, ocp_s2m.Tag, state, mptl128_hvalid)
  begin  -- process fsm_outputs
    -- Control Resp Accept
    if state = send_burst_data and mptl128_hvalid = '0' then
      resp_accept <= '1';
    else
      resp_accept <= '0';
    end if;

    -- Check Tags for out of sequence data
    if next_bl_valid = '1' and ocp_s2m.Resp = OCP_RESP_VALID and ocp_s2m.Tag /= next_tag then
      tag_mismatch <= '1';
    else
      tag_mismatch <= '0';
    end if;
    
  end process fsm_outputs0;


  mptl128_valid_i <= (mptl128_hvalid) or (mptl128_dvalid and to_std_logic(ocp_s2m.Resp = OCP_RESP_VALID));

  mptl128_data_i <= actual_packet_header when mptl128_hvalid = '1' else
                   ocp_s2m.Data;


--   gen_better_timing : if IMPROVE_TIMING_ON_MPTL128 generate
--     -- Register MPTL Output
--     -- Improves timing, but at cost of 1 clock cycle 
--     reg_mptl0 : process (ocp_clk)
--     begin  -- process reg_mptl0
--       if ocp_clk'event and ocp_clk = '1' then  -- rising clock edge
--         if mptl128_valid_i = '1' then
--           mptl128_data <= mptl128_data_i;
--         end if;
--         mptl128_valid <= (mptl128_hvalid and not mptl128_accept_i) or (mptl128_dvalid and to_std_logic(ocp_s2m.Resp = OCP_RESP_VALID));

--       end if;
--     end process reg_mptl0;
--     mptl128_accept_q <=  mptl128_accept_i and not (mptl128_hvalid and not prev_mptl128_hvalid);
--   end generate gen_better_timing;

--   gen_better_throughput : if not IMPROVE_TIMING_ON_MPTL128 generate
--     mptl128_data  <= mptl128_data_i;
--     mptl128_valid <= mptl128_valid_i;
--     mptl128_accept_q <=  mptl128_accept_i;
--   end generate gen_better_throughput;

--   mptl128_accept_i <= mptl128_accept;


  -- packet header only valid for 1 clock cycle, if not instantly accapted,
  -- use latched version
  actual_packet_header <= packet_header when prev_mptl128_hvalid = '0' else
                          prev_packet_header;

  -- Select Packet Header contents before registering
  -- Packet word (header or data) selected after registering
  next_packet_hdr : process (ocp_s2m, next_burst_packet_size, current_packet)
  begin  -- process next_packet_hdr
    packet_header               <= (others => '0');  -- Assign default values if not
                                                     -- overwritten in process
    -- Set Fixed Header values for all packets 
    packet_header(7 downto 4)   <= CONV_STD_LOGIC_VECTOR(channel, 4);  -- CHANNEL
    packet_header(23 downto 16) <= current_packet;   -- PACKET_ID
    packet_header(3 downto 0)   <= MPTL_READ_DATA;   -- TYPE
    packet_header(12 downto 8)  <= next_burst_packet_size;
    packet_header(47 downto 40) <= ocp_s2m.Tag;

    -- Set Checksum
    packet_header(39 downto 36) <= MPTL_READ_DATA xor CONV_STD_LOGIC_VECTOR(channel, 4) xor current_packet(3 downto 0) xor current_packet(7 downto 4)
                                   xor next_burst_packet_size(3 downto 0) xor ("000" & next_burst_packet_size(4))  xor ocp_s2m.Tag(3 downto 0) xor ocp_s2m.Tag(7 downto 4);
 
    
  end process next_packet_hdr;

  -- Latch Packet Header if not immediately accepted
  phr0 : process (ocp_clk)
  begin  -- process phr0
    if ocp_clk'event and ocp_clk = '1' then  -- rising clock edge
      if mptl128_hvalid = '1' and prev_mptl128_hvalid = '0' then  -- 
        prev_packet_header <= packet_header;
      end if;
      prev_mptl128_hvalid <= mptl128_hvalid;
      
    end if;
  end process phr0;

  -- Compute Burst Sizes
  burst_size0 : process (ocp_clk)
    variable nbps_var : std_logic_vector(11 downto 0);
    variable bl_var   : std_logic_vector(11 downto 0);
    variable burst_accepted_var : std_logic;
  begin  -- process burst_size0
    if ocp_clk'event and ocp_clk = '1' then  -- rising clock edge
      if state = send_burst_data and mptl128_accept_i = '1' and mptl128_hvalid = '1' then
        burst_accepted <= '1';
        burst_accepted_var := '1';
      else
        burst_accepted <= '0';
        burst_accepted_var := '0';
      end if;
      if state = idle then
        burst_length_left <= next_burst_length;

        -- next_burst_packet_size must be calculated during idle cycle
        if next_burst_length > MPTL128_MAX_BURST_PAYLOAD_SIZE then
          next_burst_packet_size <= CONV_STD_LOGIC_VECTOR(MPTL128_MAX_BURST_PAYLOAD_PACKET_SIZE, 5);
          beat_limit_calc        <= CONV_STD_LOGIC_VECTOR(MPTL128_MAX_BURST_PAYLOAD_SIZE-1, 4);
          next_burst_last        <= '0';
        else
          nbps_var               := next_burst_length + 1;
          next_burst_packet_size <= nbps_var(4 downto 0);
          bl_var                 := next_burst_length -1;
          beat_limit_calc        <= bl_var(3 downto 0);
          next_burst_last        <= '1';
        end if;
      else
        if burst_accepted_var = '1' then
          if next_burst_last = '1' then
            burst_length_left <= (others => '0');
          else
            burst_length_left <= burst_length_left-MPTL128_MAX_BURST_PAYLOAD_SIZE;
          end if;
          last_burst <= next_burst_last;
          beat_limit <= beat_limit_calc;
        end if;

        if burst_length_left > MPTL128_MAX_BURST_PAYLOAD_SIZE then
          next_burst_packet_size <= CONV_STD_LOGIC_VECTOR(MPTL128_MAX_BURST_PAYLOAD_PACKET_SIZE, 5);
          beat_limit_calc        <= CONV_STD_LOGIC_VECTOR(MPTL128_MAX_BURST_PAYLOAD_SIZE-1, 4);
          next_burst_last        <= '0';
        else
          nbps_var               := burst_length_left + 1;
          next_burst_packet_size <= nbps_var(4 downto 0);
          bl_var                 := burst_length_left -1;
          beat_limit_calc        <= bl_var(3 downto 0);
          next_burst_last        <= '1';
        end if;
      end if;
      
      

    end if;
  end process burst_size0;


  mptl128_ofifo_d <= mptl128_hvalid & mptl128_data_i;
  mptl128_ofifo_wadv <= (mptl128_hvalid and (not mptl128_ofifo_nfull_r1) and (not eop_fifo_full)) or (mptl128_dvalid and to_std_logic(ocp_s2m.Resp = OCP_RESP_VALID));
  mptl128_accept_i <= (not mptl128_ofifo_nfull_r1) and (not eop_fifo_full);
  
  mptl128_ofifo_radv <= (not mptl128_ofifo_empty) and
                       ((mptl128_accept and mptl128_ofifo_q(128) and mptl128_pvalid_q) or
                        not mptl128_ofifo_q(128));

  mptl128_valid <= (not mptl128_ofifo_empty) and
                  ((mptl128_ofifo_q(128) and mptl128_pvalid_q) or
                  not mptl128_ofifo_q(128));
  
  mptl128_data <= mptl128_ofifo_q(127 downto 0);
  
  small_sfifo1 : small_sfifo64
    generic map (
      width => 129)
    port map (
      clk    => ocp_clk,
      rst    => fifo_rst,
      wadv   => mptl128_ofifo_wadv,
      wdata  => mptl128_ofifo_d,
      nfull  => mptl128_ofifo_nfull,
      radv   => mptl128_ofifo_radv,
      rempty => mptl128_ofifo_empty,
      rdata  => mptl128_ofifo_q);

  -- 2 word, 1 bit deep FIFO for tracking end of packets
  eop_fifo: process (ocp_clk) is
  begin  -- process eop_fifo
    if rising_edge(ocp_clk) then        -- rising clock edge
      if online_r1 = '0' then
        eop_count <= (others => '0');
      elsif mptl128_pvalid = '1' and eop_fifo_radv  = '0' then
        eop_count <= eop_count+1;
      elsif mptl128_pvalid = '0' and eop_fifo_radv = '1' then
        eop_count <= eop_count-1;
      end if;

      mptl128_ofifo_nfull_r1 <= mptl128_ofifo_nfull;
    end if;
  end process eop_fifo;
  
  mptl128_pvalid_q <= OR_reduce(eop_count);

  eop_fifo_radv <= (not mptl128_ofifo_empty) and (mptl128_accept and mptl128_ofifo_q(128) and mptl128_pvalid_q);
  eop_fifo_full <= eop_count(1) and eop_count(0);

  
end rtl;
