--------------------------------------------------------------------------------
--
-- Module Name      : mptl128_to_adb3_ocp
--
--                    Copyright 2009 by
--                    Alpha Data Parallel Systems Ltd.
--                    All Rights Reserved.
--
-- Original Author  : Andrew McCormick
-- Created          : 
--
-- Modified By      : 
-- Date             :
-- Change Notes     :     
--
-- Description      : Bridge module for converting MPTL packets into ADB3_OCP transactions
--
--
-- Dependencies     : adb3_ocp, mptl
--
-- Disclaimer       : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                    WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                    WARRANTIES IMPLIED OF MERCHANTABILITY, FITNESS FOR A
--                    PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.mptl.all;

entity mptl128_to_adb3_ocp is

  port (
    -- OCP Interface
    ocp_clk            : in  std_logic;
    ocp_m2s            : out adb3_ocp_m2sT;
    ocp_s2m            : in  adb3_ocp_s2mT;
    -- MPTL interface
    mptl128_data        : in  std_logic_vector(127 downto 0);
    mptl128_valid       : in  std_logic;
    online             : in  std_logic;
    -- Last read from FIFO
    remote_last_packet : out std_logic_vector(7 downto 0);
    -- Connection to OCP RESP for reads
    local_ocp_read     : out std_logic;
    local_burst_length : out std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
    local_tag          : out std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
    local_ocp_nfull    : in  std_logic);

end mptl128_to_adb3_ocp;

architecture rtl of mptl128_to_adb3_ocp is
  
  type state_type is (idle, parse_header,parse_read_request, parse_single_write, parse_burst_cmd, parse_burst_data);
  signal state         : state_type := idle; 
  signal cmd_accepted : std_logic := '0';
  signal data_accepted : std_logic := '0';
  signal data_valid : std_logic := '0';
  signal first_data : std_logic := '0';

  signal radv   : std_logic;
  signal rempty : std_logic;
  signal rdata  : std_logic_vector(127 downto 0);

  signal rdata_reg : std_logic_vector(127 downto 0);

  signal radv_reg   : std_logic;
  signal rdata_reg1 : std_logic_vector(127 downto 0);
  signal rdata_reg2 : std_logic_vector(127 downto 0);

  signal packet_length : std_logic_vector(4 downto 0);
  signal burst_hdr : std_logic;

  signal cmd : std_logic_vector(2 downto 0) := OCP_CMD_IDLE;
 
  signal byte_en : std_logic_vector(ADB3_OCP_DATA_WIDTH/8-1 downto 0);
  signal addr : std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
  signal blen : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
  signal tag : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
  
  component dpram_sfifo
    generic (
      width : integer);
    port (
      clk    : in  std_logic;
      rst    : in  std_logic;
      wadv   : in  std_logic;
      wdata  : in  std_logic_vector(width-1 downto 0);
      nfull  : out std_logic;
      radv   : in  std_logic;
      rempty : out std_logic;
      rdata  : out std_logic_vector(width-1 downto 0));
  end component;

  signal mptl128_valid_reg : std_logic := '0';
  signal mptl128_data_reg : std_logic_vector(127 downto 0) := (others => '0');

  signal fifo_rst : std_logic := '1';
  signal fifo_rst1 : std_logic := '1';
  signal fifo_rst2 : std_logic := '1';

  signal rdata_chked_valid : std_logic := '0';
  signal chksum : std_logic_vector(3 downto 0);
  signal remote_last_packet_i : std_logic_vector(7 downto 0) := (others => '1');

begin  -- rtl

  -- Don't drive response accept 
  ocp_m2s.RespAccept  <= 'Z';
  ocp_m2s.Cmd         <= cmd;
  ocp_m2s.Addr        <= addr;
  ocp_m2s.BurstLength <= blen;
  ocp_m2s.Data        <= rdata_reg;
  ocp_m2s.DataByteEn  <= byte_en when first_data = '1'  else (others => '1');
  ocp_m2s.DataValid   <= data_valid;
  ocp_m2s.Tag         <= tag;

  remote_last_packet <= remote_last_packet_i;
                        
  
  reg_mptl64_data0: process (ocp_clk)
  begin  -- process reg_mptl64_data0
    if ocp_clk'event and ocp_clk = '1' then 
      mptl128_valid_reg <= mptl128_valid;
      mptl128_data_reg <= mptl128_data;
      --fifo_rst <= fifo_rst1;
      --fifo_rst1 <= fifo_rst2;
      --fifo_rst2 <= not online;
     
    
    end if;
  end process reg_mptl64_data0;

  fifo_rst <= not online;
  
  input_sfifo0 : dpram_sfifo
    generic map (
      width => 128)
    port map (
      clk    => ocp_clk,
      rst    => fifo_rst,
      wadv   => mptl128_valid,
      wdata  => mptl128_data,
      nfull  => open,
      radv   => radv,
      rempty => rempty,
      rdata  => rdata);




  -- FSM
  fsm0 : process (ocp_clk)
  begin  -- process fsm0
    if ocp_clk'event and ocp_clk = '1' then  -- rising clock edge
      case state is
        when idle =>                             
          if rempty = '0' then
            state <= parse_header;
          end if;
          if online = '0' then
            remote_last_packet_i <= (others => '1'); 
          end if;
          cmd   <= OCP_CMD_IDLE;
          cmd_accepted <= '0';
          data_accepted <= '0';
          data_valid <= '0';
        when parse_header =>
          if rdata_chked_valid = '0' then
            state <= idle;
          elsif rdata_reg(3 downto 0) = MPTL_READ_REQUEST and local_ocp_nfull = '0' then
            state <= parse_read_request;
            cmd   <= OCP_CMD_READ;           
          elsif rdata_reg(3 downto 0) = MPTL_READ_REQUEST and local_ocp_nfull = '1' then
            state <= parse_header;
            cmd   <= OCP_CMD_IDLE; 
          elsif rdata_reg(3 downto 0) = MPTL_SINGLE_BEAT_WRITE then
            state <= parse_single_write;
            cmd   <= OCP_CMD_WRITE;
            data_valid <= not rempty;
          elsif rdata_reg(3 downto 0) = MPTL_BURST_WRITE_CMD then
            cmd   <= OCP_CMD_WRITE;
            state <= parse_burst_cmd;             
          elsif rdata_reg(3 downto 0) = MPTL_BURST_WRITE_DATA then
            state <= parse_burst_data;
            data_valid <= not rempty;
          else
            state <= idle;
          end if;
          packet_length      <= rdata_reg(12 downto 8);
          if (rdata_reg(3 downto 0) = MPTL_READ_REQUEST or rdata_reg(3 downto 0) = MPTL_SINGLE_BEAT_WRITE or rdata_reg(3 downto 0) = MPTL_BURST_WRITE_CMD or rdata_reg(3 downto 0) = MPTL_BURST_WRITE_DATA) and rdata_chked_valid = '1' then
            remote_last_packet_i <= rdata_reg(23 downto 16);
          end if;
          byte_en <= rdata_reg(127 downto 112);
          cmd_accepted <= '0';
          data_accepted <= '0';
          first_data <= '1';
          addr <= rdata_reg(63 downto 48) & rdata_reg(111 downto 64);
          blen <= rdata_reg(35 downto 24);
          tag <= rdata_reg(47 downto 40);
        when parse_read_request =>
          if ocp_s2m.CmdAccept = '1' then
            if rempty = '1' then
              state <= idle;
            else
              state <= parse_header;
            end if;           
            cmd   <= OCP_CMD_IDLE;
          end if;
        when parse_single_write =>
          if rempty = '0' and data_valid = '0' and data_accepted = '0' then
            data_valid <= '1';
          end if;
          if cmd_accepted = '0' then
            if ocp_s2m.CmdAccept = '1' then
              cmd   <= OCP_CMD_IDLE;
              cmd_accepted <= '1';
              if (data_valid = '1' and ocp_s2m.DataAccept = '1') or data_accepted = '1'  then
                data_valid <= '0';
                data_accepted <= '0';
                if rempty = '1' then
                  state <= idle;
                else
                  state <= parse_header;
                end if;
              end if;
            else
              if data_valid = '1' and ocp_s2m.DataAccept = '1' then
                data_valid <= '0';
                data_accepted <= '1';
              end if;
            end if;
          else
            if (data_valid = '1' and ocp_s2m.DataAccept = '1') or data_accepted = '1' then
              data_valid <= '0';
              data_accepted <= '0';
              if rempty = '1' then
                state <= idle;
              else
                state <= parse_header;
              end if;
            end if;
          end if;                                       
        when parse_burst_cmd =>
          if ocp_s2m.CmdAccept = '1' then
            if rempty = '1' then
              state <= idle;
            else
              state <= parse_header;
            end if;           
            cmd   <= OCP_CMD_IDLE;
          end if;
        when parse_burst_data =>
          cmd   <= OCP_CMD_IDLE;

          if (data_valid = '1' and ocp_s2m.DataAccept = '1')  then
            first_data <= '0';
            if packet_length = "00010" then
              data_valid <= '0';
              if rempty = '1' then
                state <= idle;
              else
                state <= parse_header;
              end if;
            else
              packet_length <= packet_length-1;
              data_valid <= not rempty;
            end if;            
          elsif data_valid = '0' then
            data_valid <= not rempty;
          end if;

        when others =>
          state <= idle;
      end case;
    end if;
  end process fsm0;

  fsm_outputs0 : process (state, rempty, ocp_s2m,rdata_reg,local_ocp_nfull,data_valid,cmd_accepted,data_accepted)
  begin  -- process fsm_outputs0
    -- Default value
    radv       <= '0';
    case state is
      when idle =>
        if rempty = '0' then                
          radv <= '1';
        end if;        
      when parse_header =>
        if rdata_reg(3 downto 0) = MPTL_SINGLE_BEAT_WRITE or rdata_reg(3 downto 0) = MPTL_BURST_WRITE_DATA then
          if rempty = '0' then
            radv <= '1';
          end if;        
        end if;
      when parse_read_request =>
        if ocp_s2m.CmdAccept = '1' then
          if rempty = '0' then
            radv <= '1';
          end if;
        end if;         
      when parse_single_write =>
        if rempty = '0' and data_valid = '0'  and data_accepted = '0' then
          radv <= '1';
        end if;
        if cmd_accepted = '0' then
          if ocp_s2m.CmdAccept = '1' then
            if (data_valid = '1' and ocp_s2m.DataAccept = '1') or data_accepted = '1' then
              if rempty = '0' then
                radv <= '1';
              end if;
            end if;
          end if;
        else
          if (data_valid = '1' and ocp_s2m.DataAccept = '1') or data_accepted = '1' then
            if rempty = '0' then
              radv <= '1';
            end if;
          end if;   
        end if;
      when parse_burst_cmd =>
        if ocp_s2m.CmdAccept = '1' then
          if rempty = '0' then
            radv <= '1';
          end if;
        end if;             
      when parse_burst_data =>
        if data_valid = '1' and ocp_s2m.DataAccept = '1' then
          radv <= not rempty;
        elsif data_valid = '0' then
          radv <= not rempty;
        end if;               
      when others =>      
        radv       <= '0';
    end case; 
  end process fsm_outputs0;

  rdata_reg <= rdata_reg1 when radv_reg = '1' else rdata_reg2;

  chksum <= rdata(3 downto 0) xor rdata(7 downto 4) xor rdata(11 downto 8) xor rdata(15 downto 12) xor
            rdata(19 downto 16) xor rdata(23 downto 20) xor rdata(27 downto 24) xor rdata(31 downto 28) xor
            rdata(35 downto 32) xor rdata(43 downto 40) xor rdata(47 downto 44);
  
  phr0 : process (ocp_clk)
  begin  -- process phr0
    if ocp_clk'event and ocp_clk = '1' then  -- rising clock edge
--       Implement rdata_reg using registered radv dur to high radv fanout
--       if radv = '1' then
--         rdata_reg <= rdata;
--       end if;
      radv_reg <= radv;
      rdata_reg1 <= rdata;
      if radv_reg = '1' then
         rdata_reg2 <= rdata_reg1;
      end if;

      if radv = '1' then
        if chksum = rdata(39 downto 36) and (remote_last_packet_i+1) = rdata(23 downto 16) then
          rdata_chked_valid <= '1';
        else
          rdata_chked_valid <= '0';
        end if;
      end if;
      
      -- Latch Tag and Burst Size For OCP response
      if rdata_reg(3 downto 0) = MPTL_READ_REQUEST and local_ocp_nfull = '0' and state = parse_header and rdata_chked_valid = '1' then
        local_ocp_read <= '1';
        local_burst_length <= rdata_reg(35 downto 24);
        local_tag <= rdata_reg(47 downto 40);
      else
        local_ocp_read <= '0';
      end if; 
    end if;
  end process phr0;



end rtl;
