--
-- Behavioural simulation model of ADM-XRC-KU1-MPTL, Target FPGA side
--
-- IMPORTANT: Simulator resolution must be 1 ps or better.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

entity mptl_tgt_ku_slv_bm is
  generic(
    enable_bridge_master  : boolean := true;
    enable_target_master  : boolean := false;
    enable_dma            : boolean := true;
    wrapper_reset_signals : boolean := true;
    break_out_ocp         : boolean := false
  );
  port(
    -- MPTL
    mptl_tx_p                : out   std_logic_vector(3 downto 0);
    mptl_tx_n                : out   std_logic_vector(3 downto 0);
    mptl_rx_p                : in    std_logic_vector(3 downto 0);
    mptl_rx_n                : in    std_logic_vector(3 downto 0);
    mptl_clk_p               : in    std_logic;
    mptl_clk_n               : in    std_logic;
    mptl_clk_out             : out   std_logic;
    -- MPTL initialisation
    mptl_target_configured_n : out   std_logic;
    mptl_target_gtp_online_n : out   std_logic;
    mptl_bridge_gtp_online_n : in    std_logic;
    -- Extra Wrapper Signals
    ocp_ready                : in    std_logic;
    mptl_ready               : out   std_logic;

    -- Signals that are replicated between Target and Host via mptl_{tx,rx}
    -- OCP
    ocp_clk   : in    std_logic;
    ds_m2s    : out   std_logic_vector(232 downto 0);
    ds_s2m    : in    std_logic_vector(139 downto 0) := (others => '0');
    dma_m2s   : out   std_logic_vector(465 downto 0);
    dma_s2m   : in    std_logic_vector(279 downto 0) := (others => '0');
    dma_abort : in    std_logic_vector(1 downto 0) := (others => '0');
    abmi_m2s  : in    std_logic_vector(232 downto 0) := (others => '0');
    abmi_s2m  : out   std_logic_vector(139 downto 0);
    -- MPTL sideband
    gpio_t2b  : in    std_logic_vector(15 downto 0);
    gpio_b2t  : out   std_logic_vector(15 downto 0);
    
    -- Broken Out OCP Signals
    ds_m2s_cmd           : out   std_logic_vector(2 downto 0);
    ds_m2s_addr          : out   std_logic_vector(63 downto 0);
    ds_m2s_burstlength   : out   std_logic_vector(11 downto 0);
    ds_m2s_data          : out   std_logic_vector(127 downto 0);
    ds_m2s_databyteen    : out   std_logic_vector(15 downto 0);
    ds_m2s_datavalid     : out   std_logic;
    ds_m2s_respaccept    : out   std_logic;
    ds_m2s_tag           : out   std_logic_vector(7 downto 0);
    ds_s2m_cmdaccept     : in    std_logic := '0';
    ds_s2m_dataaccept    : in    std_logic := '0';
    ds_s2m_data          : in    std_logic_vector(127 downto 0) := (others => '0');
    ds_s2m_resp          : in    std_logic_vector(1 downto 0) := (others => '0');
    ds_s2m_tag           : in    std_logic_vector(7 downto 0) := (others => '0');
                                 
    dma0_m2s_cmd         : out   std_logic_vector(2 downto 0);
    dma0_m2s_addr        : out   std_logic_vector(63 downto 0);
    dma0_m2s_burstlength : out   std_logic_vector(11 downto 0);
    dma0_m2s_data        : out   std_logic_vector(127 downto 0);
    dma0_m2s_databyteen  : out   std_logic_vector(15 downto 0);
    dma0_m2s_datavalid   : out   std_logic;
    dma0_m2s_respaccept  : out   std_logic;
    dma0_m2s_tag         : out   std_logic_vector(7 downto 0);
    dma0_s2m_cmdaccept   : in    std_logic := '0';
    dma0_s2m_dataaccept  : in    std_logic := '0';
    dma0_s2m_data        : in    std_logic_vector(127 downto 0) := (others => '0');
    dma0_s2m_resp        : in    std_logic_vector(1 downto 0) := (others => '0');
    dma0_s2m_tag         : in    std_logic_vector(7 downto 0) := (others => '0');
                                 
    dma1_m2s_cmd         : out   std_logic_vector(2 downto 0);
    dma1_m2s_addr        : out   std_logic_vector(63 downto 0);
    dma1_m2s_burstlength : out   std_logic_vector(11 downto 0);
    dma1_m2s_data        : out   std_logic_vector(127 downto 0);
    dma1_m2s_databyteen  : out   std_logic_vector(15 downto 0);
    dma1_m2s_datavalid   : out   std_logic;
    dma1_m2s_respaccept  : out   std_logic;
    dma1_m2s_tag         : out   std_logic_vector(7 downto 0);
    dma1_s2m_cmdaccept   : in    std_logic := '0';
    dma1_s2m_dataaccept  : in    std_logic := '0';
    dma1_s2m_data        : in    std_logic_vector(127 downto 0) := (others => '0');
    dma1_s2m_resp        : in    std_logic_vector(1 downto 0) := (others => '0');
    dma1_s2m_tag         : in    std_logic_vector(7 downto 0) := (others => '0');

    abmi_m2s_cmd         : in    std_logic_vector(2 downto 0) := (others => '0');
    abmi_m2s_addr        : in    std_logic_vector(63 downto 0) := (others => '0');
    abmi_m2s_burstlength : in    std_logic_vector(11 downto 0) := (others => '0');
    abmi_m2s_data        : in    std_logic_vector(127 downto 0) := (others => '0');
    abmi_m2s_databyteen  : in    std_logic_vector(15 downto 0) := (others => '0');
    abmi_m2s_datavalid   : in    std_logic := '0';
    abmi_m2s_respaccept  : in    std_logic := '0';
    abmi_m2s_tag         : in    std_logic_vector(7 downto 0) := (others => '0');
    abmi_s2m_cmdaccept   : out   std_logic;
    abmi_s2m_dataaccept  : out   std_logic;
    abmi_s2m_data        : out   std_logic_vector(127 downto 0);
    abmi_s2m_resp        : out   std_logic_vector(1 downto 0);
    abmi_s2m_tag         : out   std_logic_vector(7 downto 0)
  );
end entity;


architecture behav of mptl_tgt_ku_slv_bm is

  constant handshake_quantum : time := 1 ps;
  -- If true, handshake protocol delays assignments to handshake signals by 'handshake_quantum'.
  constant handshake_use_quantum : boolean := false;
  
  constant transmit_quantum : time := 1 ps;
  -- If true, transmit process waits for 'transmit_quantum' between each item.
  constant transmit_use_quantum : boolean := true;
  
  -- Receive a header, which may include a single-bit std_logic value along with signal index
  procedure receive_header(
    signal   mptl_exp_rxn : in    std_logic_vector(3 downto 0);
    signal   mptl_exp_rxp : in    std_logic_vector(2 downto 0);
    signal   mptl_exp_txp : out   std_logic_vector(3 downto 3);
    variable index        : out   natural;
    variable val          : out   std_logic)
  is
    variable index_vec : std_logic_vector(10 downto 0);
  begin
    -- Header1, giving ms index of signal & std_logic value
    wait until mptl_exp_rxp(2) = '1';
    val := mptl_exp_rxp(1);
    index_vec(10 downto 10) := mptl_exp_rxp(0 downto 0);
    index_vec(9 downto 6)   := mptl_exp_rxn;
    if handshake_use_quantum then
      mptl_exp_txp(3) <= '1' after handshake_quantum;
      wait until mptl_exp_rxp(2) = '0';
      mptl_exp_txp(3) <= '0' after handshake_quantum;
    else
      mptl_exp_txp(3) <= '1';
      wait until mptl_exp_rxp(2) = '0';
      mptl_exp_txp(3) <= '0';
    end if;

    -- Header2, giving ls index of signal
    wait until mptl_exp_rxp(2) = '1';
    index_vec(5 downto 4) := mptl_exp_rxp(1 downto 0);
    index_vec(3 downto 0) := mptl_exp_rxn;
    index := to_integer(unsigned(index_vec));
    if handshake_use_quantum then
      mptl_exp_txp(3) <= '1' after handshake_quantum;
      wait until mptl_exp_rxp(2) = '0';
      mptl_exp_txp(3) <= '0' after handshake_quantum;
    else
      mptl_exp_txp(3) <= '1';
      wait until mptl_exp_rxp(2) = '0';
      mptl_exp_txp(3) <= '0';
    end if;
  end;

  -- Receive a std_logic_vector value
  procedure receive(
    signal   mptl_exp_rxn : in    std_logic_vector(3 downto 0);
    signal   mptl_exp_rxp : in    std_logic_vector(2 downto 0);
    signal   mptl_exp_txp : out   std_logic_vector(3 downto 3);
    signal   vec          : out   std_logic_vector)
  is
    constant n : natural := vec'length;
    constant nword : natural := n / 6;
    constant remainder : natural := n mod 6;
    variable tmp : std_logic_vector(5 downto 0);
    variable low : natural;
  begin
    -- Payload, giving a series of 6-bit vectors making up the std_logic_vector value
    low := 0;
    for i in 1 to nword loop
      wait until mptl_exp_rxp(2) = '1';
      vec(low + 3 downto low) <= mptl_exp_rxn;
      vec(low + 5 downto low + 4) <= mptl_exp_rxp(1 downto 0);
      if handshake_use_quantum then
        mptl_exp_txp(3) <= '1' after handshake_quantum;
        wait until mptl_exp_rxp(2) = '0';
        mptl_exp_txp(3) <= '0' after handshake_quantum;
      else
        mptl_exp_txp(3) <= '1';
        wait until mptl_exp_rxp(2) = '0';
        mptl_exp_txp(3) <= '0';
      end if;
      low := low + 6;
    end loop;
    if remainder > 0 then
      -- Remainder of payload, i.e. final slice (less than 6 bits long) of std_logic_vector value
      wait until mptl_exp_rxp(2) = '1';
      tmp(3 downto 0) := mptl_exp_rxn;
      tmp(5 downto 4) := mptl_exp_rxp(1 downto 0);
      vec(low + remainder - 1 downto low) <= tmp(remainder - 1 downto 0);
      if handshake_use_quantum then
        mptl_exp_txp(3) <= '1' after handshake_quantum;
        wait until mptl_exp_rxp(2) = '0';
        mptl_exp_txp(3) <= '0' after handshake_quantum;
      else
        mptl_exp_txp(3) <= '1';
        wait until mptl_exp_rxp(2) = '0';
        mptl_exp_txp(3) <= '0';
      end if;
    end if;
  end;

  -- Transmit a std_logic signal
  procedure transmit(
    signal   mptl_exp_txn : out   std_logic_vector(3 downto 0);
    signal   mptl_exp_txp : out   std_logic_vector(2 downto 0);
    signal   mptl_exp_rxp : in    std_logic_vector(3 downto 3);
    constant index        : in    natural;
    signal   val          : in    std_logic)
  is
    variable index_vec : std_logic_vector(10 downto 0);
  begin
    -- Work around to resolve XSIM bug with constants with initial values
    index_vec := std_logic_vector(to_unsigned(index, 11));

    -- Header1, giving ms index of signal & std_logic value
    mptl_exp_txp(1 downto 0) <= val & index_vec(10 downto 10);
    mptl_exp_txn(3 downto 0) <= index_vec(9 downto 6);
    if handshake_use_quantum then
      mptl_exp_txp(2) <= '1' after handshake_quantum;
      wait until mptl_exp_rxp(3) = '1';
      mptl_exp_txp(2) <= '0' after handshake_quantum;
    else
      mptl_exp_txp(2) <= '1';
      wait until mptl_exp_rxp(3) = '1';
      mptl_exp_txp(2) <= '0';
    end if;
    wait until mptl_exp_rxp(3) = '0';

    -- Header2, giving ls index of signal
    mptl_exp_txp(1 downto 0) <= index_vec(5 downto 4);
    mptl_exp_txn(3 downto 0) <= index_vec(3 downto 0);
    if handshake_use_quantum then
      mptl_exp_txp(2) <= '1' after handshake_quantum;
      wait until mptl_exp_rxp(3) = '1';
      mptl_exp_txp(2) <= '0' after handshake_quantum;
    else
      mptl_exp_txp(2) <= '1';
      wait until mptl_exp_rxp(3) = '1';
      mptl_exp_txp(2) <= '0';
    end if;
    wait until mptl_exp_rxp(3) = '0';
  end;

  -- Transmit a std_logic_vector signal, 6 bits at a time
  procedure transmit(
    signal   mptl_exp_txn : out   std_logic_vector(3 downto 0);
    signal   mptl_exp_txp : out   std_logic_vector(2 downto 0);
    signal   mptl_exp_rxp : in    std_logic_vector(3 downto 3);
    constant index        : in    natural;
    signal   vec          : in    std_logic_vector)
  is
    variable index_vec : std_logic_vector(10 downto 0);
    constant n : natural := vec'length;
    constant nword : natural := n / 6;
    constant remainder : natural := n mod 6;
    variable tmp : std_logic_vector(5 downto 0);
    variable low : natural;
  begin
    -- Work around to resolve XSIM bug with constants with initial values
    index_vec := std_logic_vector(to_unsigned(index, 11));

    -- Header1, giving ms index of signal
    mptl_exp_txp(1 downto 0) <= "0" & index_vec(10 downto 10);
    mptl_exp_txn(3 downto 0) <= index_vec(9 downto 6);
    if handshake_use_quantum then
      mptl_exp_txp(2) <= '1' after handshake_quantum;
      wait until mptl_exp_rxp(3) = '1';
      mptl_exp_txp(2) <= '0' after handshake_quantum;
    else
      mptl_exp_txp(2) <= '1';
      wait until mptl_exp_rxp(3) = '1';
      mptl_exp_txp(2) <= '0';
    end if;
    wait until mptl_exp_rxp(3) = '0';

    -- Header2, giving ls index of signal
    mptl_exp_txp(1 downto 0) <= index_vec(5 downto 4);
    mptl_exp_txn(3 downto 0) <= index_vec(3 downto 0);
    if handshake_use_quantum then
      mptl_exp_txp(2) <= '1' after handshake_quantum;
      wait until mptl_exp_rxp(3) = '1';
      mptl_exp_txp(2) <= '0' after handshake_quantum;
    else
      mptl_exp_txp(2) <= '1';
      wait until mptl_exp_rxp(3) = '1';
      mptl_exp_txp(2) <= '0';
    end if;
    wait until mptl_exp_rxp(3) = '0';

    -- Payload, giving a series of 6-bit vectors making up the std_logic_vector value
    low := 0;
    for i in 1 to nword loop
      mptl_exp_txp(1 downto 0) <= vec(low + 5 downto low + 4);
      mptl_exp_txn(3 downto 0) <= vec(low + 3 downto low);
      if handshake_use_quantum then
        mptl_exp_txp(2) <= '1' after handshake_quantum;
        wait until mptl_exp_rxp(3) = '1';
        mptl_exp_txp(2) <= '0' after handshake_quantum;
      else
        mptl_exp_txp(2) <= '1';
        wait until mptl_exp_rxp(3) = '1';
        mptl_exp_txp(2) <= '0';
      end if;
      wait until mptl_exp_rxp(3) = '0';
      low := low + 6;
    end loop;
    if remainder > 0 then
      -- Remainder of payload, i.e. final slice (less than 6 bits long) of std_logic_vector value
      tmp := std_logic_vector(resize(unsigned(vec(low + remainder - 1 downto low)), 6));
      mptl_exp_txp(1 downto 0) <= tmp(5 downto 4);
      mptl_exp_txn(3 downto 0) <= tmp(3 downto 0);
      if handshake_use_quantum then
        mptl_exp_txp(2) <= '1' after handshake_quantum;
        wait until mptl_exp_rxp(3) = '1';
        mptl_exp_txp(2) <= '0' after handshake_quantum;
      else
        mptl_exp_txp(2) <= '1';
        wait until mptl_exp_rxp(3) = '1';
        mptl_exp_txp(2) <= '0';
      end if;
      wait until mptl_exp_rxp(3) = '0';
    end if;
  end;

  type port_in_enum_t is (
    port_ocp_clk,
    --
    port_ds_s2m_CmdAccept,
    port_ds_s2m_DataAccept,
    port_ds_s2m_Resp,
    port_ds_s2m_Data,
    port_ds_s2m_Tag,
    --
    port_dma0_s2m_CmdAccept,
    port_dma0_s2m_DataAccept,
    port_dma0_s2m_Resp,
    port_dma0_s2m_Data,
    port_dma0_s2m_Tag,
    --
    port_dma1_s2m_CmdAccept,
    port_dma1_s2m_DataAccept,
    port_dma1_s2m_Resp,
    port_dma1_s2m_Data,
    port_dma1_s2m_Tag,
    --
    port_dma_abort,
    --
    port_abmi_m2s_Cmd,
    port_abmi_m2s_Addr,
    port_abmi_m2s_BurstLength,
    port_abmi_m2s_Tag,
    port_abmi_m2s_DataValid,
    port_abmi_m2s_Data,
    port_abmi_m2s_DataByteEn,
    port_abmi_m2s_RespAccept,
    --
    port_gpio_t2b
  );

  type port_out_enum_t is (
    port_ds_m2s_Cmd,
    port_ds_m2s_Addr,
    port_ds_m2s_BurstLength,
    port_ds_m2s_Tag,
    port_ds_m2s_DataValid,
    port_ds_m2s_Data,
    port_ds_m2s_DataByteEn,
    port_ds_m2s_RespAccept,
    --
    port_dma0_m2s_Cmd,
    port_dma0_m2s_Addr,
    port_dma0_m2s_BurstLength,
    port_dma0_m2s_Tag,
    port_dma0_m2s_DataValid,
    port_dma0_m2s_Data,
    port_dma0_m2s_DataByteEn,
    port_dma0_m2s_RespAccept,
    --
    port_dma1_m2s_Cmd,
    port_dma1_m2s_Addr,
    port_dma1_m2s_BurstLength,
    port_dma1_m2s_Tag,
    port_dma1_m2s_DataValid,
    port_dma1_m2s_Data,
    port_dma1_m2s_DataByteEn,
    port_dma1_m2s_RespAccept,
    --
    port_abmi_s2m_CmdAccept,
    port_abmi_s2m_DataAccept,
    port_abmi_s2m_Resp,
    port_abmi_s2m_Data,
    port_abmi_s2m_Tag,
    --
    port_gpio_b2t
  );

  constant num_port_in  : natural := port_in_enum_t'pos(port_in_enum_t'high) + 1;
  constant num_port_out : natural := port_out_enum_t'pos(port_out_enum_t'high) + 1;

  type boolean_array_t is array(natural range <>) of boolean;

  signal pending_in : boolean_array_t(0 to num_port_in - 1) := (others => false);
  signal ack_in     : boolean_array_t(0 to num_port_in - 1) := (others => false);

  constant inertial_delay : time := 1 ns;

  -- Input ports
  signal i_ocp_clk : std_logic;

  signal i_ds_s2m_CmdAccept  : std_logic;
  signal i_ds_s2m_DataAccept : std_logic;
  signal i_ds_s2m_Resp       : std_logic_vector(1 downto 0);
  signal i_ds_s2m_Data       : std_logic_vector(127 downto 0);
  signal i_ds_s2m_Tag        : std_logic_vector(7 downto 0);

  signal i_dma0_s2m_CmdAccept  : std_logic;
  signal i_dma0_s2m_DataAccept : std_logic;
  signal i_dma0_s2m_Resp       : std_logic_vector(1 downto 0);
  signal i_dma0_s2m_Data       : std_logic_vector(127 downto 0);
  signal i_dma0_s2m_Tag        : std_logic_vector(7 downto 0);

  signal i_dma1_s2m_CmdAccept  : std_logic;
  signal i_dma1_s2m_DataAccept : std_logic;
  signal i_dma1_s2m_Resp       : std_logic_vector(1 downto 0);
  signal i_dma1_s2m_Data       : std_logic_vector(127 downto 0);
  signal i_dma1_s2m_Tag        : std_logic_vector(7 downto 0);

  signal i_dma_abort : std_logic_vector(1 downto 0);

  signal i_abmi_m2s_Cmd         : std_logic_vector(2 downto 0);
  signal i_abmi_m2s_Addr        : std_logic_vector(63 downto 0);
  signal i_abmi_m2s_BurstLength : std_logic_vector(11 downto 0);
  signal i_abmi_m2s_Data        : std_logic_vector(127 downto 0);
  signal i_abmi_m2s_DataByteEn  : std_logic_vector(15 downto 0);
  signal i_abmi_m2s_DataValid   : std_logic;
  signal i_abmi_m2s_RespAccept  : std_logic;
  signal i_abmi_m2s_Tag         : std_logic_vector(7 downto 0);

  signal i_gpio_t2b : std_logic_vector(15 downto 0);

  -- Output ports
  signal o_ds_m2s_Cmd         : std_logic_vector(2 downto 0);
  signal o_ds_m2s_Addr        : std_logic_vector(63 downto 0);
  signal o_ds_m2s_BurstLength : std_logic_vector(11 downto 0);
  signal o_ds_m2s_Data        : std_logic_vector(127 downto 0);
  signal o_ds_m2s_DataByteEn  : std_logic_vector(15 downto 0);
  signal o_ds_m2s_DataValid   : std_logic;
  signal o_ds_m2s_RespAccept  : std_logic;
  signal o_ds_m2s_Tag         : std_logic_vector(7 downto 0);

  signal o_dma0_m2s_Cmd         : std_logic_vector(2 downto 0);
  signal o_dma0_m2s_Addr        : std_logic_vector(63 downto 0);
  signal o_dma0_m2s_BurstLength : std_logic_vector(11 downto 0);
  signal o_dma0_m2s_Data        : std_logic_vector(127 downto 0);
  signal o_dma0_m2s_DataByteEn  : std_logic_vector(15 downto 0);
  signal o_dma0_m2s_DataValid   : std_logic;
  signal o_dma0_m2s_RespAccept  : std_logic;
  signal o_dma0_m2s_Tag         : std_logic_vector(7 downto 0);

  signal o_dma1_m2s_Cmd         : std_logic_vector(2 downto 0);
  signal o_dma1_m2s_Addr        : std_logic_vector(63 downto 0);
  signal o_dma1_m2s_BurstLength : std_logic_vector(11 downto 0);
  signal o_dma1_m2s_Data        : std_logic_vector(127 downto 0);
  signal o_dma1_m2s_DataByteEn  : std_logic_vector(15 downto 0);
  signal o_dma1_m2s_DataValid   : std_logic;
  signal o_dma1_m2s_RespAccept  : std_logic;
  signal o_dma1_m2s_Tag         : std_logic_vector(7 downto 0);

  signal o_abmi_s2m_CmdAccept  : std_logic;
  signal o_abmi_s2m_DataAccept : std_logic;
  signal o_abmi_s2m_Resp       : std_logic_vector(1 downto 0);
  signal o_abmi_s2m_Data       : std_logic_vector(127 downto 0);
  signal o_abmi_s2m_Tag        : std_logic_vector(7 downto 0);

  signal o_gpio_b2t : std_logic_vector(15 downto 0);

  -- mptl_clk_out 125MHz  
  constant mptl_clk_out_period : time := 8 ns;

  signal i_mptl_ready : std_logic := '0';
  signal mptl_clk_out_i : std_logic := '0';

begin

  -- Bridge to Target Sideband Signals
  gen_no_reset_sigs: if not wrapper_reset_signals generate
    mptl_target_configured_n <= '0' after 50 us;
    mptl_target_gtp_online_n <= mptl_bridge_gtp_online_n;
  end generate gen_no_reset_sigs;

  gen_reset_sigs: if wrapper_reset_signals generate
    -- Hold off MPTL target configured until all OCP clocks are stable (resets inactive)
    mptl_target_configured_n <= NOT ocp_ready;
    mptl_target_gtp_online_n <= mptl_bridge_gtp_online_n;
    i_mptl_ready <= NOT mptl_bridge_gtp_online_n;
  end generate gen_reset_sigs;

  mptl_ready <= i_mptl_ready;


  -- GTHE3_CHANNEL_TXOUTCLK[0] (125 MHz)
  mptl_clk_out_i <= not mptl_clk_out_i after mptl_clk_out_period;
  mptl_clk_out <= mptl_clk_out_i;

  -- Input ports
  i_ocp_clk   <= ocp_clk   after inertial_delay;
  i_gpio_t2b  <= gpio_t2b  after inertial_delay;
  i_dma_abort <= dma_abort after inertial_delay;

  gen_packed_signals_in: if not break_out_ocp generate
    i_ds_s2m_CmdAccept     <= ds_s2m(0)              after inertial_delay; -- CmdAccept 
    i_ds_s2m_DataAccept    <= ds_s2m(1)              after inertial_delay; -- DataAccept
    i_ds_s2m_Data          <= ds_s2m(129 downto 2)   after inertial_delay; -- Data      
    i_ds_s2m_Resp          <= ds_s2m(131 downto 130) after inertial_delay; -- Resp      
    i_ds_s2m_Tag           <= ds_s2m(139 downto 132) after inertial_delay; -- Tag       
    
    i_dma0_s2m_CmdAccept   <= dma_s2m(0)              after inertial_delay; -- CmdAccept 
    i_dma0_s2m_DataAccept  <= dma_s2m(1)              after inertial_delay; -- DataAccept
    i_dma0_s2m_Data        <= dma_s2m(129 downto 2)   after inertial_delay; -- Data      
    i_dma0_s2m_Resp        <= dma_s2m(131 downto 130) after inertial_delay; -- Resp      
    i_dma0_s2m_Tag         <= dma_s2m(139 downto 132) after inertial_delay; -- Tag       
    
    i_dma1_s2m_CmdAccept   <= dma_s2m(140 + 0)                    after inertial_delay; -- CmdAccept 
    i_dma1_s2m_DataAccept  <= dma_s2m(140 + 1)                    after inertial_delay; -- DataAccept
    i_dma1_s2m_Data        <= dma_s2m(140 + 129 downto 140 + 2)   after inertial_delay; -- Data      
    i_dma1_s2m_Resp        <= dma_s2m(140 + 131 downto 140 + 130) after inertial_delay; -- Resp      
    i_dma1_s2m_Tag         <= dma_s2m(140 + 139 downto 140 + 132) after inertial_delay; -- Tag       
        
    i_abmi_m2s_Cmd         <= abmi_m2s(2 downto 0)     after inertial_delay; -- Cmd        
    i_abmi_m2s_Addr        <= abmi_m2s(66 downto 3)    after inertial_delay; -- Addr       
    i_abmi_m2s_BurstLength <= abmi_m2s(78 downto 67)   after inertial_delay; -- BurstLength
    i_abmi_m2s_Data        <= abmi_m2s(206 downto 79)  after inertial_delay; -- Data       
    i_abmi_m2s_DataByteEn  <= abmi_m2s(222 downto 207) after inertial_delay; -- DataByteEn 
    i_abmi_m2s_DataValid   <= abmi_m2s(223)            after inertial_delay; -- DataValid  
    i_abmi_m2s_RespAccept  <= abmi_m2s(224)            after inertial_delay; -- RespAccept 
    i_abmi_m2s_Tag         <= abmi_m2s(232 downto 225) after inertial_delay; -- Tag        
  end generate gen_packed_signals_in;

  gen_broken_out_signals_in: if break_out_ocp generate
    i_ds_s2m_CmdAccept     <= ds_s2m_cmdaccept     after inertial_delay; -- CmdAccept 
    i_ds_s2m_DataAccept    <= ds_s2m_dataaccept    after inertial_delay; -- DataAccept
    i_ds_s2m_Data          <= ds_s2m_data          after inertial_delay; -- Data      
    i_ds_s2m_Resp          <= ds_s2m_resp          after inertial_delay; -- Resp      
    i_ds_s2m_Tag           <= ds_s2m_tag           after inertial_delay; -- Tag       
                           
    i_dma0_s2m_CmdAccept   <= dma0_s2m_cmdaccept   after inertial_delay; -- CmdAccept 
    i_dma0_s2m_DataAccept  <= dma0_s2m_dataaccept  after inertial_delay; -- DataAccept
    i_dma0_s2m_Data        <= dma0_s2m_data        after inertial_delay; -- Data      
    i_dma0_s2m_Resp        <= dma0_s2m_resp        after inertial_delay; -- Resp      
    i_dma0_s2m_Tag         <= dma0_s2m_tag         after inertial_delay; -- Tag       
                           
    i_dma1_s2m_CmdAccept   <= dma1_s2m_cmdaccept   after inertial_delay; -- CmdAccept 
    i_dma1_s2m_DataAccept  <= dma1_s2m_dataaccept  after inertial_delay; -- DataAccept
    i_dma1_s2m_Data        <= dma1_s2m_data        after inertial_delay; -- Data      
    i_dma1_s2m_Resp        <= dma1_s2m_resp        after inertial_delay; -- Resp      
    i_dma1_s2m_Tag         <= dma1_s2m_tag         after inertial_delay; -- Tag       
        
    i_abmi_m2s_Cmd         <= abmi_m2s_cmd         after inertial_delay; -- Cmd        
    i_abmi_m2s_Addr        <= abmi_m2s_addr        after inertial_delay; -- Addr       
    i_abmi_m2s_BurstLength <= abmi_m2s_burstlength after inertial_delay; -- BurstLength
    i_abmi_m2s_Data        <= abmi_m2s_data        after inertial_delay; -- Data       
    i_abmi_m2s_DataByteEn  <= abmi_m2s_databyteen  after inertial_delay; -- DataByteEn 
    i_abmi_m2s_DataValid   <= abmi_m2s_datavalid   after inertial_delay; -- DataValid  
    i_abmi_m2s_RespAccept  <= abmi_m2s_respaccept  after inertial_delay; -- RespAccept 
    i_abmi_m2s_Tag         <= abmi_m2s_tag         after inertial_delay; -- Tag   
  end generate gen_broken_out_signals_in;

  -- Output ports
  gpio_b2t <= o_gpio_b2t;

  gen_packed_signals_out: if not break_out_ocp generate
    ds_m2s(2 downto 0)    <= o_ds_m2s_Cmd;
    ds_m2s(66 downto 3)   <= o_ds_m2s_Addr;
    ds_m2s(78 downto 67)  <= o_ds_m2s_BurstLength;
    ds_m2s(206 downto 79) <= o_ds_m2s_Data;
    ds_m2s(222 downto 207)<= o_ds_m2s_DataByteEn;
    ds_m2s(223)           <= o_ds_m2s_DataValid;
    ds_m2s(224)           <= o_ds_m2s_RespAccept;
    ds_m2s(232 downto 225)<= o_ds_m2s_Tag;
    
    dma_m2s(2 downto 0)     <= o_dma0_m2s_Cmd;
    dma_m2s(66 downto 3)    <= o_dma0_m2s_Addr;
    dma_m2s(78 downto 67)   <= o_dma0_m2s_BurstLength;
    dma_m2s(206 downto 79)  <= o_dma0_m2s_Data;
    dma_m2s(222 downto 207) <= o_dma0_m2s_DataByteEn;
    dma_m2s(223)            <= o_dma0_m2s_DataValid;
    dma_m2s(224)            <= o_dma0_m2s_RespAccept;
    dma_m2s(232 downto 225) <= o_dma0_m2s_Tag;
    
    dma_m2s(233 + 2 downto 233 + 0)     <= o_dma1_m2s_Cmd;
    dma_m2s(233 + 66 downto 233 + 3)    <= o_dma1_m2s_Addr;
    dma_m2s(233 + 78 downto 233 + 67)   <= o_dma1_m2s_BurstLength;
    dma_m2s(233 + 206 downto 233 + 79)  <= o_dma1_m2s_Data;
    dma_m2s(233 + 222 downto 233 + 207) <= o_dma1_m2s_DataByteEn;
    dma_m2s(233 + 223)                  <= o_dma1_m2s_DataValid;
    dma_m2s(233 + 224)                  <= o_dma1_m2s_RespAccept;
    dma_m2s(233 + 232 downto 233 + 225) <= o_dma1_m2s_Tag;
    
    abmi_s2m(0)              <= o_abmi_s2m_CmdAccept;
    abmi_s2m(1)              <= o_abmi_s2m_DataAccept;
    abmi_s2m(129 downto 2)   <= o_abmi_s2m_Data;
    abmi_s2m(131 downto 130) <= o_abmi_s2m_Resp;
    abmi_s2m(139 downto 132) <= o_abmi_s2m_Tag;

    -- Unused broken_out outputs
    ds_m2s_cmd           <= (others => '0');
    ds_m2s_addr          <= (others => '0');
    ds_m2s_burstlength   <= (others => '0');
    ds_m2s_data          <= (others => '0');
    ds_m2s_databyteen    <= (others => '0');
    ds_m2s_datavalid     <= '0';
    ds_m2s_respaccept    <= '0';
    ds_m2s_tag           <= (others => '0');

    dma0_m2s_cmd         <= (others => '0');
    dma0_m2s_addr        <= (others => '0');
    dma0_m2s_burstlength <= (others => '0');
    dma0_m2s_data        <= (others => '0');
    dma0_m2s_databyteen  <= (others => '0');
    dma0_m2s_datavalid   <= '0';
    dma0_m2s_respaccept  <= '0';
    dma0_m2s_tag         <= (others => '0');

    dma1_m2s_cmd         <= (others => '0');
    dma1_m2s_addr        <= (others => '0');
    dma1_m2s_burstlength <= (others => '0');
    dma1_m2s_data        <= (others => '0');
    dma1_m2s_databyteen  <= (others => '0');
    dma1_m2s_datavalid   <= '0';
    dma1_m2s_respaccept  <= '0';
    dma1_m2s_tag         <= (others => '0');

    abmi_s2m_cmdaccept   <= '0';
    abmi_s2m_dataaccept  <= '0';
    abmi_s2m_data        <= (others => '0');
    abmi_s2m_resp        <= (others => '0');
    abmi_s2m_tag         <= (others => '0');
  end generate gen_packed_signals_out;

  gen_broken_out_signals_out: if break_out_ocp generate
    ds_m2s_cmd           <= o_ds_m2s_Cmd;
    ds_m2s_addr          <= o_ds_m2s_Addr;
    ds_m2s_burstlength   <= o_ds_m2s_BurstLength;
    ds_m2s_data          <= o_ds_m2s_Data;
    ds_m2s_databyteen    <= o_ds_m2s_DataByteEn;
    ds_m2s_datavalid     <= o_ds_m2s_DataValid;
    ds_m2s_respaccept    <= o_ds_m2s_RespAccept;
    ds_m2s_tag           <= o_ds_m2s_Tag;

    dma0_m2s_cmd         <= o_dma0_m2s_Cmd;
    dma0_m2s_addr        <= o_dma0_m2s_Addr;
    dma0_m2s_burstlength <= o_dma0_m2s_BurstLength;
    dma0_m2s_data        <= o_dma0_m2s_Data;
    dma0_m2s_databyteen  <= o_dma0_m2s_DataByteEn;
    dma0_m2s_datavalid   <= o_dma0_m2s_DataValid;
    dma0_m2s_respaccept  <= o_dma0_m2s_RespAccept;
    dma0_m2s_tag         <= o_dma0_m2s_Tag;

    dma1_m2s_cmd         <= o_dma1_m2s_Cmd;
    dma1_m2s_addr        <= o_dma1_m2s_Addr;
    dma1_m2s_burstlength <= o_dma1_m2s_BurstLength;
    dma1_m2s_data        <= o_dma1_m2s_Data;
    dma1_m2s_databyteen  <= o_dma1_m2s_DataByteEn;
    dma1_m2s_datavalid   <= o_dma1_m2s_DataValid;
    dma1_m2s_respaccept  <= o_dma1_m2s_RespAccept;
    dma1_m2s_tag         <= o_dma1_m2s_Tag;

    abmi_s2m_cmdaccept   <= o_abmi_s2m_CmdAccept;
    abmi_s2m_dataaccept  <= o_abmi_s2m_DataAccept;
    abmi_s2m_data        <= o_abmi_s2m_Data;
    abmi_s2m_resp        <= o_abmi_s2m_Resp;
    abmi_s2m_tag         <= o_abmi_s2m_Tag;

    -- Unused packed outputs
    ds_m2s   <= (others => '0');
    dma_m2s  <= (others => '0');
    abmi_s2m <= (others => '0');
  end generate gen_broken_out_signals_out;


  -- ocp_clk
  propagate_ocp_clk : process
    constant index : natural := port_in_enum_t'pos(port_ocp_clk);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ocp_clk;
    end loop;
  end process;

  -- ds_s2m
  propagate_ds_s2m_CmdAccept : process
    constant index : natural := port_in_enum_t'pos(port_ds_s2m_CmdAccept);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_s2m_CmdAccept;
    end loop;
  end process;

  propagate_ds_s2m_DataAccept : process
    constant index : natural := port_in_enum_t'pos(port_ds_s2m_DataAccept);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_s2m_DataAccept;
    end loop;
  end process;

  propagate_ds_s2m_Resp : process
    constant index : natural := port_in_enum_t'pos(port_ds_s2m_Resp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_s2m_Resp;
    end loop;
  end process;

  propagate_ds_s2m_Data : process
    constant index : natural := port_in_enum_t'pos(port_ds_s2m_Data);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_s2m_Data;
    end loop;
  end process;

  propagate_ds_s2m_Tag : process
    constant index : natural := port_in_enum_t'pos(port_ds_s2m_Tag);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_s2m_Tag;
    end loop;
  end process;

  -- dma0_s2m
  propagate_dma0_s2m_CmdAccept : process
    constant index : natural := port_in_enum_t'pos(port_dma0_s2m_CmdAccept);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_s2m_CmdAccept;
    end loop;
  end process;

  propagate_dma0_s2m_DataAccept : process
    constant index : natural := port_in_enum_t'pos(port_dma0_s2m_DataAccept);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_s2m_DataAccept;
    end loop;
  end process;

  propagate_dma0_s2m_Resp : process
    constant index : natural := port_in_enum_t'pos(port_dma0_s2m_Resp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_s2m_Resp;
    end loop;
  end process;

  propagate_dma0_s2m_Data : process
    constant index : natural := port_in_enum_t'pos(port_dma0_s2m_Data);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_s2m_Data;
    end loop;
  end process;

  propagate_dma0_s2m_Tag : process
    constant index : natural := port_in_enum_t'pos(port_dma0_s2m_Tag);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_s2m_Tag;
    end loop;
  end process;

  -- dma1_s2m
  propagate_dma1_s2m_CmdAccept : process
    constant index : natural := port_in_enum_t'pos(port_dma1_s2m_CmdAccept);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_s2m_CmdAccept;
    end loop;
  end process;

  propagate_dma1_s2m_DataAccept : process
    constant index : natural := port_in_enum_t'pos(port_dma1_s2m_DataAccept);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_s2m_DataAccept;
    end loop;
  end process;

  propagate_dma1_s2m_Resp : process
    constant index : natural := port_in_enum_t'pos(port_dma1_s2m_Resp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_s2m_Resp;
    end loop;
  end process;

  propagate_dma1_s2m_Data : process
    constant index : natural := port_in_enum_t'pos(port_dma1_s2m_Data);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_s2m_Data;
    end loop;
  end process;

  propagate_dma1_s2m_Tag : process
    constant index : natural := port_in_enum_t'pos(port_dma1_s2m_Tag);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_s2m_Tag;
    end loop;
  end process;

  -- dma_abort
  propagate_dma_abort : process
    constant index : natural := port_in_enum_t'pos(port_dma_abort);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma_abort;
    end loop;
  end process;

  -- abmi_m2s
  propagate_abmi_m2s_Cmd : process
    constant index : natural := port_in_enum_t'pos(port_abmi_m2s_Cmd);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_m2s_Cmd;
    end loop;
  end process;

  propagate_abmi_m2s_Addr : process
    constant index : natural := port_in_enum_t'pos(port_abmi_m2s_Addr);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_m2s_Addr;
    end loop;
  end process;

  propagate_abmi_m2s_BurstLength : process
    constant index : natural := port_in_enum_t'pos(port_abmi_m2s_BurstLength);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_m2s_BurstLength;
    end loop;
  end process;

  propagate_abmi_m2s_Tag : process
    constant index : natural := port_in_enum_t'pos(port_abmi_m2s_Tag);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_m2s_Tag;
    end loop;
  end process;

  propagate_abmi_m2s_DataValid : process
    constant index : natural := port_in_enum_t'pos(port_abmi_m2s_DataValid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_m2s_DataValid;
    end loop;
  end process;

  propagate_abmi_m2s_Data : process
    constant index : natural := port_in_enum_t'pos(port_abmi_m2s_Data);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_m2s_Data;
    end loop;
  end process;

  propagate_abmi_m2s_DataByteEn : process
    constant index : natural := port_in_enum_t'pos(port_abmi_m2s_DataByteEn);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_m2s_DataByteEn;
    end loop;
  end process;

  propagate_abmi_m2s_RespAccept : process
    constant index : natural := port_in_enum_t'pos(port_abmi_m2s_RespAccept);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_abmi_m2s_RespAccept;
    end loop;
  end process;

  -- gpio_t2b
  propagate_gpio_t2b : process
    constant index : natural := port_in_enum_t'pos(port_gpio_t2b);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_gpio_t2b;
    end loop;
  end process;


  transmit_signals : process
    variable port_in : port_in_enum_t;
    variable work_done : boolean;
  begin
    loop
      loop
        work_done := false;
        for index in pending_in'range loop
          if pending_in(index) then
            work_done := true;

            -- Send current value of port to other side of link
            port_in := port_in_enum_t'val(index);
--            report "bm transmiting index " & integer'image(index) & " port " & port_in_enum_t'image(port_in);

            case port_in is
              -- 
              when port_ocp_clk              => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_ocp_clk);
              -- 
              when port_ds_s2m_CmdAccept     => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_ds_s2m_CmdAccept);
              when port_ds_s2m_DataAccept    => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_ds_s2m_DataAccept);
              when port_ds_s2m_Resp          => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_ds_s2m_Resp);
              when port_ds_s2m_Data          => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_ds_s2m_Data);
              when port_ds_s2m_Tag           => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_ds_s2m_Tag);
              -- 
              when port_dma0_s2m_CmdAccept   => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma0_s2m_CmdAccept);
              when port_dma0_s2m_DataAccept  => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma0_s2m_DataAccept);
              when port_dma0_s2m_Resp        => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma0_s2m_Resp);
              when port_dma0_s2m_Data        => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma0_s2m_Data);
              when port_dma0_s2m_Tag         => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma0_s2m_Tag);
              --
              when port_dma1_s2m_CmdAccept   => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma1_s2m_CmdAccept);
              when port_dma1_s2m_DataAccept  => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma1_s2m_DataAccept);
              when port_dma1_s2m_Resp        => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma1_s2m_Resp);
              when port_dma1_s2m_Data        => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma1_s2m_Data);
              when port_dma1_s2m_Tag         => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma1_s2m_Tag);
              --
              when port_dma_abort            => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_dma_abort);
              --
              when port_abmi_m2s_Cmd         => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_m2s_Cmd);
              when port_abmi_m2s_Addr        => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_m2s_Addr);
              when port_abmi_m2s_BurstLength => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_m2s_BurstLength);
              when port_abmi_m2s_Tag         => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_m2s_Tag);
              when port_abmi_m2s_DataValid   => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_m2s_DataValid);
              when port_abmi_m2s_Data        => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_m2s_Data);
              when port_abmi_m2s_DataByteEn  => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_m2s_DataByteEn);
              when port_abmi_m2s_RespAccept  => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_abmi_m2s_RespAccept);
              --
              when port_gpio_t2b             => transmit(mptl_tx_n(3 downto 0), mptl_tx_p(2 downto 0), mptl_rx_p(3 downto 3), index, i_gpio_t2b);
              --
              when others => report "transmit: case " & port_in_enum_t'image(port_in) & " is not handled.";-- severity failure;
            end case;

            -- Indicate that signal has been updated
            if handshake_use_quantum then
              ack_in(index) <= true after handshake_quantum;
            else
              ack_in(index) <= true;
            end if;
            
            -- Complete the handshake
            wait until not pending_in(index);
            if handshake_use_quantum then
              ack_in(index) <= false after handshake_quantum;
            else
              ack_in(index) <= false;
            end if;

            if transmit_use_quantum then
              -- Avoid excessive numbers of delta cycles at a given simulation time.
              wait for transmit_quantum;
            end if;
          end if;
        end loop;
        exit when not work_done;
      end loop;
      wait on pending_in;
    end loop;
  end process;


  receive_signals : process
    variable index : natural;
    variable val : std_logic;
    variable port_out : port_out_enum_t;
  begin
    loop
      receive_header(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), index, val);
      port_out := port_out_enum_t'val(index);
--      report "bm receiving index " & integer'image(index) & " port " & port_out_enum_t'image(port_out) & " val " & std_logic'image(val);

      case port_out is
        when port_ds_m2s_Cmd           => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_ds_m2s_Cmd);
        when port_ds_m2s_Addr          => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_ds_m2s_Addr);
        when port_ds_m2s_BurstLength   => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_ds_m2s_BurstLength);
        when port_ds_m2s_Tag           => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_ds_m2s_Tag);
        when port_ds_m2s_DataValid     => o_ds_m2s_DataValid <= val;
        when port_ds_m2s_Data          => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_ds_m2s_Data);
        when port_ds_m2s_DataByteEn    => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_ds_m2s_DataByteEn);
        when port_ds_m2s_RespAccept    => o_ds_m2s_RespAccept <= val;
        --
        when port_dma0_m2s_Cmd         => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_dma0_m2s_Cmd);
        when port_dma0_m2s_Addr        => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_dma0_m2s_Addr);
        when port_dma0_m2s_BurstLength => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_dma0_m2s_BurstLength);
        when port_dma0_m2s_Tag         => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_dma0_m2s_Tag);
        when port_dma0_m2s_DataValid   => o_dma0_m2s_DataValid <= val;
        when port_dma0_m2s_Data        => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_dma0_m2s_Data);
        when port_dma0_m2s_DataByteEn  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_dma0_m2s_DataByteEn);
        when port_dma0_m2s_RespAccept  => o_dma0_m2s_RespAccept <= val;
        --
        when port_dma1_m2s_Cmd         => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_dma1_m2s_Cmd);
        when port_dma1_m2s_Addr        => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_dma1_m2s_Addr);
        when port_dma1_m2s_BurstLength => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_dma1_m2s_BurstLength);
        when port_dma1_m2s_Tag         => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_dma1_m2s_Tag);
        when port_dma1_m2s_DataValid   => o_dma1_m2s_DataValid <= val;
        when port_dma1_m2s_Data        => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_dma1_m2s_Data);
        when port_dma1_m2s_DataByteEn  => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_dma1_m2s_DataByteEn);
        when port_dma1_m2s_RespAccept  => o_dma1_m2s_RespAccept <= val;
        --
        when port_abmi_s2m_CmdAccept   => o_abmi_s2m_CmdAccept <= val;
        when port_abmi_s2m_DataAccept  => o_abmi_s2m_DataAccept <= val;
        when port_abmi_s2m_Resp        => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_abmi_s2m_Resp);
        when port_abmi_s2m_Data        => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_abmi_s2m_Data);
        when port_abmi_s2m_Tag         => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_abmi_s2m_Tag);
        --
        when port_gpio_b2t             => receive(mptl_rx_n(3 downto 0), mptl_rx_p(2 downto 0), mptl_tx_p(3 downto 3), o_gpio_b2t);
        --
        when others => report "receive: case " & port_out_enum_t'image(port_out) & " is not handled.";-- severity failure;
      end case;
    end loop;
  end process;

end architecture;
