/*
** loader.c - Configure target FPGA with a .bit file and then exit.
**
** (C) Copyright 2012, 2014-2016 Alpha Data
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _tscanf_s(buffer, ...) _tscanf(buffer, __VA_ARGS__) /* This is OK unless the format string contains %s */
#   if !defined(_tcserror)
#     /* For some reason, 32-bit MINGW doesn't have _tcserror */
#     define _tcserror strerror
#   endif
# endif

typedef struct _stat sys_stat_t;

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tmain main
# define _tprintf printf
# define _tcserror strerror
# define _tfopen fopen
# define _tstat stat
# define FALSE (0)
# define TRUE (1)

typedef struct stat sys_stat_t;

#endif

#if defined(__INTEL_COMPILER)
/* Prevent warnings from ICC compiler of the form "external function definition with no prior declaration" */
# pragma warning(disable:1418)
#endif

#include <admxrc3.h>

#include <sdk_common.h>

/* Exit codes */
#define EXIT_OK                      (0) /* No errors */
#define EXIT_BAD_COMMAND_LINE        (1) /* Illegal basic command-line syntax */
#define EXIT_DEVICE_OPEN_ERROR       (2) /* Failed to open ADMXRC3 device */
#define EXIT_STAT_FILE_ERROR         (3) /* Failed to stat() .bit file */
#define EXIT_FILE_TOO_LARGE          (4) /* .bit file too large to load into memory */
#define EXIT_ALLOCATION_FAILED       (5) /* Failed to allocate buffer for firmware or board config data */
#define EXIT_FILE_OPEN_ERROR         (6) /* Failed to open .bit file for reading */
#define EXIT_FILE_READ_ERROR         (7) /* Failed to read .bit file */
#define EXIT_CONFIGURATION_FAILED    (8) /* ADMXRC3_Configure{FromBuffer,FromFile} failed */

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#if !defined(SIZE_T_MAX)
# define SIZE_T_MAX ((size_t)-1)
#endif

#if !defined(_MSC_VER)
static int
_tfopen_s(
  FILE** ppFile,
  const TCHAR* pFilename,
  const TCHAR* pMode)
{
  FILE* pFile;

  pFile = _tfopen(pFilename, pMode);
  if (NULL != pFile) {
    *ppFile = pFile;
    return 0;
  } else {
    return errno;
  }
}

static int
_tcserror_s(
  TCHAR* pBuffer,
  size_t n,
  int err)
{
  const TCHAR* p = _tcserror(err);
  size_t i;

  if (n == 0) {
    return 0;
  }
  if (NULL == pBuffer) {
    return EINVAL;
  }
  for (i = 0; i < n; i++) {
    pBuffer[i] = p[i];
    if (p[i] == _T('\0')) {
      break;
    }
  }
  pBuffer[n - 1] = _T('\0');
  return 0;
}
#endif

static int
doConfiguration(
  ADMXRC3_HANDLE hDevice,
  unsigned int targetIndex,
  const TCHAR* pFilename,
#if ADMXRC3_H_VERSION >= ADMXRC3_H_MAKE_VERSION(1, 8, 0)
  bool_t bIgnoreMismatch,
#endif
  bool_t bIsBinaryFile,
  bool_t bCheckLinkStatus)
{
  ADMXRC3_STATUS status;
  FILE* pBinaryFile = NULL;
  void* pData = NULL;
  int ret = EXIT_OK;
  
  if (bIsBinaryFile) {
    size_t n, nRead;
    int err;
    TCHAR buf[128];
    sys_stat_t s;

    err = _tstat(pFilename, &s);
    if (0 != err) {
      buf[0] = _T('\0');
      _tcserror_s(buf, ARRAY_LENGTH(buf), err);
      _tprintf(_T("*** Failed to stat binary file '%s': %s\n"), pFilename, buf);
      ret = EXIT_STAT_FILE_ERROR;
      goto done;
    }
    if (s.st_size > SIZE_T_MAX) {
      _tprintf(_T("*** Binary file '%s' of %llu(0x%llx) bytes is too large to load\n"), pFilename,
        (unsigned long long)s.st_size, (unsigned long long)s.st_size);
      ret = EXIT_FILE_TOO_LARGE;
      goto done;
    }
    n = (size_t)s.st_size;
    pData = malloc(n);
    if (NULL == pData) {
      _tprintf(_T("*** Failed to allocate buffer of %llu(0x%llx) byte(s) to read binary file '%s'\n"),
        (unsigned long long)n, (unsigned long long)n, pFilename);
      ret = EXIT_ALLOCATION_FAILED;
      goto done;
    }
    err = _tfopen_s(&pBinaryFile, pFilename, _T("rb"));
    if (0 != err) {
      buf[0] = _T('\0');
      _tcserror_s(buf, ARRAY_LENGTH(buf), err);
      _tprintf(_T("*** Failed to open binary file '%s' for reading: %s\n"), pFilename, buf);
      ret = EXIT_FILE_OPEN_ERROR;
      goto done;
    }
    nRead = fread(pData, sizeof(uint8_t), n, pBinaryFile);
    if (nRead != n) {
      _tprintf(_T("*** Only managed to read %llu(0x%llx) of %llu(0x%llx) bytes for binary file '%s'\n"),
        (unsigned long long)nRead, (unsigned long long)nRead,
        (unsigned long long)n, (unsigned long long)n,
        pFilename);
      ret = EXIT_FILE_READ_ERROR;
      goto done;
    }
    fclose(pBinaryFile);
    pBinaryFile = NULL;
    status = ADMXRC3_ConfigureFromBuffer(hDevice, targetIndex, bCheckLinkStatus ? 0 : ADMXRC3_CONFIGURE_NOCHECK, pData, n);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to configure target FPGA with binary file '%s': %s\n"), pFilename, ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_CONFIGURATION_FAILED;
      goto done;
    }
  } else {
    uint32_t flags = 0;
    if (!bCheckLinkStatus) {
      flags |= ADMXRC3_CONFIGURE_NOCHECK;
    }
#if ADMXRC3_H_VERSION >= ADMXRC3_H_MAKE_VERSION(1, 8, 0)
    if (bIgnoreMismatch) {
      flags |= ADMXRC3_CONFIGURE_IGNOREMISMATCH;
    }
#endif
    status = ADMXRC3_ConfigureFromFile(hDevice, targetIndex, flags, pFilename);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to configure target FPGA with '%s': %s\n"), pFilename, ADMXRC3_GetStatusString(status, TRUE));
      ret = EXIT_CONFIGURATION_FAILED;
      goto done;
    }
  }

done:
  if (NULL != pData) {
    free(pData);
  }
  if (NULL != pBinaryFile) {
    fclose(pBinaryFile);
  }

  return ret;
}
  
#if (defined(__VXWORKS__) || defined(__vxworks))
  
/***************************************************
 * VxWorks DKM entry points                        *
 ***************************************************/
 
/* Bits for 'flags' parameter */
#define FLAG_BYSERIAL       (0x1U)  /* Interpret 'index' as a serial number */
#define FLAG_BINARY         (0x10U) /* Treat pFilename as binary SelectMap data, not a .bit file */
#define FLAG_NOLINKCHECK    (0x20U) /* Don't check link status */
#if ADMXRC3_H_VERSION >= ADMXRC3_H_MAKE_VERSION(1, 8, 0)
# define FLAG_IGNOREMISMATCH (0x40U) /* Ignore mismatch between FPGA in device and FPGA identifier string in .bit file */
#endif

int
admxrc3LoaderHelp(
  void)
{
  _tprintf(_T("Available entry points:\n"));
  _tprintf(_T("  int admxrc3LoaderHelp()\n"));
  _tprintf(_T("  int admxrc3Loader(indexOrSerial, flags, targetIndex, pFilename)\n"));
  _tprintf(_T("'flags' is a bitmask of:\n"));
  _tprintf(_T("  FLAG_BYSERIAL       0x%02X => Interpret 'indexOrSerial' as a serial number\n"), FLAG_BYSERIAL);
  _tprintf(_T("  FLAG_BINARY         0x%02X => Treat pFilename as binary SelectMap data, not a .bit file\n"), FLAG_BINARY);
  _tprintf(_T("  FLAG_NOLINKCHECK    0x%02X => Don't check link status\n"), FLAG_NOLINKCHECK);
#if ADMXRC3_H_VERSION >= ADMXRC3_H_MAKE_VERSION(1, 8, 0)
  _tprintf(_T("  FLAG_IGNOREMISMATCH 0x%02X => Ignore mismatch between FPGA in device and FPGA identifier string in .bit file\n"), FLAG_IGNOREMISMATCH);
#endif
  _tprintf(_T("A return value of 0 indicates success.\n"));
  return 0;
}

int
admxrc3Loader(
  unsigned int indexOrSerial,
  unsigned int flags,
  unsigned int targetIndex,
  const TCHAR* pFilename)
{
  ADMXRC3_STATUS status;
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  bool_t bOpenBySerial = FALSE, bIsBinaryFile = FALSE, bNoCheckLink = FALSE;
#if ADMXRC3_H_VERSION >= ADMXRC3_H_MAKE_VERSION(1, 8, 0)
  bool_t bIgnoreMismatch = FALSE;
#endif
  int ret = EXIT_OK;
  
  if (flags & FLAG_BYSERIAL) {
    bOpenBySerial = TRUE;
  }
  if (flags & FLAG_BINARY) {
    bIsBinaryFile = TRUE;
  }
  if (flags & FLAG_NOLINKCHECK) {
    bNoCheckLink = TRUE;
  }
#if ADMXRC3_H_VERSION >= ADMXRC3_H_MAKE_VERSION(1, 8, 0)
  if (flags & FLAG_IGNOREMISMATCH) {
    bIgnoreMismatch = TRUE;
  }
#endif

  status = sampleOpenCard(indexOrSerial, !bOpenBySerial, indexOrSerial, bOpenBySerial, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }

  ret = doConfiguration(
    hDevice,
    targetIndex,
    pFilename,
#if ADMXRC3_H_VERSION >= ADMXRC3_H_MAKE_VERSION(1, 8, 0)
    bIgnoreMismatch,
#endif
    bIsBinaryFile,
    !bNoCheckLink);

done:
  if (hDevice != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hDevice);
  }
  
  return ret;
}

#endif

#if !(defined(__VXWORKS__) || defined(__vxworks))

/***************************************************
 * Windows/Linux entry point                       *
 ***************************************************/
 
int
_tmain(
  int argc,
  TCHAR *argv[])
{
  ADMXRC3_STATUS status;
  unsigned int index, serial;
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
#if ADMXRC3_H_VERSION >= ADMXRC3_H_MAKE_VERSION(1, 8, 0)
  Option options[5];
#else
  Option options[4];
#endif
  OptionType paramType[2];
  TCHAR* paramInfo[2];
  TCHAR* paramMeaning[2];
  Arguments arguments = g_initArguments;
  bool_t bIsBinaryFile, bCheckLinkStatus;
#if ADMXRC3_H_VERSION >= ADMXRC3_H_MAKE_VERSION(1, 8, 0)
  bool_t bIgnoreMismatch;
#endif
  unsigned int targetIndex;
  const TCHAR* pFilename;
  int ret = EXIT_OK;

  /* Set up expected options */
  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("binary");
  options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
  options[2].meaning = _T("file contains raw SelectMap data (not a .bit file)");
  options[3].key = _T("checklink");
  options[3].type = OptionBoolean;
  options[3].def.booleanVal = TRUE;
  options[3].meaning = _T("check link status");
#if ADMXRC3_H_VERSION >= ADMXRC3_H_MAKE_VERSION(1, 8, 0)
  options[4].key = _T("force");
  options[4].type = OptionBoolean;
  options[4].def.booleanVal = FALSE;
  options[4].meaning = _T("ignore device mismatch");
#endif
  paramType[0] = OptionUInt;
  paramInfo[0] = _T("<target FPGA index>");
  paramMeaning[0] = _T("index of the target FPGA to configure");
  paramType[1] = OptionPath;
  paramInfo[1] = _T("<filename.bit>");
  paramMeaning[1] = _T("file with which to configure FPGA");
  arguments.nOption = ARRAY_LENGTH(options);
  arguments.option = options;
  arguments.nParamType = ARRAY_LENGTH(paramType);
  arguments.minNParam = ARRAY_LENGTH(paramType);
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;
  arguments.bNoStdOptions = TRUE;

  if (sampleParseCommandLine(_T("loader"), argc, argv, &arguments) < 0) {
    ret = EXIT_BAD_COMMAND_LINE;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  bIsBinaryFile = options[2].value.booleanVal;
  bCheckLinkStatus = options[3].value.booleanVal;
#if ADMXRC3_H_VERSION >= ADMXRC3_H_MAKE_VERSION(1, 8, 0)
  bIgnoreMismatch = options[4].value.booleanVal; /* If TRUE, requires ADB3 Driver 1.4.17 or later */
#endif
  targetIndex = arguments.param[0].uintVal;
  pFilename = arguments.param[1].pPathVal;

  status = sampleOpenCard(index, options[0].specified, serial, options[1].specified, FALSE, &hDevice);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = EXIT_DEVICE_OPEN_ERROR;
    goto done;
  }

  ret = doConfiguration(
    hDevice,
    targetIndex,
    pFilename,
#if ADMXRC3_H_VERSION >= ADMXRC3_H_MAKE_VERSION(1, 8, 0)
    bIgnoreMismatch,
#endif
    bIsBinaryFile,
    bCheckLinkStatus);

done:
  if (hDevice != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hDevice);
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}

#endif
