#include "clockgenwnd.h"

#if (GTK_MAJOR_VERSION >= 3)
# define APPEND_TEXT(ctl, text) ((ctl).append(text))
#else
# define APPEND_TEXT(ctl, text) ((ctl).append_text(text))
#endif

// -------------------------------------------------------------------------
// Start of ISensorInfoWnd interface
// -------------------------------------------------------------------------

void
CClockGenWnd::deviceChanged()
{
  const DeviceInfo& devInfo = m_pParent->getDeviceInfo();
  if (devInfo.m_hDevice != ADMXRC3_HANDLE_INVALID_VALUE) {
    m_model->clear();
  }
  if (ADMXRC3_HANDLE_INVALID_VALUE != devInfo.m_hDevice) {
    for (unsigned int i = 0; i < devInfo.m_numClockGen; i++) {
      Gtk::TreeModel::Row row = *m_model->append();
      row[m_modelColumns.m_index] = (unsigned int)i; 
      const TCHAR* pTmp = CAdmxrc3Help::getClockGenDescription(i, devInfo.m_deviceInfo.Model);
      if (NULL == pTmp) {
        row[m_modelColumns.m_description] = "Unknown";
      } else {
        row[m_modelColumns.m_description] = pTmp;
      }
    }
  }
}

void
CClockGenWnd::numDeviceChanged()
{
  unsigned int numDevice = m_pParent->getDeviceInfo().m_numDevice;

  if ((m_numDevice == 0) == (numDevice == 0)) {
    return;
  }

  /* Nothing to do yet */

  m_numDevice = numDevice;
}

void
CClockGenWnd::update()
{
  const DeviceInfo& devInfo = m_pParent->getDeviceInfo();
  double frequency;
  unsigned int n = devInfo.m_numClockGen;
  ADMXRC3_STATUS status;

  Glib::RefPtr<Gtk::TreeModel> model = m_listCtl.get_model();

  for (unsigned int i = 0; i < n; i++) {
    if (m_freqEditIndex != (int)i) {
      Gtk::TreeModel::Row row = m_model->children()[i];
      std::ostringstream ss;
      status = ADMXRC3_GetClockFrequency(devInfo.m_hDevice, i, &frequency);
      if (ADMXRC3_SUCCESS == status) {
        ss.precision(6);
        ss << (frequency / 1.0e6);
        row[m_modelColumns.m_frequency] = ss.str();
      } else {
        row[m_modelColumns.m_frequency] = "Error getting frequency";
      }
    }
  }
}

// -------------------------------------------------------------------------
// End of ISensorInfoWnd interface
// -------------------------------------------------------------------------

CClockGenWnd::CClockGenWnd(
  ISysMonWnd* pParent,
  bool_t bVisible)
{
  m_pParent = pParent;
  m_numDevice = 0;
  m_freqEditIndex = -1;

  set_size_request(-1, 300);

  m_model = Gtk::ListStore::create(m_modelColumns);
  m_listCtl.set_model(m_model);
  m_listCtl.set_headers_visible(TRUE);
  m_listCtl.append_column("#", m_modelColumns.m_index);
  m_listCtl.append_column("Description", m_modelColumns.m_description);
  m_listCtl.append_column_editable("Frequency (MHz)", m_modelColumns.m_frequency);
  Gtk::CellRendererText* m_pFrequencyRenderer = dynamic_cast<Gtk::CellRendererText*>(m_listCtl.get_column_cell_renderer(2));
  assert(NULL != m_pFrequencyRenderer);
  m_pFrequencyRenderer->signal_editing_started().connect(
    sigc::mem_fun(
      *this,
      &CClockGenWnd::onEditingStarted));
  m_pFrequencyRenderer->signal_edited().connect(
    sigc::mem_fun(
      *this,
      &CClockGenWnd::onEdited));
  m_pFrequencyRenderer->signal_editing_canceled().connect(
    sigc::mem_fun(
      *this,
      &CClockGenWnd::onEditingCancelled));

  m_label.set_padding(8, 8);
  m_label.set_label("If executing as root, frequency values can be changed");

  m_vboxCtl.pack_start(m_listCtl, TRUE, TRUE);
  m_vboxCtl.pack_start(m_label, FALSE, FALSE);

  add(m_vboxCtl);
  set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
}

CClockGenWnd::~CClockGenWnd()
{
}

void
CClockGenWnd::onEditingStarted(
  Gtk::CellEditable* pCell,
  const Glib::ustring& path)
{
  Gtk::TreeModel::iterator iter = m_model->get_iter(path);
  if (iter) {
    Gtk::TreeModel::Row row = *iter;
    m_freqEditIndex = (int)row[m_modelColumns.m_index];
  }
}

void
CClockGenWnd::onEdited(
  const Glib::ustring& path_string,
  const Glib::ustring& new_text)
{
  unsigned int clockIndex = (unsigned int)m_freqEditIndex;
  m_freqEditIndex = -1;

  double frequency = 0.0;
  if (sscanf(new_text.c_str(), "%lf", &frequency) == 1) {
    const DeviceInfo& devInfo = m_pParent->getDeviceInfo();
    ADMXRC3_STATUS status;
    ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
    std::ostringstream ss;
    double actualFrequency;

    status = ADMXRC3_OpenEx(devInfo.m_deviceIndex, FALSE, 0, &hDevice);
    if (ADMXRC3_SUCCESS != status) {
      ss << "Failed to reopen device in active mode\n\nIndex=" << devInfo.m_deviceIndex <<
            " Status=0x" << HEX << status << "(" << ADMXRC3_GetStatusString(status, TRUE) << ")\n";
      Gtk::MessageDialog errorDialog(m_pParent->getNativeWindow(), ss.str(), false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_CLOSE, true);
      errorDialog.set_title("Failed to set clock generator frequency");
      errorDialog.run();
      return;
    }

    status = ADMXRC3_SetClockFrequency(hDevice, clockIndex, 0, frequency * 1.0e6, &actualFrequency);
    if (ADMXRC3_SUCCESS != status) {
      ss << "Failed to set clock generator " << clockIndex << " frequency\nto " << frequency << " MHz\n\n" <<
            " Status=0x" << HEX << status << "(" << ADMXRC3_GetStatusString(status, TRUE) << ")\n";
      Gtk::MessageDialog errorDialog(m_pParent->getNativeWindow(), ss.str(), false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_CLOSE, true);
      errorDialog.set_title("Failed to set clock generator frequency");
      errorDialog.run();
    }

    ADMXRC3_Close(hDevice);
  } else {
    Gtk::MessageDialog errorDialog(m_pParent->getNativeWindow(), "Frequency must be a decimal or floating point number", false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_CLOSE, true);
    errorDialog.set_title("Not a valid frequency");
    errorDialog.run();
  }
}

void
CClockGenWnd::onEditingCancelled()
{
  m_freqEditIndex = -1;
}
